# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################                   5HP  Plugin                    #############################
############################################################################################################

# Abstract:
This Paraview's programmable source reads and processes multiple five-hole probe calibration charts. This 
information is used to compute velocity (X,Y,Z), static pressure and total pressure for a given P1,...,P5 
pressure input and for given reduction method. Calibration charts are visualized in the alpha-beta plane.


# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""


print ("")
print ("")
print ('#########################################################################################################################################')
print ('##################################                                                                     ##################################')
print ('##################################                              EPSILON 3.0                            ##################################')
print ('##################################                        Aerodynamic analysis tool                    ##################################')
print ('##################################                                                                     ##################################')
print ('#########################################################################################################################################')
print ("")
print ('Developed at ISAE-SUPAERO, France, 2022')
print ("")
print ('#DISCLAIMER:')
print ('This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee of any kind, either express')
print ('or implied. Its use for teaching, research and industrial applications is allowed under the strict responsibility of the user. The authors')
print ('of this plugin are not responsible for any damages or errors related to the use of the plugin.')

print ("")
print ("")
print ("")
print ("####################################################")
print ("###############      EPSILON 5HP     ###############")
print ("####################################################")
print ("")

############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

import vtk
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa
import numpy as np
import vtk.vtkFiltersParallel
import paraview.vtk.util.numpy_support as vnp
import glob
import os
import os.path
from vtk.util import numpy_support
import sys


############################################################################################################
############################################# STRING TO FLOAT ##############################################
############################################################################################################

# Input parameters are defined as string in Paraview GUI, thus a string to float convertion is required:
SelectStep = float(SelectStep)                             						                            # 0=Visualize calibration data, 1=5HP data reduction
File_Path_Chart_1 = str(File_Path_Chart_1)  									                            # File path to the calibration data 1
File_Path_Chart_2 = str(File_Path_Chart_2)  									                            # File path to the calibration data 2
File_Path_Chart_3 = str(File_Path_Chart_3)  									                            # File path to the calibration data 3
File_Path_Chart_4 = str(File_Path_Chart_4)  									                            # File path to the calibration data 4
File_Path_Chart_5 = str(File_Path_Chart_5)  									                            # File path to the calibration data 5
File_Path_Chart_6 = str(File_Path_Chart_6)  									                            # File path to the calibration data 6
File_Path_Chart_7 = str(File_Path_Chart_7)  									                            # File path to the calibration data 7
File_Path_Chart_8 = str(File_Path_Chart_8)  									                            # File path to the calibration data 8
ProbeManufacturer = float(ProbeManufacturer)									                            # 0=Vectoflow, 1=Aeroprobe, 2=LAyF UTN, 3=Custom
ProbeType = float(ProbeType)													                            # 0=Five-hole probe, 1=Seven-hole probe, 2=Custom
HoleNBcalibration = str(HoleNBcalibration)  									                            # Hole numbering used for calibration
ReductionMethod = float(ReductionMethod)										                            # Formulation used to reduce the 5HP calibration data into coefficients
VisualizeChart = int(VisualizeChart)											                            # Select calibration chart ID to visualize
ClipCalibrationChart = float(ClipCalibrationChart)								                            # Diameter in [degrees] of the cylinder used to clip the Alpha-Beta calibration plane
ReverseClip = float(ReverseClip)												                            # 0=Disabled, 1=Enabled
HoleNBreduction = str(HoleNBreduction)											                            # Hole numbering used for data reduction
InputP1 = float(InputP1)														                            # Pressure [Pa] of hole P1
InputP2 = float(InputP2)														                            # Pressure [Pa] of hole P2
InputP3 = float(InputP3)														                            # Pressure [Pa] of hole P3
InputP4 = float(InputP4)														                            # Pressure [Pa] of hole P4
InputP5 = float(InputP5)														                            # Pressure [Pa] of hole P5
Pamb = float(Pamb)																                            # Ambient static pressure [Pa]
Tt0 = float(Tt0)																                            # Upstream total temperature [K]
Pamb2 = float(Pamb2)																                        # Ambient static pressure [Pa]
Tt02 = float(Tt02)




Measurement_Data = Measurement_Data                                                                         # 5HP measurement data
File_Parser = File_Parser                                                                                   # 0=None    1=Labview 5HP ISAE  2=User defined parser
ReductionMode = ReductionMode                                                                               # 0=Single point    1=From measurement file
Shifted_Point = Shifted_Point

Chart_Vector = [File_Path_Chart_1,File_Path_Chart_2,File_Path_Chart_3,File_Path_Chart_4,File_Path_Chart_5,File_Path_Chart_6,File_Path_Chart_7,File_Path_Chart_8]    # Create a vector with all the calibration file paths.
P_Vector = [InputP1,InputP2,InputP3,InputP4,InputP5]     						                            # Create a vector with all measured hole pressures

HoleNBcalibration = [name.strip() for name in HoleNBcalibration.split(' ')]  	                            # It splits the tab separated strings into a list of strings
HoleNBreduction = [name.strip() for name in HoleNBreduction.split(' ')]  		                            # It splits the tab separated strings into a list of strings

if ReductionMode == 0:
	Pamb = Pamb
	Tt0 = Tt0
if ReductionMode == 1:
	Pamb = Pamb2
	Tt0 = Tt02

############################################################################################################
#######################################    DEFINITION OF FUNCTIONS    ######################################
############################################################################################################

Gamma = 1.399008873407                                                                                      # Gas constant Gamma
R = 287.04213969                                                                                            # Gas constant R [J/Kg*K]

############################################
#########  PLOT VTK LINES FUNCTION  ########
############################################

def VTK_VERTEX(PointsVector):
	import numpy as np
	import vtk

	points = vtk.vtkPoints()                                                	                            # create a vtkPoints container to store all the point coordinates.
	points.SetData(dsa.numpyTovtkDataArray(PointsVector , "Points"))   			                            # numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()

	# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
	polydata2 = vtk.vtkPolyData()
	polydata2.SetPoints(points)

	Vertices = vtk.vtkVertexGlyphFilter()
	Vertices.AddInputData(polydata2)
	Vertices.Update()
	
	return Vertices


###################################
#######  DELAUNAY FUNCTION  #######
###################################

def DELAUNAY(InputPoints): 
	import vtk
	
	pts = InputPoints.GetOutput().GetPoints()

	aPolyData = vtk.vtkPolyData()
	aPolyData.SetPoints(pts)
		 
	aCellArray = vtk.vtkCellArray()
		 
	boundary = vtk.vtkPolyData()
	boundary.SetPoints(aPolyData.GetPoints())
	boundary.SetPolys(aCellArray)

	delaunay = vtk.vtkDelaunay2D()
	delaunay.SetProjectionPlaneMode(vtk.VTK_BEST_FITTING_PLANE)

	delaunay.SetInputData(aPolyData)
	delaunay.SetSourceData(boundary)

	delaunay.Update()
	return delaunay


############################################
######### ADD FIELD DATA FUNCTION ##########
############################################

def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName):                                         # Defines a function that adds the field data to the existing points in the Paraview's output
	numberOfComponents = int(numTuples)                                       	                            # We deal only with 1-component arrays
	dataArray = vtk.vtkDoubleArray()   
	dataSetAttributes.AddArray(dataArray)
	dataArray.SetNumberOfComponents(numberOfComponents)
	dataArray.SetName(ArrayName)                                               	                            # "ArrayName" is a string used to name the array in Paraview 
	data = DataArray                                                        	                            # DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
	for value in data:
		dataArray.InsertNextValue(float(value))                              	                            # Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 
	
############################################
###### FIND OPERATING POINT FUNCTION #######
############################################

def FIND_OPERATING_POINT(MeshList,VisualizeChart,CpAlpha,CpBeta,addfielddata):                                 # Finds the (alpha,beta) position of the measurement point in the calibration chart
	import vtk

	# COMPUTES CONTOUR LINE OF CONSTANT CpAlpha:
	contourA = vtk.vtkContourFilter()
	contourA.SetInputData(MeshList[VisualizeChart].GetOutput())
	contourA.SetInputArrayToProcess(0,0,0,0, "CpAlpha")
	contourA.SetValue(0, CpAlpha)
	contourA.Update()
		
	# COMPUTES CONTOUR LINE OF CONSTANT CpBeta:
	contourB = vtk.vtkContourFilter()
	contourB.SetInputData(MeshList[VisualizeChart].GetOutput())
	contourB.SetInputArrayToProcess(0,0,0,0, "CpBeta")
	contourB.SetValue(0, CpBeta)
	contourB.Update()
		
	# EXTRUDE contourA ALONG X-AXIS:
	extrudeA = vtk.vtkLinearExtrusionFilter()
	extrudeA.CappingOn()
	extrudeA.SetExtrusionTypeToVectorExtrusion()
	extrudeA.SetVector(1,0,0)
	extrudeA.SetInputData(contourA.GetOutput())
	extrudeA.SetScaleFactor(10)
	extrudeA.Update()
		
	# CALCULATE DISTANCE TO extrudeA AT EACH contourB POINT:
	curve = vtk.vtkContourFilter()   
	curve.GetOutput().ShallowCopy(contourB.GetOutput())   			                                    # Copy contourB into a new variable to avoid loosing some arrays in contourB
		
	pdd = vtk.vtkImplicitPolyDataDistance()
	pdd.SetInput(extrudeA.GetOutput())
	numPts = curve.GetOutput().GetNumberOfPoints()          
	distance = vtk.vtkDoubleArray()
	distance.SetName("Distance")
	distance.SetNumberOfComponents(1)
	distance.SetNumberOfTuples(numPts)
	for i in range(numPts):
		pt = curve.GetOutput().GetPoint(i)
		distance.SetValue(i, pdd.EvaluateFunction(pt))
		
	curve.GetOutput().GetPointData().SetScalars(distance)  
	curve.Update()

	CurveToNumpy = dsa.WrapDataObject(curve.GetOutput())                                                # Converts VTK element into a Python object
	Distance_Array = CurveToNumpy.PointData["Distance"]
	addfielddata(1, contourB.GetOutput().GetPointData(),Distance_Array,"Distance")
		
	# GET INTERSECTION POINT OF contourA AND contourB:
	vertex = vtk.vtkContourFilter()
	vertex.SetInputData(contourB.GetOutput())
	vertex.SetInputArrayToProcess(0,0,0,0, "Distance")
	vertex.SetValue(0, 0)
	vertex.Update()

	return vertex,contourA,contourB
    

############################################
#########  PLOT VTK LINES FUNCTION  ########
############################################

def PLOT_VTK_LINES_VERTEX(PointsVector):
	import numpy as np
	import vtk

	points = vtk.vtkPoints()                                                	# create a vtkPoints container to store all the point coordinates.
	points.SetData(dsa.numpyTovtkDataArray(PointsVector , "Points"))   					# numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()

	NBpts = np.shape(PointsVector)[0]                                             		# Number of mesh points
	
	polyLine = vtk.vtkPolyLine()                                            	# Creates a polyline and set the ID for each point
	polyLine.GetPointIds().SetNumberOfIds(int(NBpts)) 
	for i in range(0, int(NBpts)):
		polyLine.GetPointIds().SetId(i, i)

	cells = vtk.vtkCellArray()                                              	# Create a cell array to store the lines in and add the lines to it
	cells.InsertNextCell(polyLine)

	polyData = vtk.vtkPolyData()                                            	# Create a polydata to store everything in
	polyData.SetPoints(points)                                              	# Add the points to the dataset
	polyData.SetLines(cells)                                                	# Add the lines to the dataset

	Survey_Line = vtk.vtkGeometryFilter()                                   	# Converts the polydata to vtkPolyData  
	Survey_Line.SetInputData(polyData)             
	Survey_Line.Update()
	
	# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
	polydata2 = vtk.vtkPolyData()
	polydata2.SetPoints(points)

	Survey_Vertices = vtk.vtkVertexGlyphFilter()
	Survey_Vertices.AddInputData(polydata2)
	Survey_Vertices.Update()
	
	return Survey_Line,Survey_Vertices


###################################
#######  READ CSV FUNCTION  #######
###################################

def READ_CSV(Measurement_Data,Scale_Factors,File_Parser,Ignore_First_Line,Ignore_Last_Line,Shifted_Point):  # Note: this READ_CSV function is different from that used in Epsilon_Wake_Survey
	import numpy as np
	import vtk
	from vtk.numpy_interface import algorithms as algs
	import sys
	sys.path.insert(0, r'C:\EPSILON\Readers')
	from Reader_Labview_5HP_ISAE import Reader_Labview_5HP_ISAE
	
	if File_Parser == 1:
		DATA, Headers = Reader_Labview_5HP_ISAE(Measurement_Data)
		DATAtoSTRING = [",".join([str(x) for x in m]) for m in DATA]
		ProbeData = np.genfromtxt(DATAtoSTRING, dtype=None, names=Headers, delimiter=",", autostrip=True)
	
	if Ignore_First_Line == 1:
		ProbeData = ProbeData[1:]

	if Ignore_Last_Line == 1:
		ProbeData = ProbeData[:-1]						
		
	if Shifted_Point==0:
		# INITIALIZE ARRAYS:
		X = 0*ProbeData[ProbeData.dtype.names[0]]                               # Initialize the array with zeros. It takes the first array of data file as reference size
		Y = 0*ProbeData[ProbeData.dtype.names[0]]
		Z = 0*ProbeData[ProbeData.dtype.names[0]]
		
		# READ ARRAY DATA:
		for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
			if array=="X":                                                      # If the X array actually exists in the input data, it updates its values, otherwise a zero vector is retained
				X = ProbeData["X"]
			if array=="Y":
				Y = ProbeData["Y"]
			if array=="Z":
				Z = ProbeData["Z"]
				
	if Shifted_Point==1:   														# This technique requires transfering the 5HP data at a point N+1 to the point N due to the measurement shift
		# INITIALIZE ARRAYS:
		X = 0*ProbeData[ProbeData.dtype.names[0]][:-1]                               # Initialize the array with zeros. It takes the first array of data file as reference size
		Y = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
		Z = 0*ProbeData[ProbeData.dtype.names[0]][:-1]

		# READ ARRAY DATA:
		for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
			if array=="X":                                                      # If the X array actually exists in the input data, it updates its values, otherwise a zero vector is retained
				X = ProbeData["X"][:-1]
			if array=="Y":
				Y = ProbeData["Y"][:-1]
			if array=="Z":
				Z = ProbeData["Z"][:-1]		
	
	# CHANGE GEOMETRY SCALE:
	X = X*Scale_Factors[0]
	Y = Y*Scale_Factors[0]	
	Z = Z*Scale_Factors[0]

	CheckInputXYZ = 1
	# OUTPUT MESSAGES:
	if np.mean(X)==0 and np.mean(Y)==0 and np.mean(Z)==0:
		print ("Input probe data: coordinate arrays are unexistant or the headers are wrong (use 'X, Y and Z')")
		CheckInputXYZ = 0


	if Shifted_Point==0: 
		# INITIALIZE PRESSURE AND TEMPERATURE ARRAYS:
		Ps_ref = 0*ProbeData[ProbeData.dtype.names[0]]
		Tt_ref = 0*ProbeData[ProbeData.dtype.names[0]]
		P1 = 0*ProbeData[ProbeData.dtype.names[0]]
		P2 = 0*ProbeData[ProbeData.dtype.names[0]]
		P3 = 0*ProbeData[ProbeData.dtype.names[0]]
		P4 = 0*ProbeData[ProbeData.dtype.names[0]]
		P5 = 0*ProbeData[ProbeData.dtype.names[0]]
		
		# READ ARRAY DATA:
		for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
			if array=="Pa": 
				Ps_ref = ProbeData["Pa"]
			if array=="Tt0":
				Tt_ref = ProbeData["Tt0"]
			if array=="P1": 
				P1 = ProbeData["P1"]
			if array=="P2": 
				P2 = ProbeData["P2"]
			if array=="P3": 
				P3 = ProbeData["P3"]
			if array=="P4": 
				P4 = ProbeData["P4"]
			if array=="P5": 
				P5 = ProbeData["P5"]
				
	if Shifted_Point==1: 		
		# INITIALIZE PRESSURE AND TEMPERATURE ARRAYS:
		Ps_ref = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		Tt_ref = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		P1 = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		P2 = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		P3 = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		P4 = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		P5 = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		
		# READ ARRAY DATA:
		for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
			if array=="Pa": 
				Ps_ref = ProbeData["Pa"][1:]
			if array=="Tt0":
				Tt_ref = ProbeData["Tt0"][1:]
			if array=="P1": 
				P1 = ProbeData["P1"][1:]
			if array=="P2": 
				P2 = ProbeData["P2"][1:]
			if array=="P3": 
				P3 = ProbeData["P3"][1:]
			if array=="P4": 
				P4 = ProbeData["P4"][1:]
			if array=="P5": 
				P5 = ProbeData["P5"][1:]
	
	# PRESSURE DIMENSIONALISATION:
	P1 = P1*Scale_Factors[1]
	P2 = P2*Scale_Factors[1]
	P3 = P3*Scale_Factors[1]
	P4 = P4*Scale_Factors[1]
	P5 = P5*Scale_Factors[1]
	
	# OUTPUT MESSAGES:
	if np.mean(Ps_ref)==0 :
		print ("Input probe data: reference static pressure array is unexistant or the headers are wrong (use 'Pa')")
	if np.mean(Tt_ref)==0:
		print ("Input probe data: reference total temperature array missing or the header is wrong (use 'Tt0')")
	
	# PRINT OUTPUT MESSAGES:
	if CheckInputXYZ == 0 :
		print ("Error while reading input 5HP data")
	if CheckInputXYZ == 1 :
		print ("Input 5HP data loaded successfully")
		print ("Activate 'Axes Grid' or 'Data Axes Grid' in the properties panel to verify if the input 5HP data dimensions are in meters")
		print ("IMPORTANT: If input 5HP CSV file is in [mm] it must be transformed to [m]: Set 'Scale XYZ' from 1 to 0.001")
        
	return X,Y,Z,P1,P2,P3,P4,P5,Ps_ref,Tt_ref



############################################################################################################
########################################## LOAD CALIBRATION CHARTS #########################################
############################################################################################################

import sys
sys.path.insert(0, r'C:\EPSILON\Readers')
from Reader_Vectoflow import *                                             	                                # Import the reader of the Vectoflow 5HP calibration file format

if SelectStep >= 0:   															                            # 0=Visualize calibration data   1=5HP data reduction
	MeshList = []   															                            # Creates a list to store a delaunay mesh with the calibration data for each calibration chart
	
	#########################################################
	# LOAD CALIBRATION DATA AND GENERATE MESH FOR EACH CHART:
	
	if ReductionMethod==0:			
		print ("Reduction method set to 'None': only the parameters available in the calibration file are shown")
	
	for Chart in Chart_Vector:
		if Chart != "": 
			Chart = r'{0}'.format(Chart)                                                                    # Converts the string to raw string format, in order to deal with the "\" on the path

			# LOAD CALIBRATION DATA:
			if ProbeManufacturer == 0:   																    # 0=Vectoflow
				DATA, Headers = Reader_Vectoflow(Chart)   												    # This file parser reads Vectoflow CSV calibration charts and gives two outputs: a list of headers and a 2D array with calibration data
				DATAtoSTRING = [",".join([str(x) for x in m]) for m in DATA]							    # Calibration data is converted to a string format (required for the next step)
				ProbeData = np.genfromtxt(DATAtoSTRING, dtype=None, names=Headers, delimiter=",", autostrip=True)	# Relates the headers with the corresponding calibration data arrays
				#print (ProbeData.dtype.names)
			
			if ProbeManufacturer == 1: 
				print ("Aeroprobe file parser YET TO BE MADE!!!")

			if ProbeManufacturer == 2: 
				print ("LAyF UTN file parser YET TO BE MADE!!!"	)	
				
			if ProbeManufacturer == 3: 
				print ("Use the existing file parser to create your custom calibration file parser"	)		

			Theta = ProbeData["Theta"]
			Phi = ProbeData["Phi"]
			Alpha = ProbeData["Alpha"]
			Beta = ProbeData["Beta"]
			P1 = ProbeData["P1"]  												                            # P1 [Pa] is relative to Pref=Pamb=Ps
			P2 = ProbeData["P2"]
			P3 = ProbeData["P3"]
			P4 = ProbeData["P4"]
			P5 = ProbeData["P5"]
			Pref = ProbeData["Pref"]
			Ps = ProbeData["Ps"]
			Pt = ProbeData["Pt"]
			Ts = ProbeData["Ts"]
			Tt = ProbeData["Tt"]
			Ttc = ProbeData["Ttc"]
			M = ProbeData["M"]
			Rho = Ps/(R*(Ts+273.15))
			a = np.sqrt(Gamma*R*(Ts+273.15))
			V=M*a

			# READ HOLE PRESSURE DATA:
			if ProbeType==0:
				Pa = ProbeData[HoleNBcalibration[1]]    												    # Assigns a pressure array to the hole A
				Pb = ProbeData[HoleNBcalibration[2]]    												    # Assigns a pressure array to the hole B
				Pc = ProbeData[HoleNBcalibration[3]]    												    # Assigns a pressure array to the hole C
				Pd = ProbeData[HoleNBcalibration[4]]    												    # Assigns a pressure array to the hole D
				Pe = ProbeData[HoleNBcalibration[5]]    												    # Assigns a pressure array to the hole E
				
				Pa = Pa + Pref    												                            # Converts relative pressure to absolute pressure
				Pb = Pb + Pref	
				Pc = Pc + Pref	
				Pd = Pd + Pref
				Pe = Pe + Pref
				
			if ProbeType==1:		
				print ("Yet to code")
				
			if ProbeType==2:		
				print ("Yet to code")
		
			# COMPUTE CALIBRATION COEFFICIENTS:			
			if ReductionMethod==1:
				Psmean = (Pb+Pc+Pd+Pe)/4
				CpAlpha = (Pb-Pc)/(Pa-Psmean)
				CpBeta = (Pd-Pe)/(Pa-Psmean)
				Cps = (Psmean-Ps)/(Pa-Psmean)
				Cpt = (Pa-Pt)/(Pa-Psmean)

			if ReductionMethod==2:		
				print ("Yet to code")
				
			if ReductionMethod==3:		
				print ("Yet to code")
				
			if ReductionMethod==4:		
				print ("Yet to code")
	
			# GENERATE DELAUNAY MESH:
			PointsVector = algs.make_vector(0*Beta,Beta,Alpha)
			Vertices = VTK_VERTEX(PointsVector)
			delaunay = DELAUNAY(Vertices)
			
			# ADD MEASURED 5HP DATA INTO THE DELAUNAY MESH:     
			addfielddata(1, delaunay.GetOutput().GetPointData(),Theta,"Theta")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Phi,"Phi")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Alpha,"Alpha")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Beta,"Beta")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Pref,"Pref")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Ps,"Ps_cal")		
			addfielddata(1, delaunay.GetOutput().GetPointData(),Pt,"Pt_cal")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Ts,"Ts_cal")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Tt,"Tt_cal")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Ttc,"T_thermocouple")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Rho,"Rho_cal")
			addfielddata(1, delaunay.GetOutput().GetPointData(),V,"V_cal")
			addfielddata(1, delaunay.GetOutput().GetPointData(),M,"M_cal")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P1,"P1")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P2,"P2")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P3,"P3")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P4,"P4")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P5,"P5")
	
			if ReductionMethod!=0:
				addfielddata(1, delaunay.GetOutput().GetPointData(),Pa,"Pa")
				addfielddata(1, delaunay.GetOutput().GetPointData(),Pb,"Pb")
				addfielddata(1, delaunay.GetOutput().GetPointData(),Pc,"Pc")
				addfielddata(1, delaunay.GetOutput().GetPointData(),Pd,"Pd")
				addfielddata(1, delaunay.GetOutput().GetPointData(),Pe,"Pe")
				addfielddata(1, delaunay.GetOutput().GetPointData(),Psmean,"Psmean")		
				addfielddata(1, delaunay.GetOutput().GetPointData(),CpAlpha,"CpAlpha")		
				addfielddata(1, delaunay.GetOutput().GetPointData(),CpBeta,"CpBeta")			
				addfielddata(1, delaunay.GetOutput().GetPointData(),Cps,"Cps")			
				addfielddata(1, delaunay.GetOutput().GetPointData(),Cpt,"Cpt")				
			
			# CLIP CALIBRATION CHART:
			if ClipCalibrationChart != 0:
				# GENERATE CYLINDER ALLIGNED WITH Z-AXIS:
				transf = vtk.vtkTransform()
				transf.RotateZ(90) 														                    # By default, VTK alligns the cylinder with Y-axis, thus, a transformation is required to allign it with the X-axis
				cylinder = vtk.vtkCylinder()
				cylinder.SetCenter(0,0,0) 
				cylinder.SetRadius(ClipCalibrationChart)
				cylinder.SetTransform(transf)
				
				# CUT DELAUNAY WITH CYLINDER:
				clipper = vtk.vtkTableBasedClipDataSet()                                                    #vtkClipPolyData   
				clipper.SetInputData(delaunay.GetOutput())
				clipper.SetClipFunction(cylinder)
				clipper.SetValue(0)
				clipper.GenerateClipScalarsOff()
				if ReverseClip==0:
					clipper.InsideOutOn()
				if ReverseClip==1:
					clipper.InsideOutOff()
				clipper.GetOutput().GetPointData().CopyScalarsOff()
				clipper.Update()
				
				# Extract surface:
				ExtractSurface = vtk.vtkGeometryFilter()                                                    # Converts vtkUnstructuredGrid to vtkPolyData 
				ExtractSurface.SetInputData(clipper.GetOutput())               
				ExtractSurface.Update()
                
				delaunay = ExtractSurface
			
			# ADD THE CHART MESH TO THE LIST CONTAINING ALL THE DELAUNAYS 
			MeshList.append(delaunay)
			
	if SelectStep == 0:                                                                                     # 0=Visualize calibration data
		#########################################################
		# VISUALIZATION OF SELECTED DELAUNAY MESH:
		self.GetOutput().ShallowCopy(MeshList[VisualizeChart].GetOutput())

		# OUTPUT MESSAGES:
		print ("Change the visualization mode to 'Surface with edges' or 'Points' in order to see the calibration points")
		print ("Y-axis=Beta [deg]   Z-axis=Alfa [deg]")


############################################################################################################
############################################ 5HP DATA REDUCTION ############################################
############################################################################################################

if SelectStep == 1:   																					    # 1=5HP data reduction
		
	if ReductionMethod==0:
		print ("Please, select a reduction method first")
		
	if ReductionMethod!=0:
			
			
		if ReductionMode == 0:                                                                              # 0=Single point   
			# READS THE PRESSURE VALUE AT EACH PORT:
			DATA = [P_Vector[0],P_Vector[1],P_Vector[2],P_Vector[3],P_Vector[4]]
			Headers = ["P1","P2","P3","P4","P5"]
			DATAtoSTRING = [",".join([str(x) for x in DATA]) ]										        # Calibration data is converted into a string format (required for the next step)	
			ProbeData = np.genfromtxt(DATAtoSTRING, dtype=None, names=Headers, delimiter=",", autostrip=True)	# Relates the headers with the corresponding calibration data arrays
			
			# ASSIGNS THE PORT PRESSURES TO EACH CONVENTIONAL HOLE:
			Pa = ProbeData[HoleNBreduction[1]]    														    # Assigns a pressure value to the hole A
			Pb = ProbeData[HoleNBreduction[2]]    														    # Assigns a pressure value to the hole B
			Pc = ProbeData[HoleNBreduction[3]]    														    # Assigns a pressure value to the hole C
			Pd = ProbeData[HoleNBreduction[4]]    														    # Assigns a pressure value to the hole D
			Pe = ProbeData[HoleNBreduction[5]]    														    # Assigns a pressure value to the hole E

			Pa = Pa + Pamb                                                                                  # 5HP pressures are always relative. Then, pressures are converted to absolute 
			Pb = Pb + Pamb	
			Pc = Pc + Pamb	
			Pd = Pd + Pamb
			Pe = Pe + Pamb

		if ReductionMode == 1:                                                                              #   1=From measurement file
			# READS THE PRESSURE VALUE AT EACH PORT:
			X,Y,Z,P1,P2,P3,P4,P5,Ps_ref,Tt_ref = READ_CSV(Measurement_Data,Scale_Factors,File_Parser,Ignore_First_Line,Ignore_Last_Line,Shifted_Point) 
			
			if File_Path_CSV_1.endswith('.csv') or File_Path_CSV_1.endswith('.moy'):                        # Add extra 5HP data
				X_new,Y_new,Z_new,P1_new,P2_new,P3_new,P4_new,P5_new,Ps_ref_new,Tt_ref_new = READ_CSV(File_Path_CSV_1,Scale_Factors,File_Parser,Ignore_First_Line,Ignore_Last_Line,Shifted_Point)
				X = np.append(X,X_new)
				Y = np.append(Y,Y_new)
				Z = np.append(Z,Z_new)
				P1 = np.append(P1,P1_new)
				P2 = np.append(P2,P2_new)
				P3 = np.append(P3,P3_new)
				P4 = np.append(P4,P4_new)
				P5 = np.append(P5,P5_new)
				Ps_ref = np.append(Ps_ref,Ps_ref_new)
				Tt_ref = np.append(Tt_ref,Tt_ref_new)

			if File_Path_CSV_2.endswith('.csv') or File_Path_CSV_2.endswith('.moy'):                        # Add extra 5HP data
				X_new,Y_new,Z_new,P1_new,P2_new,P3_new,P4_new,P5_new,Ps_ref_new,Tt_ref_new = READ_CSV(File_Path_CSV_2,Scale_Factors,File_Parser,Ignore_First_Line,Ignore_Last_Line,Shifted_Point)
				X = np.append(X,X_new)
				Y = np.append(Y,Y_new)
				Z = np.append(Z,Z_new)
				P1 = np.append(P1,P1_new)
				P2 = np.append(P2,P2_new)
				P3 = np.append(P3,P3_new)
				P4 = np.append(P4,P4_new)
				P5 = np.append(P5,P5_new)
				Ps_ref = np.append(Ps_ref,Ps_ref_new)
				Tt_ref = np.append(Tt_ref,Tt_ref_new)

			if File_Path_CSV_3.endswith('.csv') or File_Path_CSV_3.endswith('.moy'):                        # Add extra 5HP data
				X_new,Y_new,Z_new,P1_new,P2_new,P3_new,P4_new,P5_new,Ps_ref_new,Tt_ref_new = READ_CSV(File_Path_CSV_3,Scale_Factors,File_Parser,Ignore_First_Line,Ignore_Last_Line,Shifted_Point)
				X = np.append(X,X_new)
				Y = np.append(Y,Y_new)
				Z = np.append(Z,Z_new)
				P1 = np.append(P1,P1_new)
				P2 = np.append(P2,P2_new)
				P3 = np.append(P3,P3_new)
				P4 = np.append(P4,P4_new)
				P5 = np.append(P5,P5_new)
				Ps_ref = np.append(Ps_ref,Ps_ref_new)
				Tt_ref = np.append(Tt_ref,Tt_ref_new)
                
			if File_Path_CSV_4.endswith('.csv') or File_Path_CSV_4.endswith('.moy'):                        # Add extra 5HP data
				X_new,Y_new,Z_new,P1_new,P2_new,P3_new,P4_new,P5_new,Ps_ref_new,Tt_ref_new = READ_CSV(File_Path_CSV_4,Scale_Factors,File_Parser,Ignore_First_Line,Ignore_Last_Line,Shifted_Point)
				X = np.append(X,X_new)
				Y = np.append(Y,Y_new)
				Z = np.append(Z,Z_new)
				P1 = np.append(P1,P1_new)
				P2 = np.append(P2,P2_new)
				P3 = np.append(P3,P3_new)
				P4 = np.append(P4,P4_new)
				P5 = np.append(P5,P5_new)
				Ps_ref = np.append(Ps_ref,Ps_ref_new)
				Tt_ref = np.append(Tt_ref,Tt_ref_new)
            
			# ASSIGNS THE PORT PRESSURES TO EACH CONVENTIONAL HOLE:
			Pa = eval(HoleNBreduction[1])   																# Assigns a pressure value to the hole A
			Pb = eval(HoleNBreduction[2])     																# Assigns a pressure value to the hole B
			Pc = eval(HoleNBreduction[3])     																# Assigns a pressure value to the hole C
			Pd = eval(HoleNBreduction[4])     																# Assigns a pressure value to the hole D
			Pe = eval(HoleNBreduction[5])     																# Assigns a pressure value to the hole E

			Pa = Pa + Pamb                                                                                  # 5HP pressures are always relative. Then, pressures are converted to absolute 
			Pb = Pb + Pamb	
			Pc = Pc + Pamb	
			Pd = Pd + Pamb
			Pe = Pe + Pamb



		
		# COMPUTES THE RELATED COEFFICIENTS:
		if ReductionMethod==1:
			Psmean = (Pb+Pc+Pd+Pe)/4
			CpAlpha = (Pb-Pc)/(Pa-Psmean)
			CpBeta = (Pd-Pe)/(Pa-Psmean)
			
		if ReductionMethod==2:		
			print ("Yet to code")
			
		if ReductionMethod==3:		
			print ("Yet to code")
			
		if ReductionMethod==4:		
			print ("Yet to code")

		# 5HP DATA REDUCTION:	
		Alpha_wake = []
		Beta_wake = []
		Cps_wake = []
		Cpt_wake = []
		Ps_wake = []
		Pt_wake = []
		Pdynamic_wake = []
		M_wake = []
		Ts_wake = []
		Rho_wake = []
		V_wake = []
		Vx_wake = []
		Vy_wake = []
		Vz_wake = []
        
		for i in range(np.size(CpAlpha)):       

			# FIND THE OPERATING POINT OF THE MEASUREMENT IN THE CALIBRATION CHART:
			vertex,contourA,contourB = FIND_OPERATING_POINT(MeshList,VisualizeChart,CpAlpha[i],CpBeta[i],addfielddata)        

			# GET DATA FROM INTERSECTION POINT:
			vertexToNumpy = dsa.WrapDataObject(vertex.GetOutput())             # Converts VTK element into a Python object
			Alpha = vertexToNumpy.PointData["Alpha"]
			Beta = vertexToNumpy.PointData["Beta"]
			Cps = vertexToNumpy.PointData["Cps"]
			Cpt = vertexToNumpy.PointData["Cpt"]
			
			if Alpha.size > 1:
				print ("ATTENTION: ", Alpha.size ," different values of Alpha have been found --> ",Alpha,"  The mean value has been kept")
			if Beta.size > 1:
				print ("ATTENTION: ", Beta.size ," different values of Beta have been found --> ",Beta,"  The mean value has been kept")
			if Cps.size > 1:
				print ("ATTENTION: ", Cps.size ," different values of Cps have been found --> ",Cps,"  The mean value has been kept")
			if Cpt.size > 1:
				print ("ATTENTION: ", Cpt.size ," different values of Cpt have been found --> ",Cpt,"  The mean value has been kept")	
			
			Alpha = mean(Alpha)
			Beta = mean(Beta)
			Cps = mean(Cps)
			Cpt = mean(Cpt)
			
			# COMPUTE REMAINING AERODYNAMIC PARAMETERS:
			Ps = Psmean[i]-Cps*(Pa[i]-Psmean[i])
			Pt = Pa[i] - Cpt*(Pa[i]-Psmean[i])
			Pdynamic = Pt-Ps
			M = np.sqrt((2/(Gamma-1))*((Pt/Ps)**((Gamma-1)/Gamma)-1))
			Ts = Tt0/(1+0.5*(Gamma-1)*M**2)                                   		# Static temperature [K] --> assumes isoenergetic flow Tt=Tt0
			Rho = Ps/(R*Ts)
			a = np.sqrt(Gamma*R*Ts)
			V = M*a
			Tita = Alpha*np.arctan(1/np.sqrt(1+np.tan(Beta*np.pi/180)))  			# See Figure 3 paper AIAA: "A velocity decomposition method for exergy based drag prediction" https://doi.org/10.2514/1.J059414
			Vx = V*np.cos(Tita*np.pi/180)*np.cos(Beta*np.pi/180)
			Vy = V*np.cos(Tita*np.pi/180)*np.sin(Beta*np.pi/180)
			Vz = V*np.sin(Tita*np.pi/180)

			Alpha_wake.append(Alpha)
			Beta_wake.append(Beta)
			Cps_wake.append(Cps)
			Cpt_wake.append(Cpt)
			Ps_wake.append(Ps)
			Pt_wake.append(Pt)
			Pdynamic_wake.append(Pdynamic)
			M_wake.append(M)
			Ts_wake.append(Ts)
			Rho_wake.append(Rho)
			V_wake.append(V)
			Vx_wake.append(Vx)
			Vy_wake.append(Vy)
			Vz_wake.append(Vz)      

		Vx_wake = np.array(Vx_wake)
		Vy_wake = np.array(Vy_wake)
		Vz_wake = np.array(Vz_wake)
		Velocity = algs.make_vector(Vx_wake,Vy_wake,Vz_wake)

		# PRINT REDUCED 5HP DATA:
		if ReductionMode==0:	                                                                                # 0=Single point
			print ("")
			print ("# REDUCED 5HP DATA:")
			print ("CpAlpha=",("%.3f" % CpAlpha),"	CpBeta=",("%.3f" % CpBeta),"	Cps=",("%.3f" % Cps),"	Cpt=",("%.3f" % Cpt))
			print ("")
			print ("Alpha=",("%.3f" % Alpha),"[deg]	Beta=",("%.3f" % Beta),"[deg]	Tita=",("%.3f" % Tita),"[Deg]"   )
			print ("Pt=",("%.1f" % Pt),"[Pa]	Ps=",("%.1f" % Ps),"[Pa]	Pd=",("%.1f" % Pdynamic),"[Pa]")
			print ("Vx=",("%.2f" % Vx),"[m/s]	Vy=",("%.2f" % Vy),"[m/s]	Vz=",("%.2f" % Vz),"[m/s]")
			print ("V=",("%.2f" % V),"[m/s]	M=",("%.5f" % M))
			print ("Ts=",("%.1f" % Ts),"[K]	Rho=",("%.4f" % Rho),"[Kg/m^3]")
	
		# VISUALIZE OUTPUT DATA:
		if ReductionMode==0:	                                                                                # 0=Single point
			# COMBINE THE VERTEX, ISOLINES AND THE DELAUNAY MESH INTO A SINGLE VTK VARIABLE:
			appendFilter = vtk.vtkAppendFilter()                                    #Output: vtkUnstructuredGrid
			appendFilter.AddInputData(vertex.GetOutput()) 
			appendFilter.AddInputData(contourA.GetOutput())        
			appendFilter.AddInputData(contourB.GetOutput())
			appendFilter.AddInputData(MeshList[VisualizeChart].GetOutput())
			appendFilter.Update()

			# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA  
			combined = vtk.vtkGeometryFilter()                                      # Converts vtkMultiblockDataSet to vtkPolyData
			combined.SetInputData(appendFilter.GetOutput())                             
			combined.Update()
		
			# VISUALIZATION:
			self.GetOutput().ShallowCopy(combined.GetOutput())                  # Displays output data

		if ReductionMode == 1:                                                                              #   1=From measurement file
			# GENERATE THE PROBE POINTS:
			PointsVector = algs.make_vector(X,Y,Z)
			Survey_Line,Survey_Vertices = PLOT_VTK_LINES_VERTEX(PointsVector)   		
			delaunay = DELAUNAY(Survey_Vertices)
			
			# ADD MEASURE 5HP DATA TO THE VERTICES:
			addfielddata(1, delaunay.GetOutput().GetPointData(),Ps_ref,"Ps_ref")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Tt_ref,"Tt_ref")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P1,"P1")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P2,"P2")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P3,"P3")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P4,"P4")
			addfielddata(1, delaunay.GetOutput().GetPointData(),P5,"P5")
			addfielddata(3, delaunay.GetOutput().GetPointData(),Velocity.ravel(),"Velocity")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Alpha_wake,"Alpha")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Beta_wake,"Beta")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Cps_wake,"Cps")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Cpt_wake,"Cpt")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Ps_wake,"Ps")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Pt_wake,"Pt")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Pdynamic_wake,"Pd")
			addfielddata(1, delaunay.GetOutput().GetPointData(),M_wake,"M")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Ts_wake,"Ts")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Rho_wake,"Rho")
			addfielddata(1, delaunay.GetOutput().GetPointData(),V_wake,"V")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Vx_wake,"Vx")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Vy_wake,"Vy")
			addfielddata(1, delaunay.GetOutput().GetPointData(),Vz_wake,"Vz")

			# VISUALIZATION:
			self.GetOutput().ShallowCopy(delaunay.GetOutput())                  # Displays output data



############################################################################################################
#############################################    END OF CODE   #############################################
############################################################################################################