
# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################                 Aero Data Plugin                 #############################
############################################################################################################

# Abstract:
This Paraview's programmable filter takes 1 input from pipeline: a 2D/3D data (CFD or WTT data) and 
completes the remaining aerodynamic field parameters if necessary. Moreover, it calculates far-field-method 
and exergy-method parameters at all the points. Details of the formulations and how to use the plugin are 
available in the User's Manual.


# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""

print ("")
print ("")
print ('#########################################################################################################################################')
print ('##################################                                                                     ##################################')
print ('##################################                              EPSILON 3.0                            ##################################')
print ('##################################                        Aerodynamic analysis tool                    ##################################')
print ('##################################                                                                     ##################################')
print ('#########################################################################################################################################')
print ("")
print ('Developed at ISAE-SUPAERO, France, 2022')
print ("")
print ('#DISCLAIMER:')
print ('This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee of any kind, either express')
print ('or implied. Its use for teaching, research and industrial applications is allowed under the strict responsibility of the user. The authors')
print ('of this plugin are not responsible for any damages or errors related to the use of the plugin.')

print ("")
print ("")
print ("")
print ("####################################################")
print ("############      EPSILON AERO DATA     ############")
print ("####################################################")
print ("")



############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

# Loading the standard Python and Paraview modules:
import numpy as np
import vtk
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa
from paraview.modules.vtkPVVTKExtensionsMisc import vtkMergeBlocks


# Defining the path to the readers of the different input file formats:
import sys
sys.path.insert(0, r'C:\EPSILON\Readers')

from Reader_elsA_CGNS_BC import *
from Reader_elsA_Tecplot_Binary import *
from Reader_Epsilon_PIV import *
from Reader_Epsilon_Wake_Survey import *
from Reader_Fluent_CAS import * 
from Reader_Fluent_CGNS import * 
from Reader_Fluent_Ensight import *
from Reader_IC3 import *
from Reader_OpenFOAM import *
from Reader_Potential_Flow_Generator import *
from Reader_STAR_CCM_Ensight import *
from Reader_SU2 import *
from Reader_TAU_Tecplot import *
from Reader_New_Format import *
from Reader_Epsilon_5HP import *

############################################################################################################
############################################# INPUT PARAMETERS #############################################
############################################################################################################

# Input parameters are defined as string in Paraview's GUI, thus a string to float convertion is required:
AMR = AMR
ExtractBlock_ID = ExtractBlock_ID
InputFormat = float(InputFormat)
UpwardAxis = float(UpwardAxis) 
PRESSURE = float(PRESSURE) 
DENSITY = float(DENSITY)
TEMPERATURE = float(TEMPERATURE)
Timestep = float(Timestep)  
Alfa = float(Alfa) 
Beta = float(Beta)
SpeedInputType = float(SpeedInputType)
Speed = float(Speed)
TypeOfParameter = float(TypeOfParameter)
P0 = float(P0)
T0 = float(T0)
Sref = float(Sref)
GetRefData = float(GetRefData)    
Nondimensionalize = float(Nondimensionalize)
UDV = float(UDV)
AERO = float(AERO)
AnalysisMethods = float(AnalysisMethods)
NEAR_FIELD = float(NEAR_FIELD)
FAR_FIELD = float(FAR_FIELD)
EXERGY = float(EXERGY)
LAMB_VECTOR = float(LAMB_VECTOR)
Momentum = float(Momentum)
Betz = float(Betz)
Jones = float(Jones)
Oswatitsch = float(Oswatitsch)
Maskell = float(Maskell)
Onorato = float(Onorato)
VDV = float(VDV)
Giles = float(Giles)
Kusunose = float(Kusunose)
Meheut = float(Meheut)
Toubin = float(Toubin)
Wu = float(Wu)
Mele = float(Mele)
Arntz = float(Arntz)
Aguirre = float(Aguirre)
Ruscio = float(Ruscio)
Dimension = float(Dimension)
R = float(R)
Gamma = float(Gamma)
Pr_turb = float(Pr_turb)
Save_Output_Data = float(Save_Output_Data)
File_Path_Output = str(File_Path_Output)
Select_Arrays = float(Select_Arrays)     				
UserSelectedArray = UserSelectedArray


	
############################################################################################################
######################################## PRETREATMENT OF INPUT DATA ########################################
############################################################################################################

INPUT=inputs[0]                                                         # Get input data


# Extract a given block from the input:
if AMR == 1:                                                            # Input grid type is: Adaptive Mesh Refinement                                         
	Extracted=INPUT.GetBlock(int(ExtractBlock_ID))                      # Extracting the desired block from the multiple block input --> useful for Adaptive mesh refinement and also when the input CFD data contains a block for the domain as well as a block for each boundary condition (wall, far-field,...) 
	INPUT = dsa.WrapDataObject(Extracted)                               # "WrapDataObject" Converts the VTK object into a Python object (it now replaces the original INPUT data)


# Convert a multiblock data into a single block:
if INPUT.IsA("vtkCompositeDataSet"):                                    # Merge blocks: if input data is a vtkCompositeDataSet (i.e., multiblock), then it merges the related blocks
	mergeFilter = vtkMergeBlocks()
	mergeFilter.SetInputData(INPUT.VTKObject)
	mergeFilter.Update()
	INPUT = dsa.WrapDataObject(mergeFilter.GetOutput())


# Convert cell-based arrays to point-based data:
if size(INPUT.CellData.keys())>0:                                       # Detects if there are cell-based arrays at the input data
	myArrays=vtk.vtkPassArrays()                                 	    # use pass arrays to extract a copy of the input data
	myArrays.SetInputDataObject(INPUT.VTKObject)  
	
	c2p = vtk.vtkCellDataToPointData()     							    # Converting the input from Cell Data To Point Data
	c2p.SetInputConnection(myArrays.GetOutputPort())
	c2p.Update()
	INPUT = dsa.WrapDataObject(c2p.GetOutput()) 					    


# Convert vtkPolyData input to vtkStructured grid:                      # This is required to process PolyData only (e.g, EpsilonPIV, EpsilonWakeSurvey and CFD slices) because the ouput format of AeroData is a vtkUnstructuredGrid
if INPUT.IsA("vtkPolyData"): 
	print('# Input format is "vtkPolyData". It is converted to "vtkUnstructuredGrid"(if input mesh has mixed cell types or triangle strips it will fail-->try applying "Clean to grid" or "Tetrahedralize" filter before AeroData)')
	clean = vtk.vtkmCleanGrid()                         
	clean.SetInputData(INPUT.VTKObject) 
	clean.CompactPointsOn()                                             # ATTENTION: this command eliminates any tensor array existing at the input, but it is necessary because vtkmCleanGrid do not work fine without this 
	clean.Update()
	INPUT = dsa.WrapDataObject(clean.GetOutput())

# Scales geometry to [m]:
if Scale_Geometry==1:
	ScaleGeom = vtk.vtkTransform()                                                                             # Generates the VTK transformation function to scale up/down the PIV data
	ScaleGeom.Scale(Geometry_Scale_Factor, Geometry_Scale_Factor, Geometry_Scale_Factor)
	ScaleGeom.Update()
	
	TransformedVTK = vtk.vtkTransformFilter()                                                                     # Applies the scale up/down to the input PIV data and creates a transformed output
	TransformedVTK.SetTransform(ScaleGeom)
	TransformedVTK.SetInputData(INPUT.VTKObject)
	TransformedVTK.Update()
	INPUT = dsa.WrapDataObject(TransformedVTK.GetOutput())

# Send input mesh (without arrays) to the output:                       # This mesh is used as a canvas where all the new arrays (calculated below) will be added
mesh=vtk.vtkPassArrays()   
mesh.SetInputDataObject(INPUT.VTKObject)
mesh.AddPointDataArray("")                               	            # This deletes all the existing arrays at the input mesh
mesh.Update()
self.GetOutput().ShallowCopy(mesh.GetOutput())                          # This send the mesh to the output of ParaView


	
############################################################################################################
############################################ FILE FORMAT READER ############################################
############################################################################################################

CpGas = (Gamma*R)/(Gamma-1)                                             # Gas constant Cp [J/Kg*K]
CvGas = CpGas-R                                                         # Gas constant Cv [J/Kg*K]


# Input format declared by user:                                        # It executes the selected file format reader function, which launches another Python file. It reads the raw data comming from the INPUT file.
if InputFormat==0:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_elsA_CGNS_BC(INPUT)
elif InputFormat==1:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_elsA_Tecplot_Binary(INPUT,R,CvGas)
elif InputFormat==2:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY,V_star = Reader_Epsilon_PIV(INPUT)
elif InputFormat==3:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY,Pt = Reader_Epsilon_Wake_Survey(INPUT)
elif InputFormat==14:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY,Pt = Reader_Epsilon_5HP(INPUT)
elif InputFormat==4:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_Fluent_CAS(INPUT)
elif InputFormat==5:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_Fluent_CGNS(INPUT)  
elif InputFormat==6:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_Fluent_Ensight(INPUT)	
elif InputFormat==7:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_IC3(INPUT)   
elif InputFormat==8:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_OpenFOAM(INPUT)
elif InputFormat==9:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_Potential_Flow_Generator(INPUT)
elif InputFormat==10:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_STAR_CCM_Ensight(INPUT)
elif InputFormat==11:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_SU2(INPUT,R,CvGas)
elif InputFormat==12:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_TAU_Tecplot(INPUT)
elif InputFormat==13:
	POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY = Reader_New_Format(INPUT)



############################################################################################################
######################################### COMPLETING UPSTREAM DATA #########################################
############################################################################################################

if InputFormat==2 or InputFormat==3 or InputFormat==14:                                                     # 2=Epsilon PIV 3=Epsilon Wake Survey   14=Epsilon 5HP
	if GetRefData==1:                                                                                       # 1=Get reference Ps0 Ts0 V0 values from CFD data
		print('# ERROR: input is WTT data, then the option "Get reference Ps0 Ts0 V0 values from CFD data" can not be enabled')

# Manual data completion (uses the GUI info provided by the user):      # It depends on SpeedInputType (Mach or velocity) and TypeOfParameter (Total or static) selected by the user
if SpeedInputType == 0 and TypeOfParameter == 0:
	M0 = Speed                                                          # Mach number
	Pt0 = P0                                                            # Total pressure [Pa]
	Tt0 = T0                                                            # Total temperature [K]
	Ps0 = Pt0/((1+0.5*(Gamma-1)*M0**2)**(Gamma/(Gamma-1)))              # Static pressure [Pa]
	Ts0 = Tt0/(1+0.5*(Gamma-1)*M0**2)                                   # Static temperature [K]
	e0 = CvGas*Ts0                                                      # Internal energy [J/Kg]
	Rho0 = Ps0/(R*Ts0)                                                  # Density [Kg/m^3]
	a0 = np.sqrt(Gamma*R*Ts0)                                           # Speed of sound [m/s]
	V0 = M0*a0                                                          # Airspeed [m/s]
	q0 = 0.5*Rho0*V0**2                                                 # Dynamic pressure [Pa]
	Hs0 = CpGas*Ts0                                                     # Static enthalpy [J/Kg]
	Ht0 = CpGas*Tt0                                                     # Total enthalpy [J/Kg]
	
elif SpeedInputType == 0 and TypeOfParameter == 1:
	M0 = Speed                                                          
	Ps0 = P0
	Ts0 = T0      
	Pt0 = Ps0*((1+0.5*(Gamma-1)*M0**2)**(Gamma/(Gamma-1)))              
	Tt0 = Ts0*(1+0.5*(Gamma-1)*M0**2)                                   
	e0 = CvGas*Ts0                                                      
	Rho0 = Ps0/(R*Ts0)                                                  
	a0 = np.sqrt(Gamma*R*Ts0)                                          
	V0 = M0*a0                                                          
	q0 = 0.5*Rho0*V0**2                                                 
	Hs0 = CpGas*Ts0
	Ht0 = CpGas*Tt0

elif SpeedInputType == 1 and TypeOfParameter == 0:
	V0 = Speed
	Pt0 = P0
	Tt0 = T0
	Ts0 = Tt0-((0.5*V0**2)/CpGas)
	e0 = CvGas*Ts0                                                      
	a0 = np.sqrt(Gamma*R*Ts0)                                           
	M0 = V0/a0     
	Ps0 = Pt0/(1+0.5*(Gamma-1)*M0**2)**(Gamma/(Gamma-1))
	Rho0 = Ps0/(R*Ts0)
	q0 = 0.5*Rho0*V0**2
	Hs0 = CpGas*Ts0
	Ht0 = CpGas*Tt0

elif SpeedInputType == 1 and TypeOfParameter == 1:
	V0 = Speed
	Ps0 = P0
	Ts0 = T0 
	Tt0 = Ts0+((0.5*V0**2)/CpGas)
	e0 = CvGas*Ts0                                                      
	a0 = np.sqrt(Gamma*R*Ts0)                                           
	M0 = V0/a0     
	Pt0 = Ps0*(1+0.5*(Gamma-1)*M0**2)**(Gamma/(Gamma-1))
	Rho0 = Ps0/(R*Ts0)
	q0 = 0.5*Rho0*V0**2
	Hs0 = CpGas*Ts0
	Ht0 = CpGas*Tt0


# Automatic data completion:                                            # It is executed when "Get reference Ps0 Ts0 V0 values from CFD data" option is enabled

def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName):     # Defines a function that adds the field data to the existing points in the Paraview's output
	numberOfComponents = int(numTuples)                                 # We deal only with 1-component arrays
	dataArray = vtk.vtkDoubleArray()   
	dataSetAttributes.AddArray(dataArray)
	dataArray.SetNumberOfComponents(numberOfComponents)
	dataArray.SetName(ArrayName)                                        # "ArrayName" is a string used to name the array in Paraview 
	data = DataArray                                                    # DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
	for value in data:
		dataArray.InsertNextValue(float(value))                         # Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 


if UpwardAxis==0:                                                       #Slice plane normals for mesh Y-axis pointing upwards
	nx = -np.sin(Alfa*np.pi/180)
	ny = np.cos(Alfa*np.pi/180)
	nz = 0
elif UpwardAxis==1:                                                     #Slice plane normals for mesh Z-axis pointing upwards
	nx = -np.sin(Alfa*np.pi/180)
	ny = 0
	nz = np.cos(Alfa*np.pi/180)


if GetRefData==1:
	if min(Ps)>0:
		if Ts!="None": 
			
			# Add V,Ps and Ts arrays to the mesh:                       # It takes the empty mesh and add these arrays which will be used to stablish the reference values
			VelAbs = algs.mag(Velocity_Vector)
			addfielddata(1, mesh.GetOutput().GetPointData(),VelAbs,"VRef")
			addfielddata(1, mesh.GetOutput().GetPointData(),Ps,"PsRef")
			addfielddata(1, mesh.GetOutput().GetPointData(),Ts,"TsRef")
			
			# Create slice plane:
			plane = vtk.vtkPlane()										           
			plane.SetOrigin(0, 0, 0)                                    # It assumes that the aircraft geometry is always centered at the origin
			plane.SetNormal(nx, ny, nz)

			# Slices mesh data with slice plane:
			cutter = vtk.vtkCutter()											    
			cutter.SetCutFunction(plane)
			cutter.SetInputData(mesh.GetOutput())
			cutter.Update()
			SLICE = dsa.WrapDataObject(cutter.GetOutput())   				       
			
			# Get Ps/Ts/V at the most upstream point of the slice:
			PtsSlice = SLICE.Points   								    # Get points coordinates
			IDxmin = algs.where(PtsSlice[:,0]==min(PtsSlice[:,0]))      # Gives the X-coordinate array ID where X=Xmin
			Ps0 = SLICE.PointData ["PsRef"][IDxmin][0]  
			Ts0 = SLICE.PointData ["TsRef"][IDxmin][0]
			V0 = SLICE.PointData ["VRef"][IDxmin][0] 
			
			# Compute the remaining upstream parameters:
			Tt0 = Ts0+((0.5*V0**2)/CpGas)
			e0 = CvGas*Ts0                                                      
			a0 = np.sqrt(Gamma*R*Ts0)                                           
			M0 = V0/a0     
			Pt0 = Ps0*(1+0.5*(Gamma-1)*M0**2)**(Gamma/(Gamma-1))
			Rho0 = Ps0/(R*Ts0)
			q0 = 0.5*Rho0*V0**2
			Hs0 = CpGas*Ts0
			Ht0 = CpGas*Tt0
			print('# Upstream reference values: Ps0=',Ps0,'[Pa]   Ts0=',Ts0,'[K]   V0=',V0,'[m/s]')   
		else:
			print('ERROR: the option "Get reference Ps0 Ts0 V0 values from CFD data" can not be used because the static temperature array is not available at the input')
	else:
		print('ERROR: the option "Get reference Ps0 Ts0 V0 values from CFD data" can not be used because the pressure field is relative')

  



############################################################################################################
###################################### COMPUTING MAIN FIELD PARAMETERS #####################################
############################################################################################################

V = algs.sqrt(Velocity_Vector[:,0]**2+Velocity_Vector[:,1]**2+Velocity_Vector[:,2]**2)                  # Velocity magnitude [m/s]. Alternative code: V = algs.mag(Velocity_Vector). This is more efficient but raises error with LambVector (V2_GRADIENT). Left here for debugging purposes.

if Ps=="None" or Ts=="None" or Rho=="None":
	if Reconstruct==0:
		print ('ERROR: Input data do not include Ps, Ts or Rho. Enable "Reconstruct Ps Ts Rho fields" to complete missing data')

# Static pressure:
if PRESSURE==0:                                                                                         # Ps is defined as the input value
	if Ps=="None":
		print ('Input data do not include Ps. Choose other "Pressure field" option in the GUI')
	else:
		if min(Ps)<0:                                                                                   # This check if "Ps" array contains negative values
			print ('# Static pressure input is relative to atmosphere. A conversion to absolute static pressure is made') 
			if InputFormat!=8:                                                                          # Valid for all cases except OpenFOAM data
				Ps = Ps+Ps0
			else:
				if TypeOfParameter == 1:
					print ('# OpenFOAM: Upstream reference static pressure used for conversion')                 
					Ps = Ps+Ps0                                                 
				if TypeOfParameter == 0: 
					print ('# OpenFOAM: Upstream reference total pressure used for conversion')                 
					Ps = Ps+Pt0 
elif PRESSURE==1:                                                                                       # Ps is defined as constant
	Ps = Ps0*np.ones(shape(Velocity_Vector[:,0]))
elif PRESSURE==2:                                                                                       # Ps from PIV (Vorticity integration method)
	Ps = Ps0*(1+0.5*(Gamma-1)*M0**2*(1-(V_star/V0)**2))**(Gamma/(Gamma-1))



# Static temperature:
Ts_constant = Ts0*(V/V)                                                                                 # (V/V) gives an array of ones
Ts_isoenergetic = Tt0 - (V**2/(2*CpGas))
Ts_isentropic = Ts0*(1+0.5*(Gamma-1)*M0**2*(1-(V/V0)**2))

if TEMPERATURE==0:                                                                                      # Ts is defined as the input value
	if Ts=="None":
		print ('Input data do not include Ts. Choose other "Temperature field" option in the GUI')
	else:
		Ts = Ts
elif TEMPERATURE==1:                                                                                    # Ts is computed from an isoenergetic flow
	Ts = Ts_isoenergetic
elif TEMPERATURE==2:                                                                                    # Ts is computed from an isentropic flow
	Ts = Ts_isentropic
elif TEMPERATURE==3:                                                                                    # Ts is defined as constant (Case 4 is defined below Density section)
	Ts = Ts_constant                                              

# Density:
Rho_isoenergetic = Ps/(R*Ts_isoenergetic)
Rho_isentropic = Rho0*(1+0.5*(Gamma-1)*M0**2*(1-(V/V0)**2))**(1/(Gamma-1))
Rho_constant = Rho0*(V/V)

if DENSITY==0:                                                                                          # Rho is defined as the input value
	if Rho=="None":
		print ('Input data do not include Rho. Choose other "Density field" option in the GUI')
	else:
		Rho = Rho
elif DENSITY==1:                                                                                        # Rho is computed from an isoenergetic flow
	Rho = Rho_isoenergetic
elif DENSITY==2:                                                                                        # Rho is computed from an isentropic flow
	Rho = Rho_isentropic
elif DENSITY==3:                                                                                        # Rho is defined as constant
	Rho = Rho_constant
elif DENSITY==4:                                                                                        # Rho is computed by using ideal gas law (from Ps and Ts)
	if Ts=="None":
		print('ERROR: "Temperature field" must be different from ideal gas law')
	else:
		Rho_gas_law = Ps/(R*Ts)
		Rho = Rho_gas_law


# Remaining temperature case:
if TEMPERATURE==4:                                                                                      # Ts is computed by using ideal gas law (from Ps and Rho)
	Ts = Ps/(R*Rho)


# Computing viscosities:
if MU_LAM=="None":
	MU_LAM = (0.000001458*Ts**1.5)/(Ts+110.4)   							                            # Two-coefficient Sutherland equation
	print ('# Laminar viscosity array is not available at the input data. It is computed from the Ts field (Sutherland equation)')

if MU_TURB=="None":
	MU_TURB = 0*MU_LAM
	print ('# Turbulent viscosity array is not available at the input data. It is assumed to be zero')
	

MU_EFF = MU_LAM + MU_TURB                                                                               # Effective viscosity [Kg/m*s]
Keff = CpGas*((MU_LAM/0.7)+(MU_TURB/Pr_turb))                                                           # Effective dissipation coefficient



if VORTICITY=="None":
	VORTICITY = algs.vorticity(Velocity_Vector)
	WallVorticityFactor = 2                                                                             # Info for Epsilon_Near_Field (indicates if vorticity has been loaded from input data or computed internally at ParaView)
	print ('# Vorticity array is not available at the input data. It is computed from the velocity field')
else: 
	WallVorticityFactor = 1

############################################################################################################
########################################### COMPLETING AERO DATA ###########################################
############################################################################################################

# Calculating the remaining parameters	
Hs = CpGas*Ts															                                # Static enthalpy [J/Kg]
eint = CvGas*Ts                                                                                         # Internal energy [J/Kg]
a = algs.sqrt(Gamma*R*Ts)                                                                               # Speed of sound [m/s]
M = V/a                                                                                                 # Mach number
q = 0.5*Rho*V**2                                                                                        # Dynamic pressure [Pa]
Cp = (Ps-Ps0)/q0                                                                                        # Pressure coeffcient
Pt = Ps*(1+0.5*(Gamma-1)*M**2)**(Gamma/(Gamma-1))                                                       # Total pressure [Pa]
Tt = Ts*(1+0.5*(Gamma-1)*M**2)                                                                          # Total temperature [K]
Ht = Hs+0.5*V**2                                                                                        # Total enthalpy [J/Kg]
DeltaHt = Ht-Ht0                                                                                        # Total enthalpy variation respect to upstream value [J/Kg]
DeltaS = CpGas*algs.log(Ts/Ts0)-R*algs.log(Ps/Ps0)                                                      # Specific entropy variation referred to upstream value [J/Kg*K]
PtPt0 = Pt/Pt0  														                                # Total pressure ratio
DeltaPt = (Pt-Pt0)/Pt0


if PRESSURE==1:                                                                                         # Ps is defined as constant
	Ps_gradient = 0*Velocity_Vector
if PRESSURE!=1: 
	Ps_gradient = gradient(Ps)
    
    
    
Ts_gradient = gradient(Ts)

V_gradient = gradient(Velocity_Vector)  							                                    # Velocity gradient tensor in the mesh reference frame[1/s]
VORTICITYmag = algs.mag(VORTICITY)
Helicity = algs.dot(Velocity_Vector,VORTICITY)/algs.max(algs.abs(algs.dot(Velocity_Vector,VORTICITY)))  # Normalized helicity
STRAIN = strain(Velocity_Vector)											                            # Strain tensor [1/s] in the mesh reference frame
#Strain_Magnitude = np.sqrt(np.einsum('ijk,ijk->i',STRAIN,STRAIN))                                      # Magnitude of a tensor (L2_1 norm), kept here for debugging purposes
Shock_Detector = algs.dot(Velocity_Vector,Ps_gradient)/(a*algs.mag(Ps_gradient)) 					    # Shockwave detector parameter (Lovely and Haimes criterion)
Q_criterion = 0.5*(algs.mag(VORTICITY)**2-np.sqrt(np.einsum('ijk,ijk->i',STRAIN,STRAIN))**2)            # Vortex detector by Q-criterion. it uses the L2_1 norm of the strain tensor

Lambda_2=STRAIN**2+(V_gradient - STRAIN)**2
where_are_NaNs = isnan(Lambda_2)    
Lambda_2[where_are_NaNs] = 0
Lambda_2 = np.linalg.eigvals(Lambda_2[:])
Lambda_2 = np.real(Lambda_2)

Phi_eff = MU_EFF*(2*V_gradient[:,0,0]**2 + 2*V_gradient[:,1,1]**2 + 2*V_gradient[:,2,2]**2 + (V_gradient[:,0,1]+V_gradient[:,1,0])**2 + (V_gradient[:,0,2]+V_gradient[:,2,0])**2 + (V_gradient[:,2,1]+V_gradient[:,1,2])**2 - (2.0/3.0)*(V_gradient[:,0,0]+V_gradient[:,1,1]+V_gradient[:,2,2])**2)  # Effective dissipation         
Phi_eff_Vort = MU_EFF*VORTICITYmag**2                                                                   # Vorticity-based expression of Phi_eff



############################################################################################################
########################################### REPLACING NaN VALUES ###########################################
############################################################################################################

if InputFormat==2:                                                      # InputFormat=2 --> EpsilonPIV

	where_are_NaNs = isnan(V)    
	V[where_are_NaNs] = V0
	
	where_are_NaNs = isnan(Cp)    
	Cp[where_are_NaNs] = 0
	
	where_are_NaNs = isnan(Ps) 
	Ps[where_are_NaNs] = Ps0
	
	where_are_NaNs = isnan(Pt) 
	Pt[where_are_NaNs] = Pt0
	
	where_are_NaNs = isnan(PtPt0) 
	PtPt0[where_are_NaNs] = 1
	
	where_are_NaNs = isnan(Rho) 
	Rho[where_are_NaNs] = Rho0
	
	where_are_NaNs = isnan(Shock_Detector) 
	Shock_Detector[where_are_NaNs] = 0
	
	where_are_NaNs = isnan(q) 
	q[where_are_NaNs] = q0
	
	where_are_NaNs = isnan(DeltaS) 
	DeltaS[where_are_NaNs] = q0



############################################################################################################
########################################### ROTATING VECTOR DATA ###########################################
############################################################################################################

if UpwardAxis==0:                                                       # Mesh Y-axis pointing upwards
	AngleX = -90
	AngleY = Beta
	AngleZ = Alfa  	
elif UpwardAxis==1:                                                     # Mesh Z-axis pointing upwards
	AngleX = 0      
	AngleY = -Alfa
	AngleZ = Beta

Rx = np.array([(1,0,0),(0,algs.cos(AngleX*np.pi/180),-algs.sin(AngleX*np.pi/180)),(0,algs.sin(AngleX*np.pi/180),algs.cos(AngleX*np.pi/180))])   # Rotation matrix about X-axis
Ry = np.array([(algs.cos(AngleY*np.pi/180),0,algs.sin(AngleY*np.pi/180)),(0,1,0),(-algs.sin(AngleY*np.pi/180),0,algs.cos(AngleY*np.pi/180))])  	# Rotation matrix about Y-axis
Rz = np.array([(algs.cos(AngleZ*np.pi/180),-algs.sin(AngleZ*np.pi/180),0),(algs.sin(AngleZ*np.pi/180),algs.cos(AngleZ*np.pi/180),0),(0,0,1)])  	# Rotation matrix about Z-axis

Rx_T = np.transpose(Rx)
Ry_T = np.transpose(Ry)
Rz_T = np.transpose(Rz)

MX = np.dot(Rx_T,np.dot(Rx,Rx_T))                                       # Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh X-axis
MY = np.dot(Ry_T,np.dot(Ry,Ry_T))                                       # Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Y-axis
MZ = np.dot(Rz_T,np.dot(Rz,Rz_T))                                       # Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Z-axis

TRANSFORMATION = np.dot(MX,np.dot(MY,MZ))                               # Total (combined) transformation matrix

# Some drag formulations require the actual coordinate of the domain points expressed in the aerodynamic reference system:
points = np.einsum('ij,kj->ik',POINTS,TRANSFORMATION)                   # Coordinates in the aerodynamic reference syste. "einsum" is a Pythonic fashion to do a dot product: np.dot(TRANSFORMATION,POINTS)
x = points[:,0]                                                         # x-coordinate in the aerodynamic reference frame
y = points[:,1]                                                         # y-coordinate in the aerodynamic reference frame
z = points[:,2]                                                         # z-coordinate in the aerodynamic reference frame

velocity = np.einsum('ij,kj->ik',Velocity_Vector,TRANSFORMATION) 		# Velocity vector in the aerodynamic reference frame
u = velocity[:,0]
v = velocity[:,1]
w = velocity[:,2]
udelta = u-V0                                                    	    # Axial velocity variation respects to the upstream value
RhoU = Rho*u															# Local mass flow rate	

vorticity = np.einsum('ij,kj->ik',VORTICITY,TRANSFORMATION)

TRANS_strain = np.einsum('ijm,kj->ikm',STRAIN,TRANSFORMATION)  	        # Matrix product: M.STRAIN   (M=transformation matrix)
TRANS_T = np.transpose(TRANSFORMATION)							        # M transpose
strain = np.einsum('km,ijk->ijm',TRANS_T,TRANS_strain)     		        # Strain tensor transformation: M.STRAIN.Mtranspose
 
tau = MU_EFF*strain

if PRESSURE==1:                                                         # Ps is defined as constant
	Cp_gradient = 0*Velocity_Vector
if PRESSURE!=1: 
	Cp_gradient = gradient(Cp)   									    # Cp gradient in the mesh reference frame


Cp_gradient = np.einsum('ij,kj->ik',Cp_gradient,TRANSFORMATION)[:,2]	# Cp gradient in the aerodynamic reference frame

Lamb_Vector = algs.cross(velocity,vorticity)                            



############################################################################################################
########################################## COMMON TO ALL METHODS ###########################################
############################################################################################################

# We define first if the far field values will be adimentionalized or not:
if Nondimensionalize==1:
	AdimF = 0.5*Rho0*Sref*V0**2                                         # Force adimensionalization factor 
	AdimP = 0.5*Rho0*Sref*V0**3                                         # Power adimensionalization factor 
	print ('# Exergy/lift/drag output values: nondimensionalized')
else:
	AdimF = 1
	AdimP = 1
	print ('# Exergy/lift/drag output values: absolute')


############################################################################################################
########################### DEFINING THE AERODYNAMIC ANALYSIS METHODS TO COMPUTE ###########################
############################################################################################################

if AnalysisMethods == 0:                                                # Compute all analysis methods
	NEAR_FIELD = 1
	Momentum = 1
	Betz = 1
	Jones = 1
	Oswatitsch = 1
	Maskell = 1
	Onorato = 1
	VDV = 1
	Giles = 1
	Kusunose = 1
	Meheut = 1
	Toubin = 1
	Wu = 1
	Mele = 1
	Arntz = 1
	Aguirre = 1
	Ruscio = 1
	
if AnalysisMethods == 1:                                                # Do not compute any analysis method
	NEAR_FIELD = 0
	Momentum = 0
	Betz = 0
	Jones = 0
	Oswatitsch = 0
	Maskell = 0
	Onorato = 0
	VDV = 0
	Giles = 0
	Kusunose = 0
	Meheut = 0
	Toubin = 0
	Wu = 0
	Mele = 0
	Arntz = 0
	Aguirre = 0
	Ruscio = 0

if AnalysisMethods == 2:                                                # Select the analysis methods to compute
	NEAR_FIELD = NEAR_FIELD                                             # Retains value defined in the GUI by the boolean check
	if FAR_FIELD == 0:                                                  # Compute all the far-field methods
		Momentum = 1
		Betz = 1
		Jones = 1
		Oswatitsch = 1
		Maskell = 1
		Onorato = 1
		VDV = 1
		Giles = 1
		Kusunose = 1
		Meheut = 1
		Toubin = 1
	if FAR_FIELD == 1:                                                  # Do not compute any far-field method
		Momentum = 0
		Betz = 0
		Jones = 0
		Oswatitsch = 0
		Maskell = 0
		Onorato = 0
		VDV = 0
		Giles = 0
		Kusunose = 0
		Meheut = 0
		Toubin = 0
	if FAR_FIELD == 2:                                                  # Select the far-field method to compute
		Momentum = Momentum                                             # Retains value defined in the GUI by the boolean check
		Betz = Betz
		Jones = Jones
		Oswatitsch = Oswatitsch
		Maskell = Maskell
		Onorato = Onorato
		VDV = VDV
		Giles = Giles
		Kusunose = Kusunose
		Meheut = Meheut
		Toubin = Toubin
	if LAMB_VECTOR == 0:                                                # Compute all the Lamb vector methods
		Wu = 1
		Mele = 1
	if LAMB_VECTOR == 1:                                                # Do not compute any Lamb vector method
		Wu = 0
		Mele = 0    
	if LAMB_VECTOR == 2:                                                # Select the Lamb vector method to compute
		Wu = Wu                                                         # Retains value defined in the GUI by the boolean check
		Mele = Mele
	if EXERGY == 0:                                                     # Compute all the exergy methods
		Arntz = 1
		Aguirre = 1
		Ruscio = 1
	if EXERGY == 1:                                                     # Do not compute any exergy method
		Arntz = 0
		Aguirre = 0
		Ruscio = 0    
	if EXERGY == 2:                                                     # Do not compute any exergy method
		Arntz = Arntz                                                   # Retains value defined in the GUI by the boolean check
		Aguirre = Aguirre
		Ruscio = Ruscio



############################################################################################################
############################################ NEAR-FIELD METHODS ############################################
############################################################################################################

if NEAR_FIELD == 1:                                                         
	TRANS_V_gradient = np.einsum('ijm,kj->ikm',V_gradient,TRANSFORMATION)  	# Matrix product: M.Vgradient   (M=transformation matrix)
	TRANS_T = np.transpose(TRANSFORMATION)									# M transpose
	V_gradient = np.einsum('km,ijk->ijm',TRANS_T,TRANS_V_gradient)     		# Vgradient tensor transformation: M.Vgradient.Mtranspose
	DimF = 0.5*Rho0*Sref*V0**2												# "Cd-to-force" conversion factor
	NBpts = Ps.size                                                         # Number of points of the input data.
	PointID = np.arange(NBpts)                                              # Point ID used in Epsilon_Near_Field filter



############################################################################################################
############################################# FAR-FIELD METHODS ############################################
############################################################################################################

########### MOMENTUM CONSERVATION METHOD ###########
if Momentum == 1:    
	MOMENTUM_total_drag = (-1*Rho*u*(u-V0)-(Ps-Ps0))/AdimF                                      # Total drag density distribution 
	MOMENTUM_u = (-1*Rho*u*(u-V0))/AdimF                                                        # Speed-component of the total drag density
	MOMENTUM_p = (-1*(Ps-Ps0))/AdimF                                                            # Pressure-component of the total drag density
	MOMENTUM_lift = (-1*Rho*u*w)/AdimF                                                          # Lift density distribution
	MOMENTUM_sideforce = (-1*Rho*u*v)/AdimF                                                     # Sideforce density distribution
	MOMENTUM_stress = 2*MU_LAM*strain[:,0,0]


########### BETZ ###########
if Betz == 1: 
	BETZ_Vartificial = algs.sqrt(u**2+2*(Pt0-Pt)*(1/Rho))                                       # Artificial velocity
	BETZ_profile_drag = (Pt0-Pt+0.5*Rho*(BETZ_Vartificial-u)*(BETZ_Vartificial+u-2*V0))/AdimF   # Profile drag density distribution
	BETZ_p = (Pt0-Pt)/AdimF                                                                     # Pressure-component of the profile drag density
	BETZ_u = (0.5*Rho*(BETZ_Vartificial-u)*(BETZ_Vartificial+u-2*V0))/AdimF                     # Velocity-component of the profile drag density
		
		
########### JONES ###########
if Jones == 1: 
	JONES_profile_drag = (2*algs.sqrt(Pt-Ps)*(algs.sqrt(Pt0-Ps0)-algs.sqrt(Pt-Ps0)) )/AdimF     # Profile drag density distribution
		
		
########### OSWATITSCH ########### 
if Oswatitsch == 1: 
	OSWATITSCH_profile_drag = (Ts0/V0)*(Rho*u*DeltaS)/AdimF                                     # profile drag density distribution

		
########### MASKELL ########### 
if Maskell == 1: 
	"""
	ub = np.trapz(BE_Vart-u, y)/(2*AreaTunel)
	MASKELL_profile_drag = (Pt0-Pt+0.5*Rho*(BETZ_Vartificial-u)*(BETZ_Vartificial+u-2*(V0+ub)))/AdimF              # integrand of Betz drag equation
	IMPLEMENT WITH EXTERNAL FILTER  §§§§§§§
	"""
	MASKELL_vortex_drag = 0.5*Rho*(v**2+w**2)/AdimF                                             # Vortex drag density
	MASKELL_lift_zetay = Rho0*V0*vorticity[:,0]*y/AdimF                                         # Local lift density	
	MASKELL_lift = (Rho0*V0*vorticity[:,0]*y + Rho0*(V0-u)*w)/AdimF                             # Local lift density	
	MASKELL_sideforce = (Rho0*V0*vorticity[:,0]*z + Rho0*(V0-u)*v)/AdimF                        # Local sideforce density
	MASKELL_lift_zeta = Rho0*V0*vorticity[:,0]/AdimF                                            # Local vorticity flux	


########### ONORATO METHOD ###########
if Onorato == 1: 
	ONORATO_Pt = (Pt0-Pt)/AdimF  
	ONORATO_u = 0.5*Rho*(u-V0)**2/AdimF  	
	ONORATO_vw = 0.5*Rho*(v**2+w**2)/AdimF	
	ONORATO_drag = ONORATO_Pt + ONORATO_u + ONORATO_vw


########### VAN DEN VOOREN ########### 
if VDV == 1: 
	Root = 1-((np.e**((Gamma-1)*DeltaS/(Gamma*R))-1)/(0.5*(Gamma-1)*M0**2))+(2*(Ht-Ht0)/V0**2)  # Term inside the root of DeltaU
	VDV_DeltaUbar = V0*(algs.sqrt(Root)-1)                                                      # velocity deficit related to viscous effect (it ignores the potential effect)
	VDV_DeltaUstar = udelta-VDV_DeltaUbar                                                       # velocity defect related to potential effects
	VDV_vortex_drag = -1*(Rho*u*(u-V0-VDV_DeltaUbar)+Ps-Ps0)/AdimF                              # Vortex drag density
	VDV_profile_drag = -1*(Rho*u*VDV_DeltaUbar)/AdimF                                           # Profile drag density
	VDV_total_drag = VDV_profile_drag + VDV_vortex_drag                                         # Total drag density


########### GILES ###########
if Giles == 1: 
	GILES_profile_drag = ((Ps0*DeltaS/R)-(Rho0*DeltaHt))/AdimF                                  # Profile drag density


########### KUSUNOSE ###########
if Kusunose == 1: 
	KUSUNOSE_vw = (0.5*Rho*(v**2+w**2))/AdimF                                                   # Traverse flow drag density (It does not belong to the classical momentum drag) 
	KUSUNOSE_u2 = (-0.5*Rho*(u-V0)**2)/AdimF                                                    # Coupled drag density (It does not belong to the classical momentum drag) 
	KUSUNOSE_vw_u2 = KUSUNOSE_vw+KUSUNOSE_u2                                                    # Traverse+coupled drag density

	KUSUNOSE_profile_drag = ((Ps0*DeltaS/R)-(Rho0*DeltaHt)-0.5*Ps0*(DeltaS/R)**2+Rho0*DeltaHt*((DeltaS/R)-0.5*DeltaHt*(M0/V0)**2))/AdimF      # Profile drag density
	KUSUNOSE_vortex_drag = 0.5*Rho0*((v**2+w**2)-(1-M0**2)*(u-V0)**2)/AdimF                                                                   # Vortex drag density
	KUSUNOSE_lift = ( Rho0*V0*vorticity[:,0]*y - Rho0*V0**2*(1-M0**2)*(w/V0**2)*(u-V0) + M0**2 *(Gamma*Ps0*w*DeltaS/(R*V0)) - Rho0*M0**2*DeltaHt*(w/V0) )/AdimF      # Local lift density                                                  # Local lift density

	"""
	# This is the formulation provided by Meheut in his PhD report, left here for debug purposes:
	KUSUNOSE_profile_drag =(2/Sref)*(((DeltaS)/(R*Gamma*M0**2)) - ((Ht-Ht0)/V0**2) - ((DeltaS**2)/(2*R**2*Gamma*M0**2)) + ((Ht-Ht0)/V0**2)*((DeltaS/R)-0.5*M0**2*((Ht-Ht0)/V0**2)))
	KUSUNOSE_profile_drag = KUSUNOSE_profile_drag*0.5*Rho0*Sref*V0**2/AdimF                     # Profile drag density
	KUSUNOSE_vortex_drag = (0.5*Rho0*(v**2+w**2)) - 0.5*Rho0*(1-M0**2)*(VDV_DeltaUbar)**2
	KUSUNOSE_vortex_drag = KUSU_vortex_drag/AdimF                                               # Vortex drag density
	"""
		
	KUSUNOSE_total_drag = KUSUNOSE_profile_drag + KUSUNOSE_vortex_drag                          # Total drag density (note that its values are alredy affected by the threshold)

			
########### MEHEUT ###########
if Meheut == 1: 
	Raiz = 1- (((Ps/Ps0)**((Gamma-1)/Gamma)-1)/(0.5*(Gamma-1)*M0**2)) - ((v**2+w**2)/V0**2)     # Term inside the root of DeltaUstar
	DeltaUstar = algs.sqrt(Raiz)-1                                                              # Deficit of velocity related to the potential effect
	DeltaUonera = (u/V0)-1                                                                      # Total deficit of velocity 
	DeltaUline = DeltaUonera-DeltaUstar                                                         # Deficit of velocity related to the viscous effects
	DeltaPt = (Pt/Pt0)-1                                                                        # Total pressure drop ratio
	DeltaTt = (Tt/Tt0)-1                                                                        # Total temperature drop ratio
	MEHEUT_profile_drag = (1/Sref)* ((-2*DeltaPt/(Gamma*M0**2)) - DeltaTt - (1-0.25*M0**2)*DeltaTt**2 - DeltaPt*DeltaTt - (1-M0**2)*(DeltaUline**2+2*DeltaUline*DeltaUstar))
	MEHEUT_profile_drag = MEHEUT_profile_drag*0.5*Rho0*Sref*V0**2/AdimF                         # Profile drag density


  
########### TOUBIN  ###########                                                                 # Simple version: no shock, no movement
if Toubin == 1: 
	tau_x = tau[:,0]                                                                            # x component of viscous stress tensor, a vector
	Root = 1-((np.e**((Gamma-1)*DeltaS/(Gamma*R))-1)/(0.5*(Gamma-1)*M0**2))+(2*(Ht-Ht0)/V0**2)  # Term inside the root of DeltaU.
	VDV_DeltaUbar = V0*(algs.sqrt(Root)-1)                                                      # velocity deficit related to viscous effect (it ignores the potential effect). 
	TOUBIN_Dv_Sv_ste = (-1*(Rho*u*VDV_DeltaUbar)+tau_x[:,0])/AdimF                              # Viscous drag density surface term
	TOUBIN_Dv_Vv_tem = -1*Rho*(u-V0)/Timestep/AdimF                                             # Viscous drag density volume temporal term (tem)
	TOUBIN_Di_Se_ste = -1*(Rho*u*(u-V0-VDV_DeltaUbar)+Ps-Ps0)/AdimF                             # Induced drag density surface term
	TOUBIN_Di_Vc_ste = (-1*Rho*(u-V0-VDV_DeltaUbar)*(1/a-1/(VDV_DeltaUbar+V0)))/AdimF           # Viscous drag density volume non-temporal factor (ste)
	TOUBIN_Di_Vc_tem = u/Timestep                                                               # Induced drag density volume temporal factor (tem)
	TOUBIN_Dpa_Vc_ste = Rho*(VDV_DeltaUbar+V0-u)/a/AdimF                                        # Pressure-acoustic drag density volume non-temporal factor (ste)
	TOUBIN_Dpa_Vc_tem = (u+(2*a/(Gamma-1)))/Timestep                                            # Pressure-acoustic drag density volume temporal factor (tem)
	
	

############################################################################################################
############################################ LAMB VECTOR METHODS ###########################################
############################################################################################################

Plane_normal = np.array([[1,0,0]])   								                            # Plane normal in the aerodynamic reference frame
n = np.repeat(Plane_normal,algs.shape(Lamb_Vector)[0],axis=0)
r = points/(Dimension-1)

########### WU ###########
if Wu == 1:    
	WU_lift = Rho*Lamb_Vector[:,2]/AdimF 
	WU_induced_drag = Rho*Lamb_Vector[:,0]/AdimF
	WU_profile_drag = Rho*algs.cross(r,algs.cross(n,Lamb_Vector))/AdimF


########### MELE ###########
if Mele == 1: 
	RHO_GRADIENT = gradient(Rho)
	V2_GRADIENT = gradient(0.5*V**2)
	rho_gradient = np.einsum('ij,kj->ik',RHO_GRADIENT,TRANSFORMATION)
	v2_gradient = np.einsum('ij,kj->ik',V2_GRADIENT,TRANSFORMATION)
	mrho = algs.cross(r,algs.cross(rho_gradient,v2_gradient))
	
	MELE_lift = (Rho*Lamb_Vector[:,2] - mrho[:,2])/AdimF 
	MELE_induced_drag = ( Rho*Lamb_Vector[:,0] - mrho[:,0])/AdimF
	MELE_profile_drag = Rho*algs.cross(r,algs.cross(n,Lamb_Vector))/AdimF
	

	
############################################################################################################
############################################## EXERGY METHODS ##############################################
############################################################################################################


########### ARNTZ ###########                                                                           # Adiabatic version coded here ! (No heat transfer allowed)
if Arntz == 1 or Aguirre == 1:    
	ARNTZ_Eth_t = ((Rho*CvGas*Ts*u*(1-Ts0*algs.log(Ts/Ts0)/Ts))-(Rho*CvGas*Ts0*u))/AdimP       			# Thermal exergy T-component (constant volume) - exact formulation -
	ARNTZ_Eth_p = ((Ps0*(1-(Rho/Rho0)*algs.log(Rho0/Rho))*u)-(Rho*R*Ts0*u))/AdimP             			# Thermal exergy P-component (variable volume) - exact formulation -
	#ARNTZ_Eth_t = (0.5*Rho*CvGas*(Ts-Ts0)**2*u/Ts0)/AdimP                                              # Thermal exergy T-component (constant volume) - approximative formulation -
	#ARNTZ_Eth_p = ((Rho*R*Ts0*((Rho0/Rho)+algs.log(Rho/Rho0))*u)-(Rho*R*Ts0*u))/AdimP                 	# Thermal exergy P-component (variable volume) - alternative formulation -
	ARNTZ_A = Ts0*Rho*DeltaS*u/AdimP                							    					# Exact value of the anergy created 	
	ARNTZ_Aphi = (Ts0/Ts)*Phi_eff/AdimP                                             					# Viscous anergy
	ARNTZ_At = (Ts0*Keff*algs.mag(Ts_gradient)**2)/(AdimP*Ts**2)                              			# Thermal anergy
	ARNTZ_Ep = ((Ps-Ps0)*(u-V0))/AdimP                                              					# Pressure exergy
	ARNTZ_Eu = (0.5*Rho*(u-V0)**2*u)/AdimP                                          					# Axial speed exergy
	ARNTZ_Ev = (0.5*Rho*(v**2+w**2)*u)/AdimP                                        					# Transverse speed exergy
	ARNTZ_Em = ARNTZ_Ep + ARNTZ_Eu + ARNTZ_Ev                                     						# Total mechanical exergy 
	ARNTZ_Eth = ARNTZ_Eth_t + ARNTZ_Eth_p                       										# Thermal exergy 
	ARNTZ_Cd = ARNTZ_A + ARNTZ_Em + ARNTZ_Eth  															# Exergy based drag coefficient
	
	ARNTZ_Aw_div = algs.divergence(Ts0*Rho*DeltaS*Velocity_Vector)/AdimP                                # Wave anergy (divergence formulation)
	MASK = Shock_Detector > 0.9  																		# This detects the shockwave regions
	ARNTZ_Aw_div_MASK_Shock = 1*MASK*ARNTZ_Aw_div                                                       # This set to zero the entire ARNTZ_Aw_div field, except at the shockwave
	MASK = Q_criterion < 0  																		    # This detects the viscous regions	
	ARNTZ_Aw_div_MASK_Qcrit = 1*MASK*ARNTZ_Aw_div                                                       # This set to zero the entire ARNTZ_Aw_div field, except at the viscous region
 
	ARNTZ_Hprop = Rho*DeltaHt*u/AdimP                                                                   # Propulsive enthalpy
	ARNTZ_WGamma = ARNTZ_Hprop - (ARNTZ_Em + ARNTZ_Eth + ARNTZ_A)                                       # "W*Gamma" term


########### AGUIRRE ###########
if Aguirre == 1: 

	# Wave anergy with masking method:
	MASK = Q_criterion < 0.0  																            # This detects the viscous regions
	AGUIRRE_Aw_div_MASKED = 1*MASK*algs.divergence(Ts0*Rho*DeltaS*Velocity_Vector)/AdimP		        # This mask the viscous regions of the wave anergy parameter

	# Velocity decomposition method:
	ETA = 1+0.5*(Gamma-1)*M0**2*(1-((u**2+v**2+w**2)*Tt0/(V0**2*Tt)))
	RootVabsstar = 1- ((ETA*(Pt/Pt0)**((Gamma-1)/Gamma)-1)/(0.5*(Gamma-1)*M0**2))                       # Term inside the root of Vabsstar
	AGUIRRE_Vabs_star = V0*algs.sqrt(RootVabsstar)                                                      # Isentropic velocity magnitude
	
	FlowTita = algs.arctan(w/algs.sqrt(u**2+v**2))                                                      # Directional angles of the real velocity vector
	FlowAlfa = algs.arctan(w/u)
	FlowBeta = algs.arctan(v/u)
	
	if InputFormat==2:                                                                                  # InputFormat=2 --> EpsilonPIV (NaN values must be eliminated)
		where_are_NaNs = isnan(AGUIRRE_Vabs_star)    
		AGUIRRE_Vabs_star[where_are_NaNs] = V0
		
		where_are_NaNs = isnan(FlowTita)    
		FlowTita[where_are_NaNs] = 0
	
		where_are_NaNs = isnan(FlowAlfa)    
		FlowAlfa[where_are_NaNs] = 0
		
		where_are_NaNs = isnan(FlowBeta)    
		FlowBeta[where_are_NaNs] = 0

	AGUIRRE_u_star = AGUIRRE_Vabs_star*algs.cos(FlowTita)*algs.cos(FlowBeta)                            # x-component of the isentropic velocity
	AGUIRRE_v_star = AGUIRRE_Vabs_star*algs.cos(FlowTita)*algs.sin(FlowBeta)                            # y-component of the isentropic velocity
	AGUIRRE_w_star = AGUIRRE_Vabs_star*algs.sin(FlowTita)                                               # z-component of the isentropic velocity

	AGUIRRE_u_bar = u-AGUIRRE_u_star                                                                    # x-component of the non-isentropic velocity
	AGUIRRE_v_bar = v-AGUIRRE_v_star                                                                    # y-component of the non-isentropic velocity
	AGUIRRE_w_bar = w-AGUIRRE_w_star                                                                    # z-component of the non-isentropic velocity

	# Decomposition of AERO parameters:
	AGUIRRE_Ts_star = Ts0*(1+0.5*(Gamma-1)*M0**2*(1-(AGUIRRE_Vabs_star/V0)**2))                         # Isentropic static temperature
	AGUIRRE_Ts_bar = Ts-AGUIRRE_Ts_star                                                                 # Non-isentropic static temperature
	AGUIRRE_Rho_star = Rho0*(1+0.5*(Gamma-1)*M0**2*(1-(AGUIRRE_Vabs_star/V0)**2))**(1/(Gamma-1))        # Isentropic density
	AGUIRRE_Rho_bar = Rho-AGUIRRE_Rho_star                                                              # Non-isentropic density
	AGUIRRE_e_star = CvGas*AGUIRRE_Ts_star                                                              # Isentropic internal energy
	AGUIRRE_e_bar = eint-AGUIRRE_e_star                                                                 # Non-isentropic internal energy
	AGUIRRE_Hs_star = CpGas*AGUIRRE_Ts_star                                                             # Isentropic static enthalpy
	AGUIRRE_Hs_bar = Hs-AGUIRRE_Hs_star                                                                 # Non-isentropic static enthalpy
	AGUIRRE_a_star = np.sqrt(Gamma*R*AGUIRRE_Ts_star)                                                   # Isentropic speed of sound
	AGUIRRE_M_star = AGUIRRE_Vabs_star/AGUIRRE_a_star                                                   # Isentropic Mach number                                       

	"""
	# OLD EXERGY FORMULATION: this uses the definition of Em, Ep and Eth proposed by Arntz (it is kept here for testing). For more info see paper AEC2020 "A BETTER ASSESSMENT OF THE RECOVERABLE ENERGY BEHIND A BODY BY THE EXERGY METHOD"
	# Exergy decomposition
	AGUIRRE_Eu_star = (0.5*Rho*(AGUIRRE_u_star-V0)**2*AGUIRRE_u_star)/AdimP
	AGUIRRE_Eu_bar = ARNTZ_Eu - AGUIRRE_Eu_star
	AGUIRRE_Ev_star = (0.5*Rho*(AGUIRRE_v_star**2+AGUIRRE_w_star**2)*AGUIRRE_u_star)/AdimP
	AGUIRRE_Ev_bar = ARNTZ_Ev - AGUIRRE_Ev_star
	AGUIRRE_Ep_star = ((Ps-Ps0)*(AGUIRRE_u_star-V0))/AdimP 
	AGUIRRE_Ep_bar = ARNTZ_Ep - AGUIRRE_Ep_star
	AGUIRRE_Em_star = AGUIRRE_Eu_star + AGUIRRE_Ev_star + AGUIRRE_Ep_star
	AGUIRRE_Em_bar = AGUIRRE_Eu_bar + AGUIRRE_Ev_bar + AGUIRRE_Ep_bar 
	AGUIRRE_Eth_t_star = ((AGUIRRE_Rho_star*CvGas*AGUIRRE_Ts_star*AGUIRRE_u_star*(1-Ts0*algs.log(AGUIRRE_Ts_star/Ts0)/AGUIRRE_Ts_star))-(AGUIRRE_Rho_star*CvGas*Ts0*AGUIRRE_u_star))/AdimP       			
	AGUIRRE_Eth_p_star = ((Ps0*(1-(AGUIRRE_Rho_star/Rho0)*algs.log(Rho0/AGUIRRE_Rho_star))*AGUIRRE_u_star)-(AGUIRRE_Rho_star*R*Ts0*AGUIRRE_u_star))/AdimP             			
	AGUIRRE_Eth_star = AGUIRRE_Eth_t_star + AGUIRRE_Eth_p_star 
	AGUIRRE_Eth_bar = ARNTZ_Eth-AGUIRRE_Eth_star
	AGUIRRE_Cd_star = AGUIRRE_Em_star + AGUIRRE_Eth_star 
	AGUIRRE_Cd_bar = ARNTZ_A + AGUIRRE_Em_bar + AGUIRRE_Eth_bar 
	"""
	
	# NEW EXERGY FORMULATION: this uses the new definition of Em, Ep and Eth (see paper AEC2020 "A BETTER ASSESSMENT OF THE RECOVERABLE ENERGY BEHIND A BODY BY THE EXERGY METHOD")
	# (Unlike the others Aguirre equations, these particular formulations require an infinite surface integral)
	AGUIRRE_INF_Ep = ARNTZ_Ep + ARNTZ_Eth_p  								                            # Redefined Ep
	AGUIRRE_INF_Em = ARNTZ_Eu + ARNTZ_Ev + AGUIRRE_INF_Ep 				  	                            # Redefined Em
	AGUIRRE_INF_Eth = ARNTZ_Eth_t                     				   		                            # Redefined Eth
	AGUIRRE_INF_Cd = ARNTZ_A + AGUIRRE_INF_Em + AGUIRRE_INF_Eth  			                            # Exergy based drag coefficient
	
	# Exergy decomposition:                                                                             # "star"=isentropic component and "bar"=non-isentropic components
	AGUIRRE_Eth_t_star = ((AGUIRRE_Rho_star*CvGas*AGUIRRE_Ts_star*AGUIRRE_u_star*(1-Ts0*algs.log(AGUIRRE_Ts_star/Ts0)/AGUIRRE_Ts_star))-(AGUIRRE_Rho_star*CvGas*Ts0*AGUIRRE_u_star))/AdimP       			
	AGUIRRE_Eth_p_star = ((Ps0*(1-(AGUIRRE_Rho_star/Rho0)*algs.log(Rho0/AGUIRRE_Rho_star))*AGUIRRE_u_star)-(AGUIRRE_Rho_star*R*Ts0*AGUIRRE_u_star))/AdimP             			
	AGUIRRE_Eu_star = (0.5*AGUIRRE_Rho_star*(AGUIRRE_u_star-V0)**2*AGUIRRE_u_star)/AdimP
	AGUIRRE_Eu_bar = ARNTZ_Eu - AGUIRRE_Eu_star
	AGUIRRE_Ev_star = (0.5*AGUIRRE_Rho_star*(AGUIRRE_v_star**2+AGUIRRE_w_star**2)*AGUIRRE_u_star)/AdimP
	AGUIRRE_Ev_bar = ARNTZ_Ev - AGUIRRE_Ev_star
	AGUIRRE_Ep_star_original = ((Ps-Ps0)*(AGUIRRE_u_star-V0))/AdimP 
	AGUIRRE_Ep_star = AGUIRRE_Ep_star_original + AGUIRRE_Eth_p_star
	AGUIRRE_Ep_bar = AGUIRRE_INF_Ep - AGUIRRE_Ep_star
	AGUIRRE_Em_star = AGUIRRE_Eu_star + AGUIRRE_Ev_star + AGUIRRE_Ep_star
	AGUIRRE_Em_bar = AGUIRRE_Eu_bar + AGUIRRE_Ev_bar + AGUIRRE_Ep_bar 
	AGUIRRE_Eth_star = AGUIRRE_Eth_t_star  
	AGUIRRE_Eth_bar = AGUIRRE_INF_Eth - AGUIRRE_Eth_star
	AGUIRRE_Cd_star = AGUIRRE_Em_star + AGUIRRE_Eth_star 
	AGUIRRE_Cd_bar = ARNTZ_A + AGUIRRE_Em_bar + AGUIRRE_Eth_bar 
	AGUIRRE_WGamma_star =  - (AGUIRRE_Em_star + AGUIRRE_Eth_star )
	AGUIRRE_WGamma_bar = ARNTZ_Hprop - (AGUIRRE_Em_bar + AGUIRRE_Eth_bar + ARNTZ_A)
	
	#Correction method for close survey plane to body:                                                  #Expression valid ONLY in incompressible flow!!!
	AGUIRRE_Cd_bar_corrected = AGUIRRE_Cd_bar + algs.abs(AGUIRRE_Cd_star)
	AGUIRRE_Em_bar_corrected = AGUIRRE_Em_bar + algs.abs(AGUIRRE_Em_star)   
	
	# Vortex drag:
	AGUIRRE_vortex_drag = (0.5*Rho0*(AGUIRRE_v_star**2+AGUIRRE_w_star**2-(AGUIRRE_u_star-V0)**2)*V0)/AdimP
	
	# Other parameters:
	AGUIRRE_Aphi_Vort = (Ts0/Ts)*MU_EFF*VORTICITYmag**2/AdimP                                           # Vorticity-based viscous anergy


########### RUSCIO ############                                                                         # Version without shockwave
if Ruscio == 1: 
	RUSCIO_Etc_vol = (((Ps0/Rho0)*Rho)/AdimP)/Timestep
	RUSCIO_En_vol = ((Rho*(CvGas*(Ts-Ts0) + V0*(u-V0) +0.5*((u-V0)**2 + v**2+w**2)))/AdimP)/Timestep	# Total unsteady Energy
	RUSCIO_An_vol = -((Ts0*Rho*DeltaS)/AdimP)/Timestep
	RUSCIO_Etc_vol_div = -(Ps0/Rho0)*algs.divergence(Rho*Velocity_Vector)/AdimP
	RUSCIO_Mo_vol = -(V0*Rho*(u-V0)/AdimP)/Timestep						    
	RUSCIO_Ie_vol = ((Rho*(CvGas*(Ts-Ts0)))/AdimP)/Timestep				    
	RUSCIO_Eu_vol = ((Rho*0.5*((u-V0)**2))/AdimP)/Timestep					    
	RUSCIO_Ev_vol = ((Rho*0.5*(v**2+w**2))/AdimP)/Timestep					    



############################################################################################################
####################################### USER-DEFINED VARIABLES (UDV) #######################################
############################################################################################################

# Use the following code snippet to generate new UDV and send it to the output:

EpsilonIntegrationTool = []

if UDV == 1:

	# UDV_var = INPUT.PointData["Set_here_the_name_of_the_variable_to_load"]                            # uncomment this to read a specific array from the input CFD data that has not been already loaded by Epsilon_Aero_Data

	UDV_new_scalar = 0.5*Rho*V**2                                                                       # Define your desired scalar variable
	output.PointData.append(UDV_new_scalar, "UDV_new_scalar")                                           # Send this variable to the output
	EpsilonIntegrationTool.append(["UDV_new_scalar","Survey_Plane"])                                    # Information for EpsilonIntegrationTool: the user must declare if the variable has to be integrated at the survey plane ("Survey_Plane") or at the entire control volume ("Control_Volume") or if it should not be integrated ("Do_Not_Integrate")
	
	UDV_new_vector = algs.make_vector(u,v,w)                                                            # Define your desired vector variable
	output.PointData.append(UDV_new_vector, "UDV_new_vector")                                           # Send this variable to the output
	EpsilonIntegrationTool.append(["UDV_new_vector","Control_Volume"])                                  # Declare the integration type


############################################################################################################
###############################################  PRINT INFO ################################################
############################################################################################################

# This info is useful for the user in order to create the slice plane normal to the upstream direction

if UpwardAxis==0:    #Mesh Y-axis pointing upwards
	Plane_NORMAL_X = np.cos(Alfa*np.pi/180)*np.cos(Beta*np.pi/180)
	Plane_NORMAL_Y = np.sin(Alfa*np.pi/180)
	Plane_NORMAL_Z = -1*np.sin(Beta*np.pi/180)
	print ('# Survey plane normal = ',Plane_NORMAL_X, '  ',Plane_NORMAL_Y, '  ', Plane_NORMAL_Z)
	
elif UpwardAxis==1:    #Mesh Z-axis pointing upwards
	Plane_NORMAL_X = np.cos(Alfa*np.pi/180)*np.cos(Beta*np.pi/180)
	Plane_NORMAL_Y = np.sin(Beta*np.pi/180)
	Plane_NORMAL_Z = np.sin(Alfa*np.pi/180)
	print ('# Survey plane normal = ',Plane_NORMAL_X, '  ',Plane_NORMAL_Y, '  ', Plane_NORMAL_Z)

Plane_NORMAL = algs.make_vector(np.array([Plane_NORMAL_X]),np.array([Plane_NORMAL_Y]),np.array([Plane_NORMAL_Z]))[0]



############################################################################################################
###############################################  DATA OUTPUT ###############################################
############################################################################################################

# Constant values:
output.FieldData.append(R, "R")
output.FieldData.append(Gamma, "Gamma")
output.FieldData.append(CpGas, "CpGas")
output.FieldData.append(CvGas, "CvGas")
output.FieldData.append(Alfa, "Alfa")
output.FieldData.append(Beta, "Beta")
output.FieldData.append(M0, "M0")
output.FieldData.append(V0, "V0")
output.FieldData.append(Pt0, "Pt0")
output.FieldData.append(Ps0, "Ps0")
output.FieldData.append(Tt0, "Tt0")
output.FieldData.append(Ts0, "Ts0")
output.FieldData.append(Rho0, "Rho0")
output.FieldData.append(AdimP, "AdimP")
output.FieldData.append(AdimF, "AdimF")
output.FieldData.append(UpwardAxis, "UpwardAxis")
output.FieldData.append(TRANSFORMATION, "TRANSFORMATION")
output.FieldData.append(UDV, "UDV")
output.FieldData.append(Plane_NORMAL, "Plane_NORMAL")
output.FieldData.append(Timestep, "Timestep")
output.FieldData.append(Sref, "Sref")


if AERO == 0:  
	output.PointData.append(vorticity, "AERO_Vorticity")                                                #This variable is used later by Epsilon_Profile_Drag_Breakdown 

if AERO == 1:    
	output.PointData.append(MU_LAM, "AERO_Mu_Lam")
	output.PointData.append(MU_TURB, "AERO_Mu_Turb")
	output.PointData.append(MU_EFF, "AERO_Mu_Eff")
	output.PointData.append(Keff, "AERO_Keff")
	output.PointData.append(Velocity_Vector, "AERO_Velocity_Vector")
	output.PointData.append(u, "AERO_u")
	output.PointData.append(v, "AERO_v")
	output.PointData.append(w, "AERO_w")
	output.PointData.append(udelta, "AERO_udelta")
	output.PointData.append(V, "AERO_Vabs")
	output.PointData.append(M, "AERO_M")
	output.PointData.append(vorticity, "AERO_Vorticity")
	output.PointData.append(VORTICITYmag, "AERO_Vorticity_Mag")
	output.PointData.append(Helicity, "AERO_Helicity")
	output.PointData.append(tau, "AERO_Stress(tau)")
	output.PointData.append(strain, "AERO_Strain")
	output.PointData.append(Q_criterion, "AERO_Q_criterion")
	output.PointData.append(Lambda_2, "AERO_Lambda_2")
	output.PointData.append(Phi_eff, "AERO_Phi_eff")
	output.PointData.append(Phi_eff_Vort, "AERO_Phi_eff_Vort")
	output.PointData.append(Ps, "AERO_Ps")
	output.PointData.append(Pt, "AERO_Pt")
	output.PointData.append(PtPt0, "AERO_Pt/Pt0")
	output.PointData.append(DeltaPt*Pt0, "AERO_Pt-Pt0")
	output.PointData.append(Shock_Detector, "AERO_Shock_Detector")
	output.PointData.append(q, "AERO_q")
	output.PointData.append(Cp, "AERO_Cp")
	output.PointData.append(Cp_gradient, "AERO_Cp_gradient")
	output.PointData.append(Ts, "AERO_Ts")
	output.PointData.append(Ts_constant, "AERO_Ts_constant")
	output.PointData.append(Ts_isoenergetic, "AERO_Ts_isoenergetic")
	output.PointData.append(Ts_isentropic, "AERO_Ts_isentropic")
	output.PointData.append(Tt, "AERO_Tt")
	output.PointData.append(Hs, "AERO_Hs")
	output.PointData.append(Ht, "AERO_Ht")
	output.PointData.append(DeltaHt, "AERO_Ht-Ht0")
	output.PointData.append(Rho, "AERO_Rho")
	output.PointData.append(Rho_constant, "AERO_Rho_constant")
	output.PointData.append(Rho_isoenergetic, "AERO_Rho_isoenergetic")
	output.PointData.append(Rho_isentropic, "AERO_Rho_isentropic")
	output.PointData.append(RhoU, "AERO_RhoU")
	output.PointData.append(DeltaS, "AERO_s-s0")
	output.PointData.append(eint, "AERO_e")
	output.PointData.append(Lamb_Vector, "AERO_Lamb_Vector")


if NEAR_FIELD == 1:     
	output.PointData.append(V_gradient, "AERO_V_gradient")
	output.PointData.append(V_gradient[:,2,0], "AERO_V_gradient_2")
	output.PointData.append(PointID, "PointID")
	output.FieldData.append(DimF, "DimF")
	output.FieldData.append(WallVorticityFactor, "WallVorticityFactor")   
	output.FieldData.append(NEAR_FIELD, "NEAR_FIELD")                                       
	
	
	if AERO == 0: 
		output.PointData.append(Cp, "AERO_Cp")
		output.PointData.append(Rho, "AERO_Rho")
		output.PointData.append(Ps, "AERO_Ps")
		output.PointData.append(tau, "AERO_Stress(tau)")
		output.PointData.append(MU_LAM, "AERO_Mu_Lam")
		output.PointData.append(MU_EFF, "AERO_Mu_Eff")
		output.PointData.append(Velocity_Vector, "AERO_Velocity_Vector")


if Momentum == 1:    
	output.PointData.append(MOMENTUM_total_drag, "MOMENTUM_total_drag")
	output.PointData.append(MOMENTUM_u, "MOMENTUM_u")
	output.PointData.append(MOMENTUM_p, "MOMENTUM_p")
	output.PointData.append(MOMENTUM_lift, "MOMENTUM_lift")
	output.PointData.append(MOMENTUM_sideforce, "MOMENTUM_sideforce")
	output.PointData.append(MOMENTUM_stress, "MOMENTUM_stress")
	
if Onorato == 1: 
	output.PointData.append(ONORATO_drag, "ONORATO_drag")
	output.PointData.append(ONORATO_Pt, "ONORATO_Pt")
	output.PointData.append(ONORATO_u, "ONORATO_u")
	output.PointData.append(ONORATO_vw, "ONORATO_vw")

if Betz == 1: 
	output.PointData.append(BETZ_Vartificial, "BETZ_Vartificial")
	output.PointData.append(BETZ_profile_drag, "BETZ_profile_drag")
	output.PointData.append(BETZ_p, "BETZ_p")
	output.PointData.append(BETZ_u, "BETZ_u")

if Jones == 1: 
	output.PointData.append(JONES_profile_drag, "JONES_profile_drag")

if Oswatitsch == 1: 
	output.PointData.append(OSWATITSCH_profile_drag, "OSWATITSCH_profile_drag")

if Maskell == 1: 
	output.PointData.append(MASKELL_vortex_drag, "MASKELL_vortex_drag")
	output.PointData.append(MASKELL_lift, "MASKELL_lift")
	output.PointData.append(MASKELL_lift_zetay, "MASKELL_lift_zetay")
	output.PointData.append(MASKELL_sideforce, "MASKELL_sideforce")
	output.PointData.append(MASKELL_lift_zeta, "MASKELL_lift_zeta")


if VDV == 1: 
	output.PointData.append(VDV_DeltaUbar, "VDV_DeltaUbar")
	output.PointData.append(VDV_DeltaUstar, "VDV_DeltaUstar")
	output.PointData.append(VDV_vortex_drag, "VDV_vortex_drag")
	output.PointData.append(VDV_profile_drag, "VDV_profile_drag")
	output.PointData.append(VDV_total_drag, "VDV_total_drag")

if Giles == 1: 
	output.PointData.append(GILES_profile_drag, "GILES_profile_drag")

if Kusunose == 1: 
	#output.PointData.append(KUSUNOSE_vw, "KUSUNOSE_vw")
	#output.PointData.append(KUSUNOSE_u2, "KUSUNOSE_u2")
	#output.PointData.append(KUSUNOSE_vw_u2, "KUSUNOSE_vw_u2")
	output.PointData.append(KUSUNOSE_vortex_drag, "KUSUNOSE_vortex_drag")
	output.PointData.append(KUSUNOSE_profile_drag, "KUSUNOSE_profile_drag")
	output.PointData.append(KUSUNOSE_total_drag, "KUSUNOSE_total_drag")
	output.PointData.append(KUSUNOSE_lift, "KUSUNOSE_lift")
	output.PointData.append(VORTICITYmag, "AERO_Vorticity_Mag")                                         # This variable is used later by Epsilon_Drag_Breakdown

if Meheut == 1: 
	output.PointData.append(MEHEUT_profile_drag, "MEHEUT_profile_drag")
	output.PointData.append(DeltaUstar, "MEHEUT_DeltaUstar")
	output.PointData.append(DeltaUline, "MEHEUT_DeltaUline")

if Toubin == 1: 
	output.PointData.append(tau_x, "TOUBIN_tau_x")
	output.PointData.append(TOUBIN_Dv_Sv_ste, "TOUBIN_Dv_Sv_ste")
	output.PointData.append(TOUBIN_Dv_Vv_tem, "TOUBIN_Dv_Vv_tem")
	output.PointData.append(TOUBIN_Di_Se_ste, "TOUBIN_Di_Se_ste")
	output.PointData.append(TOUBIN_Di_Vc_ste, "TOUBIN_Di_Vc_ste")
	output.PointData.append(TOUBIN_Di_Vc_tem, "TOUBIN_Di_Vc_tem")
	output.PointData.append(TOUBIN_Dpa_Vc_ste, "TOUBIN_Dpa_Vc_ste")
	output.PointData.append(TOUBIN_Dpa_Vc_tem, "TOUBIN_Dpa_Vc_tem")    
	
if Wu == 1:    
	output.PointData.append(WU_lift, "WU_lift")
	output.PointData.append(WU_induced_drag, "WU_induced_drag")
	output.PointData.append(WU_profile_drag, "WU_profile_drag")
 
if Mele == 1:  
	output.PointData.append(MELE_lift, "MELE_lift")
	output.PointData.append(MELE_induced_drag, "MELE_induced_drag")
	output.PointData.append(MELE_profile_drag, "MELE_profile_drag")
	
if Arntz == 1:  
	output.PointData.append(ARNTZ_Ep, "ARNTZ_Ep")
	output.PointData.append(ARNTZ_Eu, "ARNTZ_Eu")
	output.PointData.append(ARNTZ_Ev, "ARNTZ_Ev")
	output.PointData.append(ARNTZ_Em, "ARNTZ_Em")
	output.PointData.append(ARNTZ_Eth, "ARNTZ_Eth")
	output.PointData.append(ARNTZ_Eth_p, "ARNTZ_Eth_p")
	output.PointData.append(ARNTZ_Eth_t, "ARNTZ_Eth_t")
	output.PointData.append(ARNTZ_Aphi, "ARNTZ_Aphi")
	output.PointData.append(ARNTZ_At, "ARNTZ_At")
	output.PointData.append(ARNTZ_Aw_div, "ARNTZ_Aw_div")
	output.PointData.append(ARNTZ_Aw_div_MASK_Shock, "ARNTZ_Aw_div_MASK_Shock")
	output.PointData.append(ARNTZ_Aw_div_MASK_Qcrit, "ARNTZ_Aw_div_MASK_Qcrit")   
	output.PointData.append(ARNTZ_A, "ARNTZ_A")
	output.PointData.append(ARNTZ_Cd, "ARNTZ_Cd")
	output.PointData.append(ARNTZ_Hprop, "ARNTZ_Hprop")
	output.PointData.append(ARNTZ_WGamma, "ARNTZ_WGamma")	

if Aguirre == 1:  
	output.PointData.append(AGUIRRE_u_star, "AGUIRRE_AERO_u_star")
	output.PointData.append(AGUIRRE_u_bar, "AGUIRRE_AERO_u_bar")
	output.PointData.append(AGUIRRE_v_star, "AGUIRRE_AERO_v_star")
	output.PointData.append(AGUIRRE_v_bar, "AGUIRRE_AERO_v_bar")
	output.PointData.append(AGUIRRE_w_star, "AGUIRRE_AERO_w_star")
	output.PointData.append(AGUIRRE_w_bar, "AGUIRRE_AERO_w_bar")
	output.PointData.append(AGUIRRE_Vabs_star, "AGUIRRE_AERO_Vabs_star")
	output.PointData.append(AGUIRRE_Ts_star, "AGUIRRE_AERO_Ts_star")
	output.PointData.append(AGUIRRE_Ts_bar, "AGUIRRE_AERO_Ts_bar")
	output.PointData.append(AGUIRRE_Rho_star, "AGUIRRE_AERO_Rho_star")
	output.PointData.append(AGUIRRE_Rho_bar, "AGUIRRE_AERO_Rho_bar")
	output.PointData.append(AGUIRRE_e_star, "AGUIRRE_AERO_e_star")
	output.PointData.append(AGUIRRE_e_bar, "AGUIRRE_AERO_e_bar")
	output.PointData.append(AGUIRRE_Hs_star, "AGUIRRE_AERO_Hs_star")
	output.PointData.append(AGUIRRE_Hs_bar, "AGUIRRE_AERO_Hs_bar")
	output.PointData.append(AGUIRRE_M_star, "AGUIRRE_AERO_M_star")    


	output.PointData.append(AGUIRRE_Aphi_Vort, "AGUIRRE_Aphi_Vort")
	output.PointData.append(AGUIRRE_Em_star, "AGUIRRE_Em_star")
	output.PointData.append(AGUIRRE_Em_bar, "AGUIRRE_Em_bar")
	output.PointData.append(AGUIRRE_Eu_star, "AGUIRRE_Eu_star")
	output.PointData.append(AGUIRRE_Eu_bar, "AGUIRRE_Eu_bar")
	output.PointData.append(AGUIRRE_Ev_star, "AGUIRRE_Ev_star")
	output.PointData.append(AGUIRRE_Ev_bar, "AGUIRRE_Ev_bar")
	output.PointData.append(AGUIRRE_Ep_star, "AGUIRRE_Ep_star")
	output.PointData.append(AGUIRRE_Ep_bar, "AGUIRRE_Ep_bar")
	output.PointData.append(AGUIRRE_Eth_star, "AGUIRRE_Eth_star")
	output.PointData.append(AGUIRRE_Eth_bar, "AGUIRRE_Eth_bar")
	output.PointData.append(AGUIRRE_Cd_star, "AGUIRRE_Cd_star")
	output.PointData.append(AGUIRRE_Cd_bar, "AGUIRRE_Cd_bar")
	output.PointData.append(AGUIRRE_vortex_drag, "AGUIRRE_vortex_drag")	
	
	output.PointData.append(AGUIRRE_WGamma_star, "AGUIRRE_WGamma_star")	
	output.PointData.append(AGUIRRE_WGamma_bar, "AGUIRRE_WGamma_bar")	

	output.PointData.append(AGUIRRE_Em_bar_corrected, "AGUIRRE_Em_bar_corrected")
	output.PointData.append(AGUIRRE_Cd_bar_corrected, "AGUIRRE_Cd_bar_corrected")
	output.PointData.append(AGUIRRE_Aw_div_MASKED, "AGUIRRE_Aw_div_MASKED")
	
	output.PointData.append(AGUIRRE_INF_Ep, "AGUIRRE_INF_Ep")
	output.PointData.append(AGUIRRE_INF_Em, "AGUIRRE_INF_Em")
	output.PointData.append(AGUIRRE_INF_Eth, "AGUIRRE_INF_Eth")
	output.PointData.append(AGUIRRE_INF_Cd, "AGUIRRE_INF_Cd")

if Ruscio == 1: 
	output.PointData.append(RUSCIO_En_vol, "RUSCIO_En_vol")
	output.PointData.append(RUSCIO_An_vol, "RUSCIO_An_vol")
	output.PointData.append(RUSCIO_Etc_vol, "RUSCIO_Etc_vol")
	output.PointData.append(RUSCIO_Etc_vol_div, "RUSCIO_Etc_vol_div")
	output.PointData.append(RUSCIO_Mo_vol, "RUSCIO_Mo_vol")		
	output.PointData.append(RUSCIO_Ie_vol, "RUSCIO_Ie_vol")		
	output.PointData.append(RUSCIO_Eu_vol, "RUSCIO_Eu_vol")		
	output.PointData.append(RUSCIO_Ev_vol, "RUSCIO_Ev_vol")	

if Select_Arrays == 1:
	for i in UserSelectedArray:
		output.PointData.append(INPUT.PointData[i], i)



############################################################################################################
######################################  DEFINE INTEGRATION ATTRIBUTES  #####################################
############################################################################################################

# Define how to integrate the variables:                                                                    # Information for EpsilonIntegrationTool: the user must declare if the variable has to be integrated at the survey plane ("Survey_Plane") or at the entire control volume ("Control_Volume") or if it should not be integrated ("Do_Not_Integrate")
# OPTIONS: "Survey_Plane", "Control_Volume", "Do_Not_Integrate"

if AERO == 0:  
	EpsilonIntegrationTool.append(["AERO_Vorticity","Survey_Plane"]) 


if AERO == 1:    
	EpsilonIntegrationTool.append(["AERO_Mu_Lam","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_Mu_Turb","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_Mu_Eff","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_Keff","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_Velocity_Vector","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_u","Survey_Plane"]) 
	EpsilonIntegrationTool.append(["AERO_v","Survey_Plane"]) 
	EpsilonIntegrationTool.append(["AERO_w","Survey_Plane"]) 
	EpsilonIntegrationTool.append(["AERO_udelta","Survey_Plane"]) 
	EpsilonIntegrationTool.append(["AERO_Vabs","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_M","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_Vorticity","Survey_Plane"]) 
	EpsilonIntegrationTool.append(["AERO_Vorticity_Mag","Do_Not_Integrate"]) 
	EpsilonIntegrationTool.append(["AERO_Helicity","Survey_Plane"]) 
	EpsilonIntegrationTool.append(["AERO_Stress(tau)","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Strain","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Q_criterion","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Lambda_2","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Phi_eff","Control_Volume"])
	EpsilonIntegrationTool.append(["AERO_Phi_eff_Vort","Control_Volume"])
	EpsilonIntegrationTool.append(["AERO_Ps","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Pt","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Pt/Pt0","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Pt-Pt0","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Shock_Detector","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_q","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Cp","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Cp_gradient","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Ts","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Ts_constant","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Ts_isoenergetic","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Ts_isentropic","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Tt","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Hs","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Ht","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Ht-Ht0","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Rho","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Rho_constant","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Rho","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Rho_isoenergetic","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Rho_isentropic","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_RhoU","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_s-s0","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_e","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AERO_Lamb_Vector","Control_Volume"])

if NEAR_FIELD == 1:     
	EpsilonIntegrationTool.append(["AERO_V_gradient","Survey_Plane"])

	if AERO == 0: 
		EpsilonIntegrationTool.append(["AERO_Cp","Survey_Plane"])
		EpsilonIntegrationTool.append(["AERO_Rho","Do_Not_Integrate"])
		EpsilonIntegrationTool.append(["AERO_Ps","Do_Not_Integrate"])
		EpsilonIntegrationTool.append(["AERO_Stress(tau)","Survey_Plane"])
		EpsilonIntegrationTool.append(["AERO_Mu_Lam","Do_Not_Integrate"])
		EpsilonIntegrationTool.append(["AERO_Mu_Eff","Do_Not_Integrate"])
		EpsilonIntegrationTool.append(["AERO_Velocity_Vector","Do_Not_Integrate"])

if Momentum == 1:    
	EpsilonIntegrationTool.append(["MOMENTUM_total_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["MOMENTUM_u","Survey_Plane"])
	EpsilonIntegrationTool.append(["MOMENTUM_p","Survey_Plane"])
	EpsilonIntegrationTool.append(["MOMENTUM_lift","Survey_Plane"])
	EpsilonIntegrationTool.append(["MOMENTUM_sideforce","Survey_Plane"])
	EpsilonIntegrationTool.append(["MOMENTUM_stress","Survey_Plane"])
	
if Onorato == 1: 
	EpsilonIntegrationTool.append(["ONORATO_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["ONORATO_Pt","Survey_Plane"])
	EpsilonIntegrationTool.append(["ONORATO_u","Survey_Plane"])
	EpsilonIntegrationTool.append(["ONORATO_vw","Survey_Plane"])

if Betz == 1: 
	EpsilonIntegrationTool.append(["BETZ_Vartificial","Survey_Plane"])
	EpsilonIntegrationTool.append(["BETZ_profile_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["BETZ_p","Survey_Plane"])
	EpsilonIntegrationTool.append(["BETZ_u","Survey_Plane"])

if Jones == 1: 
	EpsilonIntegrationTool.append(["JONES_profile_drag","Survey_Plane"])

if Oswatitsch == 1: 
	EpsilonIntegrationTool.append(["OSWATITSCH_profile_drag","Survey_Plane"])

if Maskell == 1: 
	EpsilonIntegrationTool.append(["MASKELL_vortex_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["MASKELL_lift","Survey_Plane"])
	EpsilonIntegrationTool.append(["MASKELL_lift_zetay","Survey_Plane"])
	EpsilonIntegrationTool.append(["MASKELL_sideforce","Survey_Plane"])
	EpsilonIntegrationTool.append(["MASKELL_lift_zeta","Survey_Plane"])
    

if VDV == 1: 
	EpsilonIntegrationTool.append(["VDV_DeltaUbar","Survey_Plane"])
	EpsilonIntegrationTool.append(["VDV_DeltaUstar","Survey_Plane"])
	EpsilonIntegrationTool.append(["VDV_vortex_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["VDV_profile_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["VDV_total_drag","Survey_Plane"])

if Giles == 1: 
	EpsilonIntegrationTool.append(["GILES_profile_drag","Survey_Plane"])

if Kusunose == 1: 
	EpsilonIntegrationTool.append(["KUSUNOSE_vortex_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["KUSUNOSE_profile_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["KUSUNOSE_total_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["KUSUNOSE_lift","Survey_Plane"])
	EpsilonIntegrationTool.append(["AERO_Vorticity_Mag","Survey_Plane"])

if Meheut == 1: 
	EpsilonIntegrationTool.append(["MEHEUT_profile_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["MEHEUT_DeltaUstar","Survey_Plane"])
	EpsilonIntegrationTool.append(["MEHEUT_DeltaUline","Survey_Plane"])

if Toubin == 1: 
	EpsilonIntegrationTool.append(["TOUBIN_tau_x","Survey_Plane"])
	EpsilonIntegrationTool.append(["TOUBIN_Dv_Sv_ste","Survey_Plane"])
	EpsilonIntegrationTool.append(["TOUBIN_Dv_Vv_tem","Control_Volume"])
	EpsilonIntegrationTool.append(["TOUBIN_Di_Se_ste","Survey_Plane"])
	EpsilonIntegrationTool.append(["TOUBIN_Di_Vc_ste","Control_Volume"])
	EpsilonIntegrationTool.append(["TOUBIN_Di_Vc_tem","Control_Volume"])
	EpsilonIntegrationTool.append(["TOUBIN_Dpa_Vc_ste","Control_Volume"])
	EpsilonIntegrationTool.append(["TOUBIN_Dpa_Vc_tem","Control_Volume"])

if Wu == 1:    
	EpsilonIntegrationTool.append(["WU_lift","Control_Volume"])
	EpsilonIntegrationTool.append(["WU_induced_drag","Control_Volume"])
	EpsilonIntegrationTool.append(["WU_profile_drag","Survey_Plane"])
 
if Mele == 1:  
	EpsilonIntegrationTool.append(["MELE_lift","Control_Volume"])
	EpsilonIntegrationTool.append(["MELE_induced_drag","Control_Volume"])
	EpsilonIntegrationTool.append(["MELE_profile_drag","Survey_Plane"])

if Arntz == 1:  
	EpsilonIntegrationTool.append(["ARNTZ_Ep","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Eu","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Ev","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Em","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Eth","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Eth_p","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Eth_t","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Aphi","Control_Volume"])
	EpsilonIntegrationTool.append(["ARNTZ_At","Control_Volume"])
	EpsilonIntegrationTool.append(["ARNTZ_Aw_div","Control_Volume"])
	EpsilonIntegrationTool.append(["ARNTZ_Aw_div_MASK_Shock","Control_Volume"])
	EpsilonIntegrationTool.append(["ARNTZ_Aw_div_MASK_Qcrit","Control_Volume"])
	EpsilonIntegrationTool.append(["ARNTZ_A","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Cd","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_Hprop","Survey_Plane"])
	EpsilonIntegrationTool.append(["ARNTZ_WGamma","Survey_Plane"])

if Aguirre == 1:  
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_u_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_u_bar","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_v_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_v_bar","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_w_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_w_bar","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_Vabs_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_Ts_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_Ts_bar","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_Rho_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_Rho_bar","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_e_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_e_bar","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_Hs_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_Hs_bar","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_AERO_M_star","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_Aphi_Vort","Control_Volume"])
	EpsilonIntegrationTool.append(["AGUIRRE_Em_star","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Em_bar","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Eu_star","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Eu_bar","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Ev_star","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Ev_bar","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Ep_star","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Ep_bar","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Eth_star","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Eth_bar","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Cd_star","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Cd_bar","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_vortex_drag","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_WGamma_star","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_WGamma_bar","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Em_bar_corrected","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Cd_bar_corrected","Survey_Plane"])
	EpsilonIntegrationTool.append(["AGUIRRE_Aw_div_MASKED","Control_Volume"])
	EpsilonIntegrationTool.append(["AGUIRRE_INF_Ep","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_INF_Em","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_INF_Eth","Do_Not_Integrate"])
	EpsilonIntegrationTool.append(["AGUIRRE_INF_Cd","Do_Not_Integrate"])

if Ruscio == 1: 
	EpsilonIntegrationTool.append(["RUSCIO_En_vol","Control_Volume"])
	EpsilonIntegrationTool.append(["RUSCIO_An_vol","Control_Volume"])
	EpsilonIntegrationTool.append(["RUSCIO_Etc_vol","Control_Volume"])
	EpsilonIntegrationTool.append(["RUSCIO_Etc_vol_div","Control_Volume"])
	EpsilonIntegrationTool.append(["RUSCIO_Mo_vol","Control_Volume"])
	EpsilonIntegrationTool.append(["RUSCIO_Ie_vol","Control_Volume"])
	EpsilonIntegrationTool.append(["RUSCIO_Eu_vol","Control_Volume"])
	EpsilonIntegrationTool.append(["RUSCIO_Ev_vol","Control_Volume"])

if Select_Arrays == 1:
	for i in UserSelectedArray:
		EpsilonIntegrationTool.append([i,"Survey_Plane"])


# Send "EpsilonIntegrationTool" array to the output:
NB_of_Arrays = np.shape(EpsilonIntegrationTool)[0]

array = vtk.vtkStringArray()
array.SetName("ARRAY")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][0])
output.GetFieldData().AddArray(array)


array = vtk.vtkStringArray()
array.SetName("ARRAY_Integration")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][1])
output.GetFieldData().AddArray(array)

############################################################################################################
###############################################  SAVE DATA  ################################################
############################################################################################################


###################################
#######  WRITE VTP FUNCTION  ######
###################################

def WRITE_VTP(outputObject,File_Path_Output):     					
	import vtk
	if File_Path_Output:
		
		OutputVTP = File_Path_Output                                    # File path to the folder where to store the output data 
		
		OutputVTK = vtk.vtkXMLUnstructuredGridWriter()                  # Alternative writer:     vtkUnstructuredGridWriter   XMLPUnstructuredGridWriter
		OutputVTK.SetInputData(outputObject.VTKObject)    			    # Alternative command: SetInput           vtkObject.GetOutput()
		OutputVTK.SetFileName(OutputVTP)
		OutputVTK.Write()                                               # "OutputVTK" is a vtkUnstructuredGrid variable
		#OutputVTK.Update()
		print ("# Output file written successfully")

if Save_Output_Data==1:
	WRITE_VTP(output,File_Path_Output)



############################################################################################################
###############################################  END OF CODE ###############################################
############################################################################################################