# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################                   Axes Plugin                    #############################
############################################################################################################

# Abstract:
This Paraview's programmable source allows to create a set of 3 vectors perpendicular to each other that 
serves as a custom system of reference. These axes can be rotated based on the angle of attack, so the x-axis
will be a wind axis. Y- and Z-axis are placed following the far field method convention: Y-axis along the
spanwise direction pointing towards the right wing, and Z-axis pointing upwards.

# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""

print ("")
print ("")
print ("")
print ("####################################################")
print ("###############      EPSILON AXES     ##############")
print ("####################################################")
print ("")
print ('# Change "Solid Color" to "Colors" in order to show x-axis in red, y-axis in yellow and z-axis in green')
print ('# Preset color map must be of type "Blue to Green Rainbow"')



############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

import vtk
import numpy as np


############################################################################################################
############################################# STRING TO FLOAT ##############################################
############################################################################################################

# Input parameters are defined as string in Paraview GUI, thus a string to float convertion is required:
Size = float(Size)  # Overal size of the aerodynamic axes vectors

# The following parameters are defined as a vector in Paraview, so each component is assigned to a separate variable:
X0 = Origin[0]  	# X-mesh coordinate of the axes origin
Y0 = Origin[1]
Z0 = Origin[2]


############################################################################################################
############################################## READ INPUT DATA #############################################
############################################################################################################

TRANSFORMATION = inputs[0].FieldData["TRANSFORMATION"]
T_t = np.transpose(TRANSFORMATION)
T_inverse = np.dot(T_t,np.dot(TRANSFORMATION,T_t))  	# It calculates the inverse of the TRANSFORMATION matrix


############################################################################################################
############################################# CREATING VECTORS #############################################
############################################################################################################

positions = [[X0, Y0, Z0],[X0, Y0, Z0],[X0, Y0, Z0],[X0, Y0, Z0]]					# Origin position of the aerodynamic axes
orientations = [[1.0, 0.0, 0.0],[0.0, 1.0, 0.0],[0.0, 0.0, 1.0],[0.0, 0.0, 1.0]]	# axes orientations in the mesh reference frame
orientations = np.einsum('ij,kj->ik',orientations,T_inverse)						# axes orientations in the aero reference frame


colors = [[100, 0, 0],[0, 72, 0],[0, 0, 0],[0, 0, 50]]  							# Color code for the arrows
heights = [1,1,1,1]   																# Used for debug purposes

# Defining the origin points:
points = vtk.vtkPoints()
points.InsertNextPoint(*positions[0])
points.InsertNextPoint(*positions[1])
points.InsertNextPoint(*positions[2])
points.InsertNextPoint(*positions[3])
polydata = vtk.vtkPolyData()
polydata.SetPoints(points)

# Defining the color of the arrows:
color_def = vtk.vtkUnsignedCharArray()
color_def.SetNumberOfComponents(3)
color_def.SetName("Colors")
color_def.InsertTuple3(0, *colors[0])
color_def.InsertTuple3(1, *colors[1])
color_def.InsertTuple3(2, *colors[2])
color_def.InsertTuple3(3, *colors[3])

polydata.GetPointData().AddArray(color_def)

# Defining arrows orientations:
pointNormalsArray = vtk.vtkDoubleArray()
pointNormalsArray.SetNumberOfComponents(3)
pointNormalsArray.SetNumberOfTuples(polydata.GetNumberOfPoints())
pointNormalsArray.SetTuple(0, orientations[0])
pointNormalsArray.SetTuple(1, orientations[1])
pointNormalsArray.SetTuple(2, orientations[2])
pointNormalsArray.SetTuple(3, orientations[3])

polydata.GetPointData().SetNormals(pointNormalsArray)

# Defining the arrow source with some attributes
arrow_source = vtk.vtkArrowSource()
arrow_source.SetTipRadius(0.05)
arrow_source.SetTipLength(0.3)
arrow_source.SetTipResolution(30)
arrow_source.SetShaftRadius(0.01)
arrow_source.SetShaftResolution(30)


# Creating the arrows:
glyph = vtk.vtkGlyph3D()
glyph.SetInputData(polydata)
glyph.SetSourceConnection(arrow_source.GetOutputPort())
glyph.SetColorModeToColorByScalar()
glyph.SetVectorModeToUseNormal()
glyph.SetScaleFactor(Size)
glyph.ScalingOn()
glyph.SetScaleModeToScaleByScalar()
#glyph.SetVectorModeToUseVector()
glyph.OrientOn()
#glyph.SetScaleModeToDataScalingOff()
glyph.ClampingOn()
#glyph.SetRange(0, 4)
#glyph.SetInputArrayToProcess(0,0,0,0,'heights')		# scalars
#glyph.SetInputArrayToProcess(3,0,0,0,'colors')		# colors
#glyph.SetInputArrayToProcess(1,0,0,0,'color_def')		# vectors
glyph.Update()

 
self.GetOutput().ShallowCopy(glyph.GetOutput())



############################################################################################################
###############################################  END OF CODE ###############################################
############################################################################################################