# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################              Circulation Tool Plugin             #############################
############################################################################################################

# Abstract:
This Paraview's programmable filer computes circulation based on two methods: a)By definition (i.e., velocity
cirulation along a closed path, and b)Alternative method (i.e., flux of vorticity across the surface defined
by the closed path). Circulation can be computed at one more sections. One section is typically for 2D cases
and multiple sections is for 3D cases (it enables determining the spawise distribution of circulation). Also
local lift coefficient can be computed from circulation.

DEFINITIONS:
• Control sections: rectangular sections (up to 8) that are used to define the assembly of circulation paths
• Circulation stations: rectangular paths used to compute the circulation. It is given by the control sections
themselves and the internal stations in between two control sections

# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""

print ("")
print ("")
print ("")
print ("####################################################")
print ("#########    EPSILON CIRCULATION TOOL      #########")
print ("####################################################")
print ("")



############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

import vtk
import numpy as np


############################################################################################################
############################################# STRING TO FLOAT ##############################################
############################################################################################################

# Input parameters are defined as string in Paraview GUI, thus a string to float convertion is required:
Mode = Mode                                                                                                 # 0=3D  1=2D planar
Control_Sections = Control_Sections                                                                         # Number of control sections
Normals = Normals                                                                                           # 0=Aerodynamic x-axis    1=Aerodynamic y-axis    2=Aerodynamic z-axis    3=Mesh X-axis   4=Mesh Y-axis   5=Mesh Z-axis   6=User defined
Center_1 = Center_1                                                                                         # Center of the control section 1
Normal_1 = Normal_1                                                                                         # Normal of the control section 1
Rotate_1 = Rotate_1                                                                                         # Boolean
Angle_1 = Angle_1                                                                                           # Rotation angle of the control section 1
Box_Size_1 = Box_Size_1                                                                                     # Width and height [m] of the control section box
Center_2 = Center_2                                                                                         # Center of the control section 2
Normal_2 = Normal_2                                                                                         # Normal of the control section 2
Box_Size_2 = Box_Size_2                                                                                     # Width and height [m] of the control section box + number of intermediate stations between section i and i-1
Rotate_2 = Rotate_2                                                                                         # Boolean
Angle_2 = Angle_2                                                                                           # Rotation angle of the control section 2
Center_3 = Center_3                                                                                         # Center of the control section 3
Normal_3 = Normal_3                                                                                         # Normal of the control section 3
Box_Size_3 = Box_Size_3                                                                                     # Width and height [m] of the control section box + number of intermediate stations between section i and i-1
Rotate_3 = Rotate_3                                                                                         # Boolean
Angle_3 = Angle_3                                                                                           # Rotation angle of the control section 3
Center_4 = Center_4                                                                                         # Center of the control section 4
Normal_4 = Normal_4                                                                                         # Normal of the control section 4
Box_Size_4 = Box_Size_4                                                                                     # Width and height [m]  of the control section box + number of intermediate stations between section i and i-1
Rotate_4 = Rotate_4                                                                                         # Boolean
Angle_4 = Angle_4                                                                                           # Rotation angle of the control section 4
Center_5 = Center_5                                                                                         # Center of the control section 5
Normal_5 = Normal_5                                                                                         # Normal of the control section 5
Box_Size_5 = Box_Size_5                                                                                     # Width and height [m]  of the control section box + number of intermediate stations between section i and i-1
Rotate_5 = Rotate_5                                                                                         # Boolean
Angle_5 = Angle_5                                                                                           # Rotation angle of the control section 5
Center_6 = Center_6                                                                                         # Center of the control section 6
Normal_6 = Normal_6                                                                                         # Normal of the control section 6
Box_Size_6 = Box_Size_6                                                                                     # Width and height [m]  of the control section box + number of intermediate stations between section i and i-1
Rotate_6 = Rotate_6                                                                                         # Boolean
Angle_6 = Angle_6                                                                                           # Rotation angle of the control section 6
Center_7 = Center_7                                                                                         # Center of the control section 7
Normal_7 = Normal_7                                                                                         # Normal of the control section 7
Box_Size_7 = Box_Size_7                                                                                     # Width and height [m]  of the control section box + number of intermediate stations between section i and i-1
Rotate_7 = Rotate_7                                                                                         # Boolean
Angle_7 = Angle_7                                                                                           # Rotation angle of the control section 7
Center_8 = Center_8                                                                                         # Center of the control section 8
Normal_8 = Normal_8                                                                                         # Normal of the control section 8
Box_Size_8 = Box_Size_8                                                                                     # Width and height [m]  of the control section box + number of intermediate stations between section i and i-1
Rotate_8 = Rotate_8                                                                                         # Boolean
Angle_8 = Angle_8                                                                                           # Rotation angle of the control section 8
Integral_Type = Integral_Type                                                                               # 0=Velocity circulation    1=vorticity flux
Show_Circulation_Paths = Show_Circulation_Paths                                                             # Boolean
Flip_Normals = Flip_Normals                                                                                 # Boolean
Lifting_Line_Method = Lifting_Line_Method                                                                   # Boolean
Chord_Law = Chord_Law                                                                                       # 0=Constant chord  1=Varying chord
Chord = Chord                                                                                               # Constant chord in [m]
Load_Body_Surface = Load_Body_Surface                                                                       # Path to the VTP body surface (used to extract the local chord at each station


if Mode == 0:                           
	Select_Action = Select_Action_3D                                                                        # 0=Visualize circulation paths   1=Integrate data
if Mode == 1:
	Select_Action = Select_Action_2D

if Mode == 0:                           
	Control_Sections = Control_Sections
if Mode == 1:
	Control_Sections = 1

if Rotate_1 == 0:                           
	Angle_1 = 0
if Rotate_2 == 0:                           
	Angle_2 = 0
if Rotate_3 == 0:                           
	Angle_3 = 0
if Rotate_4 == 0:                           
	Angle_4 = 0
if Rotate_5 == 0:                           
	Angle_5 = 0    
if Rotate_6 == 0:                           
	Angle_6 = 0
if Rotate_7 == 0:                           
	Angle_7 = 0
if Rotate_8 == 0:                           
	Angle_8 = 0



############################################################################################################
############################################## READ INPUT DATA #############################################
############################################################################################################

V0 = inputs[0].FieldData["V0"]
Rho0 = inputs[0].FieldData["Rho0"]
Sref = inputs[0].FieldData["Sref"]
AdimF = inputs[0].FieldData["AdimF"]
UpwardAxis  = inputs[0].FieldData["UpwardAxis"]
TRANSFORMATION = inputs[0].FieldData["TRANSFORMATION"]

T_t = np.transpose(TRANSFORMATION)
T_inverse = np.dot(T_t,np.dot(TRANSFORMATION,T_t))  	# It calculates the inverse of the TRANSFORMATION matrix



############################################################################################################
#######################################    DEFINITION OF FUNCTIONS    ######################################
############################################################################################################

############################################
###############  PLOT ARROW  ###############
############################################

def PLOT_ARROW(Position,Direction,Lenght):                                                                  # Defines a function that draws a VTK arrow object
	import vtk

	# Defining the origin point:
	points = vtk.vtkPoints()
	points.InsertNextPoint(Position) 

	# Defining arrow direction vector:
	ArrowDirection = vtk.vtkDoubleArray()
	ArrowDirection.SetNumberOfComponents(3)
	ArrowDirection.SetNumberOfTuples(1)              
	ArrowDirection.SetTuple(0, Direction)

	# Generate polydata:
	polydata = vtk.vtkPolyData()
	polydata.SetPoints(points)
	polydata.GetPointData().SetNormals(ArrowDirection)

	# Defining the arrow source with some attributes:
	arrow_source = vtk.vtkArrowSource()
	arrow_source.SetTipRadius(0.05)
	arrow_source.SetTipLength(0.3)
	arrow_source.SetTipResolution(30)
	arrow_source.SetShaftRadius(0.01)
	arrow_source.SetShaftResolution(30)

	# Creating the arrow:
	glyph = vtk.vtkGlyph3D()
	glyph.SetInputData(polydata)
	glyph.SetSourceConnection(arrow_source.GetOutputPort())
	glyph.SetScaleFactor(Lenght)
	glyph.ScalingOn()
	glyph.SetVectorModeToUseNormal()
	glyph.OrientOn()
	glyph.Update()

	return glyph


############################################
########  TWO VECTOR TRANSFORMATION  #######
############################################

def TWO_VECTOR_TRANSFORMATION(Vector_A,Vector_B,Point):                                                             # Defines a function that computes the transformation matrix that aligns vector A with vector B. Then uses this matrix to transform the input point coordinate
	import numpy as np
	A = Vector_A/np.linalg.norm(Vector_A)                                                                           # Input vector is normalized
	B = Vector_B/np.linalg.norm(Vector_B)
	C = np.cross(A,B)
	D = np.dot(A,B)
	I = [[1,0,0],[0,1,0],[0,0,1]]
	Vt = [[0,-C[2],C[1]],[C[2],0,-C[0]],[-C[1],C[0],0]]
	if D == -1:                                             # 180° rotation
		Mtx = [[-1,0,0],[0,-1,0],[0,0,1]]
	else:
		Last = np.dot(Vt,Vt)*(1/(1+D))
		Mtx = np.add(np.add(I , Vt) , Last)            # Mtx = I + Vt + Vt*Vt*(1/(1+D))
		Mtx_T = np.transpose(Mtx)
	#print ("A",A)
	#print ("B",B)
	#print ("Atransformed",np.dot(Mtx,A))
	Transformed_Point = np.dot(Mtx,Point)
	return Transformed_Point


############################################
#########  PLOT VTK LINES FUNCTION  ########
############################################

def PLOT_VTK_LINES_VERTEX(PointsVector):                                                                    # Defines a function that takes a numpy array with the points coordinates and generates two VTK outputs: the polyline connecting the points as well as the points themselves
	import numpy as np
	import vtk

	points = vtk.vtkPoints()                                                	                            # create a vtkPoints container to store all the point coordinates.
	points.SetData(dsa.numpyTovtkDataArray(PointsVector , "Points"))   					                    # numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()

	NBpts = np.shape(PointsVector)[0]                                             		                    # Number of mesh points

	polyLine = vtk.vtkPolyLine()                                            	                            # Creates a polyline and set the ID for each point
	polyLine.GetPointIds().SetNumberOfIds(int(NBpts)) 
	for i in range(0, int(NBpts)):
		polyLine.GetPointIds().SetId(i, i)

	cells = vtk.vtkCellArray()                                              	                            # Create a cell array to store the lines in and add the lines to it
	cells.InsertNextCell(polyLine)

	polyData = vtk.vtkPolyData()                                            	                            # Create a polydata to store everything in
	polyData.SetPoints(points)                                              	                            # Add the points to the dataset
	polyData.SetLines(cells)                                                	                            # Add the lines to the dataset

	Survey_Line = vtk.vtkGeometryFilter()                                   	                            # Converts the polydata to vtkPolyData  
	Survey_Line.SetInputData(polyData)             
	Survey_Line.Update()
	
	# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
	polydata2 = vtk.vtkPolyData()
	polydata2.SetPoints(points)

	Survey_Vertices = vtk.vtkVertexGlyphFilter()
	Survey_Vertices.AddInputData(polydata2)
	Survey_Vertices.Update()
    
	return Survey_Line,Survey_Vertices


############################################
#############  PLOT RECTANGLE  #############
############################################

def PLOT_RECTANGLE(Center,Normal,Width,Height,Angle,TWO_VECTOR_TRANSFORMATION,PLOT_VTK_LINES_VERTEX):              # Defines a function that draws a rectangle centered at the origin of the YZ plane (Normal=[1,0,0]) and transforms it to the desired position
	import vtk
	import numpy as np
	
	# Generate reference rectangle points in YZ plane:
	P1 = [0,0.5*Width,-0.5*Height]
	P2 = [0,0.5*Width,0.5*Height]    
	P3 = [0,-0.5*Width,0.5*Height]    
	P4 = [0,-0.5*Width,-0.5*Height]   

	# Rotate the points around X-axis:
	Mx = [[1,0,0],[0,np.cos(Angle*3.14159/180),-1*np.sin(Angle*3.14159/180)],[0,np.sin(Angle*3.14159/180),np.cos(Angle*3.14159/180)]]
	P1 = np.dot(Mx,P1)
	P2 = np.dot(Mx,P2)
	P3 = np.dot(Mx,P3)
	P4 = np.dot(Mx,P4)
	
	# Rotate the points using direction vectors:                                    
	P1_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P1)
	P2_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P2)
	P3_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P3)    
	P4_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P4) 

	# Translate the points:
	P1_tr = P1_rot + Center
	P2_tr = P2_rot + Center
	P3_tr = P3_rot + Center  
	P4_tr = P4_rot + Center  

	# Draw the rectangle: 
	Rectangle,Vectices = PLOT_VTK_LINES_VERTEX([P1_tr,P2_tr,P3_tr,P4_tr,P1_tr])
	
	return Rectangle


############################################
#########  LOAD VTK DATA FUNCTION  #########
############################################

def LOAD_VTK_DATA(FilePath):                                                                                # Defines a function that takes a filepath (String) pointing to a VTK object to be loaded. Accepted formats: VTU, VTP, VTS, VTI and VTK
	import vtk
	reader = None                                                                                           # Generate a vtk reader depending on the input file format
	if FilePath.endswith(".vtp"):
		reader = vtk.vtkXMLPolyDataReader()
	elif FilePath.endswith(".vtk"):    
		reader = vtk.vtkGenericDataObjectReader()
	elif FilePath.endswith(".stl"):        
		reader = vtk.vtkSTLReader()
	else:
		print ("Input file format not recognized")
	reader.SetFileName(FilePath)
	reader.Update() 
	return reader


###########################################################
##################### SLICE WITH PLANE ####################
###########################################################

def SLICE_WITH_PLANE(Object,Center,Normal):                                                                         # Defines a function that slices the VTK Object with a VTK implicit plane function. IMPORTANT: if the objects are data from the input it is not necessary to add .GetOutput() (this is required for VTK objects only)
	import vtk
    
	plane = vtk.vtkPlane()																            
	plane.SetOrigin(Center[0], Center[1], Center[2])
	plane.SetNormal(Normal[0], Normal[1], Normal[2])

	cutter = vtk.vtkCutter()														                    
	cutter.SetCutFunction(plane)
	cutter.SetInputData(Object)
	cutter.Update()
	return cutter
    

###########################################################
##################### SLICE WITH BOX ####################
###########################################################

def SLICE_WITH_BOX(Object,Center,Normal,Width,Height,Angle,TWO_VECTOR_TRANSFORMATION):                              # Defines a function that slices a plane with a box 
	import vtk
	import numpy as np

	# Define box points and their normals in YZ plane:
	P1 = [0, 0.5*Width, 0]
	P2 = [0, -0.5*Width, 0]
	P3 = [0, 0, 0.5*Height]
	P4 = [0, 0, -0.5*Height]
	P5 = [1, 0, 0]
	P6 = [-1, 0, 0]    
	N1 = [0, -1, 0]
	N2 = [0, 1, 0]
	N3 = [0, 0, -1]
	N4 = [0, 0, 1]    
	N5 = [-1, 0, 0]
	N6 = [1, 0, 0]
    
	# Rotate the points and normals around X-axis:
	Mx = [[1,0,0],[0,np.cos(Angle*3.14159/180),-1*np.sin(Angle*3.14159/180)],[0,np.sin(Angle*3.14159/180),np.cos(Angle*3.14159/180)]]
	P1 = np.dot(Mx,P1)
	P2 = np.dot(Mx,P2)
	P3 = np.dot(Mx,P3)
	P4 = np.dot(Mx,P4)
	N1 = np.dot(Mx,N1)
	N2 = np.dot(Mx,N2)
	N3 = np.dot(Mx,N3)
	N4 = np.dot(Mx,N4)

	# Rotate the points and normals using direction vectors:                                    
	P1_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P1)
	P2_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P2)
	P3_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P3)
	P4_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P4)
	P5_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P5)
	P6_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P6)
	N1_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N1)
	N2_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N2)
	N3_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N3)
	N4_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N4)
	N5_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N5)
	N6_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N6)

	# Translate the points:
	P1_tr = P1_rot + Center
	P2_tr = P2_rot + Center
	P3_tr = P3_rot + Center  
	P4_tr = P4_rot + Center 
	P5_tr = P5_rot + Center  
	P6_tr = P6_rot + Center 

	# Generate implicit plane functions:   
	plane1 = vtk.vtkPlane()
	plane1.SetOrigin(P1_tr)
	plane1.SetNormal(N1_rot)

	plane2 = vtk.vtkPlane()
	plane2.SetOrigin(P2_tr)
	plane2.SetNormal(N2_rot)

	plane3 = vtk.vtkPlane()
	plane3.SetOrigin(P3_tr)
	plane3.SetNormal(N3_rot)

	plane4 = vtk.vtkPlane()
	plane4.SetOrigin(P4_tr)
	plane4.SetNormal(N4_rot)
    
	plane5 = vtk.vtkPlane()
	plane5.SetOrigin(P5_tr)
	plane5.SetNormal(N5_rot)

	plane6 = vtk.vtkPlane()
	plane6.SetOrigin(P6_tr)
	plane6.SetNormal(N6_rot)

	# Combine the implicit plane functions: 
	clipFunction = vtk.vtkImplicitBoolean()
	clipFunction.SetOperationTypeToUnion()
	clipFunction.AddFunction(plane1)
	clipFunction.AddFunction(plane2)
	clipFunction.AddFunction(plane3)
	clipFunction.AddFunction(plane4)
	#clipFunction.AddFunction(plane5)
	#clipFunction.AddFunction(plane6)

	# Slice data with the combined implicit functions: 
	cutter = vtk.vtkCutter()
	cutter.SetCutFunction(clipFunction)
	cutter.SetInputData(Object)
	cutter.Update()

	return cutter
    
###########################################################
###################### CLIP WITH BOX ######################
###########################################################

def CLIP_WITH_BOX(Object,Center,Normal,Width,Height,Angle,TWO_VECTOR_TRANSFORMATION):                              # Defines a function that clips a plane with a box 
	import vtk
	import numpy as np

	# Define box points and their normals in YZ plane:
	P1 = [0, 0.5*Width, 0]
	P2 = [0, -0.5*Width, 0]
	P3 = [0, 0, 0.5*Height]
	P4 = [0, 0, -0.5*Height]   
	N1 = [0, -1, 0]
	N2 = [0, 1, 0]
	N3 = [0, 0, -1]
	N4 = [0, 0, 1]    
    
	# Rotate the points and normals around X-axis:
	Mx = [[1,0,0],[0,np.cos(Angle*3.14159/180),-1*np.sin(Angle*3.14159/180)],[0,np.sin(Angle*3.14159/180),np.cos(Angle*3.14159/180)]]
	P1 = np.dot(Mx,P1)
	P2 = np.dot(Mx,P2)
	P3 = np.dot(Mx,P3)
	P4 = np.dot(Mx,P4)
	N1 = np.dot(Mx,N1)
	N2 = np.dot(Mx,N2)
	N3 = np.dot(Mx,N3)
	N4 = np.dot(Mx,N4)

	# Rotate the points and normals using direction vectors:                                    
	P1_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P1)
	P2_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P2)
	P3_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P3)
	P4_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,P4)
	N1_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N1)
	N2_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N2)
	N3_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N3)
	N4_rot = TWO_VECTOR_TRANSFORMATION([1,0,0],Normal,N4)

	# Translate the points:
	P1_tr = P1_rot + Center
	P2_tr = P2_rot + Center
	P3_tr = P3_rot + Center  
	P4_tr = P4_rot + Center 

	# Generate implicit plane functions:   
	plane1 = vtk.vtkPlane()
	plane1.SetOrigin(P1_tr)
	plane1.SetNormal(N1_rot)

	plane2 = vtk.vtkPlane()
	plane2.SetOrigin(P2_tr)
	plane2.SetNormal(N2_rot)

	plane3 = vtk.vtkPlane()
	plane3.SetOrigin(P3_tr)
	plane3.SetNormal(N3_rot)

	plane4 = vtk.vtkPlane()
	plane4.SetOrigin(P4_tr)
	plane4.SetNormal(N4_rot)

	# Combine the implicit plane functions: 
	clipFunction = vtk.vtkImplicitBoolean()
	clipFunction.SetOperationTypeToUnion()
	clipFunction.AddFunction(plane1)
	clipFunction.AddFunction(plane2)
	clipFunction.AddFunction(plane3)
	clipFunction.AddFunction(plane4)

	# Clip data with the combined implicit functions:    
	clipper = vtk.vtkTableBasedClipDataSet()                         # It is the faster vtk tool. It works in all ParaView versions
	clipper.SetInputData(Object)
	clipper.SetClipFunction(clipFunction)
	clipper.SetValue(0)
	clipper.GenerateClipScalarsOff()
	clipper.InsideOutOff()
	clipper.GetOutput().GetPointData().CopyScalarsOff()
	clipper.Update()

	# Converts the vtkUnstructuredGrid to vtkPolyData:
	clipper_pd = vtk.vtkGeometryFilter()                                   	                             
	clipper_pd.SetInputData(clipper.GetOutput())             
	clipper_pd.Update()

	return clipper_pd


###########################################################
#################### LINEAR EXTRUSION #####################
###########################################################
	
def LINEAR_EXTRUSION(Object,Vector,Lenght):                                                                   # Defines a function that extrudes a VTK object along a direction defined by a vector and a extrusion lenght defined by Lenght
	import vtk
	extrusion = vtk.vtkLinearExtrusionFilter()
	extrusion.CappingOff()
	extrusion.SetExtrusionTypeToVectorExtrusion()
	extrusion.SetVector(Vector[0],Vector[1],Vector[2])
	extrusion.SetInputData(Object.GetOutput())
	extrusion.SetScaleFactor(Lenght)                                      
	extrusion.Update() 
	return extrusion
    

###########################################################
##################### SURFACE NORMALS #####################
###########################################################
	
def SURFACE_NORMALS(Object,Flip_Plane_Normal):                                                              # Defines a function that computes the normals of a VTK object at the cell centers and the points
	import vtk
	Surface_Normals = vtk.vtkPolyDataNormals()
	Surface_Normals.SetInputData(Object.GetOutput())
	Surface_Normals.SetFeatureAngle(0)
	Surface_Normals.ComputeCellNormalsOn()
	Surface_Normals.ComputePointNormalsOn()
	Surface_Normals.AutoOrientNormalsOn()                                                                   # If not all the normals are pointing towards the same direction, it flips the normals of the reversed cells (so, all the normals are consistent on the entire surface)
	if Flip_Plane_Normal == 1:
		Surface_Normals.FlipNormalsOn()
	if Flip_Plane_Normal == 0:
		Surface_Normals.FlipNormalsOff()
	Surface_Normals.Update()
	return Surface_Normals


###########################################################
############# ADD NEW VARIABLES TO THE OUTPUT #############
###########################################################
	
def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName):                                         # Defines a function that adds the field data to the existing points in the Paraview's output
	import vtk
	numberOfComponents = int(numTuples)                                                                     # We deal only with 1-component arrays
	dataArray = vtk.vtkDoubleArray()   
	dataSetAttributes.AddArray(dataArray)
	dataArray.SetNumberOfComponents(numberOfComponents)
	dataArray.SetName(ArrayName)                                                                            # "ArrayName" is a string used to name the array in Paraview 
	data = DataArray                                                                                        # DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
	for value in data:
		dataArray.InsertNextValue(float(value))                                                             # Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 



############################################################################################################
###################################    DEFINITION OF CONTROL SECTIONS    ###################################
############################################################################################################

CS_centers = [Center_1,Center_2,Center_3,Center_4,Center_5,Center_6,Center_7,Center_8]                                          # Vector with the control section (CS) centers
CS_width = [Box_Size_1[0],Box_Size_2[0],Box_Size_3[0],Box_Size_4[0],Box_Size_5[0],Box_Size_6[0],Box_Size_7[0],Box_Size_8[0]]    # Vector with the control section (CS) widths
CS_height = [Box_Size_1[1],Box_Size_2[1],Box_Size_3[1],Box_Size_4[1],Box_Size_5[1],Box_Size_6[1],Box_Size_7[1],Box_Size_8[1]]   # Vector with the control section (CS) height
CS_angle =  [Angle_1,Angle_2,Angle_3,Angle_4,Angle_5,Angle_6,Angle_7,Angle_8]                                                   # Vector with the control section (CS) angle
NB_Stations = [Box_Size_2[2],Box_Size_3[2],Box_Size_4[2],Box_Size_5[2],Box_Size_6[2],Box_Size_7[2],Box_Size_8[2]]               # Vector with the number of stations (NB_Stations) in between the control sections (CS)

# Define a vector with the control section (CS) normals:
if Normals == 0:                                                                                            # 0=Aerodynamic x-axis 
	x_aero = [1,0,0]
	NORMAL = np.dot(T_inverse,x_aero)                                                                       # Aero normal expressed in the mesh reference system
	CS_normals = [NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL]                     
if Normals == 1:                                                                                            # 1=Aerodynamic y-axis 
	y_aero = [0,1,0]
	NORMAL = np.dot(T_inverse,y_aero)  
	CS_normals = [NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL]                       
if Normals == 2:                                                                                            # 2=Aerodynamic z-axis 
	z_aero = [0,0,1]
	NORMAL = np.dot(T_inverse,z_aero)
	CS_normals = [NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL,NORMAL]   
if Normals == 3:                                                                                            # 3=Mesh X-axis 
	X_mesh = [1,0,0]
	CS_normals = [X_mesh,X_mesh,X_mesh,X_mesh,X_mesh,X_mesh,X_mesh,X_mesh]   
if Normals == 4:                                                                                            # 4=Mesh Y-axis 
	Y_mesh = [0,1,0]
	CS_normals = [Y_mesh,Y_mesh,Y_mesh,Y_mesh,Y_mesh,Y_mesh,Y_mesh,Y_mesh]      
if Normals == 5:                                                                                            # 5=Mesh Z-axis 
	Z_mesh = [0,0,1]
	CS_normals = [Z_mesh,Z_mesh,Z_mesh,Z_mesh,Z_mesh,Z_mesh,Z_mesh,Z_mesh]      
if Normals == 6:                                                                                            # 6=User defined
	CS_normals = [Normal_1,Normal_2,Normal_3,Normal_4,Normal_5,Normal_6,Normal_7,Normal_8]      

# Convert to numpy arrays:
CS_normals = np.array(CS_normals)
CS_centers = np.array(CS_centers)
CS_width = np.array(CS_width)
CS_height = np.array(CS_height)
CS_angle = np.array(CS_angle)
NB_Stations = np.array(NB_Stations)

# Invert flip normals:
if Flip_Normals == 1:
	CS_normals = -1*CS_normals
    
# Normalize normals:
CS_normals = CS_normals/np.linalg.norm(CS_normals,axis=1).reshape((8,1))





DomainBounds = inputs[0].VTKObject.GetBounds()                                                              # Xmin,Xmax,Ymin,Ymax,Zmin,Zmax of the domain

Count = 0
Count2 = 0
for i in range(Control_Sections):
	Alarm = 1
	Count = Count + 1
	if DomainBounds[0] <= CS_centers[i][0] <= DomainBounds[1] and DomainBounds[2] <= CS_centers[i][1] <= DomainBounds[3] and DomainBounds[4] <= CS_centers[i][2] <= DomainBounds[5]:      # Check if: Xmin < Xcenter < Xmax
		Alarm = 0
	if Alarm ==1:        
		print ("# ERROR: control section",Count,"lies outside the data volume")
		Count2 = Count2 + 1
    
if Count2 > 0 :
	print ("# DATA BOUNDS:")
	print (DomainBounds[0],"< X <",DomainBounds[1])
	print (DomainBounds[2],"< Y <",DomainBounds[3])
	print (DomainBounds[4],"< Z <",DomainBounds[5])

############################################################################################################
##################################    DEFINITION OF CIRCULATION STATIONS   #################################
############################################################################################################

# Initialize the station parameters with STA_1 (i.e.,CS_1):
STA_centers = [CS_centers[0]]
STA_normals = [CS_normals[0]]
STA_width = [CS_width[0]]
STA_height = [CS_height[0]]
STA_angle = [CS_angle[0]]
STA_nb = 1

STA_centers = np.array(STA_centers)
STA_normals = np.array(STA_normals)
STA_width = np.array(STA_width)
STA_height = np.array(STA_height)
STA_angle = np.array(STA_angle)

# Add the other stations:
if Control_Sections > 1:
	STA_nb = Control_Sections
	for i in range(Control_Sections-1):
		# Adds the internal stations in between two control sections:
		for j in range(NB_Stations[i]):
			STA_centers = np.append(STA_centers, [CS_centers[i]+(CS_centers[i+1]-CS_centers[i])*((j+1)/(NB_Stations[i]+1))],axis=0)
			STA_normals = np.append(STA_normals, [CS_normals[i]+(CS_normals[i+1]-CS_normals[i])*((j+1)/(NB_Stations[i]+1))],axis=0)
			STA_width = np.append(STA_width, [CS_width[i]+(CS_width[i+1]-CS_width[i])*((j+1)/(NB_Stations[i]+1))],axis=0)
			STA_height = np.append(STA_height, [CS_height[i]+(CS_height[i+1]-CS_height[i])*((j+1)/(NB_Stations[i]+1))],axis=0)
			STA_angle = np.append(STA_angle, [CS_angle[i]+(CS_angle[i+1]-CS_angle[i])*((j+1)/(NB_Stations[i]+1))],axis=0)
		# Adds the external control section:
		STA_centers = np.append(STA_centers,[CS_centers[i+1]],axis=0)
		STA_normals = np.append(STA_normals,[CS_normals[i+1]],axis=0)
		STA_width = np.append(STA_width,[CS_width[i+1]],axis=0)
		STA_height = np.append(STA_height,[CS_height[i+1]],axis=0)   
		STA_angle = np.append(STA_angle,[CS_angle[i+1]],axis=0)         
		STA_nb = STA_nb + NB_Stations[i]

# Normalize normals:
STA_normals = STA_normals/np.linalg.norm(STA_normals,axis=1).reshape((STA_nb,1))


############################################################################################################
###################################    VISUALIZE CIRCULATION STATIONS    ###################################
############################################################################################################

if Select_Action == 0:                                                                                        # 0=Visualize circulation paths

	# Generate all the CS arrows:
	CS_Arrows = vtk.vtkAppendFilter()                                                                           # Output: vtkUnstructuredGrid
	for i in range(Control_Sections):
		Lenght = 0.5*(CS_width[i]+CS_height[i])                                                                 # Arrow lenght is arbitrarily defined as 0.5*(width+height)
		Arrow = PLOT_ARROW(CS_centers[i],CS_normals[i],Lenght)
		CS_Arrows.AddInputData(Arrow.GetOutput())
	CS_Arrows.Update()

	# Generate all the CS rectangles:
	CS_Rectangles = vtk.vtkAppendFilter()                                                                       # Output: vtkUnstructuredGrid
	for i in range(Control_Sections):
		Rectangle = PLOT_RECTANGLE(CS_centers[i],CS_normals[i],CS_width[i],CS_height[i],CS_angle[i],TWO_VECTOR_TRANSFORMATION,PLOT_VTK_LINES_VERTEX)
		CS_Rectangles.AddInputData(Rectangle.GetOutput())
	CS_Rectangles.Update()

	# Generate all the STA rectangles:
	STA_Rectangles = vtk.vtkAppendFilter()                                                                       # Output: vtkUnstructuredGrid
	for i in range(STA_nb):
		Rectangle = PLOT_RECTANGLE(STA_centers[i],STA_normals[i],STA_width[i],STA_height[i],STA_angle[i],TWO_VECTOR_TRANSFORMATION,PLOT_VTK_LINES_VERTEX)
		STA_Rectangles.AddInputData(Rectangle.GetOutput())
	STA_Rectangles.Update()

	# convert to vtkPolyData:
	CS_Rectangles_PD = vtk.vtkGeometryFilter()                                                                  # Converts vtkUnstructuredGrid to vtkPolyData  
	CS_Rectangles_PD.SetInputData(CS_Rectangles.GetOutput())              
	CS_Rectangles_PD.Update()

	# Add tubes to the CS rectangles:
	Tube_Radius = 0.05*np.min((np.min(STA_width),np.min(STA_height)))
	CS_tube = vtk.vtkTubeFilter()
	CS_tube.SetInputData(CS_Rectangles_PD.GetOutput())
	CS_tube.SetRadius(Tube_Radius)
	CS_tube.SetNumberOfSides(50)
	CS_tube.Update()

	# Join arrows and CS rectangles:                                                  
	appendFilter = vtk.vtkAppendFilter()
	appendFilter.AddInputData(CS_Arrows.GetOutput())
	appendFilter.AddInputData(CS_tube.GetOutput())
	if Show_Circulation_Paths == 1:
		appendFilter.AddInputData(STA_Rectangles.GetOutput())
	appendFilter.Update()

	# convert to vtkPolyData:
	Control_Sections = vtk.vtkGeometryFilter()                                                                  # Converts vtkUnstructuredGrid to vtkPolyData  
	Control_Sections.SetInputData(appendFilter.GetOutput())              
	Control_Sections.Update()

	self.GetOutput().ShallowCopy(Control_Sections.GetOutput()) 



############################################################################################################
######################################    GET BODY SURFACE SECTIONS   ######################################
############################################################################################################

if Lifting_Line_Method == 1:                                                                   

	if Chord_Law == 1:                                                                                      # 1=Variable chord

		# Get body surface:
		Body = LOAD_VTK_DATA(Load_Body_Surface)

		# Get body section at each station:
		STA_sections = []
		append_sections = vtk.vtkAppendFilter()
		for i in range(STA_nb):
			Body_Section = SLICE_WITH_PLANE(Body.GetOutput(),STA_centers[i],STA_normals[i])
			STA_sections.append(Body_Section)
			append_sections.AddInputData(Body_Section.GetOutput())

		append_sections.Update()

		# convert to vtkPolyData:
		vtk_body_Sections = vtk.vtkGeometryFilter()                                                         # Converts vtkUnstructuredGrid to vtkPolyData  
		vtk_body_Sections.SetInputData(append_sections.GetOutput())              
		vtk_body_Sections.Update()

		# Display body sections:
		if Select_Action == 0:                                                                              # 0=Visualize circulation paths
			if Show_Body_Sections == 1:
				self.GetOutput().ShallowCopy(vtk_body_Sections.GetOutput())

		# Get chord of each section:
		STA_chords = []
		for i in range(STA_nb):
			Section_NBpts = STA_sections[i].GetOutput().GetNumberOfPoints()
			if Section_NBpts == 0:
				STA_chords.append(0)
			else:
				Sections_Py = dsa.WrapDataObject(STA_sections[i].GetOutput())                               # Converts the VTK object into a Python object     
				Section_Points = Sections_Py.Points                                                         # Get the points selected by the user
				Xmin = np.min(Section_Points[:,0])
				ID_xmin = np.argwhere(Section_Points[:,0]==Xmin)
				FWD_point = Section_Points[ID_xmin[0][0]]
				Xmax = np.max(Section_Points[:,0])
				ID_xmax = np.argwhere(Section_Points[:,0]==Xmax)
				REAR_point = Section_Points[ID_xmax[0][0]]   
				Chord = np.linalg.norm(FWD_point-REAR_point)
				STA_chords.append(Chord)

	if Chord_Law == 0:                                                                                      # 0=Constant chord 
		STA_chords = []
		for i in range(STA_nb):
			STA_chords.append(Chord)        

	STA_chords = np.array(STA_chords)

############################################################################################################
############################################    INTEGRATE DATA   ###########################################
############################################################################################################

if Select_Action == 1:                                                                                        # 1=Integrate data
	
	# Load input data and filter data:
	Data = inputs[0]

	Data_Filtered = vtk.vtkPassArrays()
	Data_Filtered.SetInputDataObject(Data.VTKObject)
	Data_Filtered.ClearArrays()
	Data_Filtered.AddPointDataArray("AERO_Velocity_Vector")
	Data_Filtered.AddPointDataArray("AERO_Vorticity")
	Data_Filtered.AddCellDataArray("")
	Data_Filtered.AddFieldDataArray("") 
	Data_Filtered.Update() 

	STA_Gamma = []
	
	for i in range(STA_nb):
		if Mode == 0:                                                                                       # 0=3D
			Data_Slice = SLICE_WITH_PLANE(Data_Filtered.GetOutput(),STA_centers[i],STA_normals[i])
			if Integral_Type == 0:                                                                          # 0=Velocity circulation
				Path = SLICE_WITH_BOX(Data_Slice.GetOutput(),STA_centers[i],STA_normals[i],STA_width[i],STA_height[i],STA_angle[i],TWO_VECTOR_TRANSFORMATION)
			if Integral_Type == 1:                                                                          # 1=Vorticity flux
				Path = CLIP_WITH_BOX(Data_Slice.GetOutput(),STA_centers[i],STA_normals[i],STA_width[i],STA_height[i],STA_angle[i],TWO_VECTOR_TRANSFORMATION) 
    
		if Mode == 1:                                                                                       # 1=2D planar
			if Integral_Type == 0:                                                                          # 0=Velocity circulation
				Path = SLICE_WITH_BOX(Data_Filtered.GetOutput(),STA_centers[i],STA_normals[i],STA_width[i],STA_height[i],STA_angle[i],TWO_VECTOR_TRANSFORMATION)
			if Integral_Type == 1:                                                                          # 1=Vorticity flux
				Path = CLIP_WITH_BOX(Data_Filtered.GetOutput(),STA_centers[i],STA_normals[i],STA_width[i],STA_height[i],STA_angle[i],TWO_VECTOR_TRANSFORMATION)


		# Compute circulation by vorticity flux:
		if Integral_Type == 1:
			Path_Normals = SURFACE_NORMALS(Path,Flip_Normals)

			# Point data to cell data:  
			p2c = vtk.vtkPointDataToCellData()     							                                # Converting the input from point Data To cell Data
			p2c.SetInputData(Path_Normals.GetOutput())
			p2c.Update()

			# convert to Python object:
			WrapCellData = dsa.WrapDataObject(p2c.GetOutput())                                              # Converts vtk object into Python object

			# Compute circulation:
			Vorticity = WrapCellData.CellData["AERO_Vorticity"]
			N = WrapCellData.CellData["Normals"]
			if Mode == 0:
				N = -1*N           
			if Mode == 1:
				if Flip_Normals == 0:
					N = -1*N

			Area = area(WrapCellData) 
			VORTdotN = np.einsum('ij,ij->i',Vorticity,N)                                                         # Dot product
			Circulation = np.sum(VORTdotN*Area)     
			STA_Gamma.append(Circulation)


		# Compute circulation by using graphical method:
		if Graphical_Method == 1 and Integral_Type == 0:                                                                           # It computes circulation by extruding the path and then computing circulation at the cell centers of the extruded surface. Robust and faster method than the line approach (i.e., by definition)

			Path_Extruded = LINEAR_EXTRUSION(Path,STA_normals[i],1)                                         # Extrusion lenght set to 1 in order to simplify the integral
			Path_Normals = SURFACE_NORMALS(Path_Extruded,1)

			# Point data to cell data:  
			p2c = vtk.vtkPointDataToCellData()     							                                # Converting the input from point Data To cell Data
			p2c.SetInputData(Path_Normals.GetOutput())
			p2c.Update()

			# convert to Python object:
			WrapCellData = dsa.WrapDataObject(p2c.GetOutput())                                              # Converts vtk object into Python object

			# Compute circulation:
			V = WrapCellData.CellData["AERO_Velocity_Vector"]                                               # Get some cell-based arrays
			N = WrapCellData.CellData["Normals"]
			Area = area(WrapCellData) 
			Path_PointsNB = np.shape(N)[0]
			Path_Norm_Cell = np.tile(STA_normals[i],(Path_PointsNB,1))
			T = algs.cross(N,Path_Norm_Cell)
			VdotT = np.einsum('ij,ij->i',V,T)                                                         # Dot product
			Circulation = np.sum(VdotT*Area)        
			STA_Gamma.append(Circulation)



		# Compute circulation by using definition:
		if Graphical_Method == 0 and Integral_Type == 0:                                                                           # It computes circulation by summing up the contribution of each line of the path to the total circulation. It iterates at each path line to compute Gamma, then it is slow. Also, when the clip do not generates a perfect square, the circulation value is not correct.

			# Point data to cell data:  
			p2c = vtk.vtkPointDataToCellData()     							                                    # Converting the input from point Data To cell Data
			p2c.SetInputData(Path.GetOutput())
			p2c.Update()

			# convert to Python object:
			WrapCellData = dsa.WrapDataObject(p2c.GetOutput())                                                  # Converts vtk object into Python object
			V = WrapCellData.CellData["AERO_Velocity_Vector"]                                                   # Reads the Velocity array     

			# Iterate at each cell of the path:
			T = []
			Gamma = []
			for j in range(WrapCellData.GetNumberOfCells()):
			
				# Compute tangent vectors:
				cell = WrapCellData.GetCell(j)      #Path.GetOutput().GetCell(j)
				Cell_Point_1 = Path.GetOutput().GetPoint(cell.GetPointId(0))
				Cell_Point_2 = Path.GetOutput().GetPoint(cell.GetPointId(1))
				Cell_Point_1 = np.array(Cell_Point_1)
				Cell_Point_2 = np.array(Cell_Point_2)
				Cell_T = Cell_Point_2-Cell_Point_1                                                              # Line tangent
				Cell_lenght = np.linalg.norm(Cell_T)
				Cell_center = 0.5*(Cell_Point_1+Cell_Point_2)
				R = Cell_center-STA_centers[i]                                                                  # Distance vector from path center to cell center
				Mom = np.cross(R,Cell_T)                                                                        # Moment generated by the tangent vector T respect to the path center
				Check = np.dot(STA_normals[i],Mom)                                                              # Verifies that the tangent moment follows the circulation direction
				if Check < 0:                                                                                   # If circulation direction is reversed, then, the tangent is corrected
					Cell_T = -1*Cell_T
				Cell_T = Cell_T/np.linalg.norm(Cell_T)                                                          # Normalized tangent vector
				T.append([Cell_T[0],Cell_T[1],Cell_T[2]])

				# Compute circulation (Gamma=V.T dl) at each cell:
				Cell_V = V[j]
				Cell_gamma = np.dot(Cell_V,Cell_T)*Cell_lenght   
				Gamma.append(Cell_gamma)           

			T = np.array(T)
			Gamma = np.array(Gamma)

			# Compute circulation and local lift force:
			Circulation = np.sum(Gamma)
			STA_Gamma.append(Circulation)

	STA_Gamma = np.array(STA_Gamma)
	STA_lines,STA_points = PLOT_VTK_LINES_VERTEX(STA_centers)

	# Add data to the output vtk object:      
	addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_Gamma,"STA_Gamma") 


	if Mode == 1:					# 0=2D planar
		Lift = Rho0*V0*STA_Gamma 
		Cl = Lift/(0.5*Rho0*V0**2*Sref) 
		print ("Circulation =",STA_Gamma[0]," [m^2/s]")
		print ("Lift =",Lift[0]," [N/m]")
		print ("Cl =",Cl[0])
		addfielddata(1, STA_lines.GetOutput().GetFieldData(),STA_Gamma,"Circulation")
		addfielddata(1, STA_lines.GetOutput().GetFieldData(),Lift,"Lift")
		addfielddata(1, STA_lines.GetOutput().GetFieldData(),Cl,"Cl")

	if Mode == 0:					# 0=3D
		if Lifting_Line_Method == 1: 
			# Add data to the output vtk object:
			addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_chords,"STA_chords")

			# Compute lift coefficient:        
			ID_No_body = numpy.where(STA_chords == 0)                       # It detects the path sections that do not intersect the body
			STA_chords[ID_No_body] = 1                                      # The local zero chord values are set to 1 in order to avoid any singularity
			STA_Cl = 2*STA_Gamma/(V0*STA_chords)
			STA_Cl[ID_No_body] = 0
			STA_Lift = 0.5*Rho0*V0**2*STA_chords*STA_Cl


			# Add data to the output vtk object:  
			addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_Cl,"STA_Cl")     
			addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_Lift,"STA_Lift") 


			# Additional Lifting Line computations:  
			if Normals != 1:                                                                                        # 1=Aerodynamic y-axis
				print('# Lifting line Method enabled --> Set "Normals"="Aerodynamic y-axis" to compute induced drag distribution')
			if Normals == 1:

				# Compute global Lift & Clw:
				Lift = []
				for i in range(STA_nb-1):
					Lift_area = 0.5*(STA_Lift[i]+STA_Lift[i+1])*np.abs(STA_centers[i+1,1]-STA_centers[i,1])
					Lift.append(Lift_area)
				Lift = np.sum(Lift)
				Clw = Lift/(0.5*Rho0[0]*V0[0]**2*Sref[0])
				print("Lift =",Lift,"[N]") 
				print("Clw =",Clw)             
				addfielddata(1, STA_lines.GetOutput().GetFieldData(),[Lift],"Lift")
				addfielddata(1, STA_lines.GetOutput().GetFieldData(),[Clw],"Clw")


				# Determine (dGamma/Dy) along the span:
				STA_DgammaDy = []
				for i in range(STA_nb): 
					if i == 0:
						GammaSlope = (STA_Gamma[i+1]-STA_Gamma[i])/(STA_centers[i+1,1]-STA_centers[i,1])
					if i == STA_nb-1:
						GammaSlope = (STA_Gamma[i]-STA_Gamma[i-1])/(STA_centers[i,1]-STA_centers[i-1,1])
					else:                   
						GammaSlope = (STA_Gamma[i+1]-STA_Gamma[i-1])/(STA_centers[i+1,1]-STA_centers[i-1,1])
					STA_DgammaDy.append(GammaSlope)
				addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_DgammaDy,"STA_DgammaDy")  

				# Determine induced velocity, induced angle and induced drag spanwise distributions:
				STA_W_ind = []


				for i in range(STA_nb):
					W_ind_y0 = []
					for j in range(STA_nb-1):
						if i != j:
							Integral = (-1/(4*3.14159))*0.5*(STA_DgammaDy[j]+STA_DgammaDy[j+1])*np.abs(STA_centers[j+1,1]-STA_centers[j,1])*(1/(STA_centers[i,1]-STA_centers[j,1]))
							W_ind_y0.append(Integral)
					W_ind_y0 = np.sum(W_ind_y0)
					STA_W_ind.append(W_ind_y0)


				STA_W_ind = np.array(STA_W_ind)                
				STA_Alpha_ind = -1*STA_W_ind/V0
				STA_Alpha_ind = np.array(STA_Alpha_ind)
				STA_Drag_ind = Rho0*V0*STA_Gamma*STA_Alpha_ind
				STA_Cdi = 2*STA_Gamma*STA_Alpha_ind/(V0*Sref)
				addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_W_ind,"STA_W_ind") 
				addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_Alpha_ind*(180/3.14159),"STA_Alpha_ind") # Induced angle in degrees
				addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_Drag_ind,"STA_Drag_ind") 
				addfielddata(1, STA_lines.GetOutput().GetPointData(),STA_Cdi,"STA_Cdi")  

				# Compute global Cdi and induced drag:
				Cdi = []
				for i in range(STA_nb-1):
					Cdi_area = 0.5*(np.abs(STA_Cdi[i])+np.abs(STA_Cdi[i+1]))*np.abs(STA_centers[i+1,1]-STA_centers[i,1])
					Cdi.append(Cdi_area)
				Cdi = np.sum(Cdi)
				Drag_ind = 0.5*Rho0[0]*V0[0]**2*Sref[0]*Cdi
				print("Drag_ind =",Drag_ind,"[N]") 
				print("Cdi =",Cdi) 
				addfielddata(1, STA_lines.GetOutput().GetFieldData(),[Drag_ind],"Drag_ind")
				addfielddata(1, STA_lines.GetOutput().GetFieldData(),[Cdi],"Cdi")

	# Show vtk object:
	self.GetOutput().ShallowCopy(STA_lines.GetOutput())  



############################################################################################################
#############################################    END OF CODE   #############################################
############################################################################################################