# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################               Grid Generator Plugin              #############################
############################################################################################################

# Abstract:
This Paraview's programmable source provides several tools for wind tunnel wake survey systems:

a) Probe trayectory calculation: creates a TXT file with the coordinates of the probe to control the traverser system
b) Survey time calculator: determines the time required by the wake survey based on the number of points, traversing 
speed, hold time and data acquisition time.
c) Creates a VTP mesh where the wake survey data will be appended in Epsilon_Wake_Survey


# IMPORTANT: accepted file format for the mesh input data are Tecplot ".dat" (for raw PIV files) and VTP/VTU/VTS (for 
PIV data from Epsilon_PIV as well as CFD data slices previously created in Paraview). Note that VTP/VTU/VTS files 
from Paraview are only accepted and NOT vtm (vtk multiblock file). Hence, prior to saving a CFD slice data into 
vtp/vtu/vts format, a merge block filter must be applied (converts multiblock to single block). Moreover, if any 
other single block data format is being used in Paraview, the simplest way to convert this format into vtp is by 
using the filter "Extract Surface" (it converts any file format into a polygonal mesh --> vtp file).

# IMPORTANT: the user must provide data already expressed in the Aerodynamic Reference System. Local reference 
system like the in-plane XY PIV reference system is not allowed or may lead to errors in the calculations. In order
to change the reference system of PIV data, use Epsilon_PIV. To change the reference system of CFD data, use Epsilon_Aero_Data

# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""

print ("")
print ("")
print ('#########################################################################################################################################')
print ('##################################                                                                     ##################################')
print ('##################################                              EPSILON 3.0                            ##################################')
print ('##################################                        Aerodynamic analysis tool                    ##################################')
print ('##################################                                                                     ##################################')
print ('#########################################################################################################################################')
print ("")
print ('Developed at DAEP/ISAE-SUPAERO, France, 2022')
print ("")
print ('#DISCLAIMER:')
print ('This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee of any kind, either express')
print ('or implied. Its use for teaching, research and industrial applications is allowed under the strict responsibility of the user. The authors')
print ('of this plugin are not responsible for any damages or errors related to the use of the plugin.')

print ("")
print ("")
print ("")
print ("####################################################")
print ("##########      EPSILON GRID GENERATOR     #########")
print ("####################################################")
print ("")

############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

import vtk
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa
import numpy as np
import vtk.vtkFiltersParallel
import paraview.vtk.util.numpy_support as vnp
import glob
import os
import os.path
from vtk.util import numpy_support



############################################################################################################
############################################# STRING TO FLOAT ##############################################
############################################################################################################

# Input parameters are defined as string in Paraview GUI, thus a string to float convertion is required:
Mode = float(Mode)                                          # 0=Survey line generator  1=Mapped survey plane generator   2=Wake-shape mapped grid generator
PhaseGUI = float(Phase)  										# 0=None  1=Generate line   2=Export TXT file
StageGUI = float(Stage)   										# 0=None    1=Generate mesh   2=Visualize & export probe path
StepGUI = float(Step)                                          # 0=None   1=Visualize input data    2=Clip wake region  3=Define offset margin  4=Check tunnel walls  5=Compute mesh  6=Check traverser path

if Mode == 0:
	Phase = PhaseGUI
	Stage = 0
	Step  = 0
if Mode == 1:
	Phase = 0
	Stage = StageGUI
	Step  = 0
if Mode == 2:
	Phase = 0
	Stage = 0
	Step  = StepGUI

Mode = 0                                                    # 0=Mesh generation




Refinement_1_GUI = float(Refinement_1)                          # 0=None   1=X-axis   2=Y-axis   3=Z-axis
Refinement_2_GUI = float(Refinement_2)                          # 0=None   1=Y-axis   2=Z-axis   3=Z-axis
Refinement_3_GUI = float(Refinement_3)                          # 0=None   1=Y-axis   2=Z-axis   3=Z-axis
Refinement_4_GUI = float(Refinement_4)                          # 0=None   1=Y-axis   2=Z-axis   3=Z-axis
Refinement_5_GUI = float(Refinement_5)                          # 0=None   1=Y-axis   2=Z-axis   3=Z-axis

if Add_Refinement == 0:
	Refinement_1 = 0
	Refinement_2 = 0
	Refinement_3 = 0
	Refinement_4 = 0
	Refinement_5 = 0

if Add_Refinement == 1:
	Refinement_1 = Refinement_1_GUI
	Refinement_2 = 0
	Refinement_3 = 0
	Refinement_4 = 0
	Refinement_5 = 0

if Add_Refinement == 2:
	Refinement_1 = Refinement_1_GUI
	Refinement_2 = Refinement_2_GUI
	Refinement_3 = 0
	Refinement_4 = 0
	Refinement_5 = 0

if Add_Refinement == 3:
	Refinement_1 = Refinement_1_GUI
	Refinement_2 = Refinement_2_GUI
	Refinement_3 = Refinement_3_GUI
	Refinement_4 = 0
	Refinement_5 = 0

if Add_Refinement == 4:
	Refinement_1 = Refinement_1_GUI
	Refinement_2 = Refinement_2_GUI
	Refinement_3 = Refinement_3_GUI
	Refinement_4 = Refinement_4_GUI
	Refinement_5 = 0
    
if Add_Refinement == 5:
	Refinement_1 = Refinement_1_GUI
	Refinement_2 = Refinement_2_GUI
	Refinement_3 = Refinement_3_GUI
	Refinement_4 = Refinement_4_GUI
	Refinement_5 = Refinement_5_GUI
    
    

File_Path_Mesh = str(File_Path_Mesh)                        # File path to the input data (wake mesh, PIV or CFD slice)  
File_Path_Output = str(File_Path_Output)     				# File path to the folder where to store the output data (Survey mesh VTP, survey path VTP and TXT for traverser)
StartXYZ = StartXYZ                                         # X-, Y- and Z-coordinates of the starting point of the survey line
EndXYZ = EndXYZ                                             # X-, Y- and Z-coordinates of the ending point of the survey line
NBpts = int(NBpts)                                          # Number of points along the survey line
Orientation = float(Orientation)                            # 1=YZ plane	2=XZ plane	3=XY plane
IaxisInfo = IaxisInfo                                       # Vector containing the minimum I-coordinate, maximum I-coordinate and number of points along I-axis
JaxisInfo = JaxisInfo                                       # Vector containing the minimum J-coordinate, maximum J-coordinate and number of points along J-axis
Rotation = Rotation                                         # Rotation angles about the X-, Y- and Z-axes
NormalDistance = float(NormalDistance)                      # Shift plane along its normal distance
Clip_parameter = str(Clip_parameter)                        # Name of the variable selected to perform the wake clip
Clip_Value = float(Clip_Value)                              # Parameter threshold value for the clip
Invert_clip = float(Invert_clip)                            # 0=No  1=Yes
Offset_margin = float(Offset_margin)                        # Offset distance from wake boundary to mesh boundary in [m]
Wind_Tunnel_Model = float(Wind_Tunnel_Model)                # 0=None   1=Full model  1=Half model Y+   2=Half model Y-
Tunnel_Width = Test_section_size[0]                         # Test section width in [m]               
Tunnel_Height = Test_section_size[1]                        # Test section height in [m]
Safety_distance_to_wall = float(Safety_distance_to_wall)    # minimum distance from probe to wall in [m]
Z_mesh_method = float(Z_mesh_method)                        # 0=Constant Z-points   1=Proportional to wake thickness
Wake_thickness = float(Wake_thickness)                      # 0=automatic detection of min/max values    1=user defined
Y_min_wake = Y_min_max_wake[0]                              # Y-coordinate of the minimum wake thichness
Y_max_wake = Y_min_max_wake[1]                              # Y-coordinate of the maximum wake thichness
Mesh_Zpts_min = Z_points_min_max[0]                         # Minimum number of points along Z-axis across the wake
Mesh_Zpts_max = Z_points_min_max[1]                         # Maximum number of points along Z-axis across the wake
Y_points_baseline = float(Y_points_baseline)                # Number of Y-points for the regular mesh used as a canvas for the mesh
Hide_Wake = float(Hide_Wake)                				# Boolean condition: 0=show survey points, path lines and wake surface      1=hide wake
Ref_1_coord = Ref_1_CW[0]                                   # Coordinate in [m] of the center of the refinement region "1"
Ref_1_width = Ref_1_CW[1]                                   # Width in [m] of the refinement region "1"
Ref_1_point = Ref_1_PS[0]                                   # Number of points of the refinement region "1"
Ref_1_sigma = Ref_1_PS[1]                                   # Normal distribution parameter of the refinement region "1"
Ref_2_coord = Ref_2_CW[0]                                   # Coordinate in [m] of the center of the refinement region "2"
Ref_2_width = Ref_2_CW[1]                                   # Width in [m] of the refinement region "2"
Ref_2_point = Ref_2_PS[0]                                   # Number of points of the refinement region "2"
Ref_2_sigma = Ref_2_PS[1]                                   # Normal distribution parameter of the refinement region "2"
Ref_3_coord = Ref_3_CW[0]                                   # Coordinate in [m] of the center of the refinement region "3"
Ref_3_width = Ref_3_CW[1]                                   # Width in [m] of the refinement region "3
Ref_3_point = Ref_3_PS[0]                                   # Number of points of the refinement region "3"
Ref_3_sigma = Ref_3_PS[1]                                   # Normal distribution parameter of the refinement region "3"
Ref_4_coord = Ref_4_CW[0]                                   # Coordinate in [m] of the center of the refinement region "4"
Ref_4_width = Ref_4_CW[1]                                   # Width in [m] of the refinement region "4"
Ref_4_point = Ref_4_PS[0]                                   # Number of points of the refinement region "4"
Ref_4_sigma = Ref_4_PS[1]                                   # Normal distribution parameter of the refinement region "4"
Ref_5_coord = Ref_5_CW[0]                                   # Coordinate in [m] of the center of the refinement region "5"
Ref_5_width = Ref_5_CW[1]                                   # Width in [m] of the refinement region "5"
Ref_5_point = Ref_5_PS[0]                                   # Number of points of the refinement region "5"
Ref_5_sigma = Ref_5_PS[1]                                   # Normal distribution parameter of the refinement region "5"
Survey_speed = float(Survey_speed)                          # Speed in [m/s] of the wind tunnel traversing system
Hold_time = float(Hold_time)                                # Time in [s] required for the multi-hole pressure ports to reach a stable pressure condition prior to the measurement
DAQ_time = float(DAQ_time)                                  # Data acquisition time in [s]
Output_TXT_in_mm = float(Output_TXT_in_mm)                  # 0=output in [m]   1=Output in [mm]
Output_TXT_without_headers = float(Output_TXT_without_headers)    # 0=headers included    1=without headers
SelectTXTdelimiter = float(SelectTXTdelimiter)    			# 0=TAB  1=CSV   2=Space
DecimalSeparator = float(DecimalSeparator)  				# 0=point   1=comma   2=semicolon
ReturnToStartingPoint = float(ReturnToStartingPoint)     	# 0=Do not return to starting point  1=return to starting point
ReturnToOrigin = float(ReturnToOrigin)   					# 0=Do not return to origin  1=return to origin
OriginCoordinates = OriginCoordinates   					# XYZ coordinates of the traverser origin

Z_points_Constant = Z_points_Constant                       # NB of points for mesh type of constant NB of z points


if Z_mesh_method == 0:
	Mesh_Zpts_min = Z_points_Constant


############################################################################################################
#######################################    DEFINITION OF FUNCTIONS    ######################################
############################################################################################################

###################################
#####  READ TECPLOT FUNCTION  #####
###################################

def READ_TECPLOT(File_Path_Mesh):
	import vtk
	MBTecplotReader = vtk.vtkTecplotReader()                                    # "vtkTecplotReader" gives a vtkMultiblockDataSet output that contains several blocks of vtkStructuredGrid type
	MBTecplotReader.SetFileName(File_Path_Mesh)
	MBTecplotReader.Update()
	
	InputMesh = vtk.vtkGeometryFilter()                                         # Converts vtkMultiblockDataSet to vtkPolyData
	InputMesh.SetInputData(MBTecplotReader.GetOutput().GetBlock(0))             # "MBTecplotReader" is vtkMultiblockDataSet type. With "GetBlock(0)" it gets the only PIV block of the data (it is vtkStructuredGrid type). Then "vtkGeometryFilter" converts it to vtkPolyData
	InputMesh.Update()
	#self.GetOutput().ShallowCopy(InputMesh.GetOutput())                        # It displays the mesh. Kept here for debugging purposes
	return InputMesh

###################################
#######  READ VTP FUNCTION  #######
###################################

def READ_VTP(File_Path_Mesh):
	import vtk
	InputMesh = vtk.vtkXMLPolyDataReader()                                      # Alternative reader: vtkPolyDataReader
	InputMesh.SetFileName(File_Path_Mesh)
	InputMesh.Update()                                                          # "InputMesh" is a vtkPolyData variable
	#output.ShallowCopy(InputMesh.GetOutput())
	return InputMesh
	

###################################
#######  READ VTU FUNCTION  #######
###################################

def READ_VTU(File_Path_Mesh):
	import vtk
	VTUreader = vtk.vtkXMLUnstructuredGridReader()                              # "vtkXMLUnstructuredGridReader" gives a vtkUnstructuredGrid output type
	VTUreader.SetFileName(File_Path_Mesh)
	VTUreader.Update()

	InputMesh = vtk.vtkGeometryFilter()                                         # Converts the vtkUnstructuredGrid to vtkPolyData
	InputMesh.SetInputData(VTUreader.GetOutput())
	InputMesh.Update()
	#self.GetOutput().ShallowCopy(InputMesh.GetOutput())                        # It displays the mesh. Kept here for debugging purposes
	return InputMesh

	
###################################
#######  READ VTS FUNCTION  #######
###################################

def READ_VTS(File_Path_Mesh):
	import vtk
	VTSreader = vtk.vtkXMLStructuredGridReader()                                # "vtkXMLStructuredGridReader" gives a vtkStructuredGrid output type
	VTSreader.SetFileName(File_Path_Mesh)
	VTSreader.Update()
	
	InputMesh = vtk.vtkGeometryFilter()                                         # Converts the vtkStructuredGrid to vtkPolyData
	InputMesh.SetInputData(VTSreader.GetOutput())
	InputMesh.Update()
	#self.GetOutput().ShallowCopy(InputMesh.GetOutput())                        # It displays the mesh. Kept here for debugging purposes
	return InputMesh

	
############################################
############  ROTATE FUNCTION  #############
############################################

def ROTATE(InputVTK, Rotation):
	import vtk

	RotX = vtk.vtkTransform()                                               	# Generates the VTK transformation function to rotate the PIV data about the x-axis
	RotX.RotateX(Rotation[0])
	RotX.Update()

	OutputVTK_Rx = vtk.vtkTransformFilter()                                    	# Applies the rotation to the PIV data and creates a new transformed output
	OutputVTK_Rx.SetTransform(RotX)
	OutputVTK_Rx.SetInputData(InputVTK.GetOutput())
	OutputVTK_Rx.Update()

	RotY = vtk.vtkTransform()                                                 	# Generates the VTK transformation function to rotate the PIV data about the y-axis
	RotY.RotateY(Rotation[1])
	RotY.Update()

	OutputVTK_Ry = vtk.vtkTransformFilter()                                   	# Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
	OutputVTK_Ry.SetTransform(RotY)
	OutputVTK_Ry.SetInputData(OutputVTK_Rx.GetOutput())
	OutputVTK_Ry.Update()

	RotZ = vtk.vtkTransform()                                               	# Generates the VTK transformation function to rotate the PIV data about the z-axis
	RotZ.RotateZ(Rotation[2])
	RotZ.Update()

	OutputVTK_Rz = vtk.vtkTransformFilter()                                   	# Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
	OutputVTK_Rz.SetTransform(RotZ)
	OutputVTK_Rz.SetInputData(OutputVTK_Ry.GetOutput())
	OutputVTK_Rz.Update()

	return OutputVTK_Rz


############################################
############ TRANSLATE FUNCTION ############
############################################

def TRANSLATE(InputVTK, Translation):
	import vtk
	
	Transform = vtk.vtkTransform()                                           	# Generates the VTK transformation function to translate the PIV data along the x-, y- and z-axes
	Transform.Translate(Translation[0], Translation[1], Translation[2])
	Transform.Update()

	OutputVTK_T = vtk.vtkTransformFilter()                                    	# Applies the translation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
	OutputVTK_T.SetTransform(Transform)
	OutputVTK_T.SetInputData(InputVTK.GetOutput())
	OutputVTK_T.Update()

	return OutputVTK_T

	
############################################
########## SCALE GEOMETRY FUNCTION #########
############################################

def SCALE_GEOMETRY(InputVTK, LenghtScaleFactor):
	import vtk

	transform1 = vtk.vtkTransform()                                           	# Generates the VTK transformation function to scale up/down the PIV data
	transform1.Scale(LenghtScaleFactor, LenghtScaleFactor, LenghtScaleFactor)
	transform1.Update()

	OutputVTK_T1 = vtk.vtkTransformFilter()                                  	# Applies the scale up/down to the input PIV data and creates a transformed output
	OutputVTK_T1.SetTransform(transform1)
	OutputVTK_T1.SetInputData(InputVTK.GetOutput())
	OutputVTK_T1.Update()
	
	return OutputVTK_T1


############################################
######### ADD FIELD DATA FUNCTION ##########
############################################

def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName):             # Defines a function that adds the field data to the existing points in the Paraview's output
	numberOfComponents = int(numTuples)                                       	# We deal only with 1-component arrays
	dataArray = vtk.vtkDoubleArray()   
	dataSetAttributes.AddArray(dataArray)
	dataArray.SetNumberOfComponents(numberOfComponents)
	dataArray.SetName(ArrayName)                                               	# "ArrayName" is a string used to name the array in Paraview 
	data = DataArray                                                        	# DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
	for value in data:
		dataArray.InsertNextValue(float(value))                              	# Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 
	

############################################
########### DRAW LINES FUNCTION ############
############################################

def Line_Function(x0,y0,z0,x1,y1,z1,resolution):
	import vtk
	Line_F=vtk.vtkLineSource()
	Line_F.SetPoint1(x0,y0,z0)
	Line_F.SetPoint2(x1,y1,z1)
	Line_F.SetResolution(resolution)
	Line_F.Update()
	return Line_F 


############################################
######### GRID REFINEMENT FUNCTION #########
############################################

def refine_by_normal_distribution(Y_points,Limit_min,Mimit_max,Ref_coord,Ref_width,Ref_point,Ref_sigma):  # LOCAL MESH REFINEMENT WITH NORMAL DISTRIBUTION:
	import numpy as np
	xpts = np.linspace(0,3*Ref_sigma,Ref_point)                                 # Creates points on the RHS of the Gauss curve with 3-sigma range
	Normal = np.exp(-0.5*((xpts)/Ref_sigma)**2)/(Ref_sigma*np.sqrt(2*np.pi))    # Computes the normal distribution for these x-points 
	Base = np.linspace(0,Ref_width,Ref_point)                                   # Creates an uniform base point distribution within a desired Y-range
	Delta = (Base[0]-Base[1])*(1-Normal)                                        # Determines the distance between points that follows a normal distribution (It is still the RHS of the Gauss curve)
	
	Y_points_normal = np.array([Ref_coord])                                     # Creates a normal point distribution by adding points at both sides of the center of the refinement, by respecting the previous distance between points
	for i in range(Ref_point-1):
		Y_points_normal = np.append(Y_points_normal,Ref_coord + np.sum(Delta[:(i+1)]))                  # It adds points on the right 
		Y_points_normal = np.insert(Y_points_normal, 0, Ref_coord - np.sum(Delta[:(i+1)]), axis=0)      # It adds points on the left
	
	Y_points_normal = np.fliplr([Y_points_normal])[0]                           # It flips the resulting array in order to have the smaller values at the beginning of the array and the higher values at the end
	
	Y_points_normal = Y_points_normal[Y_points_normal>(Ref_coord-0.5*Ref_width)]  # It clips the distribution in order to retain the points inside the desired refinement region
	Y_points_normal = Y_points_normal[Y_points_normal<(Ref_coord+0.5*Ref_width)]

	Mixed_Y = np.concatenate((Y_points[Y_points<(Ref_coord-0.5*Ref_width)],Y_points_normal, Y_points[Y_points>(Ref_coord+0.5*Ref_width)] ), axis=0) # It takes the reference uniform mesh, then it take away its points belonging to the Min_value<y<max_value range, and then it adds the normal distribution points on this gap. So we have a uniform mesh with a local refinement;

	Mixed_Y = Mixed_Y[Mixed_Y>=Limit_min]                                       #Delete the points lying outside the wake contour line
	Mixed_Y = Mixed_Y[Mixed_Y<=Mimit_max]

	return Mixed_Y


############################################
##########  CREATE GRID FUNCTION  ##########
############################################

def CREATE_GRID(PointsX,PointsY,PointsZ):
	import vtk
	import numpy as np

	xCoords = vtk.vtkFloatArray()                                            	# Creates a VTK container of the X-point coordinates
	for x, i in enumerate(PointsX):     
		xCoords.InsertNextValue(i)

	yCoords = vtk.vtkFloatArray()                                             	# Creates a VTK container of the Y-point coordinates
	for y, i in enumerate(PointsY):
		yCoords.InsertNextValue(i)

	zCoords = vtk.vtkFloatArray()                                             	# Creates a VTK container of the Z-point coordinates
	for z, i in enumerate(PointsZ):
		zCoords.InsertNextValue(i)

	rgrid = vtk.vtkRectilinearGrid()                                          	# Creates a rectilinear grid by using the VTK coordinates
	rgrid.SetDimensions(x+1, y+1, z+1)                                      	# x+1=number of points along X axis
	rgrid.SetXCoordinates(xCoords)
	rgrid.SetYCoordinates(yCoords)
	rgrid.SetZCoordinates(zCoords)

	ExtractSurface = vtk.vtkGeometryFilter()                                 	# Converts vtk rectilinear grid element to vtkPolyData  --> It is equivalent to Extract Surface
	ExtractSurface.SetInputData(rgrid)           
	ExtractSurface.Update()
	#self.GetOutput().ShallowCopy(ExtractSurface.GetOutput()) 

	return ExtractSurface


############################################
##########  CREATE PATH FUNCTION  ##########
############################################

def CREATE_PATH(XXX,YYY,ZZZ,Orientation):
	import numpy as np
	from vtk.numpy_interface import algorithms as algs
	
	XXX_path = []
	YYY_path = []
	ZZZ_path = []
	Line_ID = 0
	
	if Orientation==0: 															# it sweeps the probe along the Z-axis up and down alternativelly
		for i in YYY:
			XXX_path = np.append(XXX_path,XXX*np.ones(len(ZZZ)))
			YYY_path = np.append(YYY_path,i*np.ones(len(ZZZ)))
			if (int(Line_ID) % 2) == 0:                                     	# This stores the same points as Z_points but reverses the order to have points following a zigzag pattern to optimize the traverser path
				ZZZ_path = np.append(ZZZ_path,ZZZ)              				# The probe path will only follow the lower part of the mesh
			else:
				ZZZ_path = np.append(ZZZ_path,np.flipud(ZZZ))
			Line_ID = Line_ID+1

	if Orientation==1: 															# it sweeps the probe along the Z-axis up and down alternativelly
		for i in XXX:
			XXX_path = np.append(XXX_path,i*np.ones(len(ZZZ)))
			YYY_path = np.append(YYY_path,YYY*np.ones(len(ZZZ)))
			if (int(Line_ID) % 2) == 0:                                     	# This stores the same points as Z_points but reverses the order to have points following a zigzag pattern to optimize the traverser path
				ZZZ_path = np.append(ZZZ_path,ZZZ)              				# The probe path will only follow the lower part of the mesh
			else:
				ZZZ_path = np.append(ZZZ_path,np.flipud(ZZZ))
			Line_ID = Line_ID+1

	if Orientation==2: 															# it sweeps the probe along the Y-axis up and down alternativelly
		for i in XXX:
			XXX_path = np.append(XXX_path,i*np.ones(len(YYY)))
			ZZZ_path = np.append(ZZZ_path,ZZZ*np.ones(len(YYY)))
			if (int(Line_ID) % 2) == 0:                                     	# This stores the same points as Z_points but reverses the order to have points following a zigzag pattern to optimize the traverser path
				YYY_path = np.append(YYY_path,YYY)              				# The probe path will only follow the lower part of the mesh
			else:
				YYY_path = np.append(YYY_path,np.flipud(YYY))
			Line_ID = Line_ID+1

	Probe_Path = algs.make_vector(XXX_path,YYY_path,ZZZ_path)
	return Probe_Path


############################################
#########  ROTATE POINTS FUNCTION  #########
############################################

def ROTATE_vector(vector,Rotation):
	import numpy as np
	from vtk.numpy_interface import algorithms as algs
	
	AngleX = Rotation[0]
	AngleY = Rotation[1]
	AngleZ = Rotation[2]

	Rx = np.array([(1,0,0),(0,algs.cos(AngleX*np.pi/180),-algs.sin(AngleX*np.pi/180)),(0,algs.sin(AngleX*np.pi/180),algs.cos(AngleX*np.pi/180))])   #Rotation matrix about X-axis
	Ry = np.array([(algs.cos(AngleY*np.pi/180),0,algs.sin(AngleY*np.pi/180)),(0,1,0),(-algs.sin(AngleY*np.pi/180),0,algs.cos(AngleY*np.pi/180))])  	#Rotation matrix about Y-axis
	Rz = np.array([(algs.cos(AngleZ*np.pi/180),-algs.sin(AngleZ*np.pi/180),0),(algs.sin(AngleZ*np.pi/180),algs.cos(AngleZ*np.pi/180),0),(0,0,1)])  	#Rotation matrix about Z-axis

	Rx_T = np.transpose(Rx)
	Ry_T = np.transpose(Ry)
	Rz_T = np.transpose(Rz)

	MX = np.dot(Rx_T,np.dot(Rx,Rx_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh X-axis
	MY = np.dot(Ry_T,np.dot(Ry,Ry_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Y-axis
	MZ = np.dot(Rz_T,np.dot(Rz,Rz_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Z-axis

	TRANSFORMATION = np.dot(MX,np.dot(MY,MZ))  #Total (combined) transformation matrix

	T_t = np.transpose(TRANSFORMATION)
	T_inverse = np.dot(T_t,np.dot(TRANSFORMATION,T_t))  	# It calculates the inverse of the TRANSFORMATION matrix
	TransformedVector = np.einsum('ij,kj->ik',vector,T_inverse) 
	return TransformedVector


############################################
#########  PLOT VTK LINES FUNCTION  ########
############################################

def PLOT_VTK_LINES_VERTEX(PointsVector):
	import numpy as np
	import vtk

	points = vtk.vtkPoints()                                                	# create a vtkPoints container to store all the point coordinates.
	points.SetData(dsa.numpyTovtkDataArray(PointsVector , "Points"))   					# numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()

	NBpts = np.shape(PointsVector)[0]                                             		# Number of mesh points
	
	polyLine = vtk.vtkPolyLine()                                            	# Creates a polyline and set the ID for each point
	polyLine.GetPointIds().SetNumberOfIds(int(NBpts)) 
	for i in range(0, int(NBpts)):
		polyLine.GetPointIds().SetId(i, i)

	cells = vtk.vtkCellArray()                                              	# Create a cell array to store the lines in and add the lines to it
	cells.InsertNextCell(polyLine)

	polyData = vtk.vtkPolyData()                                            	# Create a polydata to store everything in
	polyData.SetPoints(points)                                              	# Add the points to the dataset
	polyData.SetLines(cells)                                                	# Add the lines to the dataset

	Survey_Line = vtk.vtkGeometryFilter()                                   	# Converts the polydata to vtkPolyData  
	Survey_Line.SetInputData(polyData)             
	Survey_Line.Update()
	
	# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
	polydata2 = vtk.vtkPolyData()
	polydata2.SetPoints(points)

	Survey_Vertices = vtk.vtkVertexGlyphFilter()
	Survey_Vertices.AddInputData(polydata2)
	Survey_Vertices.Update()
	
	return Survey_Line,Survey_Vertices

	
############################################
######## PRINT SURVEY TIME FUNCTION ########
############################################

def PRINT_SURVEY_TIME(PATH,Survey_speed,Hold_time,DAQ_time):
		import numpy as np
		
		# COMPUTES THE TRAYECTORY LENGHT:
		PATH = np.array(PATH) 													# Make it a numpy array
		lengths = np.sqrt(np.sum(np.diff(PATH, axis=0)**2, axis=1)) 			# Length between corners
		Total_distance = np.sum(lengths)
		
		# COMPUTES THE NUMBER OF POINTS:
		NBpts = np.shape(PATH)[0]
		
		
		Total_travel_time = Total_distance/Survey_speed
		Total_Hold = Hold_time*NBpts
		Total_DAQ = DAQ_time*NBpts
		Total_time = Total_travel_time + Total_Hold + Total_DAQ

		print ("NUMBER OF SURVEY POINTS = ", int(NBpts))
		print ("")		
		if Survey_speed==0:
			print ("Survey speed = 0 m/s --> survey time will not be calculated")

			
		if Survey_speed!=0:	
			print ("TOTAL WAKE SURVEY TIME =  ", ("%.2f" % (Total_time/60)), "[min]   ", ("%.2f" % (Total_time/3600)) , "[hours]")
			print ("-------------------------------------------------------------")
			print ("Total travel distance =    ", ("%.2f" % Total_distance), " [m]" )
			print ("Total travel time =       ", ("%.2f" % (Total_travel_time/60)), "[min]   ", ("%.2f" % (Total_travel_time/3600)) , "[hours]")
			print ("Total hold time =         ", ("%.2f" % (Total_Hold/60)), "[min]   ", ("%.2f" % (Total_Hold/3600)) , "[hours]"  )     
			print ("Total DAQ time =          ", ("%.2f" % (Total_DAQ/60)), "[min]   ", ("%.2f" % (Total_DAQ/3600)) , "[hours]" )

	

############################################
######## EXPORT PATH TO TXT FUNCTION #######
############################################	

def EXPORT_PATH_TO_TXT(POINTS,File_Path_Output,Output_TXT_in_mm,Output_TXT_without_headers,SelectTXTdelimiter,ReturnToStartingPoint,DecimalSeparator,ReturnToOrigin,OriginCoordinates):	
	import numpy as np
	
	OriginCoordinates = np.matrix(OriginCoordinates)
	
	if not File_Path_Output:                                                      # This is executed only if there is not a file path declared in the GUI
		print ("")
		print ("ERROR: the output TXT file with the probe path coordinates could not be written")
		print (r"ERROR: The output folder path is not defined in the GUI. Use this format: C:\Users\Xxxxxx\Desktop") 


	if File_Path_Output:       
		if Output_TXT_in_mm == 1:
			POINTS = 1000*POINTS
			OriginCoordinates = 1000*OriginCoordinates
	
		data = np.asarray(POINTS, dtype=np.float32)
		data = np.matrix(data)
		OutputTXT = File_Path_Output+"\OutputForTraverser.txt"
		
		if SelectTXTdelimiter == 0:
			DELIMITER = '\t'
		if SelectTXTdelimiter == 1:
			DELIMITER = ' '
		if SelectTXTdelimiter == 2:
			DELIMITER = ','
			OutputTXT = File_Path_Output+"\OutputForTraverser.csv"
					
		with open(OutputTXT, "w") as txt_file: 									# Send the points data to the output TXT file
			for line in data:	
				
				if Output_TXT_in_mm == 0:
					np.savetxt(txt_file, line, fmt='%.5f',delimiter=DELIMITER)   					# 5 significative decimals are kept for output in [m]
				if Output_TXT_in_mm == 1:
					np.savetxt(txt_file, line, fmt='%.2f',delimiter=DELIMITER)   					# 2 significative decimals are kept for output in [mm]

			if ReturnToStartingPoint==1:				
				if Output_TXT_in_mm == 0:
					np.savetxt(txt_file, data[0], fmt='%.5f',delimiter=DELIMITER)   					# 5 significative decimals are kept for output in [m]
				if Output_TXT_in_mm == 1:
					np.savetxt(txt_file, data[0], fmt='%.2f',delimiter=DELIMITER)   					# 2 significative decimals are kept for output in [mm]
		
			if ReturnToOrigin==1:				
				if Output_TXT_in_mm == 0:
					np.savetxt(txt_file, OriginCoordinates, fmt='%.5f',delimiter=DELIMITER)   					# 5 significative decimals are kept for output in [m]
				if Output_TXT_in_mm == 1:
					np.savetxt(txt_file, OriginCoordinates, fmt='%.2f',delimiter=DELIMITER)   					# 2 significative decimals are kept for output in [mm]

		if DecimalSeparator == 1: 												# Change decimal separator from point (default) to comma
			with open(OutputTXT, 'r') as original: data = original.read()
			with open(OutputTXT, 'w') as modified: modified.write(data.replace('.', ','))

		if DecimalSeparator == 2: 												# Change decimal separator from point (default) to semicolon
			with open(OutputTXT, 'r') as original: data = original.read()
			with open(OutputTXT, 'w') as modified: modified.write(data.replace('.', ';'))

		if Output_TXT_without_headers == 0: 									# Add headers to the output TXT file
			with open(OutputTXT, 'r') as original: data = original.read()
			if Output_TXT_in_mm == 0:
				with open(OutputTXT, 'w') as modified: modified.write("X[m]"+DELIMITER+"Y[m]"+DELIMITER+"Z[m]\n" + data)
			if Output_TXT_in_mm == 1:
				with open(OutputTXT, 'w') as modified: modified.write("X[mm]"+DELIMITER+"Y[mm]"+DELIMITER+"Z[mm]\n" + data)
		
		
		if SelectTXTdelimiter != 2:
			print ("TXT file with probe coordinates for traverser written successfully ('OutputForTraverser.txt')")
		if SelectTXTdelimiter == 2:
			print ("CSV file with probe coordinates for traverser written successfully ('OutputForTraverser.csv')")

	
	
###################################
#######  WRITE VTP FUNCTION  #######
###################################

def WRITE_VTP(vtkObject,File_Path_Output,Path_or_Mesh):     					# Path_or_Mesh=0(Path)    Path_or_Mesh=1(mesh)
	import vtk
	if File_Path_Output:
		if Path_or_Mesh==0:
			OutputVTP = File_Path_Output+"\OutputProbePath.vtp"
		if Path_or_Mesh==1:
			OutputVTP = File_Path_Output+"\OutputSurveyMesh.vtp"
		
		OutputVTK = vtk.vtkXMLPolyDataWriter()                               	# Alternative reader: vtkPolyDataWriter
		OutputVTK.SetInputData(vtkObject.GetOutput())    						# Alternative command: SetInput
		OutputVTK.SetFileName(OutputVTP)
		OutputVTK.Write()                                                    	# "OutputVTK" is a vtkPolyData variable
		#OutputVTK.Update()

		if Path_or_Mesh==0:
			print ("Probe path VTP file written successfully ('OutputProbePath.vtp')")
		if Path_or_Mesh==1:
			print ("Mesh wake VTP file written successfully ('OutputSurveyMesh.vtp')")


	
###################################
#######  DELAUNAY FUNCTION  #######
###################################

def DELAUNAY(InputPoints): 
	import vtk
	
	pts = InputPoints.GetOutput().GetPoints()

	aPolyData = vtk.vtkPolyData()
	aPolyData.SetPoints(pts)
		 
	aCellArray = vtk.vtkCellArray()
		 
	boundary = vtk.vtkPolyData()
	boundary.SetPoints(aPolyData.GetPoints())
	boundary.SetPolys(aCellArray)

	delaunay = vtk.vtkDelaunay2D()
	delaunay.SetProjectionPlaneMode(vtk.VTK_BEST_FITTING_PLANE)
	if vtk.VTK_MAJOR_VERSION <= 5:
		delaunay.SetInput(aPolyData.GetOutput())
		delaunay.SetSource(boundary)
	else:
		delaunay.SetInputData(aPolyData)
		delaunay.SetSourceData(boundary)

	delaunay.Update()
	return delaunay
	
	

############################################################################################################
########################################## ERROR MESSAGE FOR USER ##########################################
############################################################################################################

if Phase==0 and Stage==0 and Step==0:
	if Mode == 0:
		print ("GUI is not yet properly defined by the user: NO OUTPUT EXPECTED!")
		print ("")
		print ("To create a survey line: change the phase")
		print ("To create a mapped survey plane: change stage")
		print ("To create a wake-shape mapped grid: change step")
		
	if Mode == 1:
		if not (File_Path_Mesh and File_Path_CSV):
				print ("GUI is not yet properly defined by the user: NO OUTPUT EXPECTED!")
				print ("")
				print ("The input mesh, probe data and output folder must be selected beforehand (and a file parser if required)")
				
	if Mode == 2:
		print ("GUI is not yet properly defined by the user: NO OUTPUT EXPECTED!")
		print ("")
		print ("To visualize line data: change the visualization mode")
		print ("To visualize plane data: change the mesh type")



############################################################################################################
#################################### READ AND TRANSFORM INPUT MESH DATA ####################################
############################################################################################################

# This section reads any input mesh file format and converts it into a vtkPolyData format. Then it transforms the mesh geometry

# READ INPUT MESH:
if File_Path_Mesh != "":

	if File_Path_Mesh.endswith('.dat'):
		InputMesh = READ_TECPLOT(File_Path_Mesh)

	elif File_Path_Mesh.endswith('.vtp'):
		InputMesh = READ_VTP(File_Path_Mesh)

	elif File_Path_Mesh.endswith('.vtu'):
		InputMesh = READ_VTU(File_Path_Mesh)

	elif File_Path_Mesh.endswith('.vts'):
		InputMesh = READ_VTS(File_Path_Mesh)	

	else:
		print("# ERROR: the data format of the input wake file must be DAT, VTP, VTU or VTS")

	self.GetOutput().ShallowCopy(InputMesh.GetOutput())



############################################################################################################
########################################## SURVEY LINE GENERATOR ###########################################
############################################################################################################

if Phase>0:


	
	# CREATES THE BASELINE POINT COORDINATES USING A PARAMETRIC LINE EQUATION:
	Direction = [EndXYZ[0]-StartXYZ[0],EndXYZ[1]-StartXYZ[1],EndXYZ[2]-StartXYZ[2]] # Line direction vector
	t = np.linspace(0,1,NBpts)                                                  # Line parameter: 0=start  1=end
	X = StartXYZ[0]+t*Direction[0]
	Y = StartXYZ[1]+t*Direction[1]
	Z = StartXYZ[2]+t*Direction[2]





	# THE FOLLOWING 3 BLOCKS OF CODE MUST BE COMPACTED BY USING FUNCTION. THIS FUNCTION MUST BE COMPATIBLE WITH THE REFINEMENT IN THE MAPPED SURVEY GRID GENERATOR IN LINE 555!!!!!




	# REFINES THE BASELINE POINTS ALONG X-AXIS IF REQUIRED BY THE USER:
	if Refinement_1==1 or Refinement_2==1 or Refinement_3==1 or Refinement_4==1 or Refinement_5==1:   #Refinements along X-axis activated
		if Refinement_1==1:
			X = refine_by_normal_distribution(X,np.min(X),np.max(X),Ref_1_coord,Ref_1_width,Ref_1_point,Ref_1_sigma)
		if Refinement_2==1:
			X = refine_by_normal_distribution(X,np.min(X),np.max(X),Ref_2_coord,Ref_2_width,Ref_2_point,Ref_2_sigma)
		if Refinement_3==1:
			X = refine_by_normal_distribution(X,np.min(X),np.max(X),Ref_3_coord,Ref_3_width,Ref_3_point,Ref_3_sigma)
		if Refinement_4==1:
			X = refine_by_normal_distribution(X,np.min(X),np.max(X),Ref_4_coord,Ref_4_width,Ref_4_point,Ref_4_sigma)
		if Refinement_5==1:
			X = refine_by_normal_distribution(X,np.min(X),np.max(X),Ref_5_coord,Ref_5_width,Ref_5_point,Ref_5_sigma)
		t = (X-StartXYZ[0])/Direction[0]                                        # Recomputes line paramater
		Y = StartXYZ[1]+t*Direction[1]                                          # Calculates Y points
		Z = StartXYZ[2]+t*Direction[2]                                          # Calculates Z points

	# REFINES THE BASELINE POINTS ALONG Y-AXIS IF REQUIRED BY THE USER:
	if Refinement_1==2 or Refinement_2==2 or Refinement_3==2 or Refinement_4==2 or Refinement_5==2:   #Refinements along Y-axis activated
		if Refinement_1==2:
			Y = refine_by_normal_distribution(Y,np.min(Y),np.max(Y),Ref_1_coord,Ref_1_width,Ref_1_point,Ref_1_sigma)
		if Refinement_2==2:
			Y = refine_by_normal_distribution(Y,np.min(Y),np.max(Y),Ref_2_coord,Ref_2_width,Ref_2_point,Ref_2_sigma)
		if Refinement_3==2:
			Y = refine_by_normal_distribution(Y,np.min(Y),np.max(Y),Ref_3_coord,Ref_3_width,Ref_3_point,Ref_3_sigma)
		if Refinement_4==2:
			Y = refine_by_normal_distribution(Y,np.min(Y),np.max(Y),Ref_4_coord,Ref_4_width,Ref_4_point,Ref_4_sigma)
		if Refinement_5==2:
			Y = refine_by_normal_distribution(Y,np.min(Y),np.max(Y),Ref_5_coord,Ref_5_width,Ref_5_point,Ref_5_sigma)
		t = (Y-StartXYZ[1])/Direction[1]                                        # Recomputes line paramater
		X = StartXYZ[0]+t*Direction[0]                                          # Calculates X points
		Z = StartXYZ[2]+t*Direction[2]                                          # Calculates Z points

	# REFINES THE BASELINE POINTS ALONG Z-AXIS IF REQUIRED BY THE USER:
	if Refinement_1==3 or Refinement_2==3 or Refinement_3==3 or Refinement_4==3 or Refinement_5==3:   #Refinements along Z-axis activated
		if Refinement_1==3:
			Z = refine_by_normal_distribution(Z,np.min(Z),np.max(Z),Ref_1_coord,Ref_1_width,Ref_1_point,Ref_1_sigma)
		if Refinement_2==3:
			Z = refine_by_normal_distribution(Z,np.min(Z),np.max(Z),Ref_2_coord,Ref_2_width,Ref_2_point,Ref_2_sigma)
		if Refinement_3==3:
			Z = refine_by_normal_distribution(Z,np.min(Z),np.max(Z),Ref_3_coord,Ref_3_width,Ref_3_point,Ref_3_sigma)
		if Refinement_4==3:
			Z = refine_by_normal_distribution(Z,np.min(Z),np.max(Z),Ref_4_coord,Ref_4_width,Ref_4_point,Ref_4_sigma)
		if Refinement_5==3:
			Z = refine_by_normal_distribution(Z,np.min(Z),np.max(Z),Ref_5_coord,Ref_5_width,Ref_5_point,Ref_5_sigma)
		t = (Z-StartXYZ[2])/Direction[2]                                        # Recomputes line paramater
		X = StartXYZ[0]+t*Direction[0]                                          # Calculates X points
		Y = StartXYZ[1]+t*Direction[1]                                          # Calculates Y points    

	# PLOT THE SURVEY LINE/POINT:
	POINTS = algs.make_vector(X,Y,Z)   											# Creates a vector with the point coordinates
	Survey_Line,Survey_Vertices = PLOT_VTK_LINES_VERTEX(POINTS)   				# Generate the VTK elements with survey line and the points
	
	# COMBINE THE VERTICES AND THE SURVEY PATH LINE INTO A SINGLE VARIABLE:
	appendFilter = vtk.vtkAppendFilter()                                    #Output: vtkUnstructuredGrid
	appendFilter.AddInputData(Survey_Line.GetOutput())        
	appendFilter.AddInputData(Survey_Vertices.GetOutput())
	appendFilter.Update()

	# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA  
	combined = vtk.vtkGeometryFilter()                                      # Converts vtkMultiblockDataSet to vtkPolyData
	combined.SetInputData(appendFilter.GetOutput())                             
	combined.Update()
		
	self.GetOutput().ShallowCopy(combined.GetOutput())
	
	
	if Phase==1: 
		# PRINT OUTPUT INFORMATION:
		print ("Survey line generated successfully")
		print ("")
		print ("To improve the visualization of the points:")
		print ("A) Increase the 'point size' of the 'styling' option in the properties panel")
		print ("B) Change the visualization mode from 'Surface' to 'Points'")

		# COMPUTES SURVEY TIME:
		PRINT_SURVEY_TIME(POINTS,Survey_speed,Hold_time,DAQ_time)

	if Phase==2:
		# CREATES OUTPUT TXT FILE WITH POINTS COORDINATES FOR TRAVERSER:
		EXPORT_PATH_TO_TXT(POINTS,File_Path_Output,Output_TXT_in_mm,Output_TXT_without_headers,SelectTXTdelimiter,ReturnToStartingPoint,DecimalSeparator,ReturnToOrigin,OriginCoordinates)
		WRITE_VTP(combined,File_Path_Output,0)  							# 0=survey line
		
############################################################################################################
###################################### MAPPED SURVEY PLANE GENERATOR #######################################
############################################################################################################

if Stage != 0:   																# Create a mapped grid or visualize probe path
	
	# GENERATE THE BASELINE MESH POINTS:
	if Orientation==0:
		XXX = [0]
		YYY = np.linspace(IaxisInfo[0], IaxisInfo[1], IaxisInfo[2])
		ZZZ = np.linspace(JaxisInfo[0], JaxisInfo[1], JaxisInfo[2])
		
	if Orientation==1:
		XXX = np.linspace(IaxisInfo[0], IaxisInfo[1], IaxisInfo[2])
		YYY = [0]
		ZZZ = np.linspace(JaxisInfo[0], JaxisInfo[1], JaxisInfo[2])		
		
	if Orientation==2:
		XXX = np.linspace(IaxisInfo[0], IaxisInfo[1], IaxisInfo[2])
		YYY = np.linspace(JaxisInfo[0], JaxisInfo[1], JaxisInfo[2])
		ZZZ = [0]		
		
	# REFINES THE BASELINE POINTS ALONG X-AXIS IF REQUIRED BY THE USER:
	if Refinement_1==1 or Refinement_2==1 or Refinement_3==1 or Refinement_4==1 or Refinement_5==1:   #Refinements along X-axis activated
		if Refinement_1==1:
			XXX = refine_by_normal_distribution(XXX,np.min(XXX),np.max(XXX),Ref_1_coord,Ref_1_width,Ref_1_point,Ref_1_sigma)
		if Refinement_2==1:
			XXX = refine_by_normal_distribution(XXX,np.min(XXX),np.max(XXX),Ref_2_coord,Ref_2_width,Ref_2_point,Ref_2_sigma)
		if Refinement_3==1:
			XXX = refine_by_normal_distribution(XXX,np.min(XXX),np.max(XXX),Ref_3_coord,Ref_3_width,Ref_3_point,Ref_3_sigma)
		if Refinement_4==1:
			XXX = refine_by_normal_distribution(XXX,np.min(XXX),np.max(XXX),Ref_4_coord,Ref_4_width,Ref_4_point,Ref_4_sigma)
		if Refinement_5==1:
			XXX = refine_by_normal_distribution(XXX,np.min(XXX),np.max(XXX),Ref_5_coord,Ref_5_width,Ref_5_point,Ref_5_sigma)

	# REFINES THE BASELINE POINTS ALONG Y-AXIS IF REQUIRED BY THE USER:
	if Refinement_1==2 or Refinement_2==2 or Refinement_3==2 or Refinement_4==2 or Refinement_5==2:   #Refinements along Y-axis activated
		if Refinement_1==2:
			YYY = refine_by_normal_distribution(YYY,np.min(YYY),np.max(YYY),Ref_1_coord,Ref_1_width,Ref_1_point,Ref_1_sigma)
		if Refinement_2==2:
			YYY = refine_by_normal_distribution(YYY,np.min(YYY),np.max(YYY),Ref_2_coord,Ref_2_width,Ref_2_point,Ref_2_sigma)
		if Refinement_3==2:
			YYY = refine_by_normal_distribution(YYY,np.min(YYY),np.max(YYY),Ref_3_coord,Ref_3_width,Ref_3_point,Ref_3_sigma)
		if Refinement_4==2:
			YYY = refine_by_normal_distribution(YYY,np.min(YYY),np.max(YYY),Ref_4_coord,Ref_4_width,Ref_4_point,Ref_4_sigma)
		if Refinement_5==2:
			YYY = refine_by_normal_distribution(YYY,np.min(YYY),np.max(YYY),Ref_5_coord,Ref_5_width,Ref_5_point,Ref_5_sigma)

	# REFINES THE BASELINE POINTS ALONG Z-AXIS IF REQUIRED BY THE USER:
	if Refinement_1==3 or Refinement_2==3 or Refinement_3==3 or Refinement_4==3 or Refinement_5==3:   #Refinements along Z-axis activated
		if Refinement_1==3:
			ZZZ = refine_by_normal_distribution(ZZZ,np.min(ZZZ),np.max(ZZZ),Ref_1_coord,Ref_1_width,Ref_1_point,Ref_1_sigma)
		if Refinement_2==3:
			ZZZ = refine_by_normal_distribution(ZZZ,np.min(ZZZ),np.max(ZZZ),Ref_2_coord,Ref_2_width,Ref_2_point,Ref_2_sigma)
		if Refinement_3==3:
			ZZZ = refine_by_normal_distribution(ZZZ,np.min(ZZZ),np.max(ZZZ),Ref_3_coord,Ref_3_width,Ref_3_point,Ref_3_sigma)
		if Refinement_4==3:
			ZZZ = refine_by_normal_distribution(ZZZ,np.min(ZZZ),np.max(ZZZ),Ref_4_coord,Ref_4_width,Ref_4_point,Ref_4_sigma)
		if Refinement_5==3:
			ZZZ = refine_by_normal_distribution(ZZZ,np.min(ZZZ),np.max(ZZZ),Ref_5_coord,Ref_5_width,Ref_5_point,Ref_5_sigma)
 	
	# GENERATE THE GRID AND THE PROBE PATH FROM THE POINTS:
	PLANE = CREATE_GRID(XXX,YYY,ZZZ)
	PATH = CREATE_PATH(XXX,YYY,ZZZ,Orientation)
	
	PLANE = ROTATE(PLANE, Rotation)
	PATH = ROTATE_vector(PATH,Rotation)
	
	if Orientation==0:
		PLANE = TRANSLATE(PLANE, [NormalDistance,0,0])
		PATH = np.array(PATH)+np.array([NormalDistance,0,0])
	if Orientation==1:
		PLANE = TRANSLATE(PLANE, [0,NormalDistance,0])
		PATH = np.array(PATH)+np.array([0,NormalDistance,0])
	if Orientation==2:
		PLANE = TRANSLATE(PLANE, [0,0,NormalDistance])
		PATH = np.array(PATH)+np.array([0,0,NormalDistance])
	
	self.GetOutput().ShallowCopy(PLANE.GetOutput())
	
	if Stage == 1:
		# PRINT OUTPUT INFORMATION:
		print ("Mesh generated successfully")
		print ("Change the visualization mode from 'Surface' to 'Surface with edges' to see the cells:")
		
		# COMPUTES SURVEY TIME:
		PRINT_SURVEY_TIME(PATH,Survey_speed,Hold_time,DAQ_time)
		
if Stage == 2 or Stage == 3:   													# Visualize or export probe path

	# CREATES A LINE JOINING THE POINTS ALONG THE TRAVERSER PATH
	Survey_Line,Survey_Vertices = PLOT_VTK_LINES_VERTEX(PATH)

	# COMBINE THE VERTICES AND THE SURVEY PATH LINE INTO A SINGLE VARIABLE:
	appendFilter = vtk.vtkAppendFilter()                                    #Output: vtkUnstructuredGrid
	appendFilter.AddInputData(Survey_Line.GetOutput())        
	appendFilter.AddInputData(Survey_Vertices.GetOutput())
	appendFilter.Update()

	# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA  
	combined = vtk.vtkGeometryFilter()                                      # Converts vtkMultiblockDataSet to vtkPolyData
	combined.SetInputData(appendFilter.GetOutput())                             
	combined.Update()

	# VISUALIZATION
	self.GetOutput().ShallowCopy(combined.GetOutput())                  # Displays output data

	if Stage == 2:
		# PRINT OUTPUT INFO:
		print ("Probe path generated successfully")
		print ("To improve the visualization of the points: increase the 'point size' of the 'styling' option in the properties panel")
		
		# COMPUTES SURVEY TIME:	
		PRINT_SURVEY_TIME(PATH,Survey_speed,Hold_time,DAQ_time)
	
	if Stage == 3:	
		# CREATES OUTPUT TXT FILE WITH POINTS COORDINATES FOR TRAVERSER:
		EXPORT_PATH_TO_TXT(PATH,File_Path_Output,Output_TXT_in_mm,Output_TXT_without_headers,SelectTXTdelimiter,ReturnToStartingPoint,DecimalSeparator,ReturnToOrigin,OriginCoordinates)
		WRITE_VTP(combined,File_Path_Output,0)  							# 0=survey line
		WRITE_VTP(PLANE,File_Path_Output,1)  							# 1=survey mesh
	


############################################################################################################
#######################################    WAKE SHAPE MAPPED GRID    #######################################
############################################################################################################


"""
Generalizarlo para incluir plano XZ y XY y quizas para rotar el plano resultante como en mapped grid
"""





if Step > 0:                     # Mode = Mesh generation

	if Step == 1:                                                               # Visualize input data
		self.GetOutput().ShallowCopy(InputMesh.GetOutput())
		print ("# Inputa data correctly loaded")


	if Step == 2 or Step == 3 or Step == 4 or Step == 5 or Step == 6 or Step == 7:           # Step=2(Clip wake region)  Step=3(Define offset margin)  Step=4(Check tunnel walls) Step=5 (Compute mesh)   Step=6(Check traverser path)  Step=7(export TXT)

		# CLIP DATA SET BY SCALAR:  
		clipper = vtk.vtkClipDataSet()                                          #Output: vtkUnstructuredGrid
		clipper.SetInputData(InputMesh.GetOutput())
		clipper.SetInputArrayToProcess(0,0,0,0, Clip_parameter)
		clipper.SetValue(Clip_Value)
		if Invert_clip == 1:
			clipper.InsideOutOn()
		if Invert_clip == 0:
			clipper.InsideOutOff()
		clipper.Update()

		# CONVERT CLIP FILTER OUTPUT FORMAT TO POLYMESH   
		CLIP = vtk.vtkGeometryFilter()                                          # Converts vtkUnstructuredGrid to vtkPolyData
		CLIP.SetInputData(clipper.GetOutput())                             
		CLIP.Update()

		if Step == 2:
			# DISPLAY CLIPPED MESH:
			self.GetOutput().ShallowCopy(CLIP.GetOutput())                      # Displays output data
			
			# PRINT OUTPUT INFO:
			if Clip_parameter=="Type_the_variable_name_here":
				print ("NO OUTPUT EXPECTED! Make sure to type the correct variable name in 'Clip parameter'")

			else:
				print ("Input mesh clipped successfully")
			

	if Step == 3 or Step == 4 or Step == 5 or Step == 6 or Step == 7:                        # Step=3(Define offset margin)  Step=4(Check tunnel walls)  Step=5 (Compute mesh)   Step=6(Check traverser path)   Step=7(export TXT file)

		# CALCULATE DISTANCE FROM INPUT MESH POINTS TO THE CLIPPED POINTS  
		pdd = vtk.vtkImplicitPolyDataDistance()
		pdd.SetInput(CLIP.GetOutput())
		numPts = InputMesh.GetOutput().GetNumberOfPoints()          
		distance = vtk.vtkDoubleArray()
		distance.SetName("Distance")
		distance.SetNumberOfComponents(1)
		distance.SetNumberOfTuples(numPts)
		for i in range(numPts):
			pt = InputMesh.GetOutput().GetPoint(i)
			distance.SetValue(i, pdd.EvaluateFunction(pt))

		InputMesh.GetOutput().GetPointData().SetScalars(distance)  
		InputMesh.Update()


		# COMPUTE THE ABSOLUTE VALUE OF THE DISTANCE (to avoid having negative values)
		InputMeshPython = dsa.WrapDataObject(InputMesh.GetOutput())             # Converts VTK element into a Python object
		Distance_Array = InputMeshPython.PointData["Distance"]                  # Get the array with the distance values 
		ABS_Distance = np.abs(Distance_Array)                                   # Computes the absolute value


		# APPENDS THE ABSOLUTE DISTANCE ARRAY TO THE INPUT MESH:
		ABS_distance = vtk.vtkDoubleArray()
		ABS_distance.SetName("Distance")
		ABS_distance.SetNumberOfComponents(1)
		ABS_distance.SetNumberOfTuples(numPts)
		for i in range(numPts):
			ABS_distance.SetValue(i, ABS_Distance[i])

		InputMesh.GetOutput().GetPointData().SetScalars(ABS_distance)  
		InputMesh.Update()


		# COMPUTES CONTOUR LINE OF ISO-DISTANCE TO CLIPPED REGION 
		contour = vtk.vtkContourFilter()
		contour.SetInputData(InputMesh.GetOutput())
		contour.SetInputArrayToProcess(0,0,0,0, "Distance")
		contour.SetValue(0, Offset_margin)
		contour.Update()


		# COMBINE THE OUTPUTS FROM THE CLIP AND CONTOUR INTO A SINGLE VARIABLE
		appendFilter = vtk.vtkAppendFilter()                                    #Output: vtkUnstructuredGrid
		appendFilter.AddInputData(CLIP.GetOutput())
		appendFilter.AddInputData(contour.GetOutput())
		appendFilter.Update()


		# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA  
		combined = vtk.vtkGeometryFilter()                                      # Converts vtkMultiblockDataSet to vtkPolyData
		combined.SetInputData(appendFilter.GetOutput())                             
		combined.Update()

		if Step == 3:
			# DISPLAY OUTPUT MESH:
			self.GetOutput().ShallowCopy(combined.GetOutput())                  # Displays output data
			
			# PRINT OUTPUT INFO:
			print ("The wake survey limit defined by the offset margin was successfully computed")
			if Offset_margin == 0:
				print ("The offset margin is currently set to zero: the wake survey limit is coincident with the clipped wake boundary, thus not visible.")
			
			
			

	if Step == 4 or Step == 5 or Step == 6 or Step == 7:                      	# Step=4(Check tunnel walls)   Step=5 (Compute mesh)  Step=6(Check traverser path)   Step=7(export TXT file)

		# GET POINT COORDINATES FROM INPUT FILE:
		dataInputMesh = dsa.WrapDataObject(InputMesh.GetOutput())               # Converts VTK element into a Python object
		POINTS = dataInputMesh.Points                                           # Get points coordinates
		X = POINTS[:,0]

		# DRAW LINES TO REPRESENTATE THE WIND TUNNEL WALLS AND THE SAFETY MARGIN TO THE WALLS:
		if Wind_Tunnel_Model == 1:                                              # Full model          
			SDW = Safety_distance_to_wall
			Tunnel_Ceiling = Line_Function(X[0],-0.5*Tunnel_Width,0.5*Tunnel_Height,X[0],0.5*Tunnel_Width,0.5*Tunnel_Height,100)
			Tunnel_Floor = Line_Function(X[0],-0.5*Tunnel_Width,-0.5*Tunnel_Height,X[0],0.5*Tunnel_Width,-0.5*Tunnel_Height,100)
			Tunnel_Left = Line_Function(X[0],-0.5*Tunnel_Width,-0.5*Tunnel_Height,X[0],-0.5*Tunnel_Width,0.5*Tunnel_Height,100)
			Tunnel_Right = Line_Function(X[0],0.5*Tunnel_Width,-0.5*Tunnel_Height,X[0],0.5*Tunnel_Width,0.5*Tunnel_Height,100)
			Tunnel_margin_ceiling = Line_Function(X[0],-0.5*Tunnel_Width+SDW,0.5*Tunnel_Height-SDW,X[0],0.5*Tunnel_Width-SDW,0.5*Tunnel_Height-SDW,100)
			Tunnel_margin_floor = Line_Function(X[0],-0.5*Tunnel_Width+SDW,-0.5*Tunnel_Height+SDW,X[0],0.5*Tunnel_Width-SDW,-0.5*Tunnel_Height+SDW,100)      
			Tunnel_margin_left = Line_Function(X[0],-0.5*Tunnel_Width+SDW,-0.5*Tunnel_Height+SDW,X[0],-0.5*Tunnel_Width+SDW,0.5*Tunnel_Height-SDW,100)
			Tunnel_margin_right = Line_Function(X[0],0.5*Tunnel_Width-SDW,-0.5*Tunnel_Height+SDW,X[0],0.5*Tunnel_Width-SDW,0.5*Tunnel_Height-SDW,100)
			Ymin = -0.5*Tunnel_Width+SDW
			Ymax = 0.5*Tunnel_Width-SDW
			Zmin = -0.5*Tunnel_Height+SDW
			Zmax = 0.5*Tunnel_Height-SDW

		if Wind_Tunnel_Model == 2:                                              # Half model Y+
			SDW = Safety_distance_to_wall
			Tunnel_Ceiling = Line_Function(X[0],0,0.5*Tunnel_Height,X[0],Tunnel_Width,0.5*Tunnel_Height,100)
			Tunnel_Floor = Line_Function(X[0],0,-0.5*Tunnel_Height,X[0],Tunnel_Width,-0.5*Tunnel_Height,100)
			Tunnel_Left = Line_Function(X[0],0,-0.5*Tunnel_Height,X[0],0,0.5*Tunnel_Height,100)
			Tunnel_Right = Line_Function(X[0],Tunnel_Width,-0.5*Tunnel_Height,X[0],Tunnel_Width,0.5*Tunnel_Height,100)
			Tunnel_margin_ceiling = Line_Function(X[0],0+SDW,0.5*Tunnel_Height-SDW,X[0],Tunnel_Width-SDW,0.5*Tunnel_Height-SDW,100)
			Tunnel_margin_floor = Line_Function(X[0],0+SDW,-0.5*Tunnel_Height+SDW,X[0],Tunnel_Width-SDW,-0.5*Tunnel_Height+SDW,100)      
			Tunnel_margin_left = Line_Function(X[0],0+SDW,-0.5*Tunnel_Height+SDW,X[0],0+SDW,0.5*Tunnel_Height-SDW,100)
			Tunnel_margin_right = Line_Function(X[0],Tunnel_Width-SDW,-0.5*Tunnel_Height+SDW,X[0],Tunnel_Width-SDW,0.5*Tunnel_Height-SDW,100)  
			Ymin = 0+SDW
			Ymax = Tunnel_Width-SDW
			Zmin = -0.5*Tunnel_Height+SDW
			Zmax = 0.5*Tunnel_Height-SDW

		if Wind_Tunnel_Model == 3:                                              # Half model Y-
			SDW = Safety_distance_to_wall
			Tunnel_Ceiling = Line_Function(X[0],0,0.5*Tunnel_Height,X[0],-Tunnel_Width,0.5*Tunnel_Height,100)
			Tunnel_Floor = Line_Function(X[0],0,-0.5*Tunnel_Height,X[0],-Tunnel_Width,-0.5*Tunnel_Height,100)
			Tunnel_Left = Line_Function(X[0],0,-0.5*Tunnel_Height,X[0],0,0.5*Tunnel_Height,100)
			Tunnel_Right = Line_Function(X[0],-Tunnel_Width,-0.5*Tunnel_Height,X[0],-Tunnel_Width,0.5*Tunnel_Height,100)
			Tunnel_margin_ceiling = Line_Function(X[0],0-SDW,0.5*Tunnel_Height-SDW,X[0],-Tunnel_Width+SDW,0.5*Tunnel_Height-SDW,100)
			Tunnel_margin_floor = Line_Function(X[0],0-SDW,-0.5*Tunnel_Height+SDW,X[0],-Tunnel_Width+SDW,-0.5*Tunnel_Height+SDW,100)      
			Tunnel_margin_left = Line_Function(X[0],0-SDW,-0.5*Tunnel_Height+SDW,X[0],0-SDW,0.5*Tunnel_Height-SDW,100)
			Tunnel_margin_right = Line_Function(X[0],-Tunnel_Width+SDW,-0.5*Tunnel_Height+SDW,X[0],-Tunnel_Width+SDW,0.5*Tunnel_Height-SDW,100) 
			Ymin = -Tunnel_Width+SDW
			Ymax = 0-SDW
			Zmin = -0.5*Tunnel_Height+SDW
			Zmax = 0.5*Tunnel_Height-SDW

		if Step == 4 and Wind_Tunnel_Model > 0 :                        # Step=4(Check tunnel walls)    Only displays data if tunnel wall constraint is not required

			# COMBINE THE LINES INTO A SINGLE VARIABLE
			WindTunnel = vtk.vtkAppendFilter()                              # Output: vtkUnstructuredGrid
			WindTunnel.AddInputData(Tunnel_Ceiling.GetOutput())
			WindTunnel.AddInputData(Tunnel_Floor.GetOutput())
			WindTunnel.AddInputData(Tunnel_Left.GetOutput())
			WindTunnel.AddInputData(Tunnel_Right.GetOutput())
			WindTunnel.AddInputData(Tunnel_margin_ceiling.GetOutput())
			WindTunnel.AddInputData(Tunnel_margin_floor.GetOutput())
			WindTunnel.AddInputData(Tunnel_margin_left.GetOutput())
			WindTunnel.AddInputData(Tunnel_margin_right.GetOutput())
			WindTunnel.Update()

			# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
			WindTunnel_VTP = vtk.vtkGeometryFilter()                                      # Converts vtkMultiblockDataSet to vtkPolyData
			WindTunnel_VTP.SetInputData(WindTunnel.GetOutput())                             
			WindTunnel_VTP.Update()        

			# ASSIGN VARIABLE VALUES FROM INPUT DATA TO THE WIND TUNNEL LINES:
			interpolation_WTT = vtk.vtkProbeFilter()
			interpolation_WTT.SetInputConnection(WindTunnel_VTP.GetOutputPort())
			interpolation_WTT.SetSourceData(InputMesh.GetOutput())  
			interpolation_WTT.Update()        

			# ADD INTERPOLATED DATA ARRAYS TO THE MESH
			WrapData = dsa.WrapDataObject(interpolation_WTT.GetOutput())
			Headers = WrapData.PointData.keys()

			for L in range(interpolation_WTT.GetOutput().GetPointData().GetNumberOfArrays()):
				Variable = WrapData.PointData[Headers[L]]

				if interpolation_WTT.GetOutput().GetPointData().GetArray(L).GetNumberOfComponents() == 1:      # Add scalar variables to the mesh
					addfielddata(1, output.GetPointData(),Variable,Headers[L])

				if interpolation_WTT.GetOutput().GetPointData().GetArray(L).GetNumberOfComponents() == 3:      # Add vector variables to the mesh
					addfielddata(3, output.GetPointData(),Variable.ravel(),Headers[L])

			# COMBINE THE WIND TUNNEL WITH THE CLIP AND THE CONTOUR:
			WTT_combined = vtk.vtkAppendFilter()                                            #Output: vtkUnstructuredGrid
			WTT_combined.AddInputData(CLIP.GetOutput())
			WTT_combined.AddInputData(contour.GetOutput())
			WTT_combined.AddInputData(interpolation_WTT.GetOutput())        
			WTT_combined.Update()

			# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:  
			Tunnel_Check_Walls = vtk.vtkGeometryFilter()                                    # Converts vtkMultiblockDataSet to vtkPolyData
			Tunnel_Check_Walls.SetInputData(WTT_combined.GetOutput())                             
			Tunnel_Check_Walls.Update()        
		
			self.GetOutput().ShallowCopy(Tunnel_Check_Walls.GetOutput())
		
		# PRINT OUTPUT INFO:
		if Step == 4:
			if Wind_Tunnel_Model == 0:
				print ("NO OUTPUT EXPECTED! The wind tunnel model configuration is not yet defined")
			else:
				print ("Wind tunnel boundaries successfully determined")
				print ("If tunnel walls are not visible, zoom out")

	if Step == 5 or Step == 6 or Step == 7:                                 	# Step=5 (Compute mesh)  Step=6(Check traverser path)   Step=7(export TXT file)

		# GET POINT COORDINATES FROM CONTOUR LINE:
		data = dsa.WrapDataObject(contour.GetOutput())                          # Converts VTK element into a Python object
		POINTS = data.Points                                                    # Get points coordinates
		X = POINTS[:,0] 
		Y = POINTS[:,1] 
		Z = POINTS[:,2]

		# BASELINE Y-POINTS DISTRIBUTION:
		Stretch_Y = 0.1*Offset_margin                                           # Small stretch value that avoids placing Y-points at the Y-boundaries of the contour, which leads to a single Z min/max point
		Limit_min = np.min(Y)+Stretch_Y
		Mimit_max = np.max(Y)-Stretch_Y
		Y_points = np.linspace(Limit_min,Mimit_max,int(Y_points_baseline))           # Defines a uniform distribution of meshpoints along the Y-direction contained between the contour Y-limits


		# Y-MESH LOCAL REFINEMENT:

		"""
		def refine_by_normal_distribution(Y_points,Limit_min,Mimit_max,Ref_coord,Ref_width,Ref_point,Ref_sigma):
			import numpy as np
			xpts = np.linspace(0,3*Ref_sigma,Ref_point)                         # Creates points on the RHS of the Gauss curve with 3-sigma range
			Normal = np.exp(-0.5*((xpts)/Ref_sigma)**2)/(Ref_sigma*np.sqrt(2*np.pi))   # Computes the normal distribution for these x-points 
			Base = np.linspace(0,Ref_width,Ref_point)                           # Creates an uniform base point distribution within a desired Y-range
			Delta = (Base[0]-Base[1])*(1-Normal)                                # Determines the distance between points that follows a normal distribution (It is still the RHS of the Gauss curve)

			Y_points_normal = np.array([Ref_coord])                             # Creates a normal point distribution by adding points at both sides of the center of the refinement, by respecting the previous distance between points
			for i in range(Ref_point-1):
				Y_points_normal = np.append(Y_points_normal,Ref_coord + np.sum(Delta[:(i+1)]))                  # It adds points on the right 
				Y_points_normal = np.insert(Y_points_normal, 0, Ref_coord - np.sum(Delta[:(i+1)]), axis=0)      # It adds points on the left

			Y_points_normal = np.fliplr([Y_points_normal])[0]                   # It flips the resulting array in order to have the smaller values at the beginning of the array and the higher values at the end

			Y_points_normal = Y_points_normal[Y_points_normal>(Ref_coord-0.5*Ref_width)]  # It clips the distribution in order to retain the points inside the desired refinement region
			Y_points_normal = Y_points_normal[Y_points_normal<(Ref_coord+0.5*Ref_width)]

			Mixed_Y = np.concatenate((Y_points[Y_points<(Ref_coord-0.5*Ref_width)],Y_points_normal, Y_points[Y_points>(Ref_coord+0.5*Ref_width)] ), axis=0) # It takes the reference uniform mesh, then it take away its points belonging to the Min_value<y<max_value range, and then it adds the normal distribution points on this gap. So we have a uniform mesh with a local refinement;

			Mixed_Y = Mixed_Y[Mixed_Y>=Limit_min]                               #Delete the points lying outside the wake contour line
			Mixed_Y = Mixed_Y[Mixed_Y<=Mimit_max]

			return Mixed_Y
		"""

		if Refinement_1 == 2:                                                   # Refinement 1 along Y-axis activated
			Y_points = refine_by_normal_distribution(Y_points,Limit_min,Mimit_max,Ref_1_coord,Ref_1_width,Ref_1_point,Ref_1_sigma)
			Y_points_baseline = Y_points.size

		if Refinement_2 == 2:                                                   # Refinement 2 along Y-axis activated     
			Y_points = refine_by_normal_distribution(Y_points,Limit_min,Mimit_max,Ref_2_coord,Ref_2_width,Ref_2_point,Ref_2_sigma)
			Y_points_baseline = Y_points.size       

		if Refinement_3 == 2:                                                   # Refinement 3 along Y-axis activated     
			Y_points = refine_by_normal_distribution(Y_points,Limit_min,Mimit_max,Ref_3_coord,Ref_3_width,Ref_3_point,Ref_3_sigma)
			Y_points_baseline = Y_points.size        

		if Refinement_4 == 2:                                                   # Refinement 4 along Y-axis activated     
			Y_points = refine_by_normal_distribution(Y_points,Limit_min,Mimit_max,Ref_4_coord,Ref_4_width,Ref_4_point,Ref_4_sigma)
			Y_points_baseline = Y_points.size       

		if Refinement_5 == 2:                                                   # Refinement 5 along Y-axis activated     
			Y_points = refine_by_normal_distribution(Y_points,Limit_min,Mimit_max,Ref_5_coord,Ref_5_width,Ref_5_point,Ref_5_sigma)
			Y_points_baseline = Y_points.size   


		# DELETE THE Y-POINTS LYING OUTSIDE THE TUNNEL WALL MARGIN LIMITS: 

		if Wind_Tunnel_Model > 0:                                       # Value > 0 takes into account the presence of walls
			if np.size(Y_points[Y_points<Ymin])!=0:
				Y_points = Y_points[Ymin<Y_points]
				Y_points = np.insert(Y_points, 0, Ymin, axis=0)

			if np.size(Y_points[Y_points>Ymax])!=0:
				Y_points = Y_points[Y_points<Ymax]
				Y_points = np.append(Y_points, Ymax)           

			Y_points_baseline = Y_points.size     


		# Z-MESH LOCAL REFINEMENT:
		Xmesh = []                                                              # Arrays to store the wake shape mapped XYZ grid coordinates
		Xmesh_path = []                                                         # Array to store the X-grid points to be surveyed by the probe
		Ymesh = []
		Ymesh_path = []       
		Zmesh = []
		Zmesh_path = []                                                         # Array to store the Z-grid points to be surveyed by the probe
		LenghtZ = []                                                            # Array to store the lenght along Z-axis traveled by the probe during the wake survey
		Zmax = []                                                               # Array to store the maximum coordinate of the contour line at each Y-station
		Zmin = []                                                               # Array to store the minimum coordinate of the contour line at each Y-station

		for i in Y_points:                                                      # Create the grid points by analyzing each Y-coordinate of the baseline distribution

			# DETERMINES DISTANCE BETWEEN THE UPPER/LOWER WAKE LIMITS FOR EACH Y-POINT:
			plane = vtk.vtkPlane()                                              # Create a plane perpendicular to Y-axis
			plane.SetOrigin(0, i, 0)
			plane.SetNormal(0, 1, 0)

			cutter = vtk.vtkCutter()                                            # The plane is used to cut the wake contour line in order to find the intersection points
			cutter.SetCutFunction(plane)
			cutter.SetInputConnection(contour.GetOutputPort())
			cutter.Update()

			Wake_Limits = dsa.WrapDataObject(cutter.GetOutput())                # Converts VTK element into a Python object
			Wake_Points = Wake_Limits.Points                                    # Get intersection points coordinates 
			X_wake = Wake_Points[:,0] 
			Y_wake = Wake_Points[:,1] 
			Z_wake = Wake_Points[:,2]

			LenghtZ = np.append(LenghtZ,np.max(Z_wake)-np.min(Z_wake))          # Calculates the vertical distance travelled by a probe
			Zmax = np.append(Zmax,np.max(Z_wake))
			Zmin = np.append(Zmin,np.min(Z_wake))

		if Z_mesh_method == 0:                                                  # Constant Z-points for each spanwise station

			# DETERMINES Z-POINTS DISTRIBUTION:
			Line_ID = 0                                                         # Flag to know if a survey line along Z-axis goes upwards or downwards
			for i in range(np.size(Y_points)):           

				Z_points = np.linspace(Zmin[i],Zmax[i],Mesh_Zpts_min)           # Creates a uniform distribution of points along the Z-direction between the lower and upper intersection points

				if (int(Line_ID) % 2) == 0:                                     # This stores the same points as Z_points but reverses the order to have points following a zigzag pattern to optimize the traverser path
					Zmesh_path = np.append(Zmesh_path,Z_points)
				else:
					Zmesh_path = np.append(Zmesh_path,np.flipud(Z_points))
				Line_ID = Line_ID+1

				for j in Z_points:                                              # Creates the X-, Y- and Z-coordinates of the wake shape mapped grid
					Xmesh = np.append(Xmesh,X[0])
					Ymesh = np.append(Ymesh,Y_points[i])
					Zmesh = np.append(Zmesh,j)

				for k in range(np.size(Z_points)):                              # Creates the X-, Y- and Z-coordinates of the traverser path
					Xmesh_path = np.append(Xmesh_path,X[0])
					Ymesh_path = np.append(Ymesh_path,Y_points[i])


		if Z_mesh_method == 1:                                                  # Number of local Z-points proportional to the local wake thicknes

			if Wake_thickness == 0:      #0=automatic detection of minimum and maximum wake thickness                 
				# USES THE MIN/MAX WAKE THICKNESS ALREADY CALCULATED
				Thickness_max = np.max(LenghtZ)
				Thickness_min = np.min(LenghtZ)


			if Wake_thickness == 1:      #0=automatic detection of minimum and maximum wake thickness
				# CALCULATES THE WAKE THICKNESS AT THE PRESCRIBED Y-COORDINATES BY THE USER:

				def WAKE_THICKNESS_FUNCTION(Ycoordinate,contour):
					import vtk
					import numpy as np
					from vtk.numpy_interface import dataset_adapter as dsa

					plane = vtk.vtkPlane()                                              # Create a plane perpendicular to Y-axis
					plane.SetOrigin(0, Ycoordinate, 0)
					plane.SetNormal(0, 1, 0)

					cutter = vtk.vtkCutter()                                            # The plane is used to cut the wake contour line in order to find the intersection points
					cutter.SetCutFunction(plane)
					cutter.SetInputConnection(contour.GetOutputPort())
					cutter.Update()

					Wake_Limits = dsa.WrapDataObject(cutter.GetOutput())                # Converts VTK element into a Python object
					Wake_Points = Wake_Limits.Points                                    # Get intersection points coordinates 
					X_wake = Wake_Points[:,0] 
					Y_wake = Wake_Points[:,1] 
					Z_wake = Wake_Points[:,2]

					LenghtZ = np.max(Z_wake)-np.min(Z_wake)                             # Calculates the vertical distance travelled by a probe
					return LenghtZ

				Thickness_max = WAKE_THICKNESS_FUNCTION(Y_max_wake,contour) 
				Thickness_min = WAKE_THICKNESS_FUNCTION(Y_min_wake,contour)             


			Line_ID = 0                                                         # Flag to know if a survey line along Z-axis goes upwards or downwards
			for i in range(np.size(Y_points)):           

				# DETERMINES Z-POINTS DISTRIBUTION:
				NBzPts_low = ((Mesh_Zpts_max-Mesh_Zpts_min)/(Thickness_max-Thickness_min))*(LenghtZ[i]-Thickness_min)+Mesh_Zpts_min  # Number of points inside the wake proportional to the wake thickness: Y = ((y2-y1)/(x2-x1))*(X-x1)+y1                   
				NBzPts_low = int(NBzPts_low)                                    # Number of points inside the wake (lower part of the mesh) 

				if NBzPts_low<Mesh_Zpts_min:                                    # This avoids having less nodes than the minimum prescribed value
					NBzPts_low = Mesh_Zpts_min

				if NBzPts_low>Mesh_Zpts_max:                                    # This avoids having more nodes than the maximum prescribed value
					NBzPts_low = Mesh_Zpts_max               


				NBzPts_up = Mesh_Zpts_max-NBzPts_low                            # Number of points outside the wake (upper part of the mesh) 

				Z_points_low = np.linspace(Zmin[i],Zmax[i],NBzPts_low)          # Creates a uniform distribution of points along the Z-direction between the lower and upper intersection points
				DeltaPTS = Z_points_low[1]-Z_points_low[0]                      # Distance between points in the lower part of the mesh

				if NBzPts_up == 0:              
					Z_points_up = []                                            # Empty array to define Z_points_up for the case where there are no points in the upper zone of the mesh
				if NBzPts_up != 0: 
					Z_points_up = np.linspace(Zmax[i],Zmax[i]+DeltaPTS*NBzPts_up,NBzPts_up+1)
					#Z_points_up = np.linspace(Zmax[i],Zmin[i]+np.max(LenghtZ),NBzPts_up+1)  # The upper limit of the mash is defined in such a way that the mesh thickness is constant

				Z_points_up = Z_points_up[1:]                                   # The first point of the array is deleted because it is coincident with the last point of Z_points_low
				Z_points = np.concatenate((Z_points_low,Z_points_up), axis=0) 

				if (int(Line_ID) % 2) == 0:                                     # This stores the same points as Z_points but reverses the order to have points following a zigzag pattern to optimize the traverser path
					Zmesh_path = np.append(Zmesh_path,Z_points_low)              # The probe path will only follow the lower part of the mesh
				else:
					Zmesh_path = np.append(Zmesh_path,np.flipud(Z_points_low))
				Line_ID = Line_ID+1

				for j in Z_points:                                              # Creates the X-, Y- and Z-coordinates of the wake shape mapped grid
					Xmesh = np.append(Xmesh,X[0])
					Ymesh = np.append(Ymesh,Y_points[i])
					Zmesh = np.append(Zmesh,j)

				for k in range(np.size(Z_points_low)):                              # Creates the X-, Y- and Z-coordinates of the traverser path
					Xmesh_path = np.append(Xmesh_path,X[0])
					Ymesh_path = np.append(Ymesh_path,Y_points[i])


		MESH_coordinates = algs.make_vector(Xmesh, Ymesh, Zmesh)                # convert the 3 arrays into a single three-component array to define the mesh points.
		PROBE_coordinates = algs.make_vector(Xmesh_path, Ymesh_path, Zmesh_path) # vector with the points for the traverser path line


		"""
		IMPORTANT: PLANE INCLINATION
		Input plane was assumed to be parallel to the axis reference system
		If the plane is inclined, the variables Xmesh and Xmesh_path WILL NOT BE OK!!!! 
		This can be solved by transforming (rotation) the input plane to get a plane aligned with the coordinate system, then computing the mesh coordinates and finally
		rotating back the output mesh to the original inclination. This affects both Z-meshing techniques.    
		"""


		#if Step == 5 or Step == 6 or Step == 7:
		# SEND COORDINATES DATA TO THE OUTPUT ARRAY:
		output = self.GetOutput()                                           # It calls the output port of the Paraview source
		pts = vtk.vtkPoints()                                               # create a vtkPoints container to store all the point coordinates.
		pts.SetData(dsa.numpyTovtkDataArray(MESH_coordinates , "Points"))   # numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()
		output.SetPoints(pts)                                               # It sends the points to the output


		# ALLOCATE CELL DATA:
		if Z_mesh_method == 0: 
			NB_Z_PTS = Mesh_Zpts_min
		if Z_mesh_method == 1: 
			NB_Z_PTS = Mesh_Zpts_max

		IDlist = []
		for i in range(int(Y_points_baseline)-1):                           # We iterate over each cell in order to define the counter-clockwise point ID that defines the cell connectivity
			for j in range(int(NB_Z_PTS)-1):
				IDlist.append([i*NB_Z_PTS+j,(i+1)*NB_Z_PTS+j,(i+1)*NB_Z_PTS+(j+1),i*NB_Z_PTS+j+1])
		IDlist = np.array(IDlist)

		NBcells=(Y_points_baseline-1)*(NB_Z_PTS-1)                     # Number of cells

		numCells = int(NBcells)
		output.Allocate(numCells, 1000)                                     # Allocate the cells in the memory
		for cell in range(numCells):
			numCellPoints = int(4)                                          # There are 4 points for each cell (vtk_quad cell type)
			pointIds = vtk.vtkIdList()                                      # Creates an empty array to store the cells ID's
			for pointId in range(numCellPoints):
				pointIds.InsertId(pointId, int(IDlist[cell][pointId]))
			output.InsertNextCell(9, pointIds)                              #9 stands for VTK_quad cell type

		# ADD FIELD DATA:
		output = self.GetOutput()
		NBpts = np.size(Zmesh_path)   
		Total_distance = np.sum(LenghtZ)+(np.max(Y_points)-np.min(Y_points))

		addfielddata(1, output.GetFieldData(),np.array([Total_distance]),"Traveled_distance[m]")            
		addfielddata(1, output.GetFieldData(),np.array([int(NBpts)]),"Mesh_Points")
		
		# CREATES VERTICES AT EACH MEASH POINT TO USE IT AS PROBES FOR INTERPOLATION (To interpolate input mesh data into the wake survey mesh)
		polydataOK = vtk.vtkPolyData()
		polydataOK.SetPoints(pts)

		VerticesMesh = vtk.vtkVertexGlyphFilter()
		VerticesMesh.AddInputData(polydataOK)
		VerticesMesh.Update()  

		# ASSIGN VARIABLE VALUES FROM INPUT DATA TO THE VERTICES
		interpolation_vertices = vtk.vtkProbeFilter()
		interpolation_vertices.SetInputConnection(VerticesMesh.GetOutputPort())
		interpolation_vertices.SetSourceData(InputMesh.GetOutput())  
		interpolation_vertices.Update()        

		# ADD INTERPOLATED DATA ARRAYS TO THE WAKE MESH
		WrapData = dsa.WrapDataObject(interpolation_vertices.GetOutput())
		Headers = WrapData.PointData.keys()

		for L in range(interpolation_vertices.GetOutput().GetPointData().GetNumberOfArrays()):
			Variable = WrapData.PointData[Headers[L]]

			if interpolation_vertices.GetOutput().GetPointData().GetArray(L).GetNumberOfComponents() == 1:      # Add scalar variables to the mesh
				addfielddata(1, output.GetPointData(),Variable,Headers[L])

			if interpolation_vertices.GetOutput().GetPointData().GetArray(L).GetNumberOfComponents() == 3:      # Add vector variables to the mesh
				addfielddata(3, output.GetPointData(),Variable.ravel(),Headers[L])
		
		#SCALE THE MESH TO [MM]:
		Survey_Plane_Mesh = SCALE_GEOMETRY(self,1)
		self.GetOutput().ShallowCopy(Survey_Plane_Mesh.GetOutput())					
		
		if Step == 5:
			# PRINT OUTPUT INFORMATION:
			print ("Mesh generated successfully")
			print ("Change the visualization mode from 'Surface' to 'Surface with edges' to see the cells:")
			print ("The current mesh displayed in the screen has the input data values interpolated into the output mesh points: use this to check if the grid refinement is enough to capture the field variations. If not, change the Z-mesh method and/or add local mesh refinements")
			print ("Points along Y-axis =", np.size(Y_points))
			PRINT_SURVEY_TIME(PROBE_coordinates,Survey_speed,Hold_time,DAQ_time)

	if Step == 6 or Step == 7:                                                	# Step=6(Check traverser path)   Step=7(export TXT file)

		# CREATES A LINE JOINING THE POINTS ALONG THE TRAVERSER PATH
		points = vtk.vtkPoints()                                                # create a vtkPoints container to store all the point coordinates.
		points.SetData(dsa.numpyTovtkDataArray(PROBE_coordinates , "Points"))   # numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()

		NBpts = np.size(Zmesh_path)                                             # Number of mesh points

		polyLine = vtk.vtkPolyLine()                                            # Creates a polyline and set the ID for each point
		polyLine.GetPointIds().SetNumberOfIds(int(NBpts)) 
		for i in range(0, int(NBpts)):
			polyLine.GetPointIds().SetId(i, i)

		cells = vtk.vtkCellArray()                                              # Create a cell array to store the lines in and add the lines to it
		cells.InsertNextCell(polyLine)

		polyData = vtk.vtkPolyData()                                            # Create a polydata to store everything in
		polyData.SetPoints(points)                                              # Add the points to the dataset
		polyData.SetLines(cells)                                                # Add the lines to the dataset

		Survey_Line = vtk.vtkGeometryFilter()                                   # Converts the polydata to vtkPolyData  
		Survey_Line.SetInputData(polyData)             
		Survey_Line.Update()


		# ASSIGN VARIABLE VALUES FROM INPUT DATA TO THE LINE
		probe_line = vtk.vtkProbeFilter()
		probe_line.SetInputConnection(Survey_Line.GetOutputPort())
		probe_line.SetSourceData(InputMesh.GetOutput())  
		probe_line.Update()  


		# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
		polydata2 = vtk.vtkPolyData()
		polydata2.SetPoints(points)

		Survey_Vertices = vtk.vtkVertexGlyphFilter()
		Survey_Vertices.AddInputData(polydata2)
		Survey_Vertices.Update()  


		# ASSIGN VARIABLE VALUES FROM INPUT DATA TO THE VERTICES
		probe_vertices = vtk.vtkProbeFilter()
		probe_vertices.SetInputConnection(Survey_Vertices.GetOutputPort())
		probe_vertices.SetSourceData(InputMesh.GetOutput())  
		probe_vertices.Update()        


		# COMBINE THE OUTPUTS FROM THE CLIP, VERTICES AND SURVEY PATH LINE INTO A SINGLE VARIABLE:
		appendFilter = vtk.vtkAppendFilter()                                    #Output: vtkUnstructuredGrid
		appendFilter.AddInputData(probe_line.GetOutput())        
		appendFilter.AddInputData(probe_vertices.GetOutput())
		if Hide_Wake == 0:
			appendFilter.AddInputData(CLIP.GetOutput())
		appendFilter.Update()


		# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA  
		combined = vtk.vtkGeometryFilter()                                      # Converts vtkMultiblockDataSet to vtkPolyData
		combined.SetInputData(appendFilter.GetOutput())                             
		combined.Update()

		if Step == 6:
			# MESH VISUALIZATION:
			self.GetOutput().ShallowCopy(combined.GetOutput())                  # Displays output data

			# SURVEY MESH STATISTICS:
			print ("Probe path generated successfully")
			if Hide_Wake == 0:
				print ("The wake shown behind the probe path is the input clipped wake data (Use 'Hide wake' to see the path only)")
			if Hide_Wake == 1:
				print ("The current path displayed in the screen has the input data values interpolated into the survey points: use this to check if the grid refinement is enough to capture the field variations")
				
			print ("Points along Y-axis =", np.size(Y_points))
			PRINT_SURVEY_TIME(PROBE_coordinates,Survey_speed,Hold_time,DAQ_time)
		
		if Step == 7:                                                	# Step=7(export TXT file)
			# CREATES OUTPUT TXT FILE WITH POINTS COORDINATES FOR TRAVERSER:
			EXPORT_PATH_TO_TXT(PROBE_coordinates,File_Path_Output,Output_TXT_in_mm,Output_TXT_without_headers,SelectTXTdelimiter,ReturnToStartingPoint,DecimalSeparator,ReturnToOrigin,OriginCoordinates)
			#WRITE_VTP(combined,File_Path_Output,0)  									# 0=survey line   (This saves the probe path with the input data interpolated --> kept for debugging)
			#WRITE_VTP(Survey_Plane_Mesh,File_Path_Output,1)  							# 1=survey mesh   (This saves the survey mesh with the input data interpolated --> kept for debugging)
			
			
			# DELETE ALL THE ARRAYS FOR THE PROBE PATH:
			PathWithoutArrays=vtk.vtkPassArrays()
			PathWithoutArrays.SetInputDataObject(combined.GetOutput())
			PathWithoutArrays.ClearArrays()
			PathWithoutArrays.AddPointDataArray("")
			PathWithoutArrays.AddCellDataArray("")
			PathWithoutArrays.AddFieldDataArray("")
			PathWithoutArrays.Update()
			
			WRITE_VTP(PathWithoutArrays,File_Path_Output,0)  							# 0=survey line
			

			# LET PASS THE FIELD DATA ARRAYS ONLY FOR THE MESH:
			MeshWithoutArrays=vtk.vtkPassArrays()
			MeshWithoutArrays.SetInputDataObject(Survey_Plane_Mesh.GetOutput())
			MeshWithoutArrays.ClearArrays()
			MeshWithoutArrays.AddPointDataArray("")
			MeshWithoutArrays.AddCellDataArray("")
			#MeshWithoutArrays.AddFieldDataArray("")
			MeshWithoutArrays.AddFieldDataArray("Mesh_Points")
			MeshWithoutArrays.AddFieldDataArray("Traveled_distance[m]")
			MeshWithoutArrays.Update()
			
			WRITE_VTP(MeshWithoutArrays,File_Path_Output,1)  							# 1=survey mesh
			
			# OUTPUT MESSAGES:
			print ("BE SURE TO SAVE THE CURRENT GUI SETTINGS OF 'Epsilon_Grid_Generator': File --> Save state")



############################################################################################################
#############################################    END OF CODE   #############################################
############################################################################################################