# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################              Integration Tool Plugin             #############################
############################################################################################################

# Abstract:
This filter takes the output from Epsilon_Aero_Data, Epsilon_Near_Field or Epsilon_Wave_Drag and integrates 
its variables depending upon the type of variable (survey plane integral or volume integral). Details of the 
formulations and how to use the plugin are available in the User's Manual.


# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""

print ("")
print ("")
print ("")
print ("####################################################")
print ("#########    EPSILON INTEGRATION TOOL      #########")
print ("####################################################")
print ("")



############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

# We load the Python and Paraview standard modules:
import vtk
import numpy as np
import vtk.vtkFiltersParallel
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa



############################################################################################################
############################################# INPUT PARAMETERS #############################################
############################################################################################################

# Input parameters are defined as string in Paraview's GUI, thus a string to float convertion is required:
Input_Module = Input_Module                                                                                 # 0=Epsilon Aero Data   1=Epsilon Near Field    2=Epsilon Poisson Solver    3=Epsilon Wave Drag
Integral_Type = Integral_Type				                                                                # 0=Single survey plane 1=Sweep survey plane  2=Spanwise distribution 3=3D to 2D mapping
Select_Action_0 = Select_Action_0                                                                           # 1=Visualize survey plane  2=Visualize control volume  3=Integrate variables
Select_Action_1 = Select_Action_1                                                                           # 0=Visualize sweep line    1=Visualize survey plane  2=Visualize control volume  3=Perform automatic integration
Select_Action_2 = Select_Action_2                                                                           # 1= Visualize survey plane 7=Visualize spanwise sweep line 4=Visualize integration sections    5=Visualize integration surface 3=Perform automatic integration
Select_Action_3 = Select_Action_3                                                                           # 6=Visualize mapping plane 3=Perform automatic integration
Input_Is_Survey_Plane = Input_Is_Survey_Plane                                                               # Boolean 
Plane_shape = Plane_shape                                                                                   # 0=Planar  1=Extruded curve    2=Custom surface(hidden option)
Load_curve = Load_curve                                                                                     # Path to the file of the curve that defines the survey plane extrusion profile
Sort_Direction = Sort_Direction                                                                             # 0=X-axis    1=Y-axis  2=Z-axis
Load_surface = Load_surface                                                                                 # Path to the file of the curstom survey plane surface
Plane_normal = Plane_normal                                                                                 # 0=Aerodynamic x-axis  1=User defined
Set_normal = Set_normal                                                                                     # User-defined value for the surface normal
X_single_plane = X_single_plane                                                                             # X-coordinate [m] of the center of the survey plane
Sweep_Axis_Survey = Sweep_Axis_Survey                                                                       # 0=Aerodynamic x-axis  1=Mesh X-axis
START_Survey = START_Survey                                                                                 # Initial position of the survey plane [m]
END_Survey = END_Survey                                                                                     # Final position of the survey plane [m]
NUMPTS_Survey = NUMPTS_Survey                                                                               # Number of points along the distribution line
Plane_Offset_Distance = Plane_Offset_Distance                                                               # Distance in [m] between the extruded survey plane and the actual domain slice/clip plane
Clip_Wake_Region = Clip_Wake_Region                                                                         # Boolean
Invert_Clip_Wake = Invert_Clip_Wake                                                                         # Boolean
Parameter = Parameter                                                                                       # Variable name used to clip the wake
Clip_Value = Clip_Value                                                                                     # Variable value used to clip the wake
Sweep_Axis_Spanwise = Sweep_Axis_Spanwise                                                                   # 0=Aerodynamic y-axis  1=Aerodynamic z-axis 2=Mesh Y-axis  3=Mesh Z-axis
NUMPTS_Spanwise = NUMPTS_Spanwise                                                                           # Number of points along the spanwise survey line
START_Spanwise = START_Spanwise                                                                             # Initial position of the spanwise survey line [m]
END_Spanwise = END_Spanwise                                                                                 # Final position of the spanwise survey line [m]
Mapping_Normal = Mapping_Normal                                                                             # 0=Aerodynamic x-axis  1=Aerodynamic y-axis  2=Aerodynamic z-axis  3=Mesh X-axis  4=Mesh Y-axis  5=Mesh Z-axis
NUMPTS_IJ = NUMPTS_IJ								                                                        # Number of points along the IJ axes of the mapping plane
START_IJ = START_IJ								                                                            # Start coordinates [m] of the IJ axes
END_IJ = END_IJ									                                                            # End coordinates [m] of the IJ axes
FAR_FIELD = FAR_FIELD                                                                                       # 0=None    1=All   2=Select
Momentum = Momentum                                                                                         # Boolean
Betz = Betz                                                                                                 # Boolean
Jones = Jones                                                                                               # Boolean
Oswatitsch = Oswatitsch                                                                                     # Boolean
Maskell = Maskell                                                                                           # Boolean
Onorato = Onorato                                                                                           # Boolean
VDV = VDV                                                                                                   # Boolean
Giles = Giles                                                                                               # Boolean
Kusunose = Kusunose                                                                                         # Boolean
Meheut = Meheut                                                                                             # Boolean
Toubin = Toubin                                                                                             # Boolean
EXERGY = EXERGY                                                                                             # 0=None    1=All   2=Select
Arntz = Arntz                                                                                               # Boolean
Aguirre = Aguirre                                                                                           # Boolean
Ruscio = Ruscio                                                                                             # Boolean
LAMB_VECTOR = LAMB_VECTOR                                                                                   # 0=None    1=All   2=Select
Wu = Wu                                                                                                     # Boolean
Mele = Mele                                                                                                 # Boolean
UDV = UDV                                                                                                   # Boolean
Flip_Plane_Normal = Flip_Plane_Normal                                                                       # Boolean:  0=Do not reverse plane normal     1=Reverse plane normal
Invert_Clip = Invert_Clip                                                                                   # Boolean:  0=Keep current side of the plane   1=Keep the other side
Save_Output_Data = Save_Output_Data                                                                         # Boolean:  0=Do not save data    1=Save data
File_Path_Output = File_Path_Output                                                                         # Path to the output VTP file
Plane_Body_Intersection = Plane_Body_Intersection                                                           # Boolean:  0=Plane data only     1=Plane + upstream body surface data
Load_Body_Surface = Load_Body_Surface                                                                       # VTP file with the body surface mesh from Epsilon_Near_Field

if Integral_Type == 0:                                                                                      # 0=Single survey plane 
	Select_Action = Select_Action_0
if Integral_Type == 1:                                                                                      # 1=Sweep survey plane
	Select_Action = Select_Action_1
if Integral_Type == 2:                                                                                      # 2=Spanwise distribution
	Select_Action = Select_Action_2
if Integral_Type == 3:                                                                                      # 3=3D to 2D mapping
	Select_Action = Select_Action_3


"""
• Plane-body intersection : testear en 3D y si funca bien, ver como carajo generalizarlo a 2D...
• Custom surface: currently HIDDEN. There is aproblem with clip by scalar by dustance
• Spanwise distribution: divide by chord. Yet to finish
• Output message error in SPLIT_VECTOR_ARRAYS
"""



############################################################################################################
#######################################    DEFINITION OF FUNCTIONS    ######################################
############################################################################################################

############################################
#########  LOAD VTK DATA FUNCTION  #########
############################################

def LOAD_VTK_DATA(FilePath):                                                                                # Defines a function that takes a filepath (String) pointing to a VTK object to be loaded. Accepted formats: VTU, VTP, VTS, VTI and VTK
	import vtk
	reader = None                                                                                           # Generate a vtk reader depending on the input file format
	if FilePath.endswith(".vtu"):
		reader = vtk.vtkXMLUnstructuredGridReader()
	elif FilePath.endswith(".vtp"):
		reader = vtk.vtkXMLPolyDataReader()
	elif FilePath.endswith(".vts"):
		reader = vtk.vtkXMLStructuredGridReader()
	elif FilePath.endswith(".vti"):
		reader = vtk.vtkXMLImageDataReader()
	elif FilePath.endswith(".vtk"):    
		reader = vtk.vtkGenericDataObjectReader()
	elif FilePath.endswith(".stl"):        
		reader = vtk.vtkSTLReader()
	else:
		print ("Input file format not recognized")
	reader.SetFileName(FilePath)
	reader.Update() 
	return reader


############################################
#########  PLOT VTK LINES FUNCTION  ########
############################################

def PLOT_VTK_LINES_VERTEX(PointsVector):                                                                    # Defines a function that takes a numpy array with the points coordinates and generates two VTK outputs: the polyline connecting the points as well as the points themselves
	import numpy as np
	import vtk

	points = vtk.vtkPoints()                                                	                            # create a vtkPoints container to store all the point coordinates.
	points.SetData(dsa.numpyTovtkDataArray(PointsVector , "Points"))   					                    # numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()
	
	# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
	polydata2 = vtk.vtkPolyData()
	polydata2.SetPoints(points)

	Survey_Vertices = vtk.vtkVertexGlyphFilter()
	Survey_Vertices.AddInputData(polydata2)
	Survey_Vertices.Update()
	
	# GENERATE LINE LINKING VERTICES    
	NBpts = np.shape(PointsVector)[0]                                             		                    # Number of mesh points

	if NBpts > 1:
		polyLine = vtk.vtkPolyLine()                                            	                            # Creates a polyline and set the ID for each point
		polyLine.GetPointIds().SetNumberOfIds(int(NBpts)) 
		for i in range(0, int(NBpts)):
			polyLine.GetPointIds().SetId(i, i)

		cells = vtk.vtkCellArray()                                              	                            # Create a cell array to store the lines in and add the lines to it
		cells.InsertNextCell(polyLine)

		polyData = vtk.vtkPolyData()                                            	                            # Create a polydata to store everything in
		polyData.SetPoints(points)                                              	                            # Add the points to the dataset
		polyData.SetLines(cells)                                                	                            # Add the lines to the dataset

		Survey_Line = vtk.vtkGeometryFilter()                                   	                            # Converts the polydata to vtkPolyData  
		Survey_Line.SetInputData(polyData)             
		Survey_Line.Update()

	else: 
		Survey_Line = Survey_Vertices

	return Survey_Line,Survey_Vertices


###########################################################
##################### SURFACE NORMALS #####################
###########################################################
	
def SURFACE_NORMALS(Object,Flip_Plane_Normal):                                                              # Defines a function that computes the normals of a VTK object at the cell centers and the points
	import vtk
	Surface_Normals = vtk.vtkPolyDataNormals()
	Surface_Normals.SetInputData(Object.GetOutput())
	Surface_Normals.SetFeatureAngle(0)
	Surface_Normals.ComputeCellNormalsOn()
	Surface_Normals.ComputePointNormalsOn()
	Surface_Normals.AutoOrientNormalsOn()                                                                   # If not all the normals are pointing towards the same direction, it flips the normals of the reversed cells (so, all the normals are consistent on the entire surface)
	Surface_Normals.FlipNormalsOn()
	if Flip_Plane_Normal == 1:
		Surface_Normals.FlipNormalsOff()
	Surface_Normals.Update()
	return Surface_Normals


###########################################################
#################### LINEAR EXTRUSION #####################
###########################################################
	
def LINEAR_EXTRUSION(Object,Vector,Lenght):                                                                   # Defines a function that extrudes a VTK object along a direction defined by a vector and a extrusion lenght defined by Lenght
	import vtk
	extrusion = vtk.vtkLinearExtrusionFilter()
	extrusion.CappingOff()
	extrusion.SetExtrusionTypeToVectorExtrusion()
	extrusion.SetVector(Vector[0],Vector[1],Vector[2])
	extrusion.SetInputData(Object.GetOutput())
	extrusion.SetScaleFactor(Lenght)                                      
	extrusion.Update() 
	return extrusion
    

###########################################################
################### APPEND VTK OBJECTS ####################
###########################################################
	
def APPEND_VTK_OBJECTS(Object1, Object2):                                                                   # Defines a function that combines two VTK objects into a single VTK object. Output: vtkUnstructuredGrid
	import vtk
	AppendFilter = vtk.vtkAppendFilter()                                                                   
	AppendFilter.AddInputData(Object1.GetOutput()) 
	AppendFilter.AddInputData(Object2.GetOutput())        
	AppendFilter.Update()
	return AppendFilter


###########################################################
##################### GEOMETRY FILTER #####################
###########################################################

def EXTRACT_SURFACE(VTK_Object):                                  	                                        # Defines a function that converts the input data to vtkPolyData. It also extract surface if the volume mesh
	import vtk
	PolyData = vtk.vtkGeometryFilter()   
	PolyData.SetInputData(VTK_Object.GetOutput())             
	PolyData.Update()
	return PolyData


###########################################################
################# DISTANCE BETWEEN OBJECTS ################
###########################################################

def DISTANCE_BETWEEN_OBJECTS(ReferenceObject,TargetObject):                                                 # Defines a function that calculates the distance from each TargetObject point to the ReferenceObject. IMPORTANT: if the objects are data from the input it is not necessary to add .GetOutput() (this is required for VTK objects only)
	import vtk
	pdd = vtk.vtkImplicitPolyDataDistance()
	pdd.SetInput(ReferenceObject)
	numPts = dsa.WrapDataObject(TargetObject).Points.shape[0]  
	distance = vtk.vtkDoubleArray()
	distance.SetName("Distance")
	distance.SetNumberOfComponents(1)
	distance.SetNumberOfTuples(numPts)
	for i in range(numPts):
		pt = TargetObject.GetPoint(i)
		distance.SetValue(i, pdd.EvaluateFunction(pt))
	TargetObject.GetPointData().SetScalars(distance)                                                        # The "Distance" array is sent to the TargetObject


###########################################################
##################### SLICE WITH PLANE ####################
###########################################################

def SLICE_WITH_PLANE(Object,plane):                                                                         # Defines a function that slices the VTK Object with a VTK implicit plane function. IMPORTANT: if the objects are data from the input it is not necessary to add .GetOutput() (this is required for VTK objects only)
	import vtk
	cutter = vtk.vtkCutter()														                    
	cutter.SetCutFunction(plane)
	cutter.SetInputData(Object)
	cutter.Update()
	return cutter


###########################################################
##################### SLICE BY SCALAR #####################
###########################################################

def SLICE_BY_SCALAR(Object,Scalar,ConstantValue):                                                           # Defines a function that slices the VTK Object with a surface of given by Scalar=ConstantValue. IMPORTANT: if the objects are data from the input it is not necessary to add .GetOutput() (this is required for VTK objects only)
	import vtk
	slicer = vtk.vtkContourFilter()
	slicer.SetInputData(Object)
	slicer.SetInputArrayToProcess(0,0,0,0, Scalar)
	slicer.SetValue(0, ConstantValue)                                                      
	slicer.Update()
	return slicer


###########################################################
##################### CLIP BY SCALAR ######################
###########################################################

def CLIP_BY_SCALAR(Object,Scalar,ConstantValue,Invert_Clip):                                                # Defines a function that clips the VTK Object with a surface of given by Scalar=ConstantValue. IMPORTANT: if the objects are data from the input it is not necessary to add .GetOutput() (this is required for VTK objects only)
	import vtk
	import numpy as np
	clipper = vtk.vtkTableBasedClipDataSet()                                                                # It is the faster vtk tool. It works in all ParaView versions
	clipper.SetInputData(Object)
	clipper.SetInputArrayToProcess(0,0,0,0, Scalar)
	clipper.SetValue(ConstantValue)                                                          
	clipper.InsideOutOff()
	if Invert_Clip == 1:
		clipper.InsideOutOn()                                                                               # Selects one side or the other of the clip                                                                               
	clipper.Update()

	if str(clipper.GetOutput().GetBounds()) == "(1e+299, -1e+299, 1e+299, -1e+299, 1e+299, -1e+299)":    # This means that the result of the clip by scalar is an empty object
		print("# ERROR: THE SURVEY PLANE DO NOT INTERSECT THE CLIPPED VOLUME. Set the initial position of the sweep line closer to the body")

	return clipper


###########################################################
###################### CLIP BY PLANE ######################
###########################################################

def CLIP_BY_PLANE(Object,plane,Invert_Clip):                                                                # Defines a function that clips the VTK Object with a VTK plane implicit function. IMPORTANT: if the objects are data from the input it is not necessary to add .GetOutput() (this is required for VTK objects only)
	import vtk    
	clipper = vtk.vtkTableBasedClipDataSet()                                                                # It is the faster vtk tool. It works in all ParaView versions
	clipper.SetClipFunction(plane)
	clipper.SetInputData(Object)          
	clipper.InsideOutOn()
	if Invert_Clip == 1:
		clipper.InsideOutOff()                                                                              # Selects one side or the other of the clip
	clipper.Update() 
	return clipper
    

###########################################################
############# ADD NEW VARIABLES TO THE OUTPUT #############
###########################################################
	
def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName):                                         # Defines a function that adds the field data to the existing points in the Paraview's output
	import vtk
	numberOfComponents = int(numTuples)                                                                     # We deal only with 1-component arrays
	dataArray = vtk.vtkDoubleArray()   
	dataSetAttributes.AddArray(dataArray)
	dataArray.SetNumberOfComponents(numberOfComponents)
	dataArray.SetName(ArrayName)                                                                            # "ArrayName" is a string used to name the array in Paraview 
	data = DataArray                                                                                        # DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
	for value in data:
		dataArray.InsertNextValue(value)                                                                    # Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 


###########################################################
######### DETECT INTEGRATION MODE OF EACH VARIABLE ########
###########################################################

def INTEGRATION_MODE(input,L):  
	VariableName = input.GetPointData().GetArray(L).GetName()
	InputWrapped = dsa.WrapDataObject(input) 
	if InputWrapped.GetFieldData().HasArray("ARRAY")==1:                                       	            # This check if "ARRAY" array exists in the input 
		NBarrays = InputWrapped.FieldData["ARRAY"].GetNumberOfValues()
	else:
		NBarrays = 0
	IntegrationMode = "None"
	for k in range(NBarrays):
		if InputWrapped.FieldData["ARRAY"].GetValue(k) == VariableName:
			IntegrationMode = InputWrapped.FieldData["ARRAY_Integration"].GetValue(k)
	return IntegrationMode


###########################################################
################## SURVEY PLANE INTEGRAL ##################
###########################################################

def SURVEY_PLANE_INTEGRAL(i,Survey_Plane_Mesh,Survey_Plane_Values,INTEGRATION_MODE,Input_Module,Sref,T_inverse,addfielddata,Input_Is_Survey_Plane,Plane_shape,Plane_normal,Plane_Body_Intersection,Integral_Type,SLICE_WITH_PLANE,plane):
	import vtk
	import numpy as np                

	# Get body chord:
	if Input_Module == 1:                                                                                   # 1=Near Field Method
		Bounds = Survey_Plane_Mesh.GetOutput().GetBounds()
		Chord = Bounds[1]-Bounds[0]
		if str(Chord) == "-2e+299":                                                                         # This value arises when the survey plane do not intersect the body, then the local chord is zero
			Chord = 1

	# Pass the survey plane variables only:
	Input_Filtered=vtk.vtkPassArrays()
	Input_Filtered.SetInputDataObject(Survey_Plane_Mesh.GetOutput())
	Input_Filtered.ClearArrays()

	CounterPointData = 0                                                                                    # Pass the point data variables that can only be integrated on the survey plane
	for L in range(Survey_Plane_Mesh.GetOutput().GetPointData().GetNumberOfArrays()):
		if INTEGRATION_MODE(Survey_Plane_Mesh.GetOutput(),L) == "Survey_Plane":
			Input_Filtered.AddPointDataArray(Survey_Plane_Mesh.GetOutput().GetPointData().GetArray(L).GetName())
			CounterPointData = CounterPointData + 1
	if CounterPointData == 0:
		Input_Filtered.AddPointDataArray("")

	CounterCellData = 0                                                                                     # Pass cell-based normals only
	for L in range(Survey_Plane_Mesh.GetOutput().GetCellData().GetNumberOfArrays()):
		if Survey_Plane_Mesh.GetOutput().GetCellData().GetArray(L).GetName().startswith('Normals')==True:
			Input_Filtered.AddCellDataArray(Survey_Plane_Mesh.GetOutput().GetCellData().GetArray(L).GetName())
			CounterCellData = CounterCellData + 1
	if CounterCellData == 0:
		Input_Filtered.AddCellDataArray("")
		
	Input_Filtered.Update()             

	if Input_Is_Survey_Plane==0 and Plane_shape==0 and Plane_normal==0 and Plane_Body_Intersection==0:      # Only for planar survey planes normal to aero x-axis (note that is the inout is already a survey plane, we cannot ensure that it is normal to aero x-axis)

		if Integral_Type == 2:                                                                              # 2=Spanwise distribution
			Input_Filtered = SLICE_WITH_PLANE(Input_Filtered.GetOutput(),plane)                             # Slices survey plane with slice plane (this gives the local spanwise station data)

		# Integrate variables on the survey plane:
		integrate = vtk.vtkIntegrateAttributes()
		integrate.SetInputData(Input_Filtered.GetOutput())       
		integrate.Update()
		
		# Fill "Survey_Plane_Values" with the variable names and integrated values:
		for j in range(integrate.GetOutput().GetPointData().GetNumberOfArrays()):
			if i == 0:
				arrayName = integrate.GetOutput().GetPointData().GetArray(j).GetName()
				Survey_Plane_Values.append(vtk.vtkDoubleArray())
				Survey_Plane_Values[j].SetName(arrayName)
			
			arrayValue = integrate.GetOutput().GetPointData().GetArray(j).GetTuple(0)[0]
			#arrayValue = arrayValue/Chord  # This is to divide the integrated value by the chord but it doesnt work... 
			Survey_Plane_Values[j].InsertNextValue(arrayValue)



	# Compute (V.N) for inclined survey planes:
	if Input_Is_Survey_Plane!=0 or Plane_shape!=0 or Plane_normal!=0 or Plane_Body_Intersection!=0: 
	
		p2c = vtk.vtkPointDataToCellData()	                                                                # POINT DATA TO CELL DATA CONVERSION:		  				
		p2c.SetInputConnection(Input_Filtered.GetOutputPort())
		p2c.Update()

		EmptyMesh=vtk.vtkPassArrays()                                                                       # Get the empty mesh without arrays
		EmptyMesh.SetInputDataObject(p2c.GetOutput())
		EmptyMesh.ClearArrays()
		EmptyMesh.AddPointDataArray("")
		EmptyMesh.AddCellDataArray("")
		EmptyMesh.AddFieldDataArray("")
		EmptyMesh.Update() 

		Wrapped = dsa.WrapDataObject(p2c.GetOutput()) 													    # CONVERT THE CELL-BASED VTK ARRAYS INTO NUMPY ARRAYS: 
		Normals_cell = Wrapped.CellData["Normals"]                                                          # Get the cell data

		normal = np.array([1,0,0])								                                            # x-axis aerodynamic normal direction
		NORMAL = np.dot(T_inverse,normal)   					                                            # Express the normal in the mesh reference frame

		Normals_cell_xaero = []                                                                             # Creates a matrix with the aerodynamic x-axis normal at each point of the survey plane
		for k in range(0, np.size(Normals_cell[:,0])):
			Normals_cell_xaero.append([NORMAL[0], NORMAL[1], NORMAL[2]])

		VdotN = np.einsum('ij,ij->i',Normals_cell,Normals_cell_xaero)                                       # This emulates the (V.N) factor of the equations involving surface flux of a scalar quantity
		
		for L in range(p2c.GetOutput().GetCellData().GetNumberOfArrays()):                                  # Add the corrected variables to the empty VTK object
			ArrayNameCell = p2c.GetOutput().GetCellData().GetArray(L).GetName()
			CellVariable = Wrapped.CellData[ArrayNameCell]
			CorrectedCellVariable = CellVariable*VdotN
			if ArrayNameCell != "Normals":
				addfielddata(1, EmptyMesh.GetOutput().GetCellData(),CorrectedCellVariable,ArrayNameCell)

		if Integral_Type == 2:                                                                              # 2=Spanwise distribution
			EmptyMesh = SLICE_WITH_PLANE(EmptyMesh.GetOutput(),plane)                                       # Slices survey plane with slice plane (this gives the local spanwise station data)

		integrate = vtk.vtkIntegrateAttributes()                                                            # Integrate variables on the survey plane: 
		integrate.SetInputData(EmptyMesh.GetOutput())   
		integrate.Update()
		
		for j in range(integrate.GetOutput().GetCellData().GetNumberOfArrays()):                            # Fill "Survey_Plane_Values" with the variable names and integrated values:
			arrayName = integrate.GetOutput().GetCellData().GetArray(j).GetName()
			if arrayName != "Length":	
				if i == 0:
					Survey_Plane_Values.append(vtk.vtkDoubleArray())
					Survey_Plane_Values[j].SetName(arrayName)
					
				arrayValue = integrate.GetOutput().GetCellData().GetArray(j).GetTuple(0)[0]
				#arrayValue = arrayValue/Chord  # This is to divide the integrated value by the chord but it doesnt work... 
				Survey_Plane_Values[j].InsertNextValue(arrayValue)


"""
Divide integrated values by chord: I can get the local chord but i cannot manage to divide the integrated values by the chord....IMPLEMENT IT by using a wrap VTK 
Attention: chord division is included in the two integrations options of this function!!! (lines 424 and 474)
"""


###########################################################
################# CONTROL VOLUME INTEGRAL #################
###########################################################

def CONTROL_VOLUME_INTEGRAL(i,Control_Volume_Mesh,Control_Volume_Values,INTEGRATION_MODE):
	import vtk
	import numpy as np   
	
	if Control_Volume_Mesh != "None":   
		# Pass the survey plane variables only:
		Input_Filtered=vtk.vtkPassArrays()
		Input_Filtered.SetInputDataObject(Control_Volume_Mesh.GetOutput())
		Input_Filtered.ClearArrays()
		Counter = 0
		for L in range(Control_Volume_Mesh.GetOutput().GetPointData().GetNumberOfArrays()):
			if INTEGRATION_MODE(Control_Volume_Mesh.GetOutput(),L) == "Control_Volume":
				Input_Filtered.AddPointDataArray(Control_Volume_Mesh.GetOutput().GetPointData().GetArray(L).GetName())
				Counter = Counter + 1
		if Counter == 0:
			Input_Filtered.AddPointDataArray("")
		Input_Filtered.Update()             
		
		# Integrate variables on the survey plane:
		integrate = vtk.vtkIntegrateAttributes()
		integrate.SetInputData(Input_Filtered.GetOutput())       
		integrate.Update()
		
		# Fill "Control_Volume_Values" with the variable names and integrated values:
		for j in range(integrate.GetOutput().GetPointData().GetNumberOfArrays()):
			if i == 0:
				arrayName = integrate.GetOutput().GetPointData().GetArray(j).GetName()
				Control_Volume_Values.append(vtk.vtkDoubleArray())
				Control_Volume_Values[j].SetName(arrayName)
			
			arrayValue = integrate.GetOutput().GetPointData().GetArray(j).GetTuple(0)[0]
			Control_Volume_Values[j].InsertNextValue(arrayValue)


###########################################################
############## SURVEY PLANE MASKELL INTEGRAL ##############
###########################################################

def SURVEY_PLANE_MASKELL_INTEGRAL(i,Survey_Plane_MASKELL,Survey_Plane_MASKELL_Values,Maskell,FAR_FIELD,Input_Is_Survey_Plane,Plane_shape,Plane_normal,Plane_Body_Intersection,T_inverse,addfielddata):
	import vtk
	import numpy as np  
	
	if Maskell == 1 or FAR_FIELD == 1:			
		# Pass the survey plane variables only:
		Input_Filtered=vtk.vtkPassArrays()
		Input_Filtered.SetInputDataObject(Survey_Plane_MASKELL.GetOutput())
		Input_Filtered.ClearArrays()
		Input_Filtered.AddPointDataArray("MASKELL_lift")
		Input_Filtered.AddPointDataArray("MASKELL_lift_zetay")
		Input_Filtered.AddPointDataArray("MASKELL_lift_zeta")
		Input_Filtered.AddPointDataArray("MASKELL_sideforce")
		Input_Filtered.AddCellDataArray("Normals")       
		Input_Filtered.Update()             

		if Input_Is_Survey_Plane==0 and Plane_shape==0 and Plane_normal==0 and Plane_Body_Intersection==0:      # Only for planar survey planes normal to aero x-axis (note that is the inout is already a survey plane, we cannot ensure that it is normal to aero x-axis)

			# Integrate variables on the survey plane:
			integrate = vtk.vtkIntegrateAttributes()
			integrate.SetInputData(Input_Filtered.GetOutput())       
			integrate.Update()
			
			# Fill "Survey_Plane_MASKELL_Values" with the variable names and integrated values:
			for j in range(integrate.GetOutput().GetPointData().GetNumberOfArrays()):
				if i == 0:
					arrayName = integrate.GetOutput().GetPointData().GetArray(j).GetName()
					Survey_Plane_MASKELL_Values.append(vtk.vtkDoubleArray())
					Survey_Plane_MASKELL_Values[j].SetName(arrayName)
				
				arrayValue = integrate.GetOutput().GetPointData().GetArray(j).GetTuple(0)[0]
				Survey_Plane_MASKELL_Values[j].InsertNextValue(arrayValue)   

		# Compute (V.N) for inclined survey planes:
		if Input_Is_Survey_Plane!=0 or Plane_shape!=0 or Plane_normal!=0 or Plane_Body_Intersection!=0: 
		
			p2c = vtk.vtkPointDataToCellData()	                                                                # POINT DATA TO CELL DATA CONVERSION:		  				
			p2c.SetInputConnection(Input_Filtered.GetOutputPort())
			p2c.Update()

			EmptyMesh=vtk.vtkPassArrays()                                                                       # Get the empty mesh without arrays
			EmptyMesh.SetInputDataObject(p2c.GetOutput())
			EmptyMesh.ClearArrays()
			EmptyMesh.AddPointDataArray("")
			EmptyMesh.AddCellDataArray("")
			EmptyMesh.AddFieldDataArray("")
			EmptyMesh.Update() 

			Wrapped = dsa.WrapDataObject(p2c.GetOutput()) 													    # CONVERT THE CELL-BASED VTK ARRAYS INTO NUMPY ARRAYS: 
			Normals_cell = Wrapped.CellData["Normals"]                                                          # Get the cell data

			normal = np.array([1,0,0])								                                            # x-axis aerodynamic normal direction
			NORMAL = np.dot(T_inverse,normal)   					                                            # Express the normal in the mesh reference frame

			Normals_cell_xaero = []                                                                             # Creates a matrix with the aerodynamic x-axis normal at each point of the survey plane
			for k in range(0, np.size(Normals_cell[:,0])):
				Normals_cell_xaero.append([NORMAL[0], NORMAL[1], NORMAL[2]])

			VdotN = np.einsum('ij,ij->i',Normals_cell,Normals_cell_xaero)                                       # This emulates the (V.N) factor of the equations involving surface flux of a scalar quantity
			
			for L in range(p2c.GetOutput().GetCellData().GetNumberOfArrays()):                                  # Add the corrected variables to the empty VTK object
				ArrayNameCell = p2c.GetOutput().GetCellData().GetArray(L).GetName()
				CellVariable = Wrapped.CellData[ArrayNameCell]
				CorrectedCellVariable = CellVariable*VdotN
				if ArrayNameCell != "Normals":
					addfielddata(1, EmptyMesh.GetOutput().GetCellData(),CorrectedCellVariable,ArrayNameCell)

			integrate = vtk.vtkIntegrateAttributes()                                                            # Integrate variables on the survey plane: 
			integrate.SetInputData(EmptyMesh.GetOutput())   
			integrate.Update()
			
			for j in range(integrate.GetOutput().GetCellData().GetNumberOfArrays()):                            # Fill "Survey_Plane_MASKELL_Values" with the variable names and integrated values:
				arrayName = integrate.GetOutput().GetCellData().GetArray(j).GetName()
				if arrayName != "Length":	
					if i == 0:
						Survey_Plane_MASKELL_Values.append(vtk.vtkDoubleArray())
						Survey_Plane_MASKELL_Values[j].SetName(arrayName)
						
					arrayValue = integrate.GetOutput().GetCellData().GetArray(j).GetTuple(0)[0]
					Survey_Plane_MASKELL_Values[j].InsertNextValue(arrayValue)



###########################################################
#################### FILTER INPUT DATA ####################
###########################################################

def FILTER_INPUT_DATA(input,Input_Module,INTEGRATION_MODE,FAR_FIELD,Momentum,Betz,Jones,Oswatitsch,Maskell,Onorato,VDV,Giles,Kusunose,Meheut,Toubin,EXERGY,Arntz,Aguirre,Ruscio,LAMB_VECTOR,Wu,Mele,UDV):
	import vtk
	import numpy as np
	
	Input_Filtered=vtk.vtkPassArrays()
	Input_Filtered.SetInputDataObject(input)
	Input_Filtered.ClearArrays()
	Input_Filtered.AddFieldDataArray("") 
	Input_Filtered.AddFieldDataArray("ARRAY") 
	Input_Filtered.AddFieldDataArray("ARRAY_Integration") 
	Input_Filtered.AddCellDataArray("")
	Input_Filtered.AddPointDataArray("")

	for L in range(input.GetPointData().GetNumberOfArrays()):

		# Load arrays from Epsilon Aero Data:
		if Input_Module == 0:                                                                               # 0=Epsilon Aero Data 
			if FAR_FIELD == 1 or Momentum == 1:                                                             # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('MOMENTUM')==True:	
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())	

			if FAR_FIELD == 1 or Betz == 1:                                                                 # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('BETZ')==True:	
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Jones == 1:                                                                # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('JONES')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Oswatitsch == 1:                                                           # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('OSWATITSCH')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Maskell == 1:                                                              # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('MASKELL')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Onorato == 1:                                                              # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('ONORATO')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or VDV == 1:                                                                  # FAR_FIELD--> 1=All                   
				if input.GetPointData().GetArray(L).GetName().startswith('VDV')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Giles == 1:                                                                # FAR_FIELD--> 1=All  
				if input.GetPointData().GetArray(L).GetName().startswith('GILES')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Kusunose == 1:                                                             # FAR_FIELD--> 1=All 
				if input.GetPointData().GetArray(L).GetName().startswith('KUSUNOSE')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Meheut == 1:                                                               # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('MEHEUT')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if FAR_FIELD == 1 or Toubin == 1:                                                               # FAR_FIELD--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('TOUBIN')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if EXERGY == 1 or Arntz == 1:                                                                   # EXERGY--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('ARNTZ')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if EXERGY == 1 or Aguirre == 1:                                                                 # EXERGY--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('AGUIRRE')==True:
					if INTEGRATION_MODE(input,L) != "Do_Not_Integrate":
						Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if EXERGY == 1 or Ruscio == 1:                                                                  # EXERGY--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('RUSCIO')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if LAMB_VECTOR == 1 or Wu == 1:                                                                 # LAMB_VECTOR--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('WU')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if LAMB_VECTOR == 1 or Mele == 1:                                                               # LAMB_VECTOR--> 1=All
				if input.GetPointData().GetArray(L).GetName().startswith('MELE')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if UDV == 1:                                                               
				if input.GetPointData().GetArray(L).GetName().startswith('UDV')==True:
					Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

		# Load arrays from Epsilon Near Field:
		if Input_Module == 1:                                                                               # 1=Epsilon Near Field  
			if input.GetPointData().GetArray(L).GetName().startswith('NEAR-FIELD')==True:	
				Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())	

		# Load arrays from Epsilon Poisson Solver:
		if Input_Module == 2:                                                                               # 2=Epsilon Poisson Solver 
			if input.GetPointData().GetArray(L).GetName().startswith('AGUIRRE')==True:	
				Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if input.GetPointData().GetArray(L).GetName().startswith('MASKELL')==True:	
				Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

			if input.GetPointData().GetArray(L).GetName().startswith('POISSON')==True:	
				Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

		# Load arrays from Epsilon Wave Drag:
		if Input_Module == 3:                                                                               # 3=Epsilon Wave Drag 
			if input.GetPointData().GetArray(L).GetName().startswith('Parameter')==True:	
				Input_Filtered.AddPointDataArray(input.GetPointData().GetArray(L).GetName())

	Input_Filtered.Update() 
	return Input_Filtered


###########################################################
################### SPLIT VECTOR ARRAYS ###################
###########################################################

def SPLIT_VECTOR_ARRAYS(Input_Filtered,addfielddata):                                                       # Detect vectors and splits it into three different variables:  
	for L in range(Input_Filtered.GetOutput().GetPointData().GetNumberOfArrays()):

		if Input_Filtered.GetOutput().GetPointData().GetArray(L).GetNumberOfComponents() == 3:

			name = Input_Filtered.GetOutput().GetPointData().GetArray(L).GetName()
			Wrapped_Input = dsa.WrapDataObject(Input_Filtered.GetOutput())                                  # Converts the VTK object into a Python object     
			Vector = Wrapped_Input.PointData[name]

			name_X = name+"_X"
			name_Y = name+"_Y"
			name_Z = name+"_Z"        

			Vector_X = Vector[:,0]
			Vector_Y = Vector[:,1]        
			Vector_Z = Vector[:,2]       

			addfielddata(1, Input_Filtered.GetOutput().GetPointData(),Vector_X,name_X)
			addfielddata(1, Input_Filtered.GetOutput().GetPointData(),Vector_Y,name_Y)
			addfielddata(1, Input_Filtered.GetOutput().GetPointData(),Vector_Z,name_Z)        

"""
ARREGLAR ERROR !!!!

Output de ParaView muestra esto cada vez que integro!!!!!


ERROR: In C:\glr\builds\paraview\paraview-ci\build\superbuild\paraview\src\VTK\Common\DataModel\vtkTable.cxx, line 353
vtkTable (0000026AAA2DE740): Column "Points_0" must have 1 rows, but has 2.

ERROR: In C:\glr\builds\paraview\paraview-ci\build\superbuild\paraview\src\VTK\Common\DataModel\vtkTable.cxx, line 353
vtkTable (0000026AAA2DE740): Column "Points_1" must have 1 rows, but has 2.

ERROR: In C:\glr\builds\paraview\paraview-ci\build\superbuild\paraview\src\VTK\Common\DataModel\vtkTable.cxx, line 353
vtkTable (0000026AAA2DE740): Column "Points_2" must have 1 rows, but has 2.

ERROR: In C:\glr\builds\paraview\paraview-ci\build\superbuild\paraview\src\VTK\Common\DataModel\vtkTable.cxx, line 353
vtkTable (0000026AAA2DE740): Column "Points_Magnitude" must have 1 rows, but has 2.

ERROR: In C:\glr\builds\paraview\paraview-ci\build\superbuild\paraview\src\VTK\Common\DataModel\vtkTable.cxx, line 353
vtkTable (0000026AAA2DE740): Column "vtkOriginalIndices" must have 1 rows, but has 2.
"""


###########################################################
################# PLANE-BODY INTERSECTION #################
###########################################################

"""
IMPORTANT:
• It works OK in full 3D
• Significant developement is required for the other formats (2D, extruded 2D and extruded 3D) --> before developing it, LET'S TEST A 3D CASE!!!!!!
• For 2D, edit Epsilon_Near_Field to deliver the edge instead of the surface. Then, compute edge normal by using the extrussion direction (or upward axis) and the line joining each cell end points. Do not forget to normalize !!!
"""

# LOAD BODY SURFACE AND FILTER UNDESIRED ARRAYS:
if Plane_Body_Intersection == 1:
	Body_Surface_Mesh = LOAD_VTK_DATA(Load_Body_Surface)                                                    # Load the body surface mesh generated with Epsilon_Near_Field (VTP format)
	Body_Surface_Mesh = FILTER_INPUT_DATA(Body_Surface_Mesh.GetOutput(),Input_Module,INTEGRATION_MODE,FAR_FIELD,Momentum,Betz,Jones,Oswatitsch,Maskell,Onorato,VDV,Giles,Kusunose,Meheut,Toubin,EXERGY,Arntz,Aguirre,Ruscio,LAMB_VECTOR,Wu,Mele,UDV)
	SPLIT_VECTOR_ARRAYS(Body_Surface_Mesh,addfielddata)
	print("# Plane-body intersection enabled: make sure that 'Keep input data' was enabled in Epsilon_Near_Field before saving the VTP file")

def PLANE_BODY_INTERSECTION_PLANAR(i,Body_Surface_Mesh,Survey_Plane_Mesh,POINTS,NORMAL,plane,Plane_shape,EXTRACT_SURFACE,APPEND_VTK_OBJECTS,CLIP_BY_PLANE,Invert_Clip):
	import vtk
	import numpy as np
	if  Plane_shape == 0 :                                                                                  #     0=Planar
		
		# CLIP BODY SURFACE WITH PLANAR SURVEY PLANE:
		#plane = vtk.vtkPlane()														                        # Creates slice plane		            
		#plane.SetOrigin(POINTS[i,0], POINTS[i,1], POINTS[i,2])
		#plane.SetNormal(NORMAL[0], NORMAL[1], NORMAL[2])
		clipper = CLIP_BY_PLANE(Body_Surface_Mesh.GetOutput(),plane,Invert_Clip)                            # Creates a clip of the input with the survey plane
		Body_Surface_Clipped = EXTRACT_SURFACE(clipper)                                                     # Converts the clip VTK format from vtkUnstructuredGrid to vtkPolyData

		# JOIN BODY SURFACE WITH SURVEY PLANE:
		appendFilter = APPEND_VTK_OBJECTS(Body_Surface_Clipped, Survey_Plane_Mesh)                          # Combine the survey plane and the body into a single VTK variable  
		Survey_Plane_Mesh = EXTRACT_SURFACE(appendFilter)                                                   # Convert the appended object into vtkPolyData format 

	return Survey_Plane_Mesh


def PLANE_BODY_INTERSECTION_EXTRUDED(Body_Surface_Mesh,Survey_Plane_Mesh,Plane_shape,DISTANCE_BETWEEN_OBJECTS,CLIP_BY_SCALAR,Invert_Clip,EXTRACT_SURFACE,APPEND_VTK_OBJECTS):
	import vtk
	import numpy as np
	if Plane_shape == 1 :                                                                                   # 1=Extruded curve 	

		# CLIP BODY SURFACE WITH EXTRUDED SURVEY PLANE:
		Survey_Plane_Mesh.FlipNormalsOff()                                                                  # The normals are reversed because vtkImplicitPolyDataDistance (included inside DISTANCE_BETWEEN_OBJECTS) reverses the normal from SetInput (so, at the end, the normal direction remains OK) 
		Survey_Plane_Mesh.Update()        
		DISTANCE_BETWEEN_OBJECTS(Survey_Plane_Mesh.GetOutput(),Body_Surface_Mesh.GetOutput())               # Calculate distance from each Body_Surface_Mesh point to the extruded survey plane. Adds an array called Distance to the VTK object
		clipper = CLIP_BY_SCALAR(Body_Surface_Mesh.GetOutput(),"Distance",-1*0,Invert_Clip)                 # Clip Body_Surface_Mesh with extruded survey plane. NOTE: Plane_Offset_Distance was set to zero since the extruded survey plane used to compute the distance is already shifted by the offset distance
		Body_Surface_Clipped = EXTRACT_SURFACE(clipper)                                                     # Converts object to vtkPolyData

		# JOIN BODY SURFACE WITH SURVEY PLANE:        
		appendFilter = APPEND_VTK_OBJECTS(Body_Surface_Clipped, Survey_Plane_Mesh)                          # Combine the survey plane and the body into a single VTK variable  
		Survey_Plane_Mesh = EXTRACT_SURFACE(appendFilter)                                                   # Convert the appended object into vtkPolyData format

	return Survey_Plane_Mesh


###################################
#######  WRITE VTP FUNCTION  ######
###################################

def WRITE_VTP(outputObject,File_Path_Output):     					
	import vtk
	if File_Path_Output:	
		OutputVTP = vtk.vtkXMLPolyDataWriter()                  
		OutputVTP.SetInputData(outputObject.VTKObject)    			                                        # Alternative command: SetInput           vtkObject.GetOutput()
		OutputVTP.SetFileName(File_Path_Output)                                                             # File path to the folder where to store the output data 
		OutputVTP.Write()                                                                                   # "OutputVTP" is a vtkUnstructuredGrid variable



############################################################################################################
############################################## READ INPUT DATA #############################################
############################################################################################################

Sref = inputs[0].FieldData["Sref"]
TRANSFORMATION = inputs[0].FieldData["TRANSFORMATION"]
T_t = np.transpose(TRANSFORMATION)
T_inverse = np.dot(T_t,np.dot(TRANSFORMATION,T_t))  	                                                    # It calculates the inverse of the TRANSFORMATION matrix


#Read input data and filter the undesired arrays:
input = self.GetInputDataObject(0, 0) 
Input_Filtered = FILTER_INPUT_DATA(input,Input_Module,INTEGRATION_MODE,FAR_FIELD,Momentum,Betz,Jones,Oswatitsch,Maskell,Onorato,VDV,Giles,Kusunose,Meheut,Toubin,EXERGY,Arntz,Aguirre,Ruscio,LAMB_VECTOR,Wu,Mele,UDV)
SPLIT_VECTOR_ARRAYS(Input_Filtered,addfielddata)



############################################################################################################
#################################  REFERENCE LINE FOR SURVEY PLANE POSITION ################################
############################################################################################################

# DEFINE LINEPOINTS: 
if Integral_Type == 0 or Integral_Type == 2:                                                                # 0=Single survey plane     2=Spanwise distribution
	numPts = 1
	Start = X_single_plane
	End = X_single_plane
	Sweep_Axis_Survey = 0

if Integral_Type == 1:                                                                                      # 1=Sweep survey plane 
	numPts = NUMPTS_Survey								                                                    # Number of points along the distribution line
	Start = START_Survey								                                                    # Initial position of the survey plane [m]
	End = END_Survey									                                                    # Final position of the survey plane [m]

if Integral_Type == 0 or Integral_Type == 1 or Integral_Type == 2:  									    # 0=Single survey plane    1=Sweep survey plane     2=Spanwise distribution

	if Sweep_Axis_Survey == 0:												                                # 0=Aerodynamic x-axis
		X = np.linspace(Start,End,numPts) 						                                            # Creates a uniform distribution of points along the mesh X-axis
		Y = np.linspace(0,0,numPts)								                                            # Creates an array filled with zeros
		Z = np.linspace(0,0,numPts)
		POINTS = algs.make_vector(X,Y,Z) 						                                            # Creates a vector with the point coordinates
		POINTS = np.einsum('ij,kj->ik',POINTS,T_inverse)  		                                            # Transforms the points from the mesh reference frame to the aero reference frame (the points are now alligned with the upstream flow direction)

	if Sweep_Axis_Survey == 1:                                                                              # 1=Mesh X-axis 
		X = np.linspace(Start,End,numPts)
		Y = np.linspace(0,0,numPts)
		Z = np.linspace(0,0,numPts)
		POINTS = algs.make_vector(X,Y,Z) 
		
# DEFINE SURVEY PLANE NORMAL:
if Integral_Type == 0 or Integral_Type == 1 or Integral_Type == 2:  									    # 0=Single survey plane     1=Sweep survey plane    2=Spanwise distribution

	if Plane_normal == 0:												                                    # 0=Aerodynamic x-axis
		normal = np.array([1,0,0])								                                            # Define the normal of the survey plane on the aero reference frame
		NORMAL = np.dot(T_inverse,normal)   					                                            # Express the normal in the mesh reference frame

	if Plane_normal == 1:												                                    # 6=User defined   
		NORMAL = Set_normal/np.linalg.norm(Set_normal)                                                      # The normal direction defined by the user is normalized

	if Flip_Plane_Normal == 1:
		NORMAL = -1*NORMAL

# DISPLAY REFERENCE LINE:
if  Integral_Type == 0 or Integral_Type == 1 or Integral_Type == 2:                                         # 0=Single survey plane     1=Sweep survey plane    2=Spanwise distribution
	
	if Select_Action == 0 or Select_Action == 3:                                                            # 0=Visualize sweep line    3=Integrate variables                                                                     
		Sweep_Line,Sweep_Vertices = PLOT_VTK_LINES_VERTEX(POINTS)                                           # Generate curve (polyline) from points:
		appendFilter = APPEND_VTK_OBJECTS(Sweep_Line, Sweep_Vertices)                                       # Combine the curve and point into a single VTK object
		Integration_Line = EXTRACT_SURFACE(appendFilter)                                                    # Convert the appended object into vtkPolyData format 
		self.GetOutput().ShallowCopy(Integration_Line.GetOutput())                                          # Displays output data

		if Input_Is_Survey_Plane == 0:                                                                      # Input data is not a survey plane
			Integration_Line,Sweep_Vertices = PLOT_VTK_LINES_VERTEX(POINTS)                                 # Generate curve (polyline) from points


############################################################################################################
#########################################  INPUT IS A SURVEY PLANE #########################################
############################################################################################################

if  Integral_Type == 1 or Integral_Type == 3:                                                               # 1=Sweep survey plane      3=3D to 2D mapping
	if Input_Is_Survey_Plane == 1:                                                                          # Input data is already a survey plane
		print ("# ERROR: input is already a survey plane. It cannot be swept or mapped")
		print ("# Only possible integral types for this case: 'Single survey plane' and 'Spanwise distribution'")


if  Integral_Type == 0 or Integral_Type == 2:                                                               # 0=Single survey plane     2=Spanwise distribution
	if Input_Is_Survey_Plane == 1:                                                                          # Input data is already a survey plane

		print ("# Warning: input is already a survey plane. Volume integrals cannot be performed")		
		Control_Volume_Mesh = "None"                                                                        # Since input data is a plane, there is no control volume data available for integration      
		Survey_Plane_Mesh = EXTRACT_SURFACE(Input_Filtered)
		
		if Clip_Wake_Region == 1:
			Survey_Plane_Mesh_Clip = CLIP_BY_SCALAR(Survey_Plane_Mesh.GetOutput(),Parameter,Clip_Value,Invert_Clip_Wake)
			Survey_Plane_Mesh = EXTRACT_SURFACE(Survey_Plane_Mesh_Clip)

		Survey_Plane_Mesh = SURFACE_NORMALS(Survey_Plane_Mesh,Flip_Plane_Normal)                            # Compute surface normals

		# VISUALIZE SURVEY PLANE:
		if Select_Action == 1:                                                                              # 1=Visualize survey Plane
			self.GetOutput().ShallowCopy(Survey_Plane_Mesh.GetOutput())   							        # Send sliced data to the output

		# VISUALIZE CONTROL VOLUME:
		if Select_Action == 2:                                                                              # 2=Visualize control volume
			print ("# ERROR: no volume mesh to display. You can only visualize survey plane or integrate variables")

		# INTEGRATE VARIABLES:
		if Select_Action == 3:                                                                              # 3=Integrate variables
			if  Integral_Type == 0:                                                                         # 0=Single survey plane
				Survey_Plane_Values = []                                                                    # Array to store the integrated values at the survey plane                                                     
				for i in range(0, numPts):   											                    # Integrate data at each point of the integration line (in this case there is only just one point)
					plane = vtk.vtkPlane()
					SURVEY_PLANE_INTEGRAL(i,Survey_Plane_Mesh,Survey_Plane_Values,INTEGRATION_MODE,Input_Module,Sref,T_inverse,addfielddata,Input_Is_Survey_Plane,Plane_shape,Plane_normal,Plane_Body_Intersection,Integral_Type,SLICE_WITH_PLANE,plane)

				# Add integrated values to the output VTK object:
				for array in Survey_Plane_Values:                                                                   
					Integration_Line.GetOutput().GetPointData().AddArray(array)     

				self.GetOutput().ShallowCopy(Integration_Line.GetOutput())                                  # Displays output data



############################################################################################################
###########################################  PLANAR SURVEY PLANE ###########################################
############################################################################################################

if  Integral_Type == 0 or Integral_Type == 1 or Integral_Type == 2:                                         # 0=Single survey plane     1=Sweep survey plane    2=Spanwise distribution

	if Input_Is_Survey_Plane == 0:                                                                          # Input data is NOT a survey plane (a slice is required)
		if Plane_shape == 0:                                                                                # 0=Planar 
			input = self.GetInputDataObject(0, 0)  												            # Reads input data

			# ITETRATE AT EACH SWEEP LINE POINT:
			if Select_Action == 0 or Select_Action == 1 or Select_Action == 2 or Select_Action == 4 or Select_Action == 5 or Select_Action == 7:        # 0=Visualize sweep line    1=Visualize survey Plane  2=Visualize control volume  4=Visualize integration sections  5=Visualize integration surface  7=Visualize spanwise sweep line
				NBplanes = 1
			if Select_Action == 3:                                                                          # 3=Integrate variables
				NBplanes = numPts

			Survey_Plane_Values = []                                                                        # Array to store the integrated values at the survey plane                                                     
			Control_Volume_Values = []                                                                      # Array to store the integrated values at the control volume
			for i in range(0, NBplanes):   											                        # Integrate data at each point of the integration line (in this case there is only just one point)

				# CREATE PLANE FUNCTION:
				if Select_Action == 1 or Select_Action == 2 or Select_Action == 3 or Select_Action == 4 or Select_Action == 5:    # 1=Visualize survey Plane  2=Visualize control volume  3=Integrate variables 4=Visualize integration sections 5=Visualize integration surface
					plane = vtk.vtkPlane()	
					plane.SetNormal(NORMAL[0], NORMAL[1], NORMAL[2])                    
					if Select_Action == 1 or Select_Action == 2:
						plane.SetOrigin(POINTS[int(numPts/2),0], POINTS[int(numPts/2),1], POINTS[int(numPts/2),2])
					if Select_Action == 3 or Select_Action == 4 or Select_Action == 5:
						plane.SetOrigin(POINTS[i,0], POINTS[i,1], POINTS[i,2])

				# VISUALIZE SURVEY PLANE:
				if Select_Action == 1 or Select_Action == 3 or Select_Action == 4 or Select_Action == 5:    # 1=Visualize survey Plane  3=Integrate variables 4=Visualize integration sections 5=Visualize integration surface
					Survey_Plane_Mesh = SLICE_WITH_PLANE(Input_Filtered.GetOutput(),plane)                  # Slices data with slice plane
					if Clip_Wake_Region == 1:
						Survey_Plane_Mesh_Clip = CLIP_BY_SCALAR(Survey_Plane_Mesh.GetOutput(),Parameter,Clip_Value,Invert_Clip_Wake)
						Survey_Plane_Mesh = EXTRACT_SURFACE(Survey_Plane_Mesh_Clip)
					if Plane_Body_Intersection == 1:
						Survey_Plane_Mesh = PLANE_BODY_INTERSECTION_PLANAR(i,Body_Surface_Mesh,Survey_Plane_Mesh,POINTS,NORMAL,plane,Plane_shape,EXTRACT_SURFACE,APPEND_VTK_OBJECTS,CLIP_BY_PLANE,Invert_Clip)

					Survey_Plane_Mesh = SURFACE_NORMALS(Survey_Plane_Mesh,Flip_Plane_Normal)                # Compute surface normals

					self.GetOutput().ShallowCopy(Survey_Plane_Mesh.GetOutput())   						    # Send sliced data to the output

				# VISUALIZE CONTROL VOLUME:
				if Select_Action == 2 or Select_Action == 3:                                                # 2=Visualize control volume    3=Integrate variables
					Control_Volume_Mesh = CLIP_BY_PLANE(Input_Filtered.GetOutput(),plane,Invert_Clip)       # Creates a clip of the input with the survey plane
					if Clip_Wake_Region == 1:
						Control_Volume_Mesh = CLIP_BY_SCALAR(Control_Volume_Mesh.GetOutput(),Parameter,Clip_Value,Invert_Clip_Wake)
					Extract = EXTRACT_SURFACE(Control_Volume_Mesh)                                          # Extract the outter surface of the clipped domain for visualization purposes 
					self.GetOutput().ShallowCopy(Extract.GetOutput())   								    # Send sliced data to the output       

				# INTEGRATE VARIABLES:
				if Select_Action == 3:                                                                      # 3=Integrate variables
					if Integral_Type != 2:                                                                  # 2=Spanwise distribution
						SURVEY_PLANE_INTEGRAL(i,Survey_Plane_Mesh,Survey_Plane_Values,INTEGRATION_MODE,Input_Module,Sref,T_inverse,addfielddata,Input_Is_Survey_Plane,Plane_shape,Plane_normal,Plane_Body_Intersection,Integral_Type,SLICE_WITH_PLANE,plane)
						CONTROL_VOLUME_INTEGRAL(i,Control_Volume_Mesh,Control_Volume_Values,INTEGRATION_MODE)

			if Select_Action == 3:                                                                          # 3=Integrate variables
				# Add integrated values to the output VTK object:
				for array in Survey_Plane_Values:                                                                   
					Integration_Line.GetOutput().GetPointData().AddArray(array)     
				for array in Control_Volume_Values:                                                                   
					Integration_Line.GetOutput().GetPointData().AddArray(array)
				self.GetOutput().ShallowCopy(Integration_Line.GetOutput())                                  # Displays output data



############################################################################################################
########################################## EXTRUDED SURVEY PLANE ###########################################
############################################################################################################

if  Integral_Type == 0 or Integral_Type == 1 or Integral_Type == 2:                                         # 0=Single survey plane     1=Sweep survey plane    2=Spanwise distribution

	if Input_Is_Survey_Plane == 0:                                                                          # Input data is NOT a survey plane (a slice is required)
		if Plane_shape == 1:                                                                                # 1=Extruded curve 

			# LOAD POINTS AND SORT ALONG DESIRED AXIS:                                                      # The points selected by the user with the graphical interaction tool are ordered by ID, so they must be re-ordered by coordinate
			reader = LOAD_VTK_DATA(Load_curve)                                                              # Load the VTK object available at the Load_curve filepath
			readerPy = dsa.WrapDataObject(reader.GetOutput())                                               # Converts the VTK object into a Python object     
			Curve_Points = readerPy.Points                                                                  # Get the points selected by the user
			Curve_Points = Curve_Points[Curve_Points[:,Sort_Direction].argsort()]                           # Orders the points along the desired axis

			# GET DOMAIN BOUNDS:
			Domain_Points = inputs[0].Points                                                                # Get the points of the input data and then determines the minimum/maximum coordinates along the X-, Y- and Z-axes
			Bound_Xmin = min(Domain_Points[:,0])                                                                    
			Bound_Xmax = max(Domain_Points[:,0])
			Bound_Ymin = min(Domain_Points[:,1])
			Bound_Ymax = max(Domain_Points[:,1])
			Bound_Zmin = min(Domain_Points[:,2])
			Bound_Zmax = max(Domain_Points[:,2])

			# EXTEND CURVE TO BOUNDS:                                                                       # Detect the bounds of the domain and the predominant direction of the curve. Then add extra points at the bounds along the predominant direction
			if Sort_Direction == 0:
				New_Point_Xmin = [Bound_Xmin,Curve_Points[0,1],Curve_Points[0,2]]
				New_Point_Xmax = [Bound_Xmax,Curve_Points[-1,1],Curve_Points[-1,2]]            
				Curve_Points = np.vstack((Curve_Points,New_Point_Xmin))                                     # Add New_Point_Xmin to Curve_Points
				Curve_Points = np.vstack((Curve_Points,New_Point_Xmax))                                     # Add New_Point_Xmax to Curve_Points
				Curve_Points = Curve_Points[Curve_Points[:,Sort_Direction].argsort()]                       # Orders the points along the desired axis (Because the points just appended are not in order)

			if Sort_Direction == 1:
				New_Point_Ymin = [Curve_Points[0,0],Bound_Ymin,Curve_Points[0,2]]
				New_Point_Ymax = [Curve_Points[-1,0],Bound_Ymax,Curve_Points[-1,2]]            
				Curve_Points = np.vstack((Curve_Points,New_Point_Ymin))                                     # Add New_Point_Ymin to Curve_Points
				Curve_Points = np.vstack((Curve_Points,New_Point_Ymax))                                     # Add New_Point_Ymax to Curve_Points
				Curve_Points = Curve_Points[Curve_Points[:,Sort_Direction].argsort()]                       # Orders the points along the desired axis (Because the points just appended are not in order)

			if Sort_Direction == 2:
				New_Point_Zmin = [Curve_Points[0,0],Curve_Points[0,1],Bound_Zmin]
				New_Point_Zmax = [Curve_Points[-1,0],Curve_Points[-1,1],Bound_Zmax]            
				Curve_Points = np.vstack((Curve_Points,New_Point_Zmin))                                     # Add New_Point_Zmin to Curve_Points
				Curve_Points = np.vstack((Curve_Points,New_Point_Zmax))                                     # Add New_Point_Zmax to Curve_Points
				Curve_Points = Curve_Points[Curve_Points[:,Sort_Direction].argsort()]                       # Orders the points along the desired axis (Because the points just appended are not in order)

			# SHIFT THE INTEGRATION LINE TO THE CLOSEST POINT TO X AXIS:                                    # Integration_Line is defined as the relative distance to the survey plane. The position X=0 must then be coincident with the survey plane, but the line was created centered at the origin, then a shift is required to bring the line in contact with the survey plane
			DistanceToXaxis = Curve_Points[:,1]**2+Curve_Points[:,2]**2
			ID_min = np.argwhere(DistanceToXaxis==np.min(DistanceToXaxis))
			ClosestPointToXaxis = Curve_Points[ID_min]
			Shifted_POINTS = POINTS[:]+ClosestPointToXaxis

			# DISPLAY THE SHIFTED REFERENCE LINE:
			if Select_Action == 0 or Select_Action == 3:                                                                # 0=Visualize sweep line    3=Integrate variables                                                                     
				Sweep_Line,Sweep_Vertices = PLOT_VTK_LINES_VERTEX(Shifted_POINTS[0])                                               # Generate curve (polyline) from points:
				appendFilter = APPEND_VTK_OBJECTS(Sweep_Line, Sweep_Vertices)                                           # Combine the curve and point into a single VTK object
				Integration_Line = EXTRACT_SURFACE(appendFilter)                                                        # Convert the appended object into vtkPolyData format 
				self.GetOutput().ShallowCopy(Integration_Line.GetOutput())                                              # Displays output data

			# GENERATE CURVE (POLYLINE) FROM POINTS: 
			Survey_Line,Survey_Vertices = PLOT_VTK_LINES_VERTEX(Curve_Points)

			# DEFINE EXTRUSION DIRECTION AND EXTRUSION LENGHT:
			if Sort_Direction == 0:
				Curve_Vector = [1,0,0]
			if Sort_Direction == 1:
				Curve_Vector = [0,1,0]
			if Sort_Direction == 2:
				Curve_Vector = [0,0,1]		

			Extrusion_Direction1 = np.cross(Curve_Vector,NORMAL)
			Extrusion_Direction2 = -1*Extrusion_Direction1
			
			if Sort_Direction == 0:
				Lenght = np.maximum(abs(Bound_Xmin),abs(Bound_Xmax))                                        # Takes the largest distance to the forward/rear bound
			if Sort_Direction == 1:
				Lenght = np.maximum(abs(Bound_Ymin),abs(Bound_Ymax))                                        # Takes the largest distance to the left/right bound
			if Sort_Direction == 2:
				Lenght = np.maximum(abs(Bound_Zmin),abs(Bound_Zmax))                                        # Takes the largest distance to the upper/lower bound

			# GENERATE EXTRUDED SURVEY PLANE:
			extrude1 = LINEAR_EXTRUSION(Survey_Line,Extrusion_Direction1,Lenght)                            # Generate extruded surface from curve upwards
			extrude2 = LINEAR_EXTRUSION(Survey_Line,Extrusion_Direction2,Lenght)                            # Generate extruded surface from curve downwards       
			appendFilter = APPEND_VTK_OBJECTS(extrude1, extrude2)                                           # Combine both extruded curves into a single VTK object
			combined = EXTRACT_SURFACE(appendFilter)                                                        # Convert the appended object into vtkPolyData format
			Survey_Plane = SURFACE_NORMALS(combined,Flip_Plane_Normal)                                      # Compute surface normals

			# COMPUTE DISTANCE TO SURVEY PLANE AT EACH INPUT MESH POINT:
			Survey_Plane.FlipNormalsOff()                                                                   # The normals of the survey plane are reversed because vtkImplicitPolyDataDistance (included inside DISTANCE_BETWEEN_OBJECTS) reverses the normal from SetInput (so, at the end, the normal direction remains OK) 
			Survey_Plane.Update()        
			if Select_Action != 0:                                                                          # DISTANCE_BETWEEN_OBJECTS is a slow operation an not required to visualize the sweep line, thus, avoided when  Select_Action == 0
				DISTANCE_BETWEEN_OBJECTS(Survey_Plane.GetOutput(),Input_Filtered.GetOutput())                   # Calculate distance from each input mesh point to the extruded survey plane

			# ITETRATE AT EACH SWEEP LINE POINT:
			if Select_Action == 0 or Select_Action == 1 or Select_Action == 2 or Select_Action == 4 or Select_Action == 5 or Select_Action == 7:      # 0=Visualize sweep line    1=Visualize survey Plane  2=Visualize control volume 4=Visualize integration sections    5=Visualize integration surface    7=Visualize spanwise sweep line
				NBplanes = 1
			if Select_Action == 3:                                                                          # 3=Integrate variables
				NBplanes = numPts

			Survey_Plane_Values = []                                                                        # Array to store the integrated values at the survey plane                                                     
			Control_Volume_Values = []                                                                      # Array to store the integrated values at the control volume
			for i in range(0, NBplanes):   											                        # Integrate data at each point of the integration line (in this case there is only just one point)

				if  Integral_Type == 0 or Integral_Type == 2:                   
					Plane_Offset_Distance = Plane_Offset_Distance
				if  Integral_Type == 1: 
					Plane_Offset_Distance = POINTS[i,0] 
					if Select_Action == 1 or Select_Action == 2:
						Plane_Offset_Distance = POINTS[int(numPts/2),0]
					if Select_Action == 3:
						Plane_Offset_Distance = POINTS[i,0] 

				# VISUALIZE SURVEY PLANE:
				if Select_Action == 1 or Select_Action == 3 or Select_Action == 4 or Select_Action == 5:    # 1=Visualize survey plane 	3=Integrate variables	4=Visualize integration sections    5=Visualize integration surface    
					contour = SLICE_BY_SCALAR(Input_Filtered.GetOutput(),"Distance",-1*Plane_Offset_Distance)       # Slices input mesh with extruded survey plane. NOTE: the sign of Plane_Offset_Distance is reversed in order to shift the plane downstream (along positive x-axis) when Plane_Offset_Distance is positive
					Survey_Plane_Mesh = SURFACE_NORMALS(contour,Flip_Plane_Normal)                          # Compute surface normals
					if Clip_Wake_Region == 1:
						Survey_Plane_Mesh_Clip = CLIP_BY_SCALAR(Survey_Plane_Mesh.GetOutput(),Parameter,Clip_Value,Invert_Clip_Wake)
						Survey_Plane_Mesh = EXTRACT_SURFACE(Survey_Plane_Mesh_Clip)                
					if Plane_Body_Intersection == 1:
						Survey_Plane_Mesh = PLANE_BODY_INTERSECTION_EXTRUDED(Body_Surface_Mesh,Survey_Plane_Mesh,Plane_shape,DISTANCE_BETWEEN_OBJECTS,CLIP_BY_SCALAR,Invert_Clip,EXTRACT_SURFACE,APPEND_VTK_OBJECTS)

					Survey_Plane_Mesh = SURFACE_NORMALS(Survey_Plane_Mesh,Flip_Plane_Normal)                # Compute surface normals

					self.GetOutput().ShallowCopy(Survey_Plane_Mesh.GetOutput())                             # Displays Survey_Plane_Mesh 

				# VISUALIZE CONTROL VOLUME:
				if Select_Action == 2 or Select_Action == 3:                                                # 2=Visualize control volume    3=Integrate variables
					Control_Volume_Mesh = CLIP_BY_SCALAR(Input_Filtered.GetOutput(),"Distance",-1*Plane_Offset_Distance,Invert_Clip)                 # Clip input mesh with extruded survey plane. NOTE: the sign of Plane_Offset_Distance is reversed in order to shift the plane downstream (along positive x-axis) when Plane_Offset_Distance is positive
					extract = EXTRACT_SURFACE(Control_Volume_Mesh)                                          # Extract external surface of the clip for visualization purposes 
					if Clip_Wake_Region == 1:
						Control_Volume_Mesh = CLIP_BY_SCALAR(Control_Volume_Mesh.GetOutput(),Parameter,Clip_Value,Invert_Clip_Wake)
						extract = EXTRACT_SURFACE(Control_Volume_Mesh)
					self.GetOutput().ShallowCopy(extract.GetOutput()) 

				# INTEGRATE VARIABLES:
				if Select_Action == 3:                                                                      # 3=Integrate variables
					if Integral_Type != 2:                                                                  # 2=Spanwise distribution
						plane = vtk.vtkPlane()	
						SURVEY_PLANE_INTEGRAL(i,Survey_Plane_Mesh,Survey_Plane_Values,INTEGRATION_MODE,Input_Module,Sref,T_inverse,addfielddata,Input_Is_Survey_Plane,Plane_shape,Plane_normal,Plane_Body_Intersection,Integral_Type,SLICE_WITH_PLANE,plane)
						CONTROL_VOLUME_INTEGRAL(i,Control_Volume_Mesh,Control_Volume_Values,INTEGRATION_MODE)

			if Select_Action == 3:                                                                          # 3=Integrate variables
				# Add integrated values to the output VTK object:
				for array in Survey_Plane_Values:                                                                   
					Integration_Line.GetOutput().GetPointData().AddArray(array)     
				for array in Control_Volume_Values:                                                                   
					Integration_Line.GetOutput().GetPointData().AddArray(array)
				self.GetOutput().ShallowCopy(Integration_Line.GetOutput())                                  # Displays output data



############################################################################################################
###################################### CUSTOM SURFACE AS SURVEY PLANE ######################################
############################################################################################################

"""
IMPORTANT:
Slice is not working correctly. The custom surface is correctly loaded but DISTANCE_BETWEEN_OBJECTS seems to generate weird spots of isosurface=0...
I also tried a fully closed surface (composed of multiple objects: plane and sphere) but still fails. It only works with simple closed surfaces (sphere only but not with box and not even ellipsoid)
This part of the code is left here for further developement (future Epsilon versions). The option Plane_shape == 2 was hidden in the XML. It must be re-enabled once this problem is solved
Open issue in ParaView Support !!!
"""

if Plane_shape == 2:         

	# LOAD CUSTOM SURFACE:
	if Load_surface.endswith(".vtp") or Load_surface.endswith(".stl"):
		Custom_Surface = LOAD_VTK_DATA(Load_surface)                                                        # Load the body surface mesh generated with Epsilon_Near_Field (VTP format)
		Custom_Surface = SURFACE_NORMALS(Custom_Surface,Flip_Plane_Normal)                                  # Compute surface normals
	else:
		print ('Custom surface format must be VTP or STL')

	# SLICE OR CLIP THE DOMAIN WITH THE CUSTOM SURFACE:
	if Select_Action == 1 or Select_Action == 2:	                                                        # 1=Visualize survey plane  2=Visualize control volume
		Custom_Surface.FlipNormalsOff()                                                                     # The normals of the custom surface are reversed because vtkImplicitPolyDataDistance (included inside DISTANCE_BETWEEN_OBJECTS) reverses the normal from SetInput (so, at the end, the normal direction remains OK) 
		Custom_Surface.Update()        
		DISTANCE_BETWEEN_OBJECTS(Custom_Surface.GetOutput(),Input_Filtered.GetOutput())                     # Calculate distance from each input mesh point to the custom surface

		if Select_Action == 1:                                                                              # 1=Visualize survey plane 		
			contour = SLICE_BY_SCALAR(Input_Filtered.GetOutput(),"Distance",-1*Plane_Offset_Distance)       # Slices input mesh with custom surface. NOTE: the sign of Plane_Offset_Distance is reversed in order to shift the plane downstream (along positive x-axis) when Plane_Offset_Distance is positive
			Survey_Plane_Mesh = SURFACE_NORMALS(contour,Flip_Plane_Normal)                                  # Compute surface normals
			self.GetOutput().ShallowCopy(Survey_Plane_Mesh.GetOutput())                                     # Displays output data 

		if Select_Action == 2:                                                                              # 2=Visualize control volume
			Control_Volume_Mesh = CLIP_BY_SCALAR(Input_Filtered.GetOutput(),"Distance",-1*Plane_Offset_Distance,Invert_Clip)                 # Clip input mesh with extruded survey plane. NOTE: the sign of Plane_Offset_Distance is reversed in order to shift the plane downstream (along positive x-axis) when Plane_Offset_Distance is positive
			extract = EXTRACT_SURFACE(Control_Volume_Mesh)                                                  # Extract external surface of the clip for visualization purposes 
			self.GetOutput().ShallowCopy(extract.GetOutput()) 



############################################################################################################
##########################################  SPANWISE DISTRIBUTION  #########################################
############################################################################################################

"""
• Agregar slice del Control_Volume_Mesh para integrar spanwise anergy generation !!!!!
"""

if Input_Module == 1:                                                                                       # 1=Epsilon Near Field
	Survey_Plane_Mesh = Input_Filtered

if Integral_Type == 2:                                                                                      # 2=Spanwise distribution

	if Input_Is_Survey_Plane == 0 or Input_Is_Survey_Plane == 1:                                                                     
		# DEFINE SPANWISE LINEPOINTS: 
		numPts = NUMPTS_Spanwise								                                            # Number of points along the spanwise distribution line
		Start = START_Spanwise								                                                # Initial position of the intersection plane [m]
		End = END_Spanwise									                                                # Final position of the intersection plane [m]

		if Sweep_Axis_Spanwise == 0:                                                                        # 0=Aerodynamic y-axis 
			X = np.linspace(0,0,numPts)
			Y = np.linspace(Start,End,numPts) 						                                        # Creates a uniform distribution of points along the mesh X-axis
			Z = np.linspace(0,0,numPts)								                                        # Creates an array filled with zeros
			POINTS = algs.make_vector(X,Y,Z) 						                                        # Creates a vector with the point coordinates 
			POINTS = np.einsum('ij,kj->ik',POINTS,T_inverse)  		                                        # Transforms the points from the mesh reference frame to the aero reference frame (the points are now alligned with the upstream flow direction)

		if Sweep_Axis_Spanwise == 1:                                                                        # 1=Aerodynamic z-axis 
			X = np.linspace(0,0,numPts)
			Y = np.linspace(0,0,numPts)
			Z = np.linspace(Start,End,numPts)
			POINTS = algs.make_vector(X,Y,Z) 
			POINTS = np.einsum('ij,kj->ik',POINTS,T_inverse)
			
		if Sweep_Axis_Spanwise == 2:                                                                        # 2=Mesh Y-axis
			X = np.linspace(0,0,numPts)
			Y = np.linspace(Start,End,numPts)
			Z = np.linspace(0,0,numPts)
			POINTS = algs.make_vector(X,Y,Z)

		if Sweep_Axis_Spanwise == 3:                                                                        # 3=Mesh Z-axis
			X = np.linspace(0,0,numPts)
			Y = np.linspace(0,0,numPts)
			Z = np.linspace(Start,End,numPts)
			POINTS = algs.make_vector(X,Y,Z) 


		# DEFINE INTERSECTION PLANE NORMAL:
		if Sweep_Axis_Spanwise == 0:                                                                        # 0=Aerodynamic y-axis 
			normal = np.array([0,1,0])								                                        # Define the normal of the survey plane on the aero reference frame
			NORMAL = np.dot(T_inverse,normal)   					                                        # Express the normal in the mesh reference frame

		if Sweep_Axis_Spanwise == 1:                                                                        # 1=Aerodynamic z-axis 
			normal = np.array([0,0,1])
			NORMAL = np.dot(T_inverse,normal)

		if Sweep_Axis_Spanwise == 2:                                                                        # 2=Mesh Y-axis
			NORMAL = np.array([0,1,0])

		if Sweep_Axis_Spanwise == 3:                                                                        # 3=Mesh Z-axis
			NORMAL = np.array([0,0,1])

		# DISPLAY REFERENCE LINE:
		if Select_Action == 7 :                                                                             # 7=Visualize spanwise sweep line                                                                     
			Sweep_Line,Sweep_Vertices = PLOT_VTK_LINES_VERTEX(POINTS)                                       # Generate curve (polyline) from points:
			appendFilter = APPEND_VTK_OBJECTS(Sweep_Line, Sweep_Vertices)                                   # Combine the curve and point into a single VTK object
			Integration_Line = EXTRACT_SURFACE(appendFilter)                                                # Convert the appended object into vtkPolyData format 
			self.GetOutput().ShallowCopy(Integration_Line.GetOutput())                                      # Displays output data
		if Select_Action == 3:                                                                              # 3=Integrate variables 
			Integration_Line,Sweep_Vertices = PLOT_VTK_LINES_VERTEX(POINTS)                                 # Generate curve (polyline) from points

		# ITETRATE AT EACH SPANWISE SWEEP LINE POINT:
		if Select_Action == 3 or Select_Action == 4 or Select_Action == 5 or Select_Action == 7 or Select_Action == 1:            # 3=Integrate variables  4=Visualize integration sections    5=Visualize integration surface  1=Visualize survey plane
			NB_spanwise_planes = 1
		if Select_Action == 3:                                                                              # 3=Integrate variables
			NB_spanwise_planes = numPts

		Survey_Plane_Values = []                                                                            # Array to store the integrated values at the survey plane                                                     
		Control_Volume_Values = []                                                                          # Array to store the integrated values at the control volume
		Survey_Plane_Maskell_Values = []                                                                    # Array to store the integrated values at the survey plane (Maskell method only)                                                     
    
		for i in range(0, NB_spanwise_planes):   											                # Integrate data at each point of the integration line (in this case there is only just one point)

			# CREATE PLANE FUNCTION:
			plane = vtk.vtkPlane()	
			plane.SetNormal(NORMAL[0], NORMAL[1], NORMAL[2])                    
			if Select_Action == 4 or Select_Action == 5:                                                    # 4=Visualize integration sections    5=Visualize integration surface
				plane.SetOrigin(POINTS[int(numPts/2),0], POINTS[int(numPts/2),1], POINTS[int(numPts/2),2])
			if Select_Action == 3:                                                                          # 3=Integrate variables 
				plane.SetOrigin(POINTS[i,0], POINTS[i,1], POINTS[i,2])

			# VISUALIZE INTEGRATION SECTIONS:
			if Select_Action == 4 or Select_Action == 3:                                                    # 4=Visualize integration sections  3=Integrate variables
				Survey_Plane_Mesh = EXTRACT_SURFACE(Survey_Plane_Mesh)
				Survey_Plane_Mesh_SLICED = SLICE_WITH_PLANE(Survey_Plane_Mesh.GetOutput(),plane)            # Slices data with slice plane
				self.GetOutput().ShallowCopy(Survey_Plane_Mesh_SLICED.GetOutput())   					    # Send sliced data to the output

			# VISUALIZE INTEGRATION SURFACE:
			if Select_Action == 5 or Select_Action == 3:                                                    # 5=Visualize integration surface   3=Integrate variables
				if FAR_FIELD == 1 or Maskell == 1:
					Survey_Plane_MASKELL = CLIP_BY_PLANE(Survey_Plane_Mesh.GetOutput(),plane,Invert_Clip)   # Creates a clip of the input with the survey plane
					Extract = EXTRACT_SURFACE(Survey_Plane_MASKELL)                                         # Extract the outter surface of the clipped domain for visualization purposes 
					self.GetOutput().ShallowCopy(Extract.GetOutput())   								    # Send sliced data to the output       
			if Select_Action == 5 and (FAR_FIELD != 1 or Maskell != 1):  
					print("ERROR: 'Visualize integration surface' can only be used with Maskell method ")


			# INTEGRATE VARIABLES:
			if Select_Action == 3:                                                                          # 3=Integrate variables
				SURVEY_PLANE_INTEGRAL(i,Survey_Plane_Mesh,Survey_Plane_Values,INTEGRATION_MODE,Input_Module,Sref,T_inverse,addfielddata,Input_Is_Survey_Plane,Plane_shape,Plane_normal,Plane_Body_Intersection,Integral_Type,SLICE_WITH_PLANE,plane)
				CONTROL_VOLUME_INTEGRAL(i,Survey_Plane_Mesh_SLICED,Control_Volume_Values,INTEGRATION_MODE)
				if FAR_FIELD == 1 or Maskell == 1:
					SURVEY_PLANE_MASKELL_INTEGRAL(i,Survey_Plane_MASKELL,Survey_Plane_Maskell_Values,Maskell,FAR_FIELD,Input_Is_Survey_Plane,Plane_shape,Plane_normal,Plane_Body_Intersection,T_inverse,addfielddata)

		if Select_Action == 3:                                                                              # 3=Integrate variables
			# Add integrated values to the output VTK object:
			for array in Survey_Plane_Values:                                                                   
				Integration_Line.GetOutput().GetPointData().AddArray(array)     
			for array in Control_Volume_Values:                                                                   
				Integration_Line.GetOutput().GetPointData().AddArray(array)
			for array in Survey_Plane_Maskell_Values:                                                                   
				Integration_Line.GetOutput().GetPointData().AddArray(array)  
			self.GetOutput().ShallowCopy(Integration_Line.GetOutput())                                      # Displays output data



############################################################################################################
#########################################  3D TO 2D MAPPING PLANE  #########################################
############################################################################################################

if Integral_Type == 3:                                                                                      # 3=3D to 2D mapping

	# DEFINE THE POINTS AND NORMAL FOR THE MESH USED IN THE 3D TO 2D MAPPING:
	numPts = NUMPTS_IJ[0]								                                                    # Number of points along the i-axis of the sweep plane
	Start = START_IJ[0]								                                                        # Start coordinate [m]
	End = END_IJ[0]									                                                        # End coordinate [m]

	numPts1 = NUMPTS_IJ[1]								                                                    # Number of points along the j-axis of the sweep plane
	Start1 = START_IJ[1]								                                                    # Star coordinate [m]
	End1 = END_IJ[1]									                                                    # End coordinate [m]

	if Mapping_Normal == 0:    																			    # 0=Aerodynamic x-axis 
		Y = np.linspace(Start,End,numPts)  																	# Creates points along mesh Y-axis
		Z = np.linspace(Start1,End1,numPts1)																# Creates points along mesh Z-axis
		Ymesh, Zmesh = np.meshgrid(Y, Z)  																	# Creates the points of the mesh
		Xmesh = 0*Ymesh  																					# Creates an array with zeros
		POINTS = algs.make_vector(Xmesh.ravel(order='F'), Ymesh.ravel(order='F'), Zmesh.ravel(order='F')) 	# Ravel converts the matrix-like arrays into vector-like arrays
		POINTS = np.einsum('ij,kj->ik',POINTS,T_inverse)													# The mesh points are converted from MESH reference frame to aero frame
		normal = np.array([0,1,0])																			# Survey plane normal in mesh reference frame
		NORMAL = np.dot(T_inverse,normal)   																# Survey plane normal in aero reference frame
		NORMAL_int = np.array([0,0,1])																		# Intersection plane normal in mesh reference frame
		NORMAL_int = np.dot(T_inverse,NORMAL_int)															# Intersection plane normal in aero reference frame

	if Mapping_Normal == 1:                                                                                 # 1=Aerodynamic y-axis  
		X = np.linspace(Start,End,numPts)
		Z = np.linspace(Start1,End1,numPts1)		
		Xmesh, Zmesh = np.meshgrid(X, Z)
		Ymesh = 0*Xmesh
		POINTS = algs.make_vector(Xmesh.ravel(order='F'), Ymesh.ravel(order='F'), Zmesh.ravel(order='F'))
		POINTS = np.einsum('ij,kj->ik',POINTS,T_inverse)
		normal = np.array([1,0,0])
		NORMAL = np.dot(T_inverse,normal)   			                                                    # Survey plane normal
		NORMAL_int = np.array([0,0,1])		
		NORMAL_int = np.dot(T_inverse,NORMAL_int)		                                                    # Intersection plane normal

	if Mapping_Normal == 2:                                                                                 # 2=Aerodynamic z-axis 
		X = np.linspace(Start,End,numPts)
		Y = np.linspace(Start1,End1,numPts1)		
		Xmesh, Ymesh = np.meshgrid(X, Y)
		Zmesh = 0*Xmesh
		POINTS = algs.make_vector(Xmesh.ravel(order='F'), Ymesh.ravel(order='F'), Zmesh.ravel(order='F'))
		POINTS = np.einsum('ij,kj->ik',POINTS,T_inverse)
		normal = np.array([1,0,0])
		NORMAL = np.dot(T_inverse,normal)   			                                                    # Survey plane normal
		NORMAL_int = np.array([0,1,0])		
		NORMAL_int = np.dot(T_inverse,NORMAL_int)		                                                    # Intersection plane normal

	if Mapping_Normal == 3:                                                                                 # 3=Mesh X-axis 
		Y = np.linspace(Start,End,numPts)
		Z = np.linspace(Start1,End1,numPts1)		
		Ymesh, Zmesh = np.meshgrid(Y, Z)
		Xmesh = 0*Ymesh
		POINTS = algs.make_vector(Xmesh.ravel(order='F'), Ymesh.ravel(order='F'), Zmesh.ravel(order='F'))
		normal = np.array([0,1,0])
		NORMAL = np.dot(T_inverse,normal)   			                                                    # Survey plane normal
		NORMAL_int = np.array([0,0,1])		
		NORMAL_int = np.dot(T_inverse,NORMAL_int)		                                                    # Intersection plane normal
	
	if Mapping_Normal == 4:                                                                                 # 4=Mesh Y-axis
		X = np.linspace(Start,End,numPts)
		Z = np.linspace(Start1,End1,numPts1)		
		Xmesh, Zmesh = np.meshgrid(X, Z)
		Ymesh = 0*Xmesh
		POINTS = algs.make_vector(Xmesh.ravel(order='F'), Ymesh.ravel(order='F'), Zmesh.ravel(order='F'))
		normal = np.array([1,0,0])
		NORMAL = np.dot(T_inverse,normal)   			                                                    # Survey plane normal
		NORMAL_int = np.array([0,0,1])		
		NORMAL_int = np.dot(T_inverse,NORMAL_int)		                                                    # Intersection plane normal
	
	if Mapping_Normal == 5:                                                                                 # 5=Mesh Z-axis
		X = np.linspace(Start,End,numPts)
		Y = np.linspace(Start1,End1,numPts1)		
		Xmesh, Ymesh = np.meshgrid(X, Y)
		Zmesh = 0*Xmesh
		POINTS = algs.make_vector(Xmesh.ravel(order='F'), Ymesh.ravel(order='F'), Zmesh.ravel(order='F'))
		normal = np.array([1,0,0])
		NORMAL = np.dot(T_inverse,normal)   			                                                    # Survey plane normal
		NORMAL_int = np.array([0,1,0])		
		NORMAL_int = np.dot(T_inverse,NORMAL_int)		                                                    # Intersection plane normal


	# DISPLAY MAPPING PLANE:
	if Select_Action == 6 or Select_Action == 3:                                                            # 6=Visualize mapping plane 3=Perform automatic integration
		
		# Send POINTS data to Paraview's output array:
		pts = vtk.vtkPoints()                                           	                                # create a vtkPoints container to store all the POINTS coordinates.
		pts.SetData(dsa.numpyTovtkDataArray(POINTS , "POINTS"))    			                                # numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()
		output = self.GetOutput()                                       	                                # It calls the output port of the Paraview source
		output.SetPoints(pts)                                           	                                # It sends the points to the output
		#If needed, these output points can be visualized by applying a "Glyph" filter to the output of this programmable filter
		
		# Create surface mesh:
		NBcells=(numPts-1)*(numPts1-1)                         				                                # Number of cells
		
		IDlist = []
		for i in range(int(numPts)-1):                           			                                # We iterate over each cell in order to define the counter-clockwise point ID that defines the cell connectivity
			for j in range(int(numPts1)-1):
				IDlist.append([i*numPts1+j,j+(i+1)*numPts1,j+(i+1)*numPts1+1,i*numPts1+j+1])
		IDlist = np.array(IDlist)
			
		numCells = int(NBcells)
		output.Allocate(numCells, numPts*numPts1)                      		                                # Allocate the cells in the memory
		for cell in range(numCells):
			numCellPoints = int(4)                                                                          # There are 4 points for each cell (vtk_quad cell type)
			pointIds = vtk.vtkIdList()                                                                      # Creates an empty array to store the cells ID's
			for pointId in range(numCellPoints):
				pointIds.InsertId(pointId, int(IDlist[cell][pointId]))
			output.InsertNextCell(9, pointIds)                                                              # 9=VTK_quad


	# AUTOMATIC INTEGRATION:
	if  Select_Action == 3:	                                                                                # 3=Perform automatic integration

		Values = []
		pdo = self.GetOutput()
		#pdo = self.GetPolyDataOutput()  										                            #Get a vtk.PolyData object for the output
		for i in range(0, numPts): 

			# Create survey plane:
			plane = vtk.vtkPlane()
			plane.SetOrigin(POINTS[i*numPts1,0],POINTS[i*numPts1,1],POINTS[i*numPts1,2])
			plane.SetNormal(NORMAL[0], NORMAL[1], NORMAL[2])
			
			survey = SLICE_WITH_PLANE(Input_Filtered.GetOutput(),plane)                                     # Slices of the input with the survey plane

			for j in range(0, numPts1):

				# Create intersection plane:
				plane_int = vtk.vtkPlane()
				plane_int.SetOrigin(POINTS[i*numPts1+j,0],POINTS[i*numPts1+j,1],POINTS[i*numPts1+j,2])
				plane_int.SetNormal(NORMAL_int[0], NORMAL_int[1], NORMAL_int[2])

				# Creates intersection line between the survey plane and the intersection plane               
				int_line = SLICE_WITH_PLANE(survey.GetOutput(),plane_int)

				# Integrate variables along the line:
				integrate = vtk.vtkIntegrateAttributes()
				integrate.SetInputData(int_line.GetOutput())
				integrate.SetDivideAllCellDataByVolume(1)
				integrate.Update()

				pd = integrate.GetOutput().GetPointData()
				#cd = integrate.GetOutput().GetCellData()

				# Add name and values to the output variables:
				for k in range(pd.GetNumberOfArrays()):
					if i == 0 and j == 0 :
						arrayName = pd.GetArray(k).GetName()
						Values.append(vtk.vtkDoubleArray())
						Values[k].SetName(arrayName)				

					arrayValue = pd.GetArray(k).GetTuple(0)[0]
					Values[k].InsertNextValue(arrayValue)

		for array in Values:
			pdo.GetPointData().AddArray(array)



############################################################################################################
###############################################  SAVE DATA  ################################################
############################################################################################################

if Save_Output_Data==1:
	WRITE_VTP(output,File_Path_Output)
	print ("# Output file written successfully")



############################################################################################################
###############################################  END OF CODE  ##############################################
############################################################################################################