# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################                 Near Field Plugin                #############################
############################################################################################################

# Abstract:
This Paraview's programmable filter calculates the near-field aerodynamic parameters. Details of the 
formulations and how to use the plugin are available in the User's Manual.


# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""

print ("")
print ("")
print ("")
print ("####################################################")
print ("############     EPSILON NEAR-FIELD     ############")
print ("####################################################")
print ("")



############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

# We load the Python and Paraview standard modules:
import vtk
import vtk.vtkFiltersParallel
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa
import numpy as np



############################################################################################################
############################################# INPUT PARAMETERS #############################################
############################################################################################################

# Input parameters are defined as string in Paraview's GUI, thus a string to float convertion is required:
Input_Type = float(Input_Type)                                                  # 0=2D  1=2D extruded (One cell)    2=3D extruded (Multiple cells)  3=3D
Cref = float(Cref)
X0 = Origin[0]     						                                        # Moment center X-coordinate in the mesh reference frame
Y0 = Origin[1]     						                                        # Moment center Y-coordinate in the mesh reference frame
Z0 = Origin[2]       					                                        # Moment center Z-coordinate in the mesh reference frame
Radius = float(Radius)                                                          # Cylinder radius (used to clip 2D data)
ExtrusionLenght = float(ExtrusionLenght)
WidthPercentage = float(WidthPercentage)
PercentageXmin = PercentageX[0]
PercentageXmax = PercentageX[1]
PercentageYmin = PercentageY[0]
PercentageYmax = PercentageY[1]
PercentageZmin = PercentageZ[0]
PercentageZmax = PercentageZ[1]
FlipNormals = float(FlipNormals)
CleanGrid = float(CleanGrid)
Load_clipped_geometry = float(Load_clipped_geometry)
Keep_Input_Data = float(Keep_Input_Data)
Set_Non_Slip_Condition = float(Set_Non_Slip_Condition)
Integrate_Forces_Moments = float(Integrate_Forces_Moments)
CellBasedCalculation = CellBasedCalculation
ForceWallVorticity = float(ForceWallVorticity)

INPUT=inputs[0]                                                                 # Get input data

ARRAY = INPUT.FieldData["ARRAY"]   
ARRAY_Integration = INPUT.FieldData["ARRAY_Integration"]  
AdimF = INPUT.FieldData["AdimF"]                       
DimF = INPUT.FieldData["DimF"]
Sref = INPUT.FieldData["Sref"]
TRANSFORMATION = INPUT.FieldData["TRANSFORMATION"]
UpwardAxis = INPUT.FieldData["UpwardAxis"][0]
WallVorticityFactor = INPUT.FieldData["WallVorticityFactor"][0]
UDV = INPUT.FieldData["UDV"]
NEAR_FIELD = INPUT.FieldData["NEAR_FIELD"]


############################################################################################################
##################################### GET BODY SURFACE FROM INPUT DATA #####################################
############################################################################################################

if Load_clipped_geometry == 0:                                                  # If the user is not providing the body surface already clipped, then the code will get that surface

    if CleanGrid == 1:                                                          # Eliminates the internal boundary surfaces separating the cell partitions used for multi-core processing in CFD
        clean = vtk.vtkmCleanGrid()                         
        clean.SetInputData(INPUT.VTKObject)  
        clean.CompactPointsOn()                                 
        clean.Update()
        INPUT = dsa.WrapDataObject(clean.GetOutput())                           # Converts the VTK object back to a Numpy object

    ########### 2D DOMAIN ###########
    if Input_Type == 0:                                                         # If input data is 2D: it detects the body gometry (a curve in the space) and extrude it normal to the CFD plane              

        if NEAR_FIELD==0:
            print("# ERROR: Near Field Method was not enabled in Epsilon Aero Data. Only the surface can be extracted but no near-field forces can be computed")
        print("# 2D CASE SELECTED: in order to see the body, be sure that the render view mode is in '3D' instead of '2D'")
        print("")

        # Extract surface
        ExtractSurface = vtk.vtkGeometryFilter()                                # Converts input format to vtkPolyData
        ExtractSurface.SetInputData(INPUT.VTKObject)               
        ExtractSurface.Update()

        # Detect edges:                                                         # It detects all the edges of the CFD domain: the body and the boundaries
        Edges = vtk.vtkFeatureEdges()
        Edges.SetInputData(ExtractSurface.GetOutput())
        Edges.FeatureEdgesOff()
        Edges.NonManifoldEdgesOff()
        Edges.Update()

        # Define normal direction to 2D plane:
        if UpwardAxis == 0:                                                     # UpwardAxis=0: Y-axis, UpwardAxis=1: Z-axis
            NormalAxis = (0,0,1)
        else:
            NormalAxis = (0,1,0)

        # Define cylinder to clip data:                                         # The cylinder must enclose the body only, so, when it is used a clip tool, it takes away the boundary from the detected edges
        Cylinder = vtk.vtkCylinder()
        Cylinder.SetCenter(X0, Y0, Z0)
        Cylinder.SetRadius(Radius)
        Cylinder.SetAxis(NormalAxis[0],NormalAxis[1],NormalAxis[2])

        # Clip by cylinder:                                                     # Eliminates the edges corresponding to the outer domain boundary. Output vtkUnstructuredGrid
        clipper = vtk.vtkTableBasedClipDataSet()
        clipper.SetInputData(Edges.GetOutput())
        clipper.SetClipFunction(Cylinder)
        clipper.SetValue(0)
        clipper.GenerateClipScalarsOff()
        clipper.InsideOutOn()
        clipper.GetOutput().GetPointData().CopyScalarsOff()
        clipper.Update() 

        # Extract surface:
        ExtractSurface = vtk.vtkGeometryFilter()                                # Converts vtkUnstructuredGrid to vtkPolyData 
        ExtractSurface.SetInputData(clipper.GetOutput())               
        ExtractSurface.Update()

        # Extrude body edges normal to the plane:
        extrude = vtk.vtkLinearExtrusionFilter()
        extrude.CappingOff()
        extrude.SetExtrusionTypeToVectorExtrusion()
        extrude.SetVector(NormalAxis[0],NormalAxis[1],NormalAxis[2])
        extrude.SetInputData(ExtractSurface.GetOutput())
        extrude.SetScaleFactor(ExtrusionLenght)                                      
        extrude.Update()
                
        # Send mesh to the output (without arrays):                             # This mesh is used as a canvas where all the new arrays (calculated below) will be added
        mesh=vtk.vtkPassArrays()   
        mesh.SetInputDataObject(extrude.GetOutput())
        if Keep_Input_Data == 0:                                                # If the user do not want to keep input data, it deletes all the existing point/cell arrays at the input mesh
            mesh.AddPointDataArray("")                               	        
            mesh.AddCellDataArray("")
            mesh.AddFieldDataArray("")
        if UDV == 1:
            for L in range(INPUT.GetPointData().GetNumberOfArrays()):
                if INPUT.GetPointData().GetArray(L).GetName().startswith('UDV')==True:	
                    mesh.AddPointDataArray(INPUT.GetPointData().GetArray(L).GetName())
        mesh.Update()
        self.GetOutput().ShallowCopy(mesh.GetOutput())                          # This send the mesh to the output of ParaView

        # Converts VTK data to Python:
        data = dsa.WrapDataObject(extrude.GetOutput())    



    ########### EXTRUDED 2D DOMAIN ###########
    if Input_Type == 1:                                                         # Simmilar procedure as the 2D case. Here the body edge is obtained by slicing the existing extruded surface of 1 cell of the mesh

        # Extract surface
        ExtractSurface = vtk.vtkGeometryFilter()                                # It extracts the external surface of the mesh volume 
        ExtractSurface.SetInputData(INPUT.VTKObject)               
        ExtractSurface.Update()

        # Get bounds of the CFM domain:
        data = dsa.WrapDataObject(ExtractSurface.GetOutput())
        POINTS = data.Points
        Xmin = np.min(POINTS[:,0])
        Xmax = np.max(POINTS[:,0])
        Ymin = np.min(POINTS[:,1])
        Ymax = np.max(POINTS[:,1])
        Zmin = np.min(POINTS[:,2])
        Zmax = np.max(POINTS[:,2])  
        Xmean = 0.5*(Xmin+Xmax)
        Ymean = 0.5*(Ymin+Ymax)
        Zmean = 0.5*(Zmin+Zmax)
        
        # Define normal direction to 2D plane:
        if UpwardAxis == 0:                                                     # UpwardAxis=0: Y-axis, UpwardAxis=1: Z-axis
            NormalAxis = (0,0,1)
        else:
            NormalAxis = (0,1,0)
        
        # Create cutting plane:                                                 # Creates a plane which will be used later as a cutting tool
        plane = vtk.vtkPlane() 
        plane.SetOrigin(Xmean, Ymean, Zmean)                                    # The plane is placed at the center point of the domain so it can slice the body cell
        plane.SetNormal(NormalAxis[0],NormalAxis[1],NormalAxis[2])

        # Slice the extracted surface with the plane:                           # This allows obtaining the edges of the body profile and the boundaries
        cutter = vtk.vtkCutter()                                                
        cutter.SetCutFunction(plane)
        cutter.SetInputData(ExtractSurface.GetOutput())
        cutter.Update()

        # Define cylinder to clip data:                                         # The cylinder must enclose the body only, so, when it is used a clip tool, it takes away the boundary from the detected edges
        Cylinder = vtk.vtkCylinder()
        Cylinder.SetCenter(X0, Y0, Z0)
        Cylinder.SetRadius(Radius)
        Cylinder.SetAxis(NormalAxis[0],NormalAxis[1],NormalAxis[2])

        # Clip by cylinder:                                                     # Eliminates the edges corresponding to the outer domain boundary. Output vtkUnstructuredGrid
        clipper = vtk.vtkTableBasedClipDataSet()                               
        clipper.SetInputData(cutter.GetOutput())
        clipper.SetClipFunction(Cylinder)
        clipper.SetValue(0)
        clipper.GenerateClipScalarsOff()
        clipper.InsideOutOn()
        clipper.GetOutput().GetPointData().CopyScalarsOff()
        clipper.Update() 

        # Extract surface:
        ExtractSurface = vtk.vtkGeometryFilter()                                # Converts vtkUnstructuredGrid to vtkPolyData 
        ExtractSurface.SetInputData(clipper.GetOutput())               
        ExtractSurface.Update()

        # Linear extrusion:
        extrude = vtk.vtkLinearExtrusionFilter()
        extrude.CappingOff()
        extrude.SetExtrusionTypeToVectorExtrusion()
        extrude.SetVector(NormalAxis[0],NormalAxis[1],NormalAxis[2])
        extrude.SetInputData(ExtractSurface.GetOutput())
        extrude.SetScaleFactor(ExtrusionLenght)                                   
        extrude.Update()
        
        # Send mesh to the output (without arrays) :                            # This mesh is used as a canvas where all the new arrays (calculated below) will be added
        mesh=vtk.vtkPassArrays()   
        mesh.SetInputDataObject(extrude.GetOutput())
        if Keep_Input_Data == 0:                                                # If the user do not want to keep input data, it deletes all the existing point/cell arrays at the input mesh
            mesh.AddPointDataArray("")                               	            
            mesh.AddCellDataArray("")
            mesh.AddFieldDataArray("")
        if UDV == 1:
            for L in range(INPUT.GetPointData().GetNumberOfArrays()):
                if INPUT.GetPointData().GetArray(L).GetName().startswith('UDV')==True:	
                    mesh.AddPointDataArray(INPUT.GetPointData().GetArray(L).GetName())
        mesh.Update()
        self.GetOutput().ShallowCopy(mesh.GetOutput())                          # This send the mesh to the output of ParaView

        # Converts VTK data to Python
        data = dsa.WrapDataObject(extrude.GetOutput())                     



    ########### EXTRUDED 3D DOMAIN ###########
    if Input_Type == 2:                                                         # Typical of LES simulation of airfoils. Here, the two parallel fluid domain surfaces are removed (as well as the outter boundaries), leaving only the body surface
        
        # Extract surface:                                                      # It extracts the external surface of the mesh volume 
        ExtractSurface = vtk.vtkGeometryFilter()                     
        ExtractSurface.SetInputData(INPUT.VTKObject)               
        ExtractSurface.Update()
        
        # Define normal direction to 2D plane:
        if UpwardAxis == 0:                                                     # UpwardAxis=0: Y-axis, UpwardAxis=1: Z-axis
            NormalAxis = (0,0,1)
        else:
            NormalAxis = (0,1,0)

        # Define cylinder to clip data:                                         # The cylinder must enclose the body only, so, when it is used a clip tool, it takes away the domain boundary from the extracted surfaces
        Cylinder = vtk.vtkCylinder()
        Cylinder.SetCenter(X0, Y0, Z0)
        Cylinder.SetRadius(Radius)
        Cylinder.SetAxis(NormalAxis[0],NormalAxis[1],NormalAxis[2])

        # Clip by cylinder:                                                     # Eliminates the faces corresponding to the outer domain boundary and part of the two parallel fluid domain surfaces
        clipper = vtk.vtkTableBasedClipDataSet()                               
        clipper.SetInputData(ExtractSurface.GetOutput())
        clipper.SetClipFunction(Cylinder)
        clipper.SetValue(0)
        clipper.GenerateClipScalarsOff()
        clipper.InsideOutOn()
        clipper.GetOutput().GetPointData().CopyScalarsOff()
        clipper.Update()           
        
        # Get bounds of the clipped zone:
        data = dsa.WrapDataObject(clipper.GetOutput())
        POINTS = data.Points
        Xmin = np.min(POINTS[:,0])
        Xmax = np.max(POINTS[:,0])
        Ymin = np.min(POINTS[:,1])
        Ymax = np.max(POINTS[:,1])
        Zmin = np.min(POINTS[:,2])
        Zmax = np.max(POINTS[:,2])  
        DeltaY = Ymax-Ymin
        DeltaZ = Zmax-Zmin 
        
        if UpwardAxis == 0:                                                     # UpwardAxis=0: Y-axis, UpwardAxis=1: Z-axis
            YminNew = Ymin
            YmaxNew = Ymax
            ZminNew = Zmin+(1-WidthPercentage)*DeltaZ                           # In order to get rid of the parallel fluid surfaces, the domain width is reduced by using the "WidthPercentage" parameter
            ZmaxNew = Zmax-(1-WidthPercentage)*DeltaZ 
            ExtrusionLenght = ZmaxNew-ZminNew                                   # This parameter is used later during the surface integral phase
        else:
            YminNew = Ymin+(1-WidthPercentage)*DeltaY
            YmaxNew = Ymax-(1-WidthPercentage)*DeltaY
            ZminNew = Zmin
            ZmaxNew = Zmax
            ExtrusionLenght = YmaxNew-YminNew
        
        # Clip by box:                                                          # This takes away the parallel fluid surfaces, leaving only the body surface
        box = vtk.vtkBox()
        box.SetBounds(Xmin,Xmax,YminNew,YmaxNew,ZminNew,ZmaxNew)

        boxclip = vtk.vtkTableBasedClipDataSet() 
        boxclip.SetInputData(ExtractSurface.GetOutput())
        boxclip.SetClipFunction(box)
        boxclip.SetValue(0)
        boxclip.GenerateClipScalarsOff()
        boxclip.InsideOutOn()
        boxclip.GetOutput().GetPointData().CopyScalarsOff()
        boxclip.Update()        
        
        """  This is an old code snippet --> it triangulates all types of 3D cells which is undesired (left here for debug purposes                                
        boxclip = vtk.vtkBoxClipDataSet()
        boxclip.GenerateClipScalarsOn()
        boxclip.SetBoxClip(Xmin,Xmax,YminNew,YmaxNew,ZminNew,ZmaxNew)
        boxclip.SetInputData(clipper.GetOutput())
        boxclip.Update()
        """

        # Extract surface:                                                      # Converts vtkUnstructuredGrid to vtkPolyData 
        ExtractSurface = vtk.vtkGeometryFilter() 
        ExtractSurface.SetInputData(boxclip.GetOutput())               
        ExtractSurface.Update()

        # Send mesh to the output (without arrays) :                            # This mes is used as a canvas where all the new arrays (calculated below) will be added
        mesh=vtk.vtkPassArrays()   
        mesh.SetInputDataObject(ExtractSurface.GetOutput())
        if Keep_Input_Data == 0:                                                # If the user do not want to keep input data, it deletes all the existing point/cell arrays at the input mesh
            mesh.AddPointDataArray("")                               	           
            mesh.AddCellDataArray("")
            mesh.AddFieldDataArray("")
        if UDV == 1:
            for L in range(INPUT.GetPointData().GetNumberOfArrays()):
                if INPUT.GetPointData().GetArray(L).GetName().startswith('UDV')==True:	
                    mesh.AddPointDataArray(INPUT.GetPointData().GetArray(L).GetName())
        mesh.Update()
        self.GetOutput().ShallowCopy(mesh.GetOutput())                          # This send the mesh to the output of ParaView

        # Converts VTK data to Python
        data = dsa.WrapDataObject(ExtractSurface.GetOutput())                     


    ########### 3D DOMAIN ###########
    if Input_Type == 3:                                                         # The body surface is obtained by clipping the domain's external surface with a box just enclosing the body

        # Extract surface:                                                      # It extracts the external surface of the mesh volume 
        ExtractSurface = vtk.vtkGeometryFilter()                     
        ExtractSurface.SetInputData(INPUT.VTKObject)               
        ExtractSurface.Update()

        # Get bounds of the domain:
        data = dsa.WrapDataObject(ExtractSurface.GetOutput())
        POINTS = data.Points
        Xmin = np.min(POINTS[:,0])
        Xmax = np.max(POINTS[:,0])
        Ymin = np.min(POINTS[:,1])
        Ymax = np.max(POINTS[:,1])
        Zmin = np.min(POINTS[:,2])
        Zmax = np.max(POINTS[:,2])  
        DeltaX = Xmax-Xmin
        DeltaY = Ymax-Ymin
        DeltaZ = Zmax-Zmin 
        ExtrusionLenght = 'None'                                                # This parameter is used later during the surface integral phase
        
        # Defining the bounds of the clipping box: 
        XminNew = Xmin + PercentageXmin*DeltaX
        XmaxNew = Xmax - (1-PercentageXmax)*DeltaX   
        YminNew = Ymin + PercentageYmin*DeltaY
        YmaxNew = Ymax - (1-PercentageYmax)*DeltaY
        ZminNew = Zmin + PercentageZmin*DeltaZ
        ZmaxNew = Zmax - (1-PercentageZmax)*DeltaZ 

        # Clip by box:                                                          # This takes away all the surfaces except the body surface
        box = vtk.vtkBox()
        box.SetBounds(XminNew,XmaxNew,YminNew,YmaxNew,ZminNew,ZmaxNew)


        boxclip = vtk.vtkTableBasedClipDataSet() 
        boxclip.SetInputData(ExtractSurface.GetOutput())
        boxclip.SetClipFunction(box)
        boxclip.SetValue(0)
        boxclip.GenerateClipScalarsOff()
        boxclip.InsideOutOn()
        boxclip.GetOutput().GetPointData().CopyScalarsOff()
        boxclip.Update()   
        
        """  This is an old code snippet --> it triangulates all types of 3D cells which is undesired (left here for debug purposes 
        boxclip = vtk.vtkBoxClipDataSet()
        boxclip.GenerateClipScalarsOn()
        boxclip.SetBoxClip(XminNew,XmaxNew,YminNew,YmaxNew,ZminNew,ZmaxNew)
        boxclip.SetInputData(ExtractSurface.GetOutput())
        boxclip.Update()
        """

        # Extract surface:                                                      # Converts vtkUnstructuredGrid to vtkPolyData
        ExtractSurface = vtk.vtkGeometryFilter()                      
        ExtractSurface.SetInputData(boxclip.GetOutput())               
        ExtractSurface.Update()

        # Send mesh to the output (without arrays) :                            # This mesh is used as a canvas where all the new arrays (calculated below) will be added
        mesh=vtk.vtkPassArrays()   
        mesh.SetInputDataObject(ExtractSurface.GetOutput())
        if Keep_Input_Data == 0:                                                # If the user do not want to keep input data, it deletes all the existing point/cell arrays at the input mesh
            mesh.AddPointDataArray("")                               	            
            mesh.AddCellDataArray("")
            mesh.AddFieldDataArray("")
        if UDV == 1:
            for L in range(INPUT.GetPointData().GetNumberOfArrays()):
                if INPUT.GetPointData().GetArray(L).GetName().startswith('UDV')==True:	
                    mesh.AddPointDataArray(INPUT.GetPointData().GetArray(L).GetName())
        mesh.Update()
        self.GetOutput().ShallowCopy(mesh.GetOutput())                          # This send the mesh to the output of ParaView

        # Converts VTK data to Python
        data = dsa.WrapDataObject(ExtractSurface.GetOutput())                     

if Load_clipped_geometry == 1:                                                  # The user provides the body surface

    # Extract surface:                                                          # Converts vtkUnstructuredGrid to vtkPolyData
    ExtractSurface = vtk.vtkGeometryFilter()                      
    ExtractSurface.SetInputData(INPUT.VTKObject)               
    ExtractSurface.Update()
    
    # Send mesh to the output (without arrays) :                                # This mesh is used as a canvas where all the new arrays (calculated below) will be added
    mesh=vtk.vtkPassArrays()   
    mesh.SetInputDataObject(ExtractSurface.GetOutput())
    if Keep_Input_Data == 0:                                                    # If the user do not want to keep input data, it deletes all the existing point/cell arrays at the input mesh
        mesh.AddPointDataArray("")                               	                
        mesh.AddCellDataArray("")
        mesh.AddFieldDataArray("")
    if UDV == 1:
        for L in range(INPUT.GetPointData().GetNumberOfArrays()):
            if INPUT.GetPointData().GetArray(L).GetName().startswith('UDV')==True:	
                mesh.AddPointDataArray(INPUT.GetPointData().GetArray(L).GetName())
    mesh.Update()
    self.GetOutput().ShallowCopy(mesh.GetOutput())                              # This send the mesh to the output of ParaView

    # Converts VTK data to Python
    data = dsa.WrapDataObject(ExtractSurface.GetOutput())  
  
    # Get extrusion span:  
    if Input_Type != 3:                                                         # If the input is a extruded geometry, then, we get the span of the extrusion
        ExtrusionLenght = 'None'
        data = dsa.WrapDataObject(ExtractSurface.GetOutput())
        POINTS = data.Points
        if UpwardAxis == 0:                                                     # UpwardAxis=0: Y-axis, UpwardAxis=1: Z-axis    
            Zmin = np.min(POINTS[:,2])
            Zmax = np.max(POINTS[:,2])    
            ExtrusionLenght = Zmax-Zmin
        if UpwardAxis == 1:      
            Ymin = np.min(POINTS[:,1])
            Ymax = np.max(POINTS[:,1])       
            ExtrusionLenght = Ymax-Ymin
                
        print ("# Extruded geometry detected: span =",ExtrusionLenght, "[m]")
        print ("")    


############################################################################################################
####################################### READ DATA FROM INPUT SURFACE #######################################
############################################################################################################

Cp = data.PointData["AERO_Cp"]
Ps = data.PointData["AERO_Ps"]
MU_LAM = data.PointData["AERO_Mu_Lam"]
Vorticity = data.PointData["AERO_Vorticity"]

POINTS = data.Points

PointID_Surface = data.PointData["PointID"]                                     # Point ID at the extracted surface
PointID_Domain = INPUT.PointData["PointID"]                                     # Point ID at the entire domain
Velocity_Vector_Surface = data.PointData["AERO_Velocity_Vector"]                # Velocity field at the extracted surface
Velocity_Vector_Domain = INPUT.PointData["AERO_Velocity_Vector"]                # Velocity field at the entire domain 


############################################################################################################
############################################# NEAR-FIELD METHOD ############################################
############################################################################################################

# Correct/recompute velocity and vorticity at wall:
if Set_Non_Slip_Condition == 1:
    # Set non-slip condition at wall:
    Velocity_Vector_Surface = 0*Velocity_Vector_Surface
    Velocity_Vector_Domain[PointID_Surface[:]] = [0,0,0]
    
    # Recompute vorticity from the new velocity field:
    VORTICITY = algs.vorticity(Velocity_Vector_Domain)
    Vorticity = np.einsum('ij,kj->ik',VORTICITY,TRANSFORMATION)
    Vorticity_wall = Vorticity[PointID_Surface[:]]                              # Get vorticity at wall points
else:
    Vorticity_wall = Vorticity                                                  # It retains the vorticity field from Epsilon Aero Data

if ForceWallVorticity == 1:                                                     # Allows the user to impose by force the wall vorticity correction (Advanced option)
    WallVorticityFactor=2

Vorticity_wall = WallVorticityFactor*Vorticity_wall                             # Wall-corrected vorticity. WallVorticityFactor=2 when vorticty is computed by ParaView to take into account the lack of cells on the other side of the wall, which reduces its computed value to the half.


# Near-field forces:
if FlipNormals==0:                                                              # The user must be sure that normals are pointing outward the body surface
 FN = 1
else:
 FN = -1


# Compute cell normals at the surface:
Surface_Normals = vtk.vtkPolyDataNormals()
Surface_Normals.SetInputData(mesh.GetOutput())                                  # "mesh" is the clipped surface
Surface_Normals.ComputeCellNormalsOn()
Surface_Normals.ComputePointNormalsOn()
Surface_Normals.AutoOrientNormalsOn()                                           # If not all the normals are pointing towards the same direction, it flips the normals of the reversed cells (so, normals are consistent on the entire surface)
Surface_Normals.SetFeatureAngle(180)
Surface_Normals.SplittingOff 
Surface_Normals.Update()
WrapData  = dsa.WrapDataObject(Surface_Normals.GetOutput())                     # Converts vtk object into Python object


# Reverse normals:
NORMALS_points = FN*WrapData.PointData["Normals"]  						        # NORMALS vector in the mesh reference frame
NORMALS_cell = FN*WrapData.CellData["Normals"]  						        # NORMALS vector in the mesh reference frame


if CellBasedCalculation == 0: 
    
    # Correct/transform normals:
    if Input_Type == 0 or Input_Type == 1:                                          # For 2D and 2D extruded only
        if NORMALS_points.shape[0] != Cp.shape[0]:                                  # Simetimes the normal computed by VTK from the extruded surface are duplicated then, the number of normals is not equal to the number of points
            NORMALS_points = NORMALS_points[1::2]                                   # By using "[1::2]" it keep the odd elements only of the NORMAL array (so, the duplicated elements are eliminated)

    normals_points = np.einsum('ij,kj->ik',NORMALS_points,TRANSFORMATION) 		    # normals vector in the aerodynamic reference frame
    
    
    #NORMALS_points = FN*vertex_normal(data)  							        # NORMALS vector in the mesh reference frame
    #normals_points = np.einsum('ij,kj->ik',NORMALS_points,TRANSFORMATION) 	    # normals vector in the aerodynamic reference frame (einsum=Pythonic way to calculate the dot product)
    F_pressure = -1*Cp*normals_points*DimF/(Sref*AdimF)
    F_friction = -algs.cross(normals_points,MU_LAM*Vorticity_wall)/AdimF        # Wu equation 18 (Paper 1989). 
    F_total = F_pressure+F_friction
    output.PointData.append(NORMALS_points,"NORMALS_points")


if CellBasedCalculation == 1:                                                   # 0=Computes forces from point data     1=from cell data(Gives highly accurate pressure drag values)

    output.CellData.append(NORMALS_cell,"NORMALS_cell")

    # convert point data to cell data:
    data.PointData.append(Vorticity_wall,"Vorticity_wall")                      # Add a missing array to "data" before the conversion

    p2c = vtk.vtkPointDataToCellData()     							            # Converting the input from point Data To cell Data
    p2c.SetInputData(data.VTKObject)
    p2c.Update()
    WrapCellData = dsa.WrapDataObject(p2c.GetOutput())                          # Converts vtk object into Python object
    Cp_cell = WrapCellData.CellData["AERO_Cp"]                                  # Get some cell-based arrays
    MU_LAM_cell = WrapCellData.CellData["AERO_Mu_Lam"]
    Vorticity_wall_cell = WrapCellData.CellData["Vorticity_wall"]
    output.CellData.append(Cp_cell,"AERO_Cp_cell")
    
    # Correct/transform normals:
    if Input_Type == 0 or Input_Type == 1:                                      # For 2D and 2D extruded only
        if NORMALS_cell.shape[0] != Cp_cell.shape[0]:                           # Simetimes the normal computed by VTK from the extruded surface are duplicated then, the number of normals is not equal to the number of points
            NORMALS_cell = NORMALS_cell[1::2]                               # By using "[1::2]" it keep the odd elements only of the NORMAL array (so, the duplicated elements are eliminated)

    normals_cell = np.einsum('ij,kj->ik',NORMALS_cell,TRANSFORMATION) 		    # normals vector in the aerodynamic reference frame 

    
    
    # Compute cell-based near-field forces:
    F_pressure_cell = -1*Cp_cell*normals_cell*DimF/(Sref*AdimF)
    F_friction_cell = -algs.cross(normals_cell,MU_LAM_cell*Vorticity_wall_cell)/AdimF                   # Wu equation 18 (Paper 1989). 
    F_total_cell = F_pressure_cell+F_friction_cell

    # Add the new arrays to the cell-data container:
    WrapCellData.CellData.append(F_pressure_cell,"F_pressure_cell")
    WrapCellData.CellData.append(F_friction_cell,"F_friction_cell")
    WrapCellData.CellData.append(F_total_cell,"F_total_cell")

    # Converts cell data to point data:
    c2p = vtk.vtkCellDataToPointData()     							   
    c2p.SetInputData(WrapCellData.VTKObject)
    c2p.Update()
    WrapPointData = dsa.WrapDataObject(c2p.GetOutput())
    
    # Get the point-based near-field forces:
    F_pressure = WrapPointData.PointData["F_pressure_cell"]
    F_friction = WrapPointData.PointData["F_friction_cell"]
    F_total = WrapPointData.PointData["F_total_cell"]


# Near-field moments:
if AdimF==1:
	AdimM = 1
else:
	AdimM = Cref
	
COORDS = algs.make_vector(POINTS[:,0]-X0,POINTS[:,1]-Y0,POINTS[:,2]-Z0)
coords = np.einsum('ij,kj->ik',COORDS,TRANSFORMATION)
M_pressure = algs.cross(coords,F_pressure)/(AdimM)                              # Since force has been already non-dimensionalized, here only the distances are non-dimensionalized
M_friction = algs.cross(coords,F_friction)/(AdimM)
M_total = M_pressure+M_friction 

# Aero parameters:
Cp_mag = abs(Cp)                                                                # Magnitude of Cp



############################################################################################################
####################################### USER-DEFINED VARIABLES (UDV) #######################################
############################################################################################################

EpsilonIntegrationTool = []


if UDV == 1:
    
    # Read UDV from input:
    UDV_new_scalar = data.PointData["UDV_new_scalar"]
    UDV_new_vector = data.PointData["UDV_new_vector"]

    # compute new UDV:
    #UDV_wall_flux = np.einsum('ij,ij->i',UDV_new_vector,normals_points)                # Dot product 
    #output.PointData.append(UDV_wall_flux,"UDV_wall_flux")
    #EpsilonIntegrationTool.append(["UDV_wall_flux","Survey_Plane"])


############################################################################################################
################################################ OUTPUT DATA ###############################################
############################################################################################################

output.FieldData.append(TRANSFORMATION, "TRANSFORMATION")

output.PointData.append(Ps, "AERO_Ps")
output.PointData.append(Cp,"AERO_Cp")
output.PointData.append(Cp_mag,"AERO_Cp_mag")
output.PointData.append(Vorticity_wall,"AERO_Vorticity_Wall")
output.PointData.append(F_total,"NEAR-FIELD_F_total")
output.PointData.append(F_pressure,"NEAR-FIELD_F_pressure")
output.PointData.append(F_friction,"NEAR-FIELD_F_friction")
output.PointData.append(M_total,"NEAR-FIELD_M_total")
output.PointData.append(M_pressure,"NEAR-FIELD_M_pressure")
output.PointData.append(M_friction,"NEAR-FIELD_M_friction")
output.PointData.append(Velocity_Vector_Surface,"Velocity_Vector_Surface")

if ExtrusionLenght == "None":
    output.FieldData.append(1,"ExtrusionLenght")
else:
    output.FieldData.append(ExtrusionLenght,"ExtrusionLenght")



############################################################################################################
######################################  DEFINE INTEGRATION ATTRIBUTES  #####################################
############################################################################################################

# Define how to integrate the variables:                                                                    # Information for EpsilonIntegrationTool: the user must declare if the variable has to be integrated at the survey plane ("Survey_Plane") or at the entire control volume ("Control_Volume") or if it should not be integrated ("Do_Not_Integrate")
# OPTIONS: "Survey_Plane", "Control_Volume", "Do_Not_Integrate"

EpsilonIntegrationTool.append(["AERO_Ps","Survey_Plane"])
EpsilonIntegrationTool.append(["AERO_Cp","Survey_Plane"])
EpsilonIntegrationTool.append(["Cp_mag","Do_Not_Integrate"]) 
EpsilonIntegrationTool.append(["AERO_Vorticity_Wall","Do_Not_Integrate"])
EpsilonIntegrationTool.append(["NEAR-FIELD_F_total","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_total_X","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_total_Y","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_total_Z","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_friction","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_friction_X","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_friction_Y","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_friction_Z","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_pressure","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_pressure_X","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_pressure_Y","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_F_pressure_Z","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_total","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_total_X","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_total_Y","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_total_Z","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_pressure","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_pressure_X","Survey_Plane"])
EpsilonIntegrationTool.append(["NEAR-FIELD_M_pressure_Y","Survey_Plane"])
EpsilonIntegrationTool.append(["NEAR-FIELD_M_pressure_Z","Survey_Plane"])
EpsilonIntegrationTool.append(["NEAR-FIELD_M_friction","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_friction_X","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_friction_Y","Survey_Plane"]) 
EpsilonIntegrationTool.append(["NEAR-FIELD_M_friction_Z","Survey_Plane"]) 
EpsilonIntegrationTool.append(["Velocity_Vector_Surface","Do_Not_Integrate"])





Filter=vtk.vtkPassArrays()   
Filter.SetInputDataObject(INPUT.VTKObject)
Filter.AddPointDataArray("")
Filter.AddCellDataArray("")
Filter.AddFieldDataArray("ARRAY")
Filter.AddFieldDataArray("ARRAY_Integration")
Filter.Update()

InputWrapped = dsa.WrapDataObject(Filter.GetOutput())
NBarrays = InputWrapped.FieldData["ARRAY"].GetNumberOfValues()

if Keep_Input_Data == 1:
	for k in range(NBarrays):
		IntegrationVariable = InputWrapped.FieldData["ARRAY"].GetValue(k)
		IntegrationMode = InputWrapped.FieldData["ARRAY_Integration"].GetValue(k)
		EpsilonIntegrationTool.append([IntegrationVariable,IntegrationMode])


# Send "EpsilonIntegrationTool" array to the output:
NB_of_Arrays = np.shape(EpsilonIntegrationTool)[0]

array = vtk.vtkStringArray()
array.SetName("ARRAY")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][0])
output.GetFieldData().AddArray(array)


array = vtk.vtkStringArray()
array.SetName("ARRAY_Integration")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][1])
output.GetFieldData().AddArray(array)


############################################################################################################
########################################### INTEGRATE VARIABLES ############################################
############################################################################################################

if Integrate_Forces_Moments == 1:

    # Select the arrays to integrate:
    Filter=vtk.vtkPassArrays()   
    Filter.SetInputDataObject(self.GetOutputDataObject(0))
    Filter.AddPointDataArray("NEAR-FIELD_F_total")
    Filter.AddPointDataArray("NEAR-FIELD_F_pressure")
    Filter.AddPointDataArray("NEAR-FIELD_F_friction")
    Filter.AddPointDataArray("NEAR-FIELD_M_total")
    Filter.AddPointDataArray("NEAR-FIELD_M_pressure")
    Filter.AddPointDataArray("NEAR-FIELD_M_friction")
    if UDV == 1:
        Filter.AddPointDataArray("UDV_wall_flux")
    Filter.AddCellDataArray("")
    Filter.AddFieldDataArray("")
    Filter.Update()

    # Integrate variables:
    integrate = vtk.vtkIntegrateAttributes()
    integrate.SetInputData(Filter.GetOutput())
    integrate.Update()

    # Add name and values to the output variables:
    integrated = integrate.GetOutput()
    pd = integrated.GetPointData()

    for j in range(pd.GetNumberOfArrays()):  
        arrayName = pd.GetArray(j).GetName()
        IntegralValue = np.array(pd.GetArray(j).GetTuple(0))
        if ExtrusionLenght != 'None':                                           # Valid only for 2D, extruded 2D and extruded 3D, but NOT for pure 3D
            IntegralValue = IntegralValue/ExtrusionLenght
        output.FieldData.append(IntegralValue, arrayName)

    #Xac = - output.FieldData["NEAR-FIELD_M_total"][1]/output.FieldData["NEAR-FIELD_F_total"][2]                                                   # Aerodynamic center
    #output.FieldData.append(Xac, "X_aero_center")

############################################################################################################
################################################ PRINT INFO ################################################
############################################################################################################

if AdimF == 1:
    print ("# NEAR-FIELD FORCES [N]:")
    Unit_F = "[N]"
else:
    print ("# NEAR-FIELD FORCE COEFFICIENTS:")
    Unit_F = "[---]"
print ("Total:      ",output.FieldData["NEAR-FIELD_F_total"])
print ("Pressure:   ",output.FieldData["NEAR-FIELD_F_pressure"])
print ("Friction:   ",output.FieldData["NEAR-FIELD_F_friction"])
print ("")
if AdimF == 1:
    print ("# NEAR-FIELD MOMENTS [N.m]:")
    Unit_M = "[N.m]"
else:
    print ("# NEAR-FIELD MOMENT COEFFICIENTS:")
    Unit_M = "[---]"
print ("Total:      ",output.FieldData["NEAR-FIELD_M_total"])
print ("Pressure:   ",output.FieldData["NEAR-FIELD_M_pressure"])
print ("Friction:   ",output.FieldData["NEAR-FIELD_M_friction"])
#print ("")
#print ("# AERODYNAMIC CENTER:")
#print ("Xac [m]:      ",output.FieldData["X_aero_center"])
print ("")
print ("Forces and moments referred to the aerodynamic reference system: [x y z]")   #. Aerodynamic center in mesh reference system: [X Y Z]


array = vtk.vtkStringArray()
array.SetName("NEAR-FIELD_F_units")
array.SetNumberOfTuples(1)
array.SetValue(0,Unit_F)
output.GetFieldData().AddArray(array)

array = vtk.vtkStringArray()
array.SetName("NEAR-FIELD_M_units")
array.SetNumberOfTuples(1)
array.SetValue(0,Unit_M)
output.GetFieldData().AddArray(array)


###################################
#######  WRITE VTP FUNCTION  ######
###################################

def WRITE_VTP(outputObject,File_Path_Output):     					
	import vtk
	if File_Path_Output:
				
		OutputVTP = vtk.vtkXMLPolyDataWriter()                  
		OutputVTP.SetInputData(outputObject.VTKObject)    			    # Alternative command: SetInput           vtkObject.GetOutput()
		OutputVTP.SetFileName(File_Path_Output)                         # File path to the folder where to store the output data 
		OutputVTP.Write()                                               # "OutputVTP" is a vtkUnstructuredGrid variable

if Save_Output_Data==1:
	WRITE_VTP(output,File_Path_Output)
	print ("")
	print ("# Output file written successfully")



############################################################################################################
###############################################  END OF CODE ###############################################
############################################################################################################


