# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################                   PIV Plugin                     #############################
############################################################################################################

# Abstract:
This Paraview's programmable source creates a plane mesh from a single or multiple PIV data files. 

Accepted file formats are CSV and Tecplot. 

The accepted headers name for the variables are: X,Y,Z,Vx,Vy,Vz. It is not accepted headers with units or lowercase (i.e., x,y,z). Also, the
data must be in SI units: [m], [m/s] and [Pa], otherwise, the calculated vorticity and gradients will be wrong.

# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""


print ("")
print ("")
print ('#########################################################################################################################################')
print ('##################################                                                                     ##################################')
print ('##################################                              EPSILON 3.0                            ##################################')
print ('##################################                        Aerodynamic analysis tool                    ##################################')
print ('##################################                                                                     ##################################')
print ('#########################################################################################################################################')
print ("")
print ('Developed at ISAE-SUPAERO, France, 2022')
print ("")
print ('#DISCLAIMER:')
print ('This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee of any kind, either express')
print ('or implied. Its use for teaching, research and industrial applications is allowed under the strict responsibility of the user. The authors')
print ('of this plugin are not responsible for any damages or errors related to the use of the plugin.')

print ("")
print ("")
print ("")
print ("####################################################")
print ("###############      EPSILON PIV     ###############")
print ("####################################################")
print ("")

############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

import vtk
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa
import numpy as np
import vtk.vtkFiltersParallel
import paraview.vtk.util.numpy_support as vnp
import glob
import os
import os.path
from vtk.util import numpy_support
import sys


############################################################################################################
############################################# STRING TO FLOAT ##############################################
############################################################################################################

# Input parameters are defined as string in Paraview GUI, thus a string to float convertion is required:
NB_PIV_Regions = NB_PIV_Regions                        # Number of PIV regions
File_Path_PIV1 = str(File_Path_PIV1)  						# File path to the PIV1 data
File_Path_PIV2 = str(File_Path_PIV2)  						# File path to the PIV2 data
File_Path_PIV3 = str(File_Path_PIV3)  						# File path to the PIV3 data
File_Path_PIV4 = str(File_Path_PIV4)  						# File path to the PIV4 data
File_Path_PIV5 = str(File_Path_PIV5)  						# File path to the PIV5 data
File_Path_PIV6 = str(File_Path_PIV6)  						# File path to the PIV6 data
File_Path_PIV7 = str(File_Path_PIV7)  						# File path to the PIV7 data
File_Path_PIV8 = str(File_Path_PIV8)  						# File path to the PIV8 data
InstantaneousPIVfiles = float(InstantaneousPIVfiles)   		# 0=Open the mean file   1=Open the entire file sequence
Select_Action = float(Select_Action)                        # 0=Modify file headers, 1=Transform PIV data, 2=Merge PIV windows, 3=Average instantaneous PIV files, 4=Cleanup PIV data, 5=Pressure from PIV, 6=Temperature from PIV
ChangeHeaders = float(ChangeHeaders)                        # 0=None, 1=Dantec Tecplot
OriginPIV1 = OriginPIV1                                     # Vector with the origin coordinates of the PIV1 window
OriginPIV2 = OriginPIV2                                     # Vector with the origin coordinates of the PIV2 window
OriginPIV3 = OriginPIV3                                     # Vector with the origin coordinates of the PIV3 window
OriginPIV4 = OriginPIV4                                     # Vector with the origin coordinates of the PIV4 window
OriginPIV5 = OriginPIV5                                     # Vector with the origin coordinates of the PIV5 window
OriginPIV6 = OriginPIV6                                     # Vector with the origin coordinates of the PIV6 window
OriginPIV7 = OriginPIV7                                     # Vector with the origin coordinates of the PIV7 window
OriginPIV8 = OriginPIV8                                     # Vector with the origin coordinates of the PIV8 window
PlanesOrientation = PlanesOrientation                       # PIV planes orientation: rotations about x-, y- and z-axis
LenghtScaleFactor = float(LenghtScaleFactor)                # Lenght conversion factor. By default (0.001) it converts from mm to m (since standard PV data is in mm)
VelocityScaleFactorPIV1 = float(VelocityScaleFactorPIV1)    # Velocity conversion factor for PIV1. By default (1) it assumes that PIV data is in m/s (standard practice)
VelocityScaleFactorPIV2 = float(VelocityScaleFactorPIV2)    # Velocity conversion factor for PIV2
VelocityScaleFactorPIV3 = float(VelocityScaleFactorPIV3)    # Velocity conversion factor for PIV3
VelocityScaleFactorPIV4 = float(VelocityScaleFactorPIV4)    # Velocity conversion factor for PIV4
VelocityScaleFactorPIV5 = float(VelocityScaleFactorPIV5)    # Velocity conversion factor for PIV5
VelocityScaleFactorPIV6 = float(VelocityScaleFactorPIV6)    # Velocity conversion factor for PIV6
VelocityScaleFactorPIV7 = float(VelocityScaleFactorPIV7)    # Velocity conversion factor for PIV7
VelocityScaleFactorPIV8 = float(VelocityScaleFactorPIV8)    # Velocity conversion factor for PIV8
#ShowReferenceSystems = float(ShowReferenceSystems)          # 0=hide axes, 1=show PIV and aerodynamic axes
SpatialResolution = float(SpatialResolution)                # 0=Keep local smallest grid size, 1=Keep overall smallest, 2=Keep PIV1 grid size, 3=Keep PIV2 grid size, 4=Keep PIV3 grid size, 5=Keep PIV4 grid size, 6=Keep PIV5 grid size, 7=Keep PIV6 grid size, 8=Keep PIV7 grid size, 9=Keep PIV8 grid size
OverlappingMethod = float(OverlappingMethod)                # 0=Keep data of smallest cell size, 1=Averaged values, 2=Keep maximum values, 3=keep minimum values
#ShowPIVplanes = float(ShowPIVplanes)                        # 0=hide planes, 1=show outline of the PIV planes
AveragingMethod = float(AveragingMethod)    				# 0=None, 1=All files together, 2=Incremental, 3=Incremental with survey points
SurveyPointsID = SurveyPointsID                             # A vector with ID of the survey points for convergence monitoring (-2=Point disabled, -1=Point at the image center,  or ID value from 0 to max number of points)
ShowSurveyPoints = float(ShowSurveyPoints)                  # 0=hide survey points, 1=show survey points
VortexWandering = float(VortexWandering)                    # 0=ignore     1=include vortex wandering correction
MaskBoundaries = float(MaskBoundaries)						# 0=Do not mask   1=Mask Measurement Boundary
ThresholdValuePhi = float(ThresholdValuePhi)
MaskVorticityNoise = float(MaskVorticityNoise)				# 0=Do not mask   1=Mask Vorticity Noise
ThresholdValueVorticity = float(ThresholdValueVorticity)
DeleteNoiseSpots = int(DeleteNoiseSpots)					# 0=None  1=1-iteration  2=2-iterations  3=3-iterations  4=4-iterations
IntegrationY = float(IntegrationY) 							# 0=Y+  1=Y-   2=Y+-_interpolated   3=Y+(no error)   4=Y-(no error)   5=Y+-_interpolated(no error)
IntegrationZ = float(IntegrationZ)
BoundaryLayerDefinition = float(BoundaryLayerDefinition) 	# 0=None   1=XY plane   2=XZ plane
IsentropicVelocity = float(IsentropicVelocity)   			# 0=Vx+Vx_bar_y   1=Vx+Vx_bar_z
Core_Pressure = float(Core_Pressure)  						# 0=ignore  1=Z+  2=Z-  3=Y+  4=Y-
Q_threshold = float(Q_threshold)   							
Ts_Calculation_Method = float(Ts_Calculation_Method)        # 0=Isoenergetic flow    1=Vorticity breakdown method
Switch_IJ_axes = float(Switch_IJ_axes)	  					# 0=No  1=Switch
Cut_Edges_PIV1 = float(Cut_Edges_PIV1)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
Cut_Edges_PIV2 = float(Cut_Edges_PIV2)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
Cut_Edges_PIV3 = float(Cut_Edges_PIV3)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
Cut_Edges_PIV4 = float(Cut_Edges_PIV4)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
Cut_Edges_PIV5 = float(Cut_Edges_PIV5)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
Cut_Edges_PIV6 = float(Cut_Edges_PIV6)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
Cut_Edges_PIV7 = float(Cut_Edges_PIV7)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
Cut_Edges_PIV8 = float(Cut_Edges_PIV8)                      # 0=None  1=All four edges   2=Left edge  3=Right edge  4=Bottom edge  5=Top edge
NB_Layers_PIV1 = float(NB_Layers_PIV1)                      # Number of cell layers to be removed from the PIV data
NB_Layers_PIV2 = float(NB_Layers_PIV2)                      # Number of cell layers to be removed from the PIV data
NB_Layers_PIV3 = float(NB_Layers_PIV3)                      # Number of cell layers to be removed from the PIV data
NB_Layers_PIV4 = float(NB_Layers_PIV4)                      # Number of cell layers to be removed from the PIV data
NB_Layers_PIV5 = float(NB_Layers_PIV5)                      # Number of cell layers to be removed from the PIV data
NB_Layers_PIV6 = float(NB_Layers_PIV6)                      # Number of cell layers to be removed from the PIV data
NB_Layers_PIV7 = float(NB_Layers_PIV7)                      # Number of cell layers to be removed from the PIV data
NB_Layers_PIV8 = float(NB_Layers_PIV8)                      # Number of cell layers to be removed from the PIV data



File_Path_Vector = [File_Path_PIV1,File_Path_PIV2,File_Path_PIV3,File_Path_PIV4,File_Path_PIV5,File_Path_PIV6,File_Path_PIV7,File_Path_PIV8]    # Create a vector with all the PIV file paths.
OriginPIV_Vector = [OriginPIV1,OriginPIV2,OriginPIV3,OriginPIV4,OriginPIV5,OriginPIV6,OriginPIV7,OriginPIV8]                                    # Create a vector with all the PIV origin vectors
Velocity_Scale_Vector = [VelocityScaleFactorPIV1, VelocityScaleFactorPIV2, VelocityScaleFactorPIV3, VelocityScaleFactorPIV4, VelocityScaleFactorPIV5, VelocityScaleFactorPIV6,VelocityScaleFactorPIV7,VelocityScaleFactorPIV8]
Cut_Edges_Vector = [Cut_Edges_PIV1,Cut_Edges_PIV2,Cut_Edges_PIV3,Cut_Edges_PIV4,Cut_Edges_PIV5,Cut_Edges_PIV6,Cut_Edges_PIV7,Cut_Edges_PIV8]
NB_Layers_Vector = [NB_Layers_PIV1,NB_Layers_PIV2,NB_Layers_PIV3,NB_Layers_PIV4,NB_Layers_PIV5,NB_Layers_PIV6,NB_Layers_PIV7,NB_Layers_PIV8]


File_Path_Vector = File_Path_Vector[0:NB_PIV_Regions]                       # Retain only the NB of PIV paths selcted by the user
OriginPIV_Vector = OriginPIV_Vector[0:NB_PIV_Regions]
Velocity_Scale_Vector = Velocity_Scale_Vector[0:NB_PIV_Regions]
Cut_Edges_Vector = Cut_Edges_Vector[0:NB_PIV_Regions]
NB_Layers_Vector = NB_Layers_Vector[0:NB_PIV_Regions]



############################################################################################################
############################################ MODIFY FILE HEADERS ###########################################
############################################################################################################


if Select_Action == 0:

    sys.path.insert(0, r'C:\EPSILON\Readers')                                                                   # Sets the path to the specific reader of different PIV formats:
    from Reader_PIV_Dantec_Tecplot import *                                                                     # Import the reader of the DANTEC Tecplot PIV file format

    if ChangeHeaders==1:                                                                                        # Modifies the input file headers only if required
        print ("Epsilon PIV: CHANGING HEADERS...")
        
        pivIDcounter = 0                                                                                        # Flag used to enumerate the PIV file number
        for PIVi in File_Path_Vector:
            pivIDcounter = pivIDcounter+1
            if PIVi != "":                                                                                      # Check if the "i-th" PIV file path exists in the GUI
                print ("Epsilon PIV: Processign PIV",pivIDcounter," input file...")
                Folder_Path=os.path.dirname(PIVi)
                
                if InstantaneousPIVfiles == 0:
                    File_Name = os.path.basename(PIVi)
                    Reader_PIV_Dantec_Tecplot(File_Name, Folder_Path)                                            # Modifies the header of the selected Tecplot file
                    
                if InstantaneousPIVfiles == 1:
                    for filename in os.listdir(Folder_Path):   
                        Reader_PIV_Dantec_Tecplot(filename, Folder_Path)                                        # Modifies the headers of all the PIV files from Dantec Tecplot present in the folder 
        
        print ("Epsilon PIV: HEADERS MODIFICATION FINISHED...")



############################################################################################################
############################################# TRANSFORM PIV DATA ###########################################
############################################################################################################



# Requerimiento adicional para Epsilon V3.0: leer todos los formatos de Output de DANTEC Dynamic Studio !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# Ver si consigo algun archivo directo de LaVision y alguno de OPENPIV !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



############################################
#########  TECPLOT READER FUNCTION  ########
############################################

def TECPLOT_READER(Time, File_Path):
    import vtk
    if File_Path[0].endswith('.dat'):
	
        MBTecplotReader = vtk.vtkTecplotReader()   															    # "vtkTecplotReader" gives a vtkMultiblockDataSet output that contains several blocks of vtkStructuredGrid type
        MBTecplotReader.SetFileName(File_Path[Time])   														    # time allows selecting a file between a file series
        MBTecplotReader.Update()
        
        Tecplot = vtk.vtkGeometryFilter()    																    # Converts vtkMultiblockDataSet to vtkPolyData
        Tecplot.SetInputData(MBTecplotReader.GetOutput().GetBlock(0))   									    # "MBTecplotReader" is vtkMultiblockDataSet type. With "GetBlock(0)" it gets the only PIV block of the data (it is vtkStructuredGrid type). Then "vtkGeometryFilter" converts it to vtkPolyData
        Tecplot.Update()
        #self.GetOutput().ShallowCopy(Tecplot.GetOutput())    												    # It displays the mesh. Kept here for debugging purposes
        
        Points_i = MBTecplotReader.GetOutput().GetBlock(0).GetDimensions()[0]   							    # Get "I" index from Tecplot file
        Points_j = MBTecplotReader.GetOutput().GetBlock(0).GetDimensions()[1]   							    # Get "J" index from Tecplot file
        print ("# Epsilon PIV: Tecplot file --> I =",Points_i,"   J =",Points_j	)

        return Tecplot , Points_i, Points_j
        
    else:
        print ("# ERROR: Tecplot file (DAT) not found!")


############################################
###########  VTP READER FUNCTION  ##########
############################################

def VTP_READER(Time, File_Path):
    import vtk
    import numpy as np
    if File_Path[0].endswith('.vtp'):

        VTPdata = vtk.vtkXMLPolyDataReader()   																	# Alternative reader: vtkPolyDataReader
        VTPdata.SetFileName(File_Path[Time])
        VTPdata.Update()      																					# "VTPdata" is a vtk variable
        #self.GetOutput().ShallowCopy(VTPdata.GetOutput())

        # GET I and J INDEX:
        data = dsa.WrapDataObject(VTPdata.GetOutput())         
        DATApts = data.Points
        Xref = DATApts[0][0]
        Yref = DATApts[0][1]
        Points_i = np.size(np.where(DATApts[:,0]==Xref))						                    # "I" index from VTP file
        Points_j = np.size(np.where(DATApts[:,1]==Yref))				                    # "J" index from VTP file
        print ("# Epsilon PIV: VTP file --> I =",Points_i,"   J =",Points_j	)

        return VTPdata, Points_i, Points_j
        
    else:
        print ("# ERROR: VTP file not found!")


############################################
#######  TRANSFORM GEOMETRY FUNCTION  ######
############################################ 

def TRANSFORM_GEOMETRY(InputVTK, LenghtScaleFactor, Origin, Orientation):
    
    import vtk

    transform1 = vtk.vtkTransform()                                                                             # Generates the VTK transformation function to scale up/down the PIV data
    transform1.Scale(LenghtScaleFactor, LenghtScaleFactor, LenghtScaleFactor)
    transform1.Update()
    
    OutputVTK_T1 = vtk.vtkTransformFilter()                                                                     # Applies the scale up/down to the input PIV data and creates a transformed output
    OutputVTK_T1.SetTransform(transform1)
    OutputVTK_T1.SetInputData(InputVTK.GetOutput())
    OutputVTK_T1.Update()
    
    transform2 = vtk.vtkTransform()                                                                             # Generates the VTK transformation function to rotate the PIV data about the x-axis
    transform2.RotateX(Orientation[0])
    transform2.Update()
    
    OutputVTK_T2 = vtk.vtkTransformFilter()                                                                     # Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_T2.SetTransform(transform2)
    OutputVTK_T2.SetInputData(OutputVTK_T1.GetOutput())
    OutputVTK_T2.Update()
    
    transform3 = vtk.vtkTransform()                                                                             # Generates the VTK transformation function to rotate the PIV data about the y-axis
    transform3.RotateY(Orientation[1])
    transform3.Update()
    
    OutputVTK_T3 = vtk.vtkTransformFilter()                                                                     # Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_T3.SetTransform(transform3)
    OutputVTK_T3.SetInputData(OutputVTK_T2.GetOutput())
    OutputVTK_T3.Update()
    
    transform4 = vtk.vtkTransform()                                                                             # Generates the VTK transformation function to rotate the PIV data about the z-axis
    transform4.RotateZ(Orientation[2])
    transform4.Update()
    
    OutputVTK_T4 = vtk.vtkTransformFilter()                                                                     # Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_T4.SetTransform(transform4)
    OutputVTK_T4.SetInputData(OutputVTK_T3.GetOutput())
    OutputVTK_T4.Update()
    
    transform5 = vtk.vtkTransform()                                                                             # Generates the VTK transformation function to translate the PIV data along the x-, y- and z-axes
    transform5.Translate(Origin[0], Origin[1], Origin[2])
    transform5.Update()

    OutputVTK_T5 = vtk.vtkTransformFilter()                                                                     # Applies the translation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_T5.SetTransform(transform5)
    OutputVTK_T5.SetInputData(OutputVTK_T4.GetOutput())
    OutputVTK_T5.Update()
    
    #self.GetOutput().ShallowCopy(OutputVTK.GetOutput())    												    # It displays the mesh. Kept here for debugging purposes
    return OutputVTK_T5


############################################
##########  TRANSFORMATION MATRIX  #########
############################################

AngleX = PlanesOrientation[0]
AngleY = PlanesOrientation[1]
AngleZ = PlanesOrientation[2]

Rx = np.array([(1,0,0),(0,algs.cos(AngleX*np.pi/180),-algs.sin(AngleX*np.pi/180)),(0,algs.sin(AngleX*np.pi/180),algs.cos(AngleX*np.pi/180))])   #Rotation matrix about X-axis
Ry = np.array([(algs.cos(AngleY*np.pi/180),0,algs.sin(AngleY*np.pi/180)),(0,1,0),(-algs.sin(AngleY*np.pi/180),0,algs.cos(AngleY*np.pi/180))])  	#Rotation matrix about Y-axis
Rz = np.array([(algs.cos(AngleZ*np.pi/180),-algs.sin(AngleZ*np.pi/180),0),(algs.sin(AngleZ*np.pi/180),algs.cos(AngleZ*np.pi/180),0),(0,0,1)])  	#Rotation matrix about Z-axis

Rx_T = np.transpose(Rx)
Ry_T = np.transpose(Ry)
Rz_T = np.transpose(Rz)

MX = np.dot(Rx_T,np.dot(Rx,Rx_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh X-axis
MY = np.dot(Ry_T,np.dot(Ry,Ry_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Y-axis
MZ = np.dot(Rz_T,np.dot(Rz,Rz_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Z-axis

TRANSFORMATION = np.dot(MX,np.dot(MY,MZ))  #Total (combined) transformation matrix

T_t = np.transpose(TRANSFORMATION)
T_inverse = np.dot(T_t,np.dot(TRANSFORMATION,T_t))  	# It calculates the inverse of the TRANSFORMATION matrix

############################################
########  TRANSFORM VECTOR FUNCTION  #######
############################################

def TRANSFORM_VECTOR(InputVTK,VelocityScaleFactor,T_inverse):
    import vtk
    from vtk.numpy_interface import algorithms as algs
    from vtk.numpy_interface import dataset_adapter as dsa
    import numpy as np

    # READ VELOCITY DATA:
    data = dsa.WrapDataObject(InputVTK.GetOutput())   															    # Converts VTK element into a Python object

    for array in data.PointData.keys():  																		    # keys() provides the list of variables headers
        if array=="U,m_s":                 
            Vx = data.PointData["U,m_s"]                                       									    # If the array exists, then reads the array data directly from input
        if array=="V,m_s":                 
            Vy = data.PointData["V,m_s"]                                       									    # If the array exists, then reads the array data directly from input
        #Vz = 0*Vx       ---->> NO FUNCA BIEN, LO IMPONE EN ZERO Y NO DEJA LEER LA COMPONENTE Vz DEL TECPLOT EN EL PROXIMO IF....             # In PIV 2D2C this velocity component is unexistant, hence, set to zero
        if array=="W,m_s":                 
            Vz = data.PointData["W,m_s"]                                       									    # If the array exists, then reads the array data directly from input

    # SCALE UP/DOWN VELOCITY:
    Vx = Vx*VelocityScaleFactor
    Vy = Vy*VelocityScaleFactor
    Vz = Vz*VelocityScaleFactor

    # TRANSFORM VELOCITY VECTOR:
    Vector_Speed = algs.make_vector(Vx,Vy,Vz)                                                                       # Velocity vector in the local (PIV) reference system
    Velocity = np.einsum('ij,kj->ik',Vector_Speed,T_inverse) 		                                                # Velocity vector in the global (aerodynamic) reference frame
    #u = Velocity[:,0]
    #v = Velocity[:,1]
    #w = Velocity[:,2]
   
    return Velocity#,u,v,w



############################################
##########  ADD SCALAR FUNCTION  ###########
############################################

def ADD_SCALAR(InputMesh,name,scalar):
    numPts = InputMesh.GetOutput().GetNumberOfPoints()   
    NewScalar = vtk.vtkDoubleArray()
    NewScalar.SetName(name)
    NewScalar.SetNumberOfComponents(1)
    NewScalar.SetNumberOfTuples(numPts)
    for i in range(numPts):
      NewScalar.SetValue(i, scalar[i])

    InputMesh.GetOutput().GetPointData().SetScalars(NewScalar)  
    InputMesh.Update()
    
    return InputMesh

    
############################################
##########  ADD VECTOR FUNCTION  ###########
############################################

def ADD_VECTOR(InputMesh,name,vector):
    numPts = InputMesh.GetOutput().GetNumberOfPoints() 
    NewVector = vtk.vtkDoubleArray()
    NewVector.SetNumberOfComponents(3)
    NewVector.SetName(name)
    for i in range(numPts):
        NewVector.InsertNextTuple(vector[i])
        
    InputMesh.GetOutput().GetPointData().SetScalars(NewVector)  
    InputMesh.Update()
    
    return InputMesh


############################################
###########  TRANSFORM PIV DATA  ###########
############################################ 

if Select_Action > 0:

    # GET CURRENT TIME STEP:
    outInfo = self.GetOutputInformation(0)      																	# Gets the output of the "Request for information" script (i.e., "Epsilon_PIV_RI.py")
    if outInfo.Has(vtk.vtkStreamingDemandDrivenPipeline.UPDATE_TIME_STEP()):
        Time = outInfo.Get(vtk.vtkStreamingDemandDrivenPipeline.UPDATE_TIME_STEP())                                 # Get current timestep for opening multiple files
    else:
        Time = 0
    Time = int(Time)


    # ITERATES OVER EACH INPUT PIV FILE TO READ/TRANSFORM/DISPLAY ALL PIV DATA:
    PIVvtkVector=[]                                                                                                 # Empty vector to store the VTK output file of each transformed PIV data
    IJvector=[]                                                                                                     # Empty vector to store the IJ index of each PIV file
    VelocityArray=[]                                                                                                # Empty array to store the velocity vector of each PIV file

    for PIVid in range(0,len(File_Path_Vector)):

        File_Path=File_Path_Vector[PIVid]
        
        if File_Path != "":                                                                                         # Empty file paths (i.e., PIV data input not used in the GUI) are ignored

            if InstantaneousPIVfiles == 1:    																	    # Modifies the File_Path variable if a sequence of instantaneous PIV files must be loaded
                    
                Folder_Path=os.path.dirname(File_Path)
                if File_Path.endswith('.dat'):
                    File_Path = Folder_Path + "\\*.dat"															    # To open multiple files, replace filename by "*", e.g.: replace "C:\Users\...\Folder\Step0001.dat" by  "C:\Users\...\Folder\*.dat"

            File_Path_IFS = glob.glob(File_Path)   	 															    # Creates a list with all the filepaths of the PIV instantaneous file sequence (IFS) for a given PIV input (e.g., PIV1). If InstantaneousPIVfiles is not activated, it contains only one element (the mean PIV file)			
            if len(File_Path_IFS)==1:
                Time = 0

            if File_Path.endswith(".vtp"):
                PIVvtk, Points_i, Points_j = VTP_READER(Time, File_Path_IFS)                                        # Reads the Tecplot data (creates a VTK element with the PIV data)

            if File_Path.endswith(".dat"):
                PIVvtk, Points_i, Points_j = TECPLOT_READER(Time, File_Path_IFS)                                        # Reads the Tecplot data (creates a VTK element with the PIV data)
            PIVvtk.Update()


            TXPIVvtk = TRANSFORM_GEOMETRY(PIVvtk, LenghtScaleFactor, OriginPIV_Vector[PIVid], PlanesOrientation)    # Transforms the vtk element
            TXPIVvtk.Update()
            
            Velocity = TRANSFORM_VECTOR(TXPIVvtk,Velocity_Scale_Vector[PIVid],T_inverse)                                     # Transform the velocity vector from the PIV (local) reference system to the aerodynamic (global) reference system
            
            TXPIVvtk = ADD_VECTOR(TXPIVvtk,"Velocity",Velocity)
            TXPIVvtk.Update()
            
            #TXPIVvtk = ADD_SCALAR(TXPIVvtk,"u",u)
            #TXPIVvtk.Update()
            
            #TXPIVvtk = ADD_SCALAR(TXPIVvtk,"v",v)
            #TXPIVvtk.Update()
            
            #TXPIVvtk = ADD_SCALAR(TXPIVvtk,"w",w)
            #TXPIVvtk.Update()
            
            PIVvtkVector.append(TXPIVvtk)                                                                           # The transformed VTK element (transformed PIV data) is stored in the output vector (the vector contains as elements as PIV inputs declared in the GUI)
            IJvector.append([Points_i,Points_j])                                                                    # The I- and J-index of each Tecplot file is stored in the IJvector


    # COMBINE THE PIVvtkVector COMPONENTS INTO A SINGLE VTK VARIABLE:
    appendFilter = vtk.vtkAppendFilter()                                                                            # The output format of this filter is vtkUnstructuredGrid
    for ZZZ in range(0,len(PIVvtkVector)):                                                                          # Appends each element of the PIVvtkVector into the appendFilter variable
        appendFilter.AddInputData(PIVvtkVector[ZZZ].GetOutput())
    appendFilter.Update()

    # CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
    combined = vtk.vtkGeometryFilter()                                                                              # Converts vtkMultiblockDataSet to vtkPolyData
    combined.SetInputData(appendFilter.GetOutput())                             
    combined.Update()

    if Select_Action == 1:
        self.GetOutput().ShallowCopy(combined.GetOutput())                                                          # This single (combined) VTK variable can is displayed only if the step "Trasform PIV data" is selected

    #Agregar flechas VTK (vectores) para representar los ejes XYZ local/global superpuesta al archivo de salida--> dejarlo para la version definitiva de Epsilon V3.0 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



############################################################################################################
############################################# MERGE PIV WINDOWS ############################################
############################################################################################################

############################################
##########  CREATE GRID FUNCTION  ##########
############################################

def CREATE_GRID(PointsX,PointsY,PointsZ):
    import vtk
    import numpy as np
    
    xCoords = vtk.vtkFloatArray()                                                                                   # Creates a VTK container of the X-point coordinates
    for x, i in enumerate(PointsX):     
        xCoords.InsertNextValue(i)

    yCoords = vtk.vtkFloatArray()                                                                                   # Creates a VTK container of the Y-point coordinates
    for y, i in enumerate(PointsY):
        yCoords.InsertNextValue(i)

    zCoords = vtk.vtkFloatArray()                                                                                   # Creates a VTK container of the Z-point coordinates
    for z, i in enumerate(PointsZ):
        zCoords.InsertNextValue(i)
    
    rgrid = vtk.vtkRectilinearGrid()                                                                                # Creates a rectilinear grid by using the VTK coordinates
    rgrid.SetDimensions(x+1, y+1, z+1)                                                                              # x+1=number of points along X axis
    rgrid.SetXCoordinates(xCoords)
    rgrid.SetYCoordinates(yCoords)
    rgrid.SetZCoordinates(zCoords)

    ExtractSurface = vtk.vtkGeometryFilter()                                                                        # Converts vtk rectilinear grid element to vtkPolyData  --> It is equivalent to Extract Surface
    ExtractSurface.SetInputData(rgrid)           
    ExtractSurface.Update()
    #self.GetOutput().ShallowCopy(ExtractSurface.GetOutput()) 
    
    return ExtractSurface


############################################
#######  REVERSE ROTATION FUNCTION  ########
############################################

def REVERSE_ROTATION(InputVTK, Orientation):
    import vtk
    
    RotZ = vtk.vtkTransform()                                                                                       # Generates the VTK transformation function to rotate the PIV data about the x-axis
    RotZ.RotateZ(-Orientation[2])
    RotZ.Update()
    
    OutputVTK_Rz = vtk.vtkTransformFilter()                                                                         # Applies the rotation to the PIV data and creates a new transformed output
    OutputVTK_Rz.SetTransform(RotZ)
    OutputVTK_Rz.SetInputData(InputVTK.GetOutput())
    OutputVTK_Rz.Update()
    
    RotY = vtk.vtkTransform()                                                                                       # Generates the VTK transformation function to rotate the PIV data about the y-axis
    RotY.RotateY(-Orientation[1])
    RotY.Update()
    
    OutputVTK_Ry = vtk.vtkTransformFilter()                                                                         # Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_Ry.SetTransform(RotY)
    OutputVTK_Ry.SetInputData(OutputVTK_Rz.GetOutput())
    OutputVTK_Ry.Update()
    
    RotX = vtk.vtkTransform()                                                                                       # Generates the VTK transformation function to rotate the PIV data about the z-axis
    RotX.RotateX(-Orientation[2])
    RotX.Update()
    
    OutputVTK_Rx = vtk.vtkTransformFilter()                                                                         # Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_Rx.SetTransform(RotX)
    OutputVTK_Rx.SetInputData(OutputVTK_Ry.GetOutput())
    OutputVTK_Rx.Update()
    
    return OutputVTK_Rx


############################################
############  ROTATE FUNCTION  #############
############################################

def ROTATE(InputVTK, Orientation):
    import vtk
    
    RotX = vtk.vtkTransform()                                                                                       # Generates the VTK transformation function to rotate the PIV data about the x-axis
    RotX.RotateX(Orientation[0])
    RotX.Update()
    
    OutputVTK_Rx = vtk.vtkTransformFilter()                                                                         # Applies the rotation to the PIV data and creates a new transformed output
    OutputVTK_Rx.SetTransform(RotX)
    OutputVTK_Rx.SetInputData(InputVTK.GetOutput())
    OutputVTK_Rx.Update()
    
    RotY = vtk.vtkTransform()                                                                                       # Generates the VTK transformation function to rotate the PIV data about the y-axis
    RotY.RotateY(Orientation[1])
    RotY.Update()
    
    OutputVTK_Ry = vtk.vtkTransformFilter()                                                                         # Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_Ry.SetTransform(RotY)
    OutputVTK_Ry.SetInputData(OutputVTK_Rx.GetOutput())
    OutputVTK_Ry.Update()
    
    RotZ = vtk.vtkTransform()                                                                                       # Generates the VTK transformation function to rotate the PIV data about the z-axis
    RotZ.RotateZ(Orientation[2])
    RotZ.Update()
    
    OutputVTK_Rz = vtk.vtkTransformFilter()                                                                         # Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
    OutputVTK_Rz.SetTransform(RotZ)
    OutputVTK_Rz.SetInputData(OutputVTK_Ry.GetOutput())
    OutputVTK_Rz.Update()
    
    return OutputVTK_Rz

############################################
###########  SET Z=ZERO FUNCTION  ##########
############################################







# agregarrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr






############################################
######  SPATIAL RESOLUTION FUNCTION  #######
############################################

def GET_SPATIAL_RESOLUTION(InputVTK):
    from vtk.numpy_interface import dataset_adapter as dsa
    import numpy as np
    
    PIVdata = dsa.WrapDataObject(InputVTK.GetOutput())   														    # Converts the VTK element (the PIV plane) into a Python object
    POINTS = PIVdata.Points   																					    # Get points coordinates
    
    REFptX = POINTS[0,0]                                                                                            # Take the first X-point of the list as reference
    REFptY = POINTS[0,1]                                                                                            # Take the first Y-point of the list as reference
    
    for XXX in POINTS[:,0]:                                                                                         # Sweeps the points list until find a variation in the X-coordinate
        if XXX != REFptX:
            DeltaXCell = np.abs(REFptX-XXX)                                                                         # Get X-cell size
            break 
    
    for YYY in POINTS[:,1]:                                                                                         # Sweeps the points list until find a variation in the Y-coordinate
        if YYY != REFptY:
            DeltaYCell = np.abs(REFptY-YYY)                                                                         # Get Y-cell size
            break 
        
    return DeltaXCell,DeltaYCell


############################################
########  ADD FIELD DATA FUNCTION  #########
############################################

def ADD_FIELD_DATA(InputMesh,name,data):
    import vtk
    
    value = vtk.vtkFloatArray()
    value.SetName(name)
    value.SetNumberOfComponents(1)
    value.InsertNextValue(data)

    InputMesh.GetOutput().GetFieldData().AddArray(value)
    InputMesh.Update()
    
    return InputMesh


############################################
#########  GET PIV BOUNDS FUNCTION  ########
############################################

def GET_PIV_BOUNDS(InputPIVdata):                                                                            # Get the XY bounds coordinates of a given PIV plane
    import vtk
    from vtk.numpy_interface import dataset_adapter as dsa
    import numpy as np
    
    outline = vtk.vtkOutlineFilter()                                                                                # Gets the outline square that sourrounds the input data
    outline.SetInputConnection(InputPIVdata.GetOutputPort())
    outline.Update()

    data = dsa.WrapDataObject(outline.GetOutput())   															    # Converts the VTK element into a Python object
    POINTS = data.Points   																						    # Get points coordinates
    XminPIV = np.min(POINTS[:,0])                                                                                  # Get minimum x-coordinate of the outline square
    XmaxPIV = np.max(POINTS[:,0])
    YminPIV = np.min(POINTS[:,1])
    YmaxPIV = np.max(POINTS[:,1])

    return XminPIV,XmaxPIV,YminPIV,YmaxPIV
    

############################################
#########  CLIP BY PLANE FUNCTION  #########
############################################

def CLIP_BY_PLANE(DataToCut,CutPlane,CutPosition,InvertDirection):
    import vtk
    
    clipper = vtk.vtkClipPolyData()
    clipper.SetInputData(DataToCut.GetOutput())
    clipper.SetClipFunction(CutPlane)
    clipper.SetValue(CutPosition)
    clipper.GenerateClipScalarsOff()
    if InvertDirection==1:
        clipper.InsideOutOn()
    if InvertDirection==0:
        clipper.InsideOutOff()
    clipper.GetOutput().GetPointData().CopyScalarsOff()
    clipper.Update()
    
    return clipper

############################################
#########  CUT PIV EDGES FUNCTION  #########
############################################

def CUTAWAY_PIV_EDGES(PIV_in_XY_plane,Cut_Edges,Xcut_min,Xcut_max,Ycut_min,Ycut_max,CLIP_BY_PLANE):

    import vtk 
    
    CuttedPlane = PIV_in_XY_plane
    if Cut_Edges == 1:
        
        plane = vtk.vtkPlane()
        plane.SetNormal(1.0, 0.0, 0.0)
        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Xcut_min,0)
        CuttedPlane.Update()

        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Xcut_max,1)
        CuttedPlane.Update()     

        plane = vtk.vtkPlane()
        plane.SetNormal(0.0, 1.0, 0.0)
        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Ycut_min,0)
        CuttedPlane.Update()

        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Ycut_max,1)
        CuttedPlane.Update()          

    if Cut_Edges == 2:
        
        plane = vtk.vtkPlane()
        plane.SetNormal(1.0, 0.0, 0.0)
        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Xcut_min,0)
        CuttedPlane.Update()
        
    if Cut_Edges == 3:

        plane = vtk.vtkPlane()
        plane.SetNormal(1.0, 0.0, 0.0)
        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Xcut_max,1)
        CuttedPlane.Update()
        
    if Cut_Edges == 4:
    
        plane = vtk.vtkPlane()
        plane.SetNormal(0.0, 1.0, 0.0)
        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Ycut_min,0)
        CuttedPlane.Update()
    
    if Cut_Edges == 5:
        
        plane = vtk.vtkPlane()
        plane.SetNormal(0.0, 1.0, 0.0)
        CuttedPlane = CLIP_BY_PLANE(CuttedPlane,plane,Ycut_max,1)
        CuttedPlane.Update() 

        #self.GetOutputDataObject(0).ShallowCopy(CuttedPlane.GetOutput())
        
    return CuttedPlane



############################################
############  MERGE PIV WINDOWS  ###########
############################################

if Select_Action > 1:


	########################################
	# DETERMINES THE BOUNDS LIMITS OF THE COMBINED PIV ENSEMBLE IN THE XY PLANE:                                    # The bounds will define the extent of the baseline grid

    # COMBINES THE PIVvtkVector COMPONENTS INTO A SINGLE VTK VARIABLE:
    appendFilter = vtk.vtkAppendFilter()                                                                            # VTK container of multiple VTK elements. The output format of this filter is vtkUnstructuredGrid
    for ZZZ in range(0,len(PIVvtkVector)):                                                                          # Appends each element of the PIVvtkVector into the appendFilter variable
        PIVreversed = REVERSE_ROTATION(PIVvtkVector[ZZZ], PlanesOrientation)                                        # First transform back each PIV plane to the XY plane
        appendFilter.AddInputData(PIVreversed.GetOutput())                                                          # Then adds each transformed PIV to the VTK container
    appendFilter.Update()

    combined = vtk.vtkGeometryFilter()                                                                              # Converts vtkMultiblockDataSet to vtkPolyData
    combined.SetInputData(appendFilter.GetOutput())                             
    combined.Update()
    
    # GET THE BOUNDS OF THE COMBINED XY PIV PLANES:
    outline = vtk.vtkOutlineFilter()                                                                                # Gets the outline square that sourrounds the input data
    outline.SetInputConnection(combined.GetOutputPort())
    outline.Update()
    
    data = dsa.WrapDataObject(outline.GetOutput())   															    # Converts the VTK element into a Python object
    POINTS = data.Points   																						    # Get points coordinates
    Xout_min = np.min(POINTS[:,0])                                                                                  # Get minimum x-coordinate of the outline square
    Xout_max = np.max(POINTS[:,0])
    Yout_min = np.min(POINTS[:,1])
    Yout_max = np.max(POINTS[:,1])
    
    
	########################################
	# GENERATE THE BASELINE GRID:

    
    if SpatialResolution == 0:                                                                                      #0=Keep local smallest grid size
        
        DeltaXvector = []
        DeltaYvector = []
        XptsVector = []
        YptsVector = []
        BoundsVector = []
        
        for ZZZ in range(0,len(PIVvtkVector)):
        
            PIVreversed = REVERSE_ROTATION(PIVvtkVector[ZZZ], PlanesOrientation)                                    # Transforms each PIV plane back to the XY plane
            XminPIV,XmaxPIV,YminPIV,YmaxPIV = GET_PIV_BOUNDS(PIVreversed)                                           # Get the XY bound coordinates of the PIV plane    
            DeltaXCell,DeltaYCell = GET_SPATIAL_RESOLUTION(PIVreversed)                                             # Get the cell size of each PIV plane
            NBcellsX = int((XmaxPIV-XminPIV)/DeltaXCell)                                                            # Number of cells along the X-axis
            NBcellsY = int((YmaxPIV-YminPIV)/DeltaYCell)                                                            # Number of cells along the Y-axis
            Xpts = np.linspace(XminPIV,XmaxPIV,NBcellsX)                                                            # Distribution of points along the X-axis for each PIV plane
            Ypts = np.linspace(YminPIV,YmaxPIV,NBcellsY)
            DeltaXvector.append(DeltaXCell)
            DeltaYvector.append(DeltaYCell)
            XptsVector.append(Xpts)
            YptsVector.append(Ypts)
            BoundsVector.append([XminPIV,XmaxPIV,YminPIV,YmaxPIV])
        
        # Sort the PIV data from coarse to fine:
        SortedIDindicesX = np.argsort(DeltaXvector)                                                                 # Get the ID of the sorted elements
        SortedIDindicesX = SortedIDindicesX[::-1]                                                                   # Reverse the list to sort from max to min
        
        SortedIDindicesY = np.argsort(DeltaYvector)                                                                 # Get the ID of the sorted elements
        SortedIDindicesY = SortedIDindicesY[::-1]
        
        # Generate an initial coarse baseline grid:
        MeshXsize = np.max(DeltaXvector)                                                                            # Get the maximum X-cell size
        MeshYsize = np.max(DeltaYvector) 
        Points_i = int(np.round((Xout_max-Xout_min)/MeshXsize)+1)                                                            # The number of points along X-axis is giben by the bounds limitsof the outline squre and the desired mesh size
        Points_j = int(np.round((Yout_max-Yout_min)/MeshYsize)+1)        
        XXX = np.linspace(Xout_min, Xout_max, Points_i)
        YYY = np.linspace(Yout_min, Yout_max, Points_j)
        ZZZ = [0]   
        
        
        # Replace the refined PIV points in the coarse point distribution
        for KKK in SortedIDindicesX:
            
            # Replace the X-points inside each PIV bound:
            Xstart = BoundsVector[KKK][0]
            Xend = BoundsVector[KKK][1]
            IDstart = np.min(np.where((Xstart <= XXX) & (XXX <= Xend)))
            IDend = np.max(np.where((Xstart <= XXX) & (XXX <= Xend)))
            XXX = np.concatenate((XXX[0:IDstart], XptsVector[KKK], XXX[(IDend+1):]), axis=0)
                      
            # Replace the Y-points inside each PIV bound:
            Ystart = BoundsVector[KKK][2]
            Yend = BoundsVector[KKK][3]
            IDstart = np.min(np.where((Ystart <= YYY) & (YYY <= Yend)))
            IDend = np.max(np.where((Ystart <= YYY) & (YYY <= Yend)))
            YYY = np.concatenate((YYY[0:IDstart], YptsVector[KKK], YYY[(IDend+1):]), axis=0)           
                    
        XXX = np.sort(XXX)  # Reorder the points because at the intersection zone there is an overlapping of points (which could lead to an error in the mesh generation)
        YYY = np.sort(YYY)
        Points_i = len(XXX)
        Points_j = len(YYY)
        BaselineGrid = CREATE_GRID(XXX,YYY,ZZZ)
        print ("# Epsilon PIV: interpolation grid --> I =",Points_i,"   J =",Points_j )
        
        
    if SpatialResolution == 1:                                                                                      # 1=Keep overall smallest grid size
    
        # Determine smallest cell size of PIV data:
        DeltaXvector = []                                                                                             # Empty list to store the X-cell sizes of all the PIV planes
        DeltaYvector = []
        for ZZZ in range(0,len(PIVvtkVector)):                                                                      # This check every element of PIVvtkVector to get each cell size
            PIVreversed = REVERSE_ROTATION(PIVvtkVector[ZZZ], PlanesOrientation)                                    # First transform back each PIV plane to the XY plane
            DeltaXCell,DeltaYCell = GET_SPATIAL_RESOLUTION(PIVreversed)                                             # Get the cell size of each PIV plane
            DeltaXvector.append(DeltaXCell)
            DeltaYvector.append(DeltaYCell)
            
        MeshXsize = np.min(DeltaXvector)                                                                              # Get the minimum X-cell size
        MeshYsize = np.min(DeltaYvector)   
    
    
    if SpatialResolution == 2:                                                                                      # 2=Keep PIV1 grid size
        PIV1reversed = REVERSE_ROTATION(PIVvtkVector[0], PlanesOrientation)                                         # First transform back the PIV1 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV1reversed)                                                  # Get the cell size of the PIV1 plane
    
    
    if SpatialResolution == 3:                                                                                      # 3=Keep PIV2 grid size
        PIV2reversed = REVERSE_ROTATION(PIVvtkVector[1], PlanesOrientation)                                         # First transform back the PIV2 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV2reversed)                                                  # Get the cell size of the PIV2 plane


    if SpatialResolution == 4:                                                                                      # 4=Keep PIV3 grid size
        PIV3reversed = REVERSE_ROTATION(PIVvtkVector[2], PlanesOrientation)                                         # First transform back the PIV3 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV3reversed)                                                  # Get the cell size of the PIV3 plane


    if SpatialResolution == 5:                                                                                      # 5=Keep PIV4 grid size
        PIV4reversed = REVERSE_ROTATION(PIVvtkVector[3], PlanesOrientation)                                         # First transform back the PIV4 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV4reversed)                                                  # Get the cell size of the PIV4 plane    
    
    if SpatialResolution == 6:                                                                                      # 6=Keep PIV5 grid size
        PIV5reversed = REVERSE_ROTATION(PIVvtkVector[4], PlanesOrientation)                                         # First transform back the PIV5 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV5reversed)                                                  # Get the cell size of the PIV5 plane

    if SpatialResolution == 7:                                                                                      # 7=Keep PIV6 grid size
        PIV6reversed = REVERSE_ROTATION(PIVvtkVector[5], PlanesOrientation)                                         # First transform back the PIV6 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV6reversed)                                                  # Get the cell size of the PIV6 plane
        
    if SpatialResolution == 8:                                                                                      # 8=Keep PIV7 grid size
        PIV6reversed = REVERSE_ROTATION(PIVvtkVector[6], PlanesOrientation)                                         # First transform back the PIV7 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV7reversed)                                                  # Get the cell size of the PIV7 plane
 
    if SpatialResolution == 9:                                                                                      # 9=Keep PIV8 grid size
        PIV6reversed = REVERSE_ROTATION(PIVvtkVector[7], PlanesOrientation)                                         # First transform back the PIV8 data to the XY plane
        MeshXsize,MeshYsize = GET_SPATIAL_RESOLUTION(PIV8reversed)                                                  # Get the cell size of the PIV8 plane
 
    if SpatialResolution > 0:                                                                                       # This generates a uniform baseline grid for all cases excep the SpatialResolution==0
        # Generate the grid points:
        Points_i = int(np.round((Xout_max-Xout_min)/MeshXsize)+1)                                                            # The number of points along X-axis is giben by the bounds limitsof the outline squre and the desired mesh size
        Points_j = int(np.round((Yout_max-Yout_min)/MeshYsize)+1)        
        print ("# Epsilon PIV: interpolation grid --> I =",Points_i,"   J =",Points_j )
            
        XXX = np.linspace(Xout_min, Xout_max, Points_i)
        YYY = np.linspace(Yout_min, Yout_max, Points_j)
        ZZZ = [0]     
        BaselineGrid = CREATE_GRID(XXX,YYY,ZZZ)
    

	########################################
	# INTERPOLATE EACH PIV INTO A DISTINCT BASELINE GRID:

    PIVinterpolatedVector = []
    for ZZZ in range(0,len(PIVvtkVector)):                                                                          # Appends each element of the PIVvtkVector into the appendFilter variable
        
        PIVreversed = REVERSE_ROTATION(PIVvtkVector[ZZZ], PlanesOrientation)                                        # Transforms each PIV plane back to the XY plane
        XminPIV,XmaxPIV,YminPIV,YmaxPIV = GET_PIV_BOUNDS(PIVreversed)                                               # Get the XY bound coordinates of the PIV plane    
        DeltaXCell,DeltaYCell = GET_SPATIAL_RESOLUTION(PIVreversed)                                                 # Get the cell size of each PIV plane
        
        Xcut_min = XminPIV + NB_Layers_Vector[ZZZ]*DeltaXCell                                                       # Minimum X-coordinate to set the cut plane
        Xcut_max = XmaxPIV - NB_Layers_Vector[ZZZ]*DeltaXCell
        Ycut_min = YminPIV + NB_Layers_Vector[ZZZ]*DeltaYCell
        Ycut_max = YmaxPIV - NB_Layers_Vector[ZZZ]*DeltaYCell
                
        PIVreversedCUTTED = CUTAWAY_PIV_EDGES(PIVreversed,Cut_Edges_Vector[ZZZ],Xcut_min,Xcut_max,Ycut_min,Ycut_max,CLIP_BY_PLANE)

        PIVinterpolation = vtk.vtkProbeFilter()                                                                     # Interpolates values from the PIV data to the baseline grid
        PIVinterpolation.SetSourceData(PIVreversedCUTTED.GetOutput())
        PIVinterpolation.SetInputConnection(BaselineGrid.GetOutputPort())        
        PIVinterpolation.Update()
        
        CellArea = DeltaXCell*DeltaYCell
        PIVinterpolation = ADD_FIELD_DATA(PIVinterpolation,"CellArea",CellArea)
        PIVinterpolation.Update()
        
        PIVinterpolatedVector.append(PIVinterpolation)

    
	########################################
	# COMBINE THE INTERPOLATED GRIDS:

    Ucombined = np.zeros(Points_i*Points_j)                                                                     # Combined x-velocity component. Starts all in zero, then data is added by each PIV plane
    Vcombined = np.zeros(Points_i*Points_j)
    Wcombined = np.zeros(Points_i*Points_j)
    Acombined = 0.1*np.ones(Points_i*Points_j)                                                                 # Current minimal cell area during combination. Starts all at 1000 and it will retain the minimum local cell area during the combination       
    
    for ZZZ in range(0,len(PIVinterpolatedVector)):
        data = dsa.WrapDataObject(PIVinterpolatedVector[ZZZ].GetOutput()) 
        
        Velocity = data.PointData["Velocity"]
        CellArea = data.FieldData["CellArea"][0]
        #Status = data.PointData["Status"]
        
        if ZZZ == 0:
            ID = algs.where(mag(Velocity) > 0)                                                                     # Gives the ID of the mesh points where velocity is not zero (i.e., where there is PIV data)            
            Acombined[ID] = CellArea
            Ucombined[ID] = Velocity[ID,0][0]
            Vcombined[ID] = Velocity[ID,1][0]
            Wcombined[ID] = Velocity[ID,2][0]
            
        if ZZZ > 0:
        
            if OverlappingMethod == 0:                                                                                      # 0=Keep data of smallest cell size
                #IDtrimmed = np.where((CellArea < Acombined) & (mag(Velocity) > 0) & (Status!=8)) 
                IDtrimmed = np.where((CellArea < Acombined) & (mag(Velocity) > 0)) 
                Acombined[IDtrimmed] = CellArea 
                Ucombined[IDtrimmed] = Velocity[IDtrimmed,0][0]
                Vcombined[IDtrimmed] = Velocity[IDtrimmed,1][0]
                Wcombined[IDtrimmed] = Velocity[IDtrimmed,2][0]   
            
            if OverlappingMethod == 1:                                                                                      # 1=Averaged values       
                ZEROS = np.where(Ucombined[:] == 0)
                Ucombined[ZEROS]=Velocity[ZEROS,0]
                Vcombined[ZEROS]=Velocity[ZEROS,1]
                Wcombined[ZEROS]=Velocity[ZEROS,2]
                
                ZEROS = np.where(Velocity[:,0] == 0)
                Velocity[ZEROS,0]=Ucombined[ZEROS]
                Velocity[ZEROS,1]=Vcombined[ZEROS]
                Velocity[ZEROS,2]=Wcombined[ZEROS]
                
                Ucombined[:] = (Velocity[:,0] + Ucombined[:]*ZZZ)/(ZZZ+1)   								                # Update averaged Vx. "ZZZ+1"=number of total samples. 
                Vcombined[:] = (Velocity[:,1] + Vcombined[:]*ZZZ)/(ZZZ+1)
                Wcombined[:] = (Velocity[:,2] + Wcombined[:]*ZZZ)/(ZZZ+1)
            
            if OverlappingMethod == 2:                                                                                      # 2=Keep maximum values
                Ucombined[:] = np.maximum(Velocity[:,0],Ucombined[:])
                Vcombined[:] = np.maximum(Velocity[:,1],Vcombined[:])
                Wcombined[:] = np.maximum(Velocity[:,2],Wcombined[:])
            
            if OverlappingMethod == 3:                                                                                      # 3=Keep minimum values    
                ZEROS = np.where(Ucombined[:] == 0)
                Ucombined[ZEROS]=Velocity[ZEROS,0]
                Vcombined[ZEROS]=Velocity[ZEROS,1]
                Wcombined[ZEROS]=Velocity[ZEROS,2]
                
                ZEROS = np.where(Velocity[:,0] == 0)
                Velocity[ZEROS,0]=Ucombined[ZEROS]
                Velocity[ZEROS,1]=Vcombined[ZEROS]
                Velocity[ZEROS,2]=Wcombined[ZEROS]
                
                Ucombined[:] = np.minimum(Velocity[:,0],Ucombined[:])
                Vcombined[:] = np.minimum(Velocity[:,1],Vcombined[:])
                Wcombined[:] = np.minimum(Velocity[:,2],Wcombined[:]) 


    Velocity = algs.make_vector(Ucombined,Vcombined,Wcombined)
    AAAAAA = ADD_VECTOR(BaselineGrid,"Velocity",Velocity)
    AAAAAA.Update()
    
    #AAAAAA = ADD_SCALAR(BaselineGrid,"Acombined",Acombined)
    #AAAAAA.Update()


	########################################
	# ALLIGN THE COMBINED BASELINE GRID TO THE PIV PLANE IN THE AERO AXES:
    
    PIVok = ROTATE(AAAAAA, PlanesOrientation)
    
    
    # agregar translacion logitudinal a ROTATE y REVERSE_ROTATION para alinear el plano PIV con X=0, sino la interpolacion falla

    
    if Select_Action == 2:
        self.GetOutput().ShallowCopy(PIVok.GetOutput())
        



############################################################################################################
########################################### AVERAGING INPUT DATA ###########################################
############################################################################################################



#REVISAR TODO ESTE MODULO!!!!!

# if VortexWandering==1:   vortex wandering correction !!!!!!


# INCLUDE calculation of STD DEV for VORTICITY !!!!!!

if Select_Action == 3:





    ###################################
    ##########  INCREMENTAL  ##########
    ###################################


    if AveragingMethod == 1:                                                                                        # AveragingMethod=0: none,  AveragingMethod=1:incremental approach,  AveragingMethod=2:all files together
        

		
		
        Output_File_Path = os.path.dirname(File_Path_Vector[0])+"\Average.vtp" 											# The output file ("Average.vtp") wil be in the same folder as the file sequence
        Started = np.array([0])   																					# "Started" is a flag variable allowing knowing if the animation has been already started										
        
        for i in np.arange(len(SurveyPointsID)):   																	# If SurveyPointsID[i]=-1 --> It takes a point at the center of the image
            if SurveyPointsID[i] == -1:
                SurveyPointsID[i] = int(0.5*(Points_i*Points_j))  													# This asigns the ID of the point at the center of the image
        
        print ("# Epsilon Wind Tunnel: Frame =",Time)

        
		
		
        Vx = Velocity[:,0]
        Vy = Velocity[:,1]
        Vz = Velocity[:,2]		
		
		
        ###################################
        # READ EXISTING AVERAGED DATA FILE:
        if os.path.exists(Output_File_Path) == True: 																# Reads "Average.vtp" only if it exists (It is a vtkPolyData file format)
        
            # Opens the averaged filed:
            reader = vtk.vtkXMLPolyDataReader()   																	# Alternative reader: vtkPolyDataReader
            reader.SetFileName(Output_File_Path)
            reader.Update()      																					# "reader" is a vtk variable
            #output.ShallowCopy(reader.GetOutput())    																# It displays the mesh. Kept here for debugging purposes
            
            data = dsa.WrapDataObject(reader.GetOutputDataObject(0))     											# It converts a vtk variable into a numpy variable
            
            # Reads mean and standard deviations of velocity and vorticity fields (obtained by averaging the previous PIV files):
            Vx_mean = data.PointData["Vx"]     			                                                            # Mean X-velocity component															
            Vy_mean = data.PointData["Vy"]
            Vz_mean = data.PointData["Vz"]
            
            Vx_sdev = data.PointData["Vx_sdev"]                                                                     # standard deviation of X-velocity component  																		
            Vy_sdev = data.PointData["Vy_sdev"]
            Vz_sdev = data.PointData["Vz_sdev"]
            
            Wx_mean = data.PointData["Wx"]     						                                                # Mean X-vorticity component												
            Wy_mean = data.PointData["Wy"]
            Wz_mean = data.PointData["Wz"]
            
            Wx_sdev = data.PointData["Wx_sdev"]     				                                                # standard deviation of X-vorticity component  														
            Wy_sdev = data.PointData["Wy_sdev"]
            Wz_sdev = data.PointData["Wz_sdev"]
            
            TimeFlag = data.FieldData["TimeFlag"]    																# Timestep stored in the output file (indicates the last PIV data file of the total sequence that has been included in the averaging)
            Started = data.FieldData["Started"]     																# Indicates if the averaging process has started
            
            # Reads the current frame value and the mean/stddev of the averaged frames for velocity/vorticity at the survey points:
            
			
            MonitorData = 0      # BORRAR ESTA VARIABLE, la puse aca para poder ejecutar la parte de abajo pero MonitorData no deberia existir mas
            if MonitorData == 1:     																				# Monitors the data convergence in one or more survey points
            
                if SurveyPointsID[0] > -2:     					 													# Reads data from the first survey point
                    V_frame_1 = data.FieldData["V_frame_1"]   														# Array of instantaneous velocity at each frame
                    V_mean_1 = data.FieldData["V_mean_1"]     														# Array of mean velocity 
                    V_sdev_1 = data.FieldData["V_sdev_1"]   	                                                    # Array of velocity standard deviation
                    
                    W_frame_1 = data.FieldData["W_frame_1"]   														# Array of instantaneous vorticity at each frame
                    W_mean_1 = data.FieldData["W_mean_1"]     														# Array of mean vorticity
                    W_sdev_1 = data.FieldData["W_sdev_1"]                                                           # Array of vorticity standard deviation
                    
                if SurveyPointsID[1] > -2:
                    V_frame_2 = data.FieldData["V_frame_2"]
                    V_mean_2 = data.FieldData["V_mean_2"]
                    V_sdev_2 = data.FieldData["V_sdev_2"]   	                                                   
                    
                    W_frame_2 = data.FieldData["W_frame_2"]		
                    W_mean_2 = data.FieldData["W_mean_2"]
                    W_sdev_2 = data.FieldData["W_sdev_2"]                
                    
                if SurveyPointsID[2] > -2:
                    V_frame_3 = data.FieldData["V_frame_3"]
                    V_mean_3 = data.FieldData["V_mean_3"]
                    V_sdev_3 = data.FieldData["V_sdev_3"]
                    
                    W_frame_3 = data.FieldData["W_frame_3"]		
                    W_mean_3 = data.FieldData["W_mean_3"]		
                    W_sdev_3 = data.FieldData["W_sdev_3"]
               

        ############################################################################################
        # IF AVERAGED DATA FILE DO NOT EXIST, IT IS CREATED FROM THE FIRST PIV FILE OF THE SEQUENCE:		
        if Started[0] == 0:
			


			
            NewBaselineGrid = ROTATE(BaselineGrid, PlanesOrientation)
			
            AAAAAA = ADD_SCALAR(NewBaselineGrid,"Vx",Vx)
            AAAAAA.Update()	
			
            """
            AAAAAA = ADD_SCALAR(NewBaselineGrid,"Vy",Vy)
            AAAAAA.Update()	
            AAAAAA = ADD_SCALAR(NewBaselineGrid,"Vz",Vz)
            AAAAAA.Update()	
            AAAAAA = ADD_SCALAR(NewBaselineGrid,"Vx_sdev",0*Vx)
            AAAAAA.Update()	
            AAAAAA = ADD_SCALAR(NewBaselineGrid,"Vy_sdev",0*Vy)
            AAAAAA.Update()	
            AAAAAA = ADD_SCALAR(NewBaselineGrid,"Vz_sdev",0*Vz)
            AAAAAA.Update()			
			"""
            
			
			
            # Create initial velocity and stddev fields:
            #VTKarray = dsa.numpyTovtkDataArray(Vx, "Vx")     														# Converts the Numpy element into a vtk element
            #AAAAAA.GetOutput().GetPointData().AddArray(VTKarray) 													# Add the new vtk element into the "Tecplot" variable (vtk element) 
            
            VTKarray = dsa.numpyTovtkDataArray(Vy, "Vy") 
            AAAAAA.GetOutput().GetPointData().AddArray(VTKarray)
            
            VTKarray = dsa.numpyTovtkDataArray(Vz, "Vz") 
            AAAAAA.GetOutput().GetPointData().AddArray(VTKarray)
            
            VTKarray = dsa.numpyTovtkDataArray(0*Vx, "Vx_sdev")                                                     # For a single measurement, the STDDEV is zero by definition
            AAAAAA.GetOutput().GetPointData().AddArray(VTKarray)
            
            VTKarray = dsa.numpyTovtkDataArray(0*Vy, "Vy_sdev") 
            AAAAAA.GetOutput().GetPointData().AddArray(VTKarray)
            
            VTKarray = dsa.numpyTovtkDataArray(0*Vz, "Vz_sdev") 
            AAAAAA.GetOutput().GetPointData().AddArray(VTKarray)
            
            
            TimeFlag = np.array([Time])
            VTKarray = dsa.numpyTovtkDataArray(TimeFlag , "TimeFlag") 
            AAAAAA.GetOutput().GetFieldData().AddArray(VTKarray)
            
            Started = np.array([1])
            VTKarray = dsa.numpyTovtkDataArray(Started , "Started")
            AAAAAA.GetOutput().GetFieldData().AddArray(VTKarray)
                        
			
			
			
            MonitorData = 0      # BORRAR ESTA VARIABLE, la puse aca para poder ejecutar la parte de abajo pero MonitorData no deberia existir mas
            if MonitorData == 1:
                
                Vector_Speed = algs.make_vector(Vx,Vy,Vz)
                Vorticity = algs.vorticity(Vector_Speed)
                
                if SurveyPointsID[0] > -2:
                    
                    VTKarray = dsa.numpyTovtkDataArray(np.array([SurveyPointsID[0]]) , "ID_Point_1") 
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vector_Speed[SurveyPointsID[0]], "V_frame_1")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)	
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vector_Speed[SurveyPointsID[0]], "V_mean_1")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                    VTKarray = dsa.numpyTovtkDataArray(np.array([0]), "V_sdev_1") 
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vorticity[SurveyPointsID[0]], "W_frame_1")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)	
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vorticity[SurveyPointsID[0]], "W_mean_1") 
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                if SurveyPointsID[1] > -2:
                
                    VTKarray = dsa.numpyTovtkDataArray(np.array([SurveyPointsID[1]]) , "ID_Point_2") 
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                
                    VTKarray = dsa.numpyTovtkDataArray(Vector_Speed[SurveyPointsID[1]], "V_frame_2")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)	
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vector_Speed[SurveyPointsID[1]], "V_mean_2")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vorticity[SurveyPointsID[1]], "W_frame_2")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)	
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vorticity[SurveyPointsID[1]], "W_mean_2") 
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                if SurveyPointsID[2] > -2:
                
                    VTKarray = dsa.numpyTovtkDataArray(np.array([SurveyPointsID[2]]) , "ID_Point_3") 
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vector_Speed[SurveyPointsID[2]], "V_frame_3")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)	
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vector_Speed[SurveyPointsID[2]], "V_mean_3")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vorticity[SurveyPointsID[2]], "W_frame_3")
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)	
                    
                    VTKarray = dsa.numpyTovtkDataArray(Vorticity[SurveyPointsID[2]], "W_mean_3") 
                    VTKarray.SetNumberOfComponents(3)
                    combined.GetOutput().GetFieldData().AddArray(VTKarray)
                    
                    
            writer = vtk.vtkXMLPolyDataWriter()  																	# Writes a VTP output file (Alternative command: vtkPolyDataWriter)
            writer.SetFileName(Output_File_Path)
            writer.SetInputData(AAAAAA.GetOutput())         														# Alternative command: SetInput     
            writer.Write()

            Started = np.array([1])    																				# Updates the flag values for the next part of the code
            TimeFlag = np.array([Time])



            #Generalizar el "writer" para que el SetInputData data sea un vtp o csv y NO unicamente "Tecplot.GetOutput()". Idem en la parte de numpyTovtkDataArray --> change Tecplot.GetOutput()
        


        



            
        ##############################################################
        # APPENDS THE NEW TIMESTEP DATA TO THE EXISTING AVERAGED DATA:
        if Started[0] == 1:                                                          								# Executes this part of the code only if the averaging process has started
            if Time > TimeFlag[-1]:   																				# Appends data to the output file only if the animation time is not included in the output file
                
                ###################################
                # GET THE CURRENT FRAME PARAMETERS:
                if MonitorData == 1:   																				# If survey points are active, it appends current frame data to their output arrays
                    Vector_Speed = algs.make_vector(Vx,Vy,Vz)    													# Instantaneous velocity vector
                    Vorticity = algs.vorticity(Vector_Speed)    													# Instantaneous vorticity vector
                    
                    if SurveyPointsID[0] > -2:
                        V_frame_1 = np.vstack((V_frame_1,Vector_Speed[SurveyPointsID[0]]))   						# Evaluates the current velocity at the first survey point and appends it to the history array
                        W_frame_1 = np.vstack((W_frame_1,Vorticity[SurveyPointsID[0]]))		    					# Evaluates the current vorticity at the first survey point and appends it to the history array			
                        
                    if SurveyPointsID[1] > -2:
                        V_frame_2 = np.vstack((V_frame_2,Vector_Speed[SurveyPointsID[1]]))
                        W_frame_2 = np.vstack((W_frame_2,Vorticity[SurveyPointsID[1]]))	

                    if SurveyPointsID[2] > -2:
                        V_frame_3 = np.vstack((V_frame_3,Vector_Speed[SurveyPointsID[2]]))
                        W_frame_3 = np.vstack((W_frame_3,Vorticity[SurveyPointsID[2]]))				
                    
                TimeFlag = np.append(TimeFlag, Time)     															# Add current Time value to the TimeFlag history array
                
                
                ##########################################
                # COMPUTE THE MEAN AND STANDARD DEVIATION:
                
                # Update mean velocity values:
                Vx_mean_new = (Vx + Vx_mean*Time)/(Time+1)   													    # Update averaged Vx. "Time+1"=number of total samples. 
                Vy_mean_new = (Vy + Vy_mean*Time)/(Time+1)
                Vz_mean_new = (Vz + Vz_mean*Time)/(Time+1)
                

                # Update standard deviation values for velocity components:
                Vx_sdev_new = np.sqrt(((Time-1)*Vx_sdev**2+(Vx-Vx_mean_new)*(Vx-Vx_mean))/(Time))		
                Vy_sdev_new = np.sqrt(((Time-1)*Vy_sdev**2+(Vy-Vy_mean_new)*(Vy-Vy_mean))/(Time))
                Vz_sdev_new = np.sqrt(((Time-1)*Vz_sdev**2+(Vz-Vz_mean_new)*(Vz-Vz_mean))/(Time))			
                
                # Change the name of the averaged variables:
                Vx = Vx_mean_new  																                    # Updated averaged Vx. Output array has been named "Vx" instead of "Vx_mean_new" for simplicity (it is compatible with the code blocks below) 
                Vy = Vy_mean_new
                Vz = Vz_mean_new
                Vx_sdev = Vx_sdev_new
                Vy_sdev = Vy_sdev_new
                Vz_sdev = Vz_sdev_new			
                
                # Get the sdev values at the survey points:
                V_sdev_1 = np.append(V_sdev_1, Vx_sdev_new[SurveyPointsID[0]])    # ---> V_sdev_1_X !!!!  --> armar vector con componentes xyz
                
                
                
                
                
                
                
                

                #Calculate Sdev !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!






                # Evaluates the averaged velocity and vorticity vectors ath the survey points:
                if MonitorData == 1:   																				# If survey points are active, it appends the updated averaged data to their output arrays
                    Vector_Speed = algs.make_vector(Vx,Vy,Vz)    													# Averaged velocity vector
                    Vorticity = algs.vorticity(Vector_Speed)   														# Averaged vorticity vector
                    
                    if SurveyPointsID[0] > -2:
                        V_mean_1 = np.vstack((V_mean_1,Vector_Speed[SurveyPointsID[0]]))   							# Evaluates the current averaged velocity at the first survey point and appends it to the history array
                        W_mean_1 = np.vstack((W_mean_1,Vorticity[SurveyPointsID[0]]))		    					# Evaluates the current averaged vorticity at the first survey point and appends it to the history array			
                        
                    if SurveyPointsID[1] > -2:
                        V_mean_2 = np.vstack((V_mean_2,Vector_Speed[SurveyPointsID[1]]))
                        W_mean_2 = np.vstack((W_mean_2,Vorticity[SurveyPointsID[1]]))	

                    if SurveyPointsID[2] > -2:
                        V_mean_3 = np.vstack((V_mean_3,Vector_Speed[SurveyPointsID[2]]))
                        W_mean_3 = np.vstack((W_mean_3,Vorticity[SurveyPointsID[2]]))	
                
                
                #####################################
                # SEND CALCULATED DATA TO THE OUTPUT:
                
                # Send the averaged velocity arrays to the output:
                VTKarray = dsa.numpyTovtkDataArray(Vx , "Vx")     													# Converts array from numpy to vtk format
                reader.GetOutput().GetPointData().AddArray(VTKarray)												# Appends array to the "reader" vtk variable
                
                VTKarray = dsa.numpyTovtkDataArray(Vy , "Vy") 
                reader.GetOutput().GetPointData().AddArray(VTKarray)
                
                VTKarray = dsa.numpyTovtkDataArray(Vz , "Vz") 
                reader.GetOutput().GetPointData().AddArray(VTKarray)
                
                
                # Send the sdev velocity arrays to the output:
                VTKarray = dsa.numpyTovtkDataArray(Vx_sdev , "Vx_sdev")     												
                reader.GetOutput().GetPointData().AddArray(VTKarray)												
                
                VTKarray = dsa.numpyTovtkDataArray(Vy_sdev , "Vy_sdev") 
                reader.GetOutput().GetPointData().AddArray(VTKarray)
                
                VTKarray = dsa.numpyTovtkDataArray(Vz_sdev , "Vz_sdev") 
                reader.GetOutput().GetPointData().AddArray(VTKarray)
                
                
                # Update the current TimeFlag value in the output file:
                VTKarray = dsa.numpyTovtkDataArray(TimeFlag , "TimeFlag")
                reader.GetOutput().GetFieldData().AddArray(VTKarray)    


                # Send the survey point data to the output: current frame values, mean values and STDEV values (for both, velocity and vorticity)
                if MonitorData == 1:
                    
                    if SurveyPointsID[0] > -2:
                    
                        VTKarray = dsa.numpyTovtkDataArray(V_frame_1, "V_frame_1")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)	
                        
                        VTKarray = dsa.numpyTovtkDataArray(V_mean_1, "V_mean_1")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)

                        VTKarray = dsa.numpyTovtkDataArray(V_sdev_1, "V_sdev_1")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)	
                        
                        
                        
                        
                        VTKarray = dsa.numpyTovtkDataArray(W_frame_1, "W_frame_1")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)	
                        
                        VTKarray = dsa.numpyTovtkDataArray(W_mean_1, "W_mean_1") 
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)
                        
                    if SurveyPointsID[1] > -2:
                    
                        VTKarray = dsa.numpyTovtkDataArray(V_frame_2, "V_frame_2")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)	
                        
                        VTKarray = dsa.numpyTovtkDataArray(V_mean_2, "V_mean_2")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)
                        
                        VTKarray = dsa.numpyTovtkDataArray(W_frame_2, "W_frame_2")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)	
                        
                        VTKarray = dsa.numpyTovtkDataArray(W_mean_2, "W_mean_2") 
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)
                        
                    if SurveyPointsID[2] > -2:
                        
                        VTKarray = dsa.numpyTovtkDataArray(V_frame_3, "V_frame_3")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)	
                        
                        VTKarray = dsa.numpyTovtkDataArray(V_mean_3, "V_mean_3")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)
                        
                        VTKarray = dsa.numpyTovtkDataArray(W_frame_3, "W_frame_3")
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)	
                        
                        VTKarray = dsa.numpyTovtkDataArray(W_mean_3, "W_mean_3") 
                        reader.GetOutput().GetFieldData().AddArray(VTKarray)
                
                
                writer = vtk.vtkXMLPolyDataWriter()  	 															# Writes a VTP output file (Alternative command: vtkXMLPolyDataWriter) 
                writer.SetFileName(Output_File_Path)
                writer.SetInputData(reader.GetOutput())    
                writer.Write()
                
                






                #Generalizar el "writer" para que el input data sea un vtp o csv y NO unicamente "Tecplot.GetOutput()"








        
        # IT OPENS AGAIN THE OUTPUT FILE WITH FRESH AVERAGED DATA: 
        readerNEW = vtk.vtkXMLPolyDataReader()   																	# Alternative: vtkPolyDataReader
        readerNEW.SetFileName(Output_File_Path)
        readerNEW.Update()
        self.GetOutput().ShallowCopy(readerNEW.GetOutput())
        

        data = dsa.WrapDataObject(readerNEW.GetOutputDataObject(0))     											# Converts the vtk element into a numpy object
        Vx = data.PointData["Vx"]
        Vy = data.PointData["Vy"]
        Vz = data.PointData["Vz"]




    ################################### 
    ##########  ALL FILES TOGETHER  ##########     yet to be coded !
    ###################################

    """
    if AveragingMethod == 2:
    """










############################################################################################################
########################################## CLEANUP VORTICITY FIELD #########################################
############################################################################################################


if Select_Action > 3:

	PLANE = PIVok

	########################################
	#####  COMPUTING AERO PARAMETERS  ######
	########################################

	#WRAP THE VTK ELEMENTS:
	if AveragingMethod == 0:
		data = dsa.WrapDataObject(PLANE.GetOutput()) 
		Velocity = data.PointData["Velocity"]
		Vx = Velocity[:,0]
		Vy = Velocity[:,1]
		Vz = Velocity[:,2]
		
		
	if AveragingMethod != 0:    #REVISAR ESTOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
		#Vx = Vx.Arrays[0]    # It takes the first block only (For PIV data, it is the only block present in the Tecplot file)
		#Vy = Vy.Arrays[0]
		#Vz = Vz.Arrays[0]	
		Vx = data.PointData["Vx"]    # .Arrays[0]
		Vy = data.PointData["Vy"]
		Vz = data.PointData["Vz"]
		#POINTS = POINTS
	

	# GRADIENT AND VORTICITY:
	POINTS = data.Points   												    # Get points coordinates
	X = POINTS[:,0] 
	Y = POINTS[:,1] 
	Z = POINTS[:,2]
	Vector_Speed = algs.make_vector(Vx,Vy,Vz)
	Vorticity = algs.vorticity(Vector_Speed) 
	VORTICITY_magnitude = algs.mag(Vorticity)
	Vx_GRADIENT = algs.gradient(Vx)
	Vy_GRADIENT = algs.gradient(Vy)
	Vz_GRADIENT = algs.gradient(Vz)
	V_gradient = algs.gradient(Vector_Speed)
	Phi_eff = (2*V_gradient[:,0,0]**2 + 2*V_gradient[:,1,1]**2 + 2*V_gradient[:,2,2]**2 + (V_gradient[:,0,1]+V_gradient[:,1,0])**2 + (V_gradient[:,0,2]+V_gradient[:,2,0])**2 + (V_gradient[:,2,1]+V_gradient[:,1,2])**2 - (2.0/3.0)*(V_gradient[:,0,0]+V_gradient[:,1,1]+V_gradient[:,2,2])**2)  # Effective dissipation         
	STRAIN = strain(Vector_Speed)	
	Q_criterion = 0.5*(algs.mag(Vorticity)**2-np.sqrt(np.einsum('ijk,ijk->i',STRAIN,STRAIN))**2)

		
        
        
        
        
	"""
   
	########################################
	#########  Y- and Z-VORTICITY  #########
	########################################	

    # The following module was tested based on the Wu suggestion for the calculation of pressure from velocity gradients. 
    # It still requires a lot of testing and development. Kept here for improving later on.

	
	if Survey_Plane==0:  # DEPRECATED VARIABLE!!!!!: 0=YZ plane	1=XZ plane	2=XY plane
		
		# COMPUTES THE REMAINING VELOCITY DERIVATIVES ALONG X-AXIS (not available in a PIV stereo measurement):
		if data.GetPointData().HasArray("AERO_Ps")==0:
			Unknown_y = 0*Vy_GRADIENT[:,0]
			Unknown_z = 0*Vy_GRADIENT[:,0]
		else:
			Ps = data.PointData["AERO_Ps"]
			Rho = data.PointData["AERO_Rho"]
			Ps_grad = algs.gradient(Ps)
			Unknown_y = 1*Ps_grad[:,1]/Rho
			Unknown_z = 1*Ps_grad[:,2]/Rho


		#Unknown_y = 0*Vy_GRADIENT[:,0]
		#Unknown_z = 0*Vy_GRADIENT[:,0]
			
		dVy_dx = -(Unknown_y + Vector_Speed[:,1]*Vy_GRADIENT[:,1] + Vector_Speed[:,2]*Vy_GRADIENT[:,2])/Vx    # From Euler equation: dVy_dx = -((dPs_dy/Rho) + Vy*dVy_dy + Vz*dVy_dz)/Vx
		dVz_dx = -(Unknown_z + Vector_Speed[:,1]*Vz_GRADIENT[:,1] + Vector_Speed[:,2]*Vz_GRADIENT[:,2])/Vx    # From Euler equation: dVz_dx = -((dPs_dz/Rho) + Vy*dVz_dy + Vz*dVz_dz)/Vx

		
		where_are_NaNs = isnan(dVy_dx)    # The previous operation gives NaN values in the masked regions, which avoids its integration on the survey plane. This eliminates this problem.
		dVy_dx[where_are_NaNs] = 0

		where_are_NaNs = isnan(dVz_dx)    
		dVz_dx[where_are_NaNs] = 0		
		
		
		# COMPUTES THE Y- AND Z-VORTICITY:
		Vorticity[:,1] = Vx_GRADIENT[:,2] #+ dVz_dx
		Vorticity[:,2] = dVy_dx - Vx_GRADIENT[:,1]

		#Vorticity[:,1] = Vx_GRADIENT[:,2] + dVz_dx
		#Vorticity[:,2] = - dVy_dx - Vx_GRADIENT[:,1]	
		
		OMEGA = data.PointData["AERO_Vorticity"]
		#Vorticity[:,1] = OMEGA[:,1]
		#Vorticity[:,2] = OMEGA[:,2]


		data.PointData.append(dVz_dx , 'dVz_dx')

	"""


	########################################
	#########  DATA CONDITIONING  ##########
	########################################

	Backup_VORTICITY_magnitude = VORTICITY_magnitude # Backup array used by "Delete Noise Spots" 
	
	########################################
	# MASKING DATA AT THE BOUNDARIES OF THE MEASUREMENT REGION:
	if MaskBoundaries==1 or MaskBoundaries==2:
	
		MASK = algs.sqrt(Vx**2+Vy**2+Vz**2) != 0   # This mask the regions where there is no velocity data (It affects the vorticity and gradient arrays)

		# Iterative boundary detection: it is a two step procedure
		if MaskBoundaries==1:
			PhiThreshold = ThresholdValuePhi*algs.max(Phi_eff)
		if MaskBoundaries==2:
			PhiThreshold = ThresholdValuePhi
		
		MASK_1 = Phi_eff < PhiThreshold    		# First detection of the boundaries by Phi_eff criterion (This marks the cells at the measurement boundary)
		
		Vector_Speed_New = 1*MASK_1*Vector_Speed
		VORTICITY_New = algs.vorticity(Vector_Speed_New) 		
		V_gradient_New = algs.gradient(Vector_Speed_New)
		Phi_eff_New = (2*V_gradient_New[:,0,0]**2 + 2*V_gradient_New[:,1,1]**2 + 2*V_gradient_New[:,2,2]**2 + (V_gradient_New[:,0,1]+V_gradient_New[:,1,0])**2 + (V_gradient_New[:,0,2]+V_gradient_New[:,2,0])**2 + (V_gradient_New[:,2,1]+V_gradient_New[:,1,2])**2 - (2.0/3.0)*(V_gradient_New[:,0,0]+V_gradient_New[:,1,1]+V_gradient_New[:,2,2])**2)  # Effective dissipation         
		
		MASK_2 = Phi_eff_New < PhiThreshold    # Second detection of the boundaries by Phi_eff criterion (This marks the internal neighbouring cells of the measurement boundary)
		
		Vx = 1*MASK_1*MASK_2*Vx
		Vy = 1*MASK_1*MASK_2*Vy
		Vz = 1*MASK_1*MASK_2*Vz
		Vector_Speed = 1*MASK_1*MASK_2*Vector_Speed
		Vorticity = 1*MASK*MASK_1*MASK_2*Vorticity
		VORTICITY_magnitude = 1*MASK*MASK_1*MASK_2*VORTICITY_magnitude
		Vx_GRADIENT = 1*MASK*MASK_1*MASK_2*Vx_GRADIENT
		Vy_GRADIENT = 1*MASK*MASK_1*MASK_2*Vy_GRADIENT
		Vz_GRADIENT = 1*MASK*MASK_1*MASK_2*Vz_GRADIENT

	########################################
	# MASKING ZONES WITH WEAK VORTICITY VALUES (Measurement noise):
	if MaskVorticityNoise==1 or MaskVorticityNoise==2:
	
		if MaskVorticityNoise==1:
			VorticityThreshold = ThresholdValueVorticity*algs.max(VORTICITY_magnitude)
		if MaskVorticityNoise==2:
			VorticityThreshold = ThresholdValueVorticity
			
		MASK_3 = VORTICITY_magnitude > VorticityThreshold
		
		Vorticity = 1*MASK_3*Vorticity
		VORTICITY_magnitude = 1*MASK_3*VORTICITY_magnitude
		Vx_GRADIENT = 1*MASK_3*Vx_GRADIENT
		Vy_GRADIENT = 1*MASK_3*Vy_GRADIENT
		Vz_GRADIENT = 1*MASK_3*Vz_GRADIENT
		

	########################################
	# MASKING HARD SPOTS OF VORTICITY NOISE:
	
	# This is the first iteration (first level of masking):
	if DeleteNoiseSpots == 1 or DeleteNoiseSpots == 2 or DeleteNoiseSpots == 3:

		# Sweep along Y-axis:
		for j in range(Points_j):   
			for i in range(Points_i-1):
				# Delete spots of vorticity noise:
				if i!=0 and i!=Points_i and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i-1+j*Points_i]==0 and VORTICITY_magnitude[i+1+j*Points_i]==0:
					VORTICITY_magnitude[i+j*Points_i]=0
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0
				# Restore vorticity in holes:
				if i!=0 and i!=Points_i and VORTICITY_magnitude[i+j*Points_i]==0 and VORTICITY_magnitude[i-1+j*Points_i]!=0 and VORTICITY_magnitude[i+1+j*Points_i]!=0:
					VORTICITY_magnitude[i+j*Points_i] = Backup_VORTICITY_magnitude[i+j*Points_i]	
		
		# Sweep along Z-axis:
		for i in range(Points_i):   
			for j in range(Points_j-1):				
				# Delete spots of vorticity noise:
				if j!=0 and j<Points_j-2 and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i+(j-1)*Points_i]==0 and VORTICITY_magnitude[i+(j+1)*Points_i]==0:
					VORTICITY_magnitude[i+j*Points_i]=0	
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0
				# Restore vorticity in holes:
				if j!=0 and j<Points_j-2 and VORTICITY_magnitude[i+j*Points_i]==0 and VORTICITY_magnitude[i+(j-1)*Points_i]!=0 and VORTICITY_magnitude[i+(j+1)*Points_i]!=0:
					VORTICITY_magnitude[i+j*Points_i] = Backup_VORTICITY_magnitude[i+j*Points_i]

	
	# This is the second iteration (second level of masking):
	if DeleteNoiseSpots == 2 or DeleteNoiseSpots == 3:

		# Sweep along Y-axis:
		for j in range(Points_j):   
			for i in range(Points_i-2):
				# Delete spots of vorticity noise:
				if i>1 and i<Points_i-1 and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i-2+j*Points_i]==0 and VORTICITY_magnitude[i+2+j*Points_i]==0 :
					VORTICITY_magnitude[i+j*Points_i]=0
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0
				# Restore vorticity in holes:
				if i>1 and i<Points_i-2 and VORTICITY_magnitude[i+j*Points_i]==0 and VORTICITY_magnitude[i-2+j*Points_i]!=0 and VORTICITY_magnitude[i+2+j*Points_i]!=0 :
					VORTICITY_magnitude[i+j*Points_i] = Backup_VORTICITY_magnitude[i+j*Points_i]

		# Sweep along Z-axis:
		for i in range(Points_i):   
			for j in range(Points_j-2):
				# Delete spots of vorticity noise:
				if j>1 and j<Points_j-1 and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i+(j-2)*Points_i]==0 and VORTICITY_magnitude[i+(j+2)*Points_i]==0 :
					VORTICITY_magnitude[i+j*Points_i]=0	
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0
				# Restore vorticity in holes:
				if j>1 and j<Points_j-2 and VORTICITY_magnitude[i+j*Points_i]==0 and VORTICITY_magnitude[i+(j-2)*Points_i]!=0 and VORTICITY_magnitude[i+(j+2)*Points_i]!=0 :
					VORTICITY_magnitude[i+j*Points_i] = Backup_VORTICITY_magnitude[i+j*Points_i]
	
	
	# This is the third iteration (third level of masking):
	if DeleteNoiseSpots == 3:
		# Sweep along Y-axis:
		for j in range(Points_j):   
			for i in range(Points_i-3):
				# Delete spots of vorticity noise:
				if i!=0 and i!=Points_i and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i-1+j*Points_i]==0 and VORTICITY_magnitude[i+1+j*Points_i]==0:
					VORTICITY_magnitude[i+j*Points_i]=0
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0
				if i>1 and i<Points_i-3 and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i-3+j*Points_i]==0 and VORTICITY_magnitude[i+3+j*Points_i]==0 :
					VORTICITY_magnitude[i+j*Points_i]=0
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0
	
		# Sweep along Z-axis:
		for i in range(Points_i):   
			for j in range(Points_j-3):
				# Delete spots of vorticity noise:
				if j!=0 and j!=Points_j and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i+(j-1)*Points_i]==0 and VORTICITY_magnitude[i+(j+1)*Points_i]==0:
					VORTICITY_magnitude[i+j*Points_i]=0
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0					
				if j>1 and j<Points_j-3 and VORTICITY_magnitude[i+j*Points_i]!=0 and VORTICITY_magnitude[i+(j-3)*Points_i]==0 and VORTICITY_magnitude[i+(j+3)*Points_i]==0 :
					VORTICITY_magnitude[i+j*Points_i]=0	
					Vorticity[i+j*Points_i,:]=0
					Vx_GRADIENT[i+j*Points_i,:]=0
					Vy_GRADIENT[i+j*Points_i,:]=0
					Vz_GRADIENT[i+j*Points_i,:]=0


	
	

	
############################################################################################################
############################################# PRESSURE FROM PIV ############################################
############################################################################################################


"""
Ver si corresponde poner Select_Action > 4 o Select_Action==5 en funcion de Temperature from PIV
"""


if Select_Action > 4:

	
	Survey_Plane=0   # DEPRECATED??? ---> Sino ver como incluirlo dentro del GUI actual    # 0=YZ plane	1=XZ plane	2=XY plane
	

	########################################
	#########  VORTICITY INTEGRAL  #########
	########################################

	#######################################################
	# INTEGRAL OF Z-VORTICITY ALONG "I" AXIS (Y-mesh axis):
	
	if Survey_Plane == 1 and Switch_IJ_axes == 1:   	# If survey plane is XZ, it uses this code structure to integrate the Y-VORTICITY along the mesh Z-axis (coincident with the J axis)
		VortComp = 1   									# Vorticity component
		Coord = 2           							# XYZ point component
		IntegrationY = IntegrationZ  					# Because we are using the Y-integration tool to integrate along the Z-axis (which has been selected by the user in the GUI)
		Sign = -1   									# This provides the good sign of the integrated vorticity for this case
	else:
		VortComp = 2
		Coord = 1	
		Sign = 1
	
	# Integration along the positive Y-axis direction:
	if IntegrationY==0 or IntegrationY==2 or IntegrationY==3 or IntegrationY==5:
		Vxbar_y_p = 0*POINTS[:,0]     					# An array of zeros is created, "p" stands for positive
		for j in range(Points_j):   
			for i in range(Points_i):
				if i==0:
					Vxbar_y_p[i+j*Points_i] = 0
				else:
					Vxbar_y_p[i+j*Points_i] = Vxbar_y_p[i-1+j*Points_i] - Sign*0.5*(Vorticity[i+j*Points_i,VortComp]+Vorticity[i-1+j*Points_i,VortComp])*(POINTS[i+j*Points_i,Coord]-POINTS[i-1+j*Points_i,Coord])
		

	# Integration along the negative Y-axis direction:	
	if IntegrationY==1 or IntegrationY==2 or IntegrationY==4 or IntegrationY==5:
		Vxbar_y_n = 0*POINTS[:,0]     					# An array of zeros is created, "n" stands for negative
		for j in range(Points_j):   
			for i in reversed(range(Points_i)):
				if i==Points_i-1:
					Vxbar_y_n[i+j*Points_i] = 0
				else:
					Vxbar_y_n[i+j*Points_i] = Vxbar_y_n[i+1+j*Points_i] + Sign*0.5*(Vorticity[i+j*Points_i,VortComp]+Vorticity[i-1+j*Points_i,VortComp])*(POINTS[i+j*Points_i,Coord]-POINTS[i-1+j*Points_i,Coord])


	# Both directions interpolated:
	if IntegrationY==0:
		Vx_bar_y = Vxbar_y_p
		
	if IntegrationY==1:
		Vx_bar_y = Vxbar_y_n
		
	if IntegrationY==2:
		Vx_bar_y = 0.5*(Vxbar_y_p + Vxbar_y_n)
		zero_p = algs.where(Vxbar_y_p == 0)  
		zero_n = algs.where(Vxbar_y_n == 0)			
		Vx_bar_y[zero_p] = 0
		Vx_bar_y[zero_n] = 0
		
	# Integration along positive Y-direction (Delete Linear Error):
	if IntegrationY==3 or IntegrationY==5:
		for j in range(Points_j):   
			Error_initial = Vxbar_y_p[0+j*Points_i]
			Error_final = Vxbar_y_p[Points_i-1+j*Points_i]
			Y_initial = 0
			Y_final = Points_i-1
			if Error_initial!= Error_final:
				while Vxbar_y_p[Y_initial+j*Points_i] == Error_initial:
					Y_initial = Y_initial+1
					
				while Vxbar_y_p[Y_final+j*Points_i] == Error_final:
					Y_final = Y_final-1
			Y_initial = Y_initial-3
			Y_final = Y_final +3
			Error_law = (Error_final-Error_initial)/(Y_final-Y_initial)
			for i in range(Points_i):
				if Y_initial<=i<=Y_final:
					Vxbar_y_p[i+j*Points_i] = Vxbar_y_p[i+j*Points_i] - Error_law*(i-Y_initial)
				if i>Y_final:
					Vxbar_y_p[i+j*Points_i] = 0
		Vx_bar_y = Vxbar_y_p

	# Integration along negative Y-direction (Delete Linear Error):
	if IntegrationY==4 or IntegrationY==5:
		for j in range(Points_j):   
			Error_initial = Vxbar_y_n[Points_i-1+j*Points_i]
			Error_final = Vxbar_y_n[0+j*Points_i]
			Y_initial = Points_i-1
			Y_final = 0
			if Error_initial!= Error_final:
				while Vxbar_y_n[Y_initial+j*Points_i] == Error_initial:
					Y_initial = Y_initial-1
					
				while Vxbar_y_n[Y_final+j*Points_i] == Error_final:
					Y_final = Y_final+1
			Y_initial = Y_initial+3
			Y_final = Y_final -3
			Error_law = (Error_final-Error_initial)/(Y_final-Y_initial)
			for i in range(Points_i):
				if Y_final<=i<=Y_initial:
					Vxbar_y_n[i+j*Points_i] = Vxbar_y_n[i+j*Points_i] - Error_law*(i-Y_initial)
				if i<Y_final:
					Vxbar_y_n[i+j*Points_i] = 0
		Vx_bar_y = Vxbar_y_n

	# Both directions interpolated (Delete Linear Error):
	if IntegrationY==5:
		Vx_bar_y = 0.5*(Vxbar_y_p + Vxbar_y_n)


	#######################################################
	# INTEGRAL OF Y-VORTICITY ALONG "J" AXIS (Z-mesh axis):
	
	if Survey_Plane == 2:   			# If survey plane is XY, it uses this code structure to integrate the Z-VORTICITY along the mesh Y-axis (coincident with the J axis)
		VortComp = 2   					# Vorticity component
		Coord = 1           			# XYZ point component
		IntegrationZ = IntegrationY  	# Because we are using the Z-integration tool to integrate along the Y-axis (which has been selected by the user in the GUI)
		Sign = -1   					# This provides the good sign of the integrated vorticity for this case
	else:
		VortComp = 1
		Coord = 2	
		Sign = 1
	
	# Integration along the positive Z-axis direction:
	if IntegrationZ==0 or IntegrationZ==2 or IntegrationZ==3 or IntegrationZ==5:
		Vx_bar_z_p = 0*POINTS[:,0]     					# An array of zeros is created, "p" stands for positive
		for i in range(Points_i):   
			for j in range(Points_j):	
				if j==0:
					Vx_bar_z_p[i+j*Points_i] = 0
				else:
					Vx_bar_z_p[i+j*Points_i] = Vx_bar_z_p[i+(j-1)*Points_i] + Sign*0.5*(Vorticity[i+j*Points_i,VortComp]+Vorticity[i+(j-1)*Points_i,VortComp])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])
		
		
	# Integration along the negative Z-axis direction:
	if IntegrationZ==1 or IntegrationZ==2 or IntegrationZ==4 or IntegrationZ==5:
		Vx_bar_z_n = 0*POINTS[:,0]     					# An array of zeros is created, "n" stands for negative
		for i in range(Points_i):   
			for j in reversed(range(Points_j)):
				if j==Points_j-1:
					Vx_bar_z_n[i+j*Points_i] = 0
				else:
					Vx_bar_z_n[i+j*Points_i] = Vx_bar_z_n[i+(j+1)*Points_i] - Sign*0.5*(Vorticity[i+j*Points_i,VortComp]+Vorticity[i+(j-1)*Points_i,VortComp])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])
	
	# Both directions interpolated:
	if IntegrationZ==0:
		Vx_bar_z = Vx_bar_z_p
		
	if IntegrationZ==1:
		Vx_bar_z = Vx_bar_z_n
		
	if IntegrationZ==2:
		Vx_bar_z = 0.5*(Vx_bar_z_p + Vx_bar_z_n)
		zero_p = algs.where(Vx_bar_z_p == 0)  
		zero_n = algs.where(Vx_bar_z_n == 0)			
		Vx_bar_z[zero_p] = 0
		Vx_bar_z[zero_n] = 0
		
	# Integration along positive Z-direction (Delete Linear Error):
	if IntegrationZ==3 or IntegrationZ==5:
		for i in range(Points_i):   
			Error_initial = Vx_bar_z_p[i]
			Error_final = Vx_bar_z_p[i+(Points_j-1)*Points_i]
			Z_initial = 0
			Z_final = Points_j-1
			if Error_initial!= Error_final:
				while Vx_bar_z_p[i+Z_initial*Points_i] == Error_initial:
					Z_initial = Z_initial+1
					
				while Vx_bar_z_p[i+Z_final*Points_i] == Error_final:
					Z_final = Z_final-1
			Z_initial = Z_initial-3
			Z_final = Z_final +3
			Error_law = (Error_final-Error_initial)/(Z_final-Z_initial)
			for j in range(Points_j):
				if Z_initial<=j<=Z_final:
					Vx_bar_z_p[i+j*Points_i] = Vx_bar_z_p[i+j*Points_i] - Error_law*(j-Z_initial)
				if j>Z_final:
					Vx_bar_z_p[i+j*Points_i] = 0		
		Vx_bar_z = Vx_bar_z_p

	# Integration along negative Z-direction (Delete Linear Error):
	if IntegrationZ==4 or IntegrationZ==5:
		for i in range(Points_i):   
			Error_initial = Vx_bar_z_n[i+(Points_j-1)*Points_i]
			Error_final = Vx_bar_z_n[i]
			Z_initial = Points_j-1
			Z_final = 0
			if Error_initial!= Error_final:
				while Vx_bar_z_n[i+Z_initial*Points_i] == Error_initial:
					Z_initial = Z_initial-1
					
				while Vx_bar_z_n[i+Z_final*Points_i] == Error_final:
					Z_final = Z_final+1
			Z_initial = Z_initial+3
			Z_final = Z_final -3
			Error_law = (Error_final-Error_initial)/(Z_final-Z_initial)
			for j in range(Points_j):
				if Z_final<=j<=Z_initial:
					Vx_bar_z_n[i+j*Points_i] = Vx_bar_z_n[i+j*Points_i] - Error_law*(j-Z_initial)
				if j<Z_final:
					Vx_bar_z_n[i+j*Points_i] = 0
		Vx_bar_z = Vx_bar_z_n
		
	# Both directions interpolated (Delete Linear Error):
	if IntegrationZ==5:
		Vx_bar_z = 0.5*(Vx_bar_z_p + Vx_bar_z_n)
		
	# For XY survey plane, it uses this code part (integration of Y-VORTICITY along J-axis) to integrate the Z-VORTICITY along the J-axis
	if Survey_Plane == 2:
		if 	Switch_IJ_axes == 0:
			Vx_bar_y = Vx_bar_z
			Vx_bar_z = 0*Vx_bar_z
	
	if Survey_Plane == 1 and Switch_IJ_axes == 1:
		Vx_bar_z = Vx_bar_y
		Vx_bar_y = 0*Vx_bar_z
		
	
	
	"""
	#######################################################
	# VORTEX KINETIC ENERGY AT THE CORE: integral of X-VORTICITY
	
	V_core = 0*POINTS[:,0]

	#Core_Pressure = 5

	
	# INTEGRAL OF X-VORTICITY ALONG "J" AXIS (Z-mesh axis):
	if Core_Pressure == 1 or Core_Pressure == 2 or Core_Pressure == 5:   	
	
		if Survey_Plane == 2:   					# If survey plane is XY, it uses this code structure to integrate the X-VORTICITY along the mesh Y-axis (coincident with the J axis)
			VortComp = 2   							# Vorticity component used to determine the sign of the local integral
			Coord = 1           					# XYZ point component
			IntegrationZ = IntegrationY  			# Because we are using the Z-integration tool to integrate along the Y-axis (which has been selected by the user in the GUI)
			Sign = -1   							# This provides the good sign of the integrated vorticity for this case
		else:
			VortComp = 1
			Coord = 2	
			Sign = 1
		
		# Integration along the positive Z-axis direction:
		if Core_Pressure == 1 or Core_Pressure == 5:  					
			V_core_z_p = 0*POINTS[:,0]     			# An array of zeros is created, "p" stands for positive
			for i in range(Points_i):   
				for j in range(Points_j):	
					if j==0 or Q_criterion[i+j*Points_i]<Q_threshold:     # This tags the points outside the vortex core
						V_core_z_p[i+j*Points_i] = 0
					else:
						V_core_z_p[i+j*Points_i] = V_core_z_p[i+(j-1)*Points_i] + np.sign(Vorticity[i+(j-1)*Points_i,VortComp])*Sign*0.5*(Vorticity[i+j*Points_i,0]+Vorticity[i+(j-1)*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])

			V_core = V_core_z_p
		
		# Integration along the negative Z-axis direction:
		if Core_Pressure == 2 or Core_Pressure == 5: 
			V_core_z_n = 0*POINTS[:,0]     					# An array of zeros is created, "n" stands for negative
			for i in range(Points_i):   
				for j in reversed(range(Points_j)):
					if j==Points_j-1 or Q_criterion[i+j*Points_i]<Q_threshold:
						V_core_z_n[i+j*Points_i] = 0
					else:
						V_core_z_n[i+j*Points_i] = V_core_z_n[i+(j+1)*Points_i] - np.sign(Vorticity[i+(j-1)*Points_i,VortComp])*Sign*0.5*(Vorticity[i+j*Points_i,0]+Vorticity[i+(j-1)*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])
			V_core = V_core_z_n
		
		
		
		
		# Integration along positive Z-direction (Delete Linear Error):
		if Core_Pressure == 1 or Core_Pressure == 5:
			for i in range(Points_i):   
				Error_initial = V_core_z_p[i]
				Error_final = V_core_z_p[i+(Points_j-1)*Points_i]
				Z_initial = 0
				Z_final = Points_j-1
				if Error_initial!= Error_final:
					while V_core_z_p[i+Z_initial*Points_i] == Error_initial:
						Z_initial = Z_initial+1
						
					while V_core_z_p[i+Z_final*Points_i] == Error_final:
						Z_final = Z_final-1
				Z_initial = Z_initial-3
				Z_final = Z_final +3
				Error_law = (Error_final-Error_initial)/(Z_final-Z_initial)
				for j in range(Points_j):
					if Z_initial<=j<=Z_final:
						V_core_z_p[i+j*Points_i] = V_core_z_p[i+j*Points_i] - Error_law*(j-Z_initial)
					if j>Z_final:
						V_core_z_p[i+j*Points_i] = 0		
			V_core = V_core_z_p
		
		
		# Integration along negative Z-direction (Delete Linear Error):
		if Core_Pressure == 2 or Core_Pressure == 5:
			for i in range(Points_i):   
				Error_initial = V_core_z_n[i+(Points_j-1)*Points_i]
				Error_final = V_core_z_n[i]
				Z_initial = Points_j-1
				Z_final = 0
				if Error_initial!= Error_final:
					while V_core_z_n[i+Z_initial*Points_i] == Error_initial:
						Z_initial = Z_initial-1
						
					while V_core_z_n[i+Z_final*Points_i] == Error_final:
						Z_final = Z_final+1
				Z_initial = Z_initial+3
				Z_final = Z_final -3
				Error_law = (Error_final-Error_initial)/(Z_final-Z_initial)
				for j in range(Points_j):
					if Z_final<=j<=Z_initial:
						V_core_z_n[i+j*Points_i] = V_core_z_n[i+j*Points_i] - Error_law*(j-Z_initial)
					if j<Z_final:
						V_core_z_n[i+j*Points_i] = 0
			V_core = V_core_z_n
		
		
		
		
		
		
	# INTEGRAL OF X-VORTICITY ALONG "I" AXIS (Y-mesh axis):
	if Core_Pressure == 3 or Core_Pressure == 4 or Core_Pressure == 5:   	
	
		if Survey_Plane == 1 and Switch_IJ_axes == 1:   	# If survey plane is XZ, it uses this code structure to integrate the X-VORTICITY along the mesh Z-axis (coincident with the J axis)
			VortComp = 1   									# Vorticity component used to determine the sign of the local integral
			Coord = 2           							# XYZ point component
			IntegrationY = IntegrationZ  					# Because we are using the Y-integration tool to integrate along the Z-axis (which has been selected by the user in the GUI)
			Sign = -1   									# This provides the good sign of the integrated vorticity for this case
		else:
			VortComp = 2
			Coord = 1	
			Sign = 1
		
		# Integration along the positive Y-axis direction:
		if Core_Pressure == 3 or Core_Pressure == 5:
			V_core_y_p = 0*POINTS[:,0]     					# An array of zeros is created, "p" stands for positive
			for j in range(Points_j):   
				for i in range(Points_i):
					if i==0 or Q_criterion[i+j*Points_i]<Q_threshold:
						V_core_y_p[i+j*Points_i] = 0
					else:
						V_core_y_p[i+j*Points_i] = V_core_y_p[i-1+j*Points_i] - np.sign(Vorticity[i+(j-1)*Points_i,VortComp])*Sign*0.5*(Vorticity[i+j*Points_i,0]+Vorticity[i-1+j*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i-1+j*Points_i,Coord])
			V_core = V_core_y_p

		# Integration along the negative Y-axis direction:	
		if Core_Pressure == 4 or Core_Pressure == 5:
			V_core_y_n = 0*POINTS[:,0]     					# An array of zeros is created, "n" stands for negative
			for j in range(Points_j):   
				for i in reversed(range(Points_i)):
					if i==Points_i-1  or Q_criterion[i+j*Points_i]<Q_threshold:
						V_core_y_n[i+j*Points_i] = 0
					else:
						V_core_y_n[i+j*Points_i] = V_core_y_n[i+1+j*Points_i] + np.sign(Vorticity[i+(j-1)*Points_i,VortComp])*Sign*0.5*(Vorticity[i+j*Points_i,0]+Vorticity[i-1+j*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i-1+j*Points_i,Coord])
			V_core = V_core_y_n


		# Integration along positive Y-direction (Delete Linear Error):
		if Core_Pressure == 3 or Core_Pressure == 5:
			for j in range(Points_j):   
				Error_initial = V_core_y_p[0+j*Points_i]
				Error_final = V_core_y_p[Points_i-1+j*Points_i]
				Y_initial = 0
				Y_final = Points_i-1
				if Error_initial!= Error_final:
					while V_core_y_p[Y_initial+j*Points_i] == Error_initial:
						Y_initial = Y_initial+1
						
					while V_core_y_p[Y_final+j*Points_i] == Error_final:
						Y_final = Y_final-1
				Y_initial = Y_initial-3
				Y_final = Y_final +3
				Error_law = (Error_final-Error_initial)/(Y_final-Y_initial)
				for i in range(Points_i):
					if Y_initial<=i<=Y_final:
						V_core_y_p[i+j*Points_i] = V_core_y_p[i+j*Points_i] - Error_law*(i-Y_initial)
					if i>Y_final:
						V_core_y_p[i+j*Points_i] = 0
			V_core = V_core_y_p


		# Integration along negative Y-direction (Delete Linear Error):
		if Core_Pressure == 4 or Core_Pressure == 5:
			for j in range(Points_j):   
				Error_initial = V_core_y_n[Points_i-1+j*Points_i]
				Error_final = V_core_y_n[0+j*Points_i]
				Y_initial = Points_i-1
				Y_final = 0
				if Error_initial!= Error_final:
					while V_core_y_n[Y_initial+j*Points_i] == Error_initial:
						Y_initial = Y_initial-1
						
					while V_core_y_n[Y_final+j*Points_i] == Error_final:
						Y_final = Y_final+1
				Y_initial = Y_initial+3
				Y_final = Y_final -3
				Error_law = (Error_final-Error_initial)/(Y_final-Y_initial)
				for i in range(Points_i):
					if Y_final<=i<=Y_initial:
						V_core_y_n[i+j*Points_i] = V_core_y_n[i+j*Points_i] - Error_law*(i-Y_initial)
					if i<Y_final:
						V_core_y_n[i+j*Points_i] = 0
			V_core = V_core_y_n





	#V_core = (V_core_z_p+V_core_z_n+V_core_y_p+V_core_y_n)/4

	
	V_core_old = V_core
	#data.PointData.append(V_core_old , 'V_core_old')
	"""
	
	
	
	
	########################################
	#######  VELOCITY DECOMPOSITION  #######
	########################################
		
	# BOUNDARY LAYER LOSSES DETECTION:
	if BoundaryLayerDefinition==0:
		Vx_bar_BL = 0*Vx_bar_z	
		Vx_no_BL = 0*Vx_bar_z
	if BoundaryLayerDefinition==1:
		Vx_bar_BL = Vx_bar_z - Vx_bar_y
		Vx_no_BL = Vx + Vx_bar_BL
	if BoundaryLayerDefinition==2:
		Vx_bar_BL =  Vx_bar_y - Vx_bar_z
		Vx_no_BL = Vx + Vx_bar_BL
	
	"""
	revisar BL definition !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	"""
	
	
	#  ISENTROPIC X-VELOCITY DEFINITION:
	if IsentropicVelocity == 0:
		Vx_star = Vx - Vx_bar_y
	if IsentropicVelocity == 1:
		Vx_star = Vx - Vx_bar_z

	# ISENTROPIC VELOCITY:
	FlowAlfa = algs.arctan(Vz/Vx)
	FlowTita = algs.arctan(Vz/algs.sqrt(Vx**2+Vy**2))
	FlowBeta = algs.arctan(Vy/Vx)

	Vy_bar = Vx_bar_z*np.tan(FlowBeta)
	Vz_bar = Vx_bar_z*np.tan(FlowAlfa)
	Vy_star = Vy - Vy_bar
	Vz_star = Vz - Vz_bar	
	Vt_star = np.sqrt(Vy_star**2+Vz_star**2)
	
	
	
	########################################
	########  VORTEX CORE VELOCITY  ########
	########################################
	
	MASK_core = Q_criterion>Q_threshold    
	V_core_Q = 1*MASK_core*Vt_star  											# Mask everywhere except the vortex core
	
	V_core = 0*POINTS[:,0]       												# Creates empty array for V_core variable
	V_core_new = 0*POINTS[:,0]
	Margin = 3     																# This ignores the first lines at the PIV domain 
	
	
	# Vortex core reconstruction along positive Z-direction:
	if Core_Pressure == 1:
		for i in range(Margin,Points_i-Margin):     											# Analyzes each vertical image line
			Z_initial = 0    													# Initial point coordinate of the line
			Z_final = Points_j-1												# Final point coordinate of the line

			while V_core_Q[i+Z_initial*Points_i] == 0:  						# It looks for regions with non-zero vortex core values: this detects the lower point of that region
				Z_initial = Z_initial+1
				if Z_initial == Points_j-1:
					break
				
			while V_core_Q[i+Z_final*Points_i] == 0: 							# It looks for regions with non-zero vortex core values: this detects the upper point of that region
				Z_final = Z_final-1
				if Z_final == 0:
					break
			
			if Z_initial!=Points_j-1 and Z_final!=0:   							# This applies only for image lines where a core exist
				Z_initial_new = Z_initial-7    								# Some extra points are included after/before the core zone
				Z_final_new = Z_final +7
				Linear_slope = (Vt_star[i+Z_final_new*Points_i]-Vt_star[i+Z_initial_new*Points_i])/(Z_final_new-Z_initial_new)  # Slope of line joining the start/end points of the core region
				
				
				for j in range(Points_j):   									# First rectification step:								 
					if Z_initial_new<=j<=Z_final_new:
						V_core[i+j*Points_i] = Vt_star[i+j*Points_i] - (Vt_star[i+Z_initial_new*Points_i]+Linear_slope*(j-Z_initial_new))  # The V_core_Q distribution is shifted downwards by the previous linear law
				
				Z_max_lower = Z_initial_new
				Z_max_upper = Z_final_new
				while V_core[i+(Z_max_lower+1)*Points_i]>V_core[i+Z_max_lower*Points_i]:
					Z_max_lower = Z_max_lower+1
				while V_core[i+(Z_max_upper-1)*Points_i]>V_core[i+Z_max_upper*Points_i]:
					Z_max_upper = Z_max_upper-1
				Linear_slope_max = (V_core[i+Z_max_upper*Points_i]-V_core[i+Z_max_lower*Points_i])/(Z_max_upper-Z_max_lower)
						
										
				for j in range(Points_j):   									# Second rectification step:
					if Z_max_lower<=j<=Z_max_upper:
						V_core_new[i+j*Points_i] = V_core[i+j*Points_i] - 2*(V_core[i+Z_max_lower*Points_i]+Linear_slope_max*(j-Z_max_lower)) 						
					if j<Z_max_lower or j>Z_max_upper:					
						V_core_new[i+j*Points_i] = -V_core[i+j*Points_i]	

	"""
	Agregar integracion segun eje Y
	Identificar regiones con vortices
	Determinar minimo de cada region
	"""

	where_are_NaNs = isnan(V_core_new)    
	V_core_new[where_are_NaNs] = 0
	
	Vt_star_new = Vt_star + np.sqrt(V_core_new**2+2*algs.min(V_core_new)*V_core_new)
	V_star = np.sqrt(Vx_star**2+Vt_star_new**2)
	#V_star = np.sqrt(Vx_star**2+Vt_star**2)
	
	"""
	#Vt_star_new = Vt_star - np.sqrt(2)*0.5*V_core_old
	#V_star = np.sqrt(Vx_star**2+Vt_star**2+0.5*1.225*V_core_new**2)	
	"""
	Vt = np.sqrt(Vy**2+Vz**2)
	
	
	
	#V_star = Vx_star
	#V_star = np.sqrt(Vx_star**2+Vt**2)
	#V_star = np.sqrt(Vx_star**2+Vt_star**2)

	#V_star = Vx_star/(np.cos(FlowTita)*np.cos(FlowBeta))

	V_abs = algs.sqrt(Vx**2+Vy**2+Vz**2)
	
	#V_star = algs.sqrt((Vx- Vx_bar_z)**2+Vy**2+Vz**2) 

	"""
	data.PointData.append(Vt , 'Vt')
	data.PointData.append(V_core , 'V_core')
	data.PointData.append(V_core_new , 'V_core_new')
	data.PointData.append(V_core_Q , 'V_core_Q')
	data.PointData.append(Vy_bar , 'Vy_bar')
	data.PointData.append(Vz_bar , 'Vz_bar')
	data.PointData.append(Vy_star , 'Vy_star')
	data.PointData.append(Vz_star , 'Vz_star')
	data.PointData.append(Vt_star , 'Vt_star')
	data.PointData.append(Vt_star_new , 'Vt_star_new')
	data.PointData.append(np.sqrt(V_core_new**2+2*algs.min(V_core_new)*V_core_new) , 'process')
	"""

	#######################################################
	# INTEGRAL OF X-VORTICITY
	
	X_vort = 1
	Vortex_Phi_Threshold = 600000
	
	"""
	Mejorar vortex core detection
	"""
	MASK_vortex = Phi_eff_New>Vortex_Phi_Threshold 
	Vorticity_V = 1*MASK_vortex*Vorticity
	
	MASK_wake = Phi_eff_New<Vortex_Phi_Threshold 
	MASK_core = Q_criterion<Q_threshold
	Vorticity_W = 1*MASK_wake*MASK_core*Vorticity
	"""
	data.PointData.append(Vorticity_V , 'Vorticity_V')
	data.PointData.append(Vorticity_W , 'Vorticity_W')
	"""
	# INTEGRAL OF X-VORTICITY ALONG "J" AXIS (Z-mesh axis):
	if X_vort == 1 or X_vort == 2:   	
	
		if Survey_Plane == 2:   					# If survey plane is XY, it uses this code structure to integrate the X-VORTICITY along the mesh Y-axis (coincident with the J axis)
			VortComp = 2   							# Vorticity component used to determine the sign of the local integral
			Coord = 1           					# XYZ point component
			IntegrationZ = IntegrationY  			# Because we are using the Z-integration tool to integrate along the Y-axis (which has been selected by the user in the GUI)
			Sign = -1   							# This provides the good sign of the integrated vorticity for this case
		else:
			VortComp = 1
			Coord = 2	
			Sign = 1
		
		# Integration along the positive Z-axis direction:
		if X_vort == 1:   					
			Vw_z_p = 0*POINTS[:,0]     			# An array of zeros is created, "p" stands for positive
			Vv_z_p = 0*POINTS[:,0]
			for i in range(Points_i):   
				for j in range(Points_j):	
					if j==0:     
						Vw_z_p[i+j*Points_i] = 0
						Vv_z_p[i+j*Points_i] = 0
					else:
						Vw_z_p[i+j*Points_i] = Vw_z_p[i+(j-1)*Points_i] + np.sign(Vy[i+(j-1)*Points_i])*Sign*0.5*(Vorticity_W[i+j*Points_i,0]+Vorticity_W[i+(j-1)*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])
						Vv_z_p[i+j*Points_i] = Vv_z_p[i+(j-1)*Points_i] + np.sign(Vy[i+(j-1)*Points_i])*Sign*0.5*(Vorticity_V[i+j*Points_i,0]+Vorticity_V[i+(j-1)*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])
			#Vy_z = Vy_z_p
			#Vy_z = Vy_z_p
	
		# Integration along the negative Z-axis direction:
		if X_vort == 2: 
			Vw_z_p = 0*POINTS[:,0]     			# An array of zeros is created, "p" stands for positive
			Vv_z_p = 0*POINTS[:,0]
			for i in range(Points_i):   
				for j in reversed(range(Points_j)):
					if j==Points_j-1:
						Vw_z_p[i+j*Points_i] = 0
						Vv_z_p[i+j*Points_i] = 0
					else:
						Vw_z_p[i+j*Points_i] = Vw_z_p[i+(j+1)*Points_i] - np.sign(Vy[i+(j-1)*Points_i])*Sign*0.5*(Vorticity_W[i+j*Points_i,0]+Vorticity_W[i+(j-1)*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])
						Vv_z_p[i+j*Points_i] = Vv_z_p[i+(j+1)*Points_i] - np.sign(Vy[i+(j-1)*Points_i])*Sign*0.5*(Vorticity_V[i+j*Points_i,0]+Vorticity_V[i+(j-1)*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i+(j-1)*Points_i,Coord])

			#V_core = V_core_z_n
	
	
	"""
	# INTEGRAL OF X-VORTICITY ALONG "I" AXIS (Y-mesh axis):
	if X_vort == 1 or X_vort == 2:   	
	
		if Survey_Plane == 1 and Switch_IJ_axes == 1:   	# If survey plane is XZ, it uses this code structure to integrate the X-VORTICITY along the mesh Z-axis (coincident with the J axis)
			VortComp = 1   									# Vorticity component used to determine the sign of the local integral
			Coord = 2           							# XYZ point component
			IntegrationY = IntegrationZ  					# Because we are using the Y-integration tool to integrate along the Z-axis (which has been selected by the user in the GUI)
			Sign = -1   									# This provides the good sign of the integrated vorticity for this case
		else:
			VortComp = 2
			Coord = 1	
			Sign = 1
		
		# Integration along the positive Y-axis direction:
		if X_vort == 1: 
			Vz_y_p = 0*POINTS[:,0]     					# An array of zeros is created, "p" stands for positive
			for j in range(Points_j):   
				for i in range(Points_i):
					if i==0:
						Vz_y_p[i+j*Points_i] = 0
					else:
						Vz_y_p[i+j*Points_i] = Vz_y_p[i-1+j*Points_i] - np.sign(Vz[i+(j-1)*Points_i])*Sign*0.5*(Vorticity[i+j*Points_i,0]+Vorticity[i-1+j*Points_i,0])*(POINTS[i+j*Points_i,Coord]-POINTS[i-1+j*Points_i,Coord])
			Vz_y = Vz_y_p
	"""	

			
			
			
			
			


	#Vz_new = Vz + Vz_y
	"""
	data.PointData.append(Vw_z_p , 'Vw_z_p')
	data.PointData.append(Vv_z_p , 'Vv_z_p')	
	"""
	#data.PointData.append(Vz_y , 'Vz_y')
	#data.PointData.append(Vz_new , 'Vz_new')

	"""
	Vt_new = np.sqrt(Vy_star**2+Vz_star**2)+ Vv_z_p +Vw_z_p
	V_star = np.sqrt(Vx_star**2+Vt_new**2)
	"""
	







# Continuar a analizar BL !!! Y generalizar para feature extraction (delete strut wake)
# Definir output: si usuario quiere quitar BL del wake, debo redefinir output
# Considerar casos de plano no alineado con flujo (ejemplo: PIV normal al ala delta)

# Leer archivos vtp desde CFD,especialmente para usar Epsilon Mesh Generator para interpolar CFD data into regular mesh y preparar campaña WTT a partir de datos CFD 
# --> Es un atout importante de Epsilon Wind Tunnel para definir la resolucion espacial requerida. Ver de incluir ruido numérico en el campo de velocidad (?)

# Incluir Vorticity decomposition method




# Calculat Ps from PIV en AeroData --> SI, porque no tengo los valores de upstream en EpsilonPIV pero si en AeroData
# comunicacion hacia AeroData (conservar vorticity vector, no recalcularlo en aerodata! Idem Phieff --> ver si otro parametro tambien !)
# Hacer validacion con datos CFD
# For validation, use the laminar nozzle PIV as a trivial case: Ps=cte across jet



"""
############################################################################################################
########################################### TEMPERATURE FROM PIV ###########################################
############################################################################################################


if Select_Action == 6:

	if Ts_Calculation_Method == 0:
		Isoenergetic = 1

	if Ts_Calculation_Method == 1:
		Isoenergetic = 0
		
"""

############################################################################################################
############################################### DATA OUTPUT ################################################
############################################################################################################

if Select_Action > 3 :																			# Output for Tecplot format
	data.PointData.append(Vector_Speed , 'Vector_Speed')
	data.PointData.append(Vx,'Vx')
	data.PointData.append(Vy,'Vy')
	data.PointData.append(Vz,'Vz')
	data.PointData.append(Vorticity , 'Vorticity')
	#data.PointData.append(Vorticity[:,0] , 'Vorticity_X')
	#data.PointData.append(Vorticity[:,1] , 'Vorticity_Y')
	#data.PointData.append(Vorticity[:,2] , 'Vorticity_Z')
	#data.PointData.append(np.sqrt(Vorticity[:,1]**2+Vorticity[:,2]**2) , 'Vorticity_YZ')
	data.PointData.append(VORTICITY_magnitude , 'Vorticity_magnitude')
	#data.PointData.append(Vx_GRADIENT , 'Vx_gradient')
	#data.PointData.append(Vy_GRADIENT , 'Vy_gradient')
	#data.PointData.append(Vz_GRADIENT , 'Vz_gradient')
	#data.PointData.append(V_gradient , 'V_gradient')
	data.PointData.append(Phi_eff , 'Phi_eff')
	#data.PointData.append(Q_criterion , 'Q_criterion')
	output.ShallowCopy(data.VTKObject)  															# This displays the data available in the Paraview output

	
	
if Select_Action > 4 :      														# Output for Tecplot format
	data.PointData.append(Vx_bar_z , 'Vx_bar_z')
	data.PointData.append(Vx_bar_y , 'Vx_bar_y')
	#data.PointData.append(Vx_bar_BL , 'Vx_bar_BL')
	data.PointData.append(Vx_star , 'Vx_star')
	#data.PointData.append(Vx_no_BL , 'Vx_no_BL')
	data.PointData.append(V_star , 'V_star')
	data.PointData.append(V_abs , 'V_abs')
	#data.PointData.append(V_core , 'V_core')	
	output.ShallowCopy(data.VTKObject)  															# This displays the data available in the Paraview output

#if Select_Action == 6:

	#data.FieldData.append(Isoenergetic,'Isoenergetic')
	#output.ShallowCopy(data.VTKObject)  															# This displays the data available in the Paraview output

"""
if MeshType !=0:    																				# Output for VTK formats (regular grid and Delaunay2D grid)
	numPoints = X.size                 																# Number of points of the  mesh

	def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName): 							# Defines a function that adds the field data to the existing points in the Paraview's output
		numberOfComponents = int(numTuples)                                       					# We deal with 1-, 3- and 6-component arrays
		dataArray = vtk.vtkDoubleArray()
		#dataSetAttributes.AddArray(dataArray)
		dataArray.SetNumberOfComponents(numberOfComponents)
		dataArray.SetName(ArrayName)                                      							# "ArrayName" is a string used to name the array in Paraview 
		data = DataArray                                                 								# DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
		for value in data:
			dataArray.InsertNextValue(float(value))                         							# Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 
		dataSetAttributes.AddArray(dataArray)

		
	output = self.GetOutput()
	
	
	addfielddata(1, output.GetFieldData(),np.array([Isoenergetic]),"Isoenergetic")  							# Now we use this function to send the scalar arrays to the output:
	addfielddata(1, output.GetPointData(),Vx,"Vx")
	addfielddata(1, output.GetPointData(),Vy,"Vy")
	addfielddata(1, output.GetPointData(),Vz,"Vz")

	if Instrument == 1:
		addfielddata(1, output.GetPointData(),Ps,"Ps")
		addfielddata(1, output.GetPointData(),Pt,"Pt")

	if MeshType==1:
		addfielddata(3, output.GetPointData(),Vector_Speed.ravel(),"Vector_Speed")
		addfielddata(3, output.GetPointData(),Vorticity.ravel(),"Vorticity")
		addfielddata(3, output.GetPointData(),Vx_GRADIENT.ravel(),"Vx_gradient")
		addfielddata(3, output.GetPointData(),Vy_GRADIENT.ravel(),"Vy_gradient")
		addfielddata(3, output.GetPointData(),Vz_GRADIENT.ravel(),"Vz_gradient")
		#addfielddata(1, output.GetPointData(),dVz_dx.ravel(),"dVz_dx")
		#addfielddata(1, output.GetPointData(),dVy_dx.ravel(),"dVy_dx")
		addfielddata(1, output.GetPointData(),Phi_eff.ravel(),"Phi_eff")
		addfielddata(1, output.GetPointData(),VORTICITY_magnitude.ravel(),"Vorticity_magnitude")
		addfielddata(1, output.GetPointData(),Vorticity[:,0].ravel(),"Vorticity_x")
		addfielddata(1, output.GetPointData(),Vorticity[:,1].ravel(),"Vorticity_y")
		addfielddata(1, output.GetPointData(),Vorticity[:,2].ravel(),"Vorticity_z")
		addfielddata(1, output.GetPointData(),Q_criterion.ravel(),"Q_criterion")	
		

	if PressureFromPIV==1 and MeshType==1:
		addfielddata(1, output.GetPointData(),Vx_bar_z,"Vx_bar_z")
		addfielddata(1, output.GetPointData(),Vx_bar_y,"Vx_bar_y")
		addfielddata(1, output.GetPointData(),Vx_bar_BL,"Vx_bar_BL")
		addfielddata(1, output.GetPointData(),Vx_star,"Vx_star")
		addfielddata(1, output.GetPointData(),Vx_no_BL,"Vx_no_BL")
		addfielddata(1, output.GetPointData(),V_star,"V_star")
		addfielddata(1, output.GetPointData(),V_abs,"V_abs")
		addfielddata(1, output.GetPointData(),V_core,"V_core")
		addfielddata(1, output.GetPointData(),Vy_bar,"Vy_bar")
		addfielddata(1, output.GetPointData(),Vz_bar,"Vz_bar")
		addfielddata(1, output.GetPointData(),Vy_star,"Vy_star")
		addfielddata(1, output.GetPointData(),Vz_star,"Vz_star")		
"""        




#Armonizar los arrays de salida entre los dos formatos!





"""
import time
start = time.time() 	
end = time.time()
SolveTime = end - start
print ("Calculation time", SolveTime	)
"""

#data = dsa.WrapDataObject(self.GetOutputDataObject(0))   													# Converts VTK element into a Python object



############################################################################################################
#############################################    END OF CODE   #############################################
############################################################################################################
