# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################			   Poisson Solver Plugin			   #############################
############################################################################################################

# Abstract:
This Paraview's programmable filter takes a slice data from Epsilon_Aero_Data filter and solves the Poisson 
equation to find both, the stream function and velocity potential function. This functions allows then to 
calculate the vortex drag density on each point of the mesh plane. The user has to indicate if he wants to
calculate or not the velocity potential (and its contribution to the vortex drag). By default it is not 
considered. The Poisson equation is solved by using three different methods:
• GREEN FUNCTIONS: it is an analytical solution wich can be used in all types of mesh distributions. It provides 
an analytical solution for free-air flow fields (full model), wall flows (half-model and/or ground) as well as
wall bounded flows (wind tuunel).
• SINGULARITIES METHOD: it is a method suited only for bounded flows (wind tunnel) and it is equivalent to the
vortex panel method. The wall panel vortices are computed to reach the condition of stream function = 0 at the 
walls. These vortices are then used to compute the total stream function.
• FINITE DIFFERENCE METHOD: xxxxxxxxxx


#DISCLAIMER:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee of any kind, either express
or implied. Its use for teaching, research and industrial applications is allowed under the strict responsibility of the user. The authors
of this plugin are not responsible for any damages or errors related to the use of the plugin.
"""

print ("")
print ("")
print ("")
print ("####################################################")
print ("##########    EPSILON POISSON SOLVER      ##########")
print ("####################################################")
print ("")



############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

# We load the Python and Paraview standard modules:
import time
import numpy as np
import vtk
import vtk.numpy_interface.dataset_adapter as dsa
from vtk.numpy_interface import algorithms as algs
import paraview.vtk.util.numpy_support as vnp
import numpy.ma as ma



############################################################################################################
############################################# INPUT PARAMETERS #############################################
############################################################################################################
		
# Input parameters are defined as string in Paraview's GUI, thus a string to float convertion is required:

Slice_Input = Slice_Input																					# Boolean
Slice_Center = Slice_Center																					# XYZ coordinates of the slice plane
Select_Normal = Select_Normal																				# 0=aerodynamic x-axis  1=user defined	  
Slice_Normal = Slice_Normal																					# Slice plane normal
Clip_By_Scalar = Clip_By_Scalar																				# Boolean
Parameter = Parameter																						# Variable used to clip by scalar
Clip_Value = Clip_Value																						# Scalar value used to clip by scalar
Invert_Clip = Invert_Clip																					# Boolean
Flow_Domain = Flow_Domain                                                                                   # 0=Free-air    1=Wind tunnel
Yw = Ywall  																								# Y-coordinate of the wall/symmetry plane
Zw = Zwall  																								# Z-coordinate of the wall/symmetry plane
Tunnel_Y = Tunnel_Y  																						# min/max Y-coordinates and the related number of panels
Tunnel_Z = Tunnel_Z  																						# min/max Z-coordinates and the related number of panels
Layers = Layers																								# Number of layers of singularities around the wind tunnel
VelocityPotential = VelocityPotential   																	# 0=Disabled   1=-du_dx  2=dv_dy+dw_dz
Save_Output_Data = Save_Output_Data																			# Boolean
File_Path_Output = File_Path_Output																			# Path to the output VTP file

if Flow_Domain == 0:
	Method = Method_1                                                                                       # 0=Greeen functions
if Flow_Domain == 1:
	Method = Method_2                                                                                       # 0=Greeen functions    1=Singularities

if Flow_Domain == 0:
	Case = Case_1   																					    # 0=Full wake   1=Wall or Symmetry at XZ plane   2=Wall or Symmetry at XY plane   3=Double wall or symmetry at XY and XZ planes	
if Flow_Domain == 1:
	Case = 4   																					            # 4=Full model	 5=Half model Y+	6=Half model Y-  (Cases 5 and 6 are coded but not exposed in GUI, because 4 works fine for all cases)

if Slice_Input == 1 and Clip_By_Scalar == 1 and Flow_Domain == 0:
	Select_Action =  Select_Action_1																		# 0=View slice plane	1=View clipped wake   2=Display wall or simmetry plane   4=Solve Poisson equation
if Slice_Input == 1 and Clip_By_Scalar == 0 and Flow_Domain == 0:
	Select_Action =  Select_Action_2
if Slice_Input == 0 and Clip_By_Scalar == 1 and Flow_Domain == 0:
	Select_Action =  Select_Action_3
if Slice_Input == 0 and Clip_By_Scalar == 0 and Flow_Domain == 0:
	Select_Action =  Select_Action_4
if Slice_Input == 1 and Clip_By_Scalar == 1 and Flow_Domain == 1:
	Select_Action =  Select_Action_5																		# 0=View slice plane	1=View clipped wake   2=Display tunnel walls   3=Solve Poisson equation on entire tunnel section   4=Solve Poisson equation on wake
if Slice_Input == 1 and Clip_By_Scalar == 0 and Flow_Domain == 1:
	Select_Action =  Select_Action_6
if Slice_Input == 0 and Clip_By_Scalar == 1 and Flow_Domain == 1:
	Select_Action =  Select_Action_7
if Slice_Input == 0 and Clip_By_Scalar == 0 and Flow_Domain == 1:
	Select_Action =  Select_Action_8



############################################################################################################
#######################################	   DEFINITION OF FUNCTIONS	  ######################################
############################################################################################################

###########################################################
##################### CLIP BY SCALAR ######################
###########################################################

def CLIP_BY_SCALAR(Object,Scalar,ConstantValue,Invert_Clip):												# Defines a function that clips the VTK Object with a surface of given by Scalar=ConstantValue. IMPORTANT: if the objects are data from the input it is not necessary to add .GetOutput() (this is required for VTK objects only)
	import vtk
	
	# Clip by scalar: 
	clipper = vtk.vtkTableBasedClipDataSet()																# It is the faster vtk tool. It works in all ParaView versions
	clipper.SetInputData(Object)
	clipper.SetInputArrayToProcess(0,0,0,0, Scalar)
	clipper.SetValue(ConstantValue)														  
	clipper.InsideOutOff()
	if Invert_Clip == 1:
		clipper.InsideOutOn()																				# Selects one side or the other of the clip																			   
	clipper.Update()

	# Convert to vtkPolyData:
	Clip_PD = vtk.vtkGeometryFilter()					 
	Clip_PD.SetInputData(clipper.GetOutput())	
	Clip_PD.Update()

	return Clip_PD


############################################
########### DRAW LINES FUNCTION ############
############################################

def Line_Function(x0,y0,z0,x1,y1,z1,resolution):
	import vtk
	Line_F=vtk.vtkLineSource()
	Line_F.SetPoint1(x0,y0,z0)
	Line_F.SetPoint2(x1,y1,z1)
	Line_F.SetResolution(resolution)
	Line_F.Update()
	return Line_F 


############################################
##########  CREATE GRID FUNCTION  ##########
############################################

def CREATE_GRID(PointsX,PointsY,PointsZ):
	import vtk
	import numpy as np

	xCoords = vtk.vtkFloatArray()																			# Creates a VTK container of the X-point coordinates
	for x, i in enumerate(PointsX):	 
		xCoords.InsertNextValue(i)

	yCoords = vtk.vtkFloatArray()											 								# Creates a VTK container of the Y-point coordinates
	for y, i in enumerate(PointsY):
		yCoords.InsertNextValue(i)

	zCoords = vtk.vtkFloatArray()											 								# Creates a VTK container of the Z-point coordinates
	for z, i in enumerate(PointsZ):
		zCoords.InsertNextValue(i)

	rgrid = vtk.vtkRectilinearGrid()										  								# Creates a rectilinear grid by using the VTK coordinates
	rgrid.SetDimensions(x+1, y+1, z+1)									  									# x+1=number of points along X axis
	rgrid.SetXCoordinates(xCoords)
	rgrid.SetYCoordinates(yCoords)
	rgrid.SetZCoordinates(zCoords)

	ExtractSurface = vtk.vtkGeometryFilter()								 								# Converts vtk rectilinear grid element to vtkPolyData  --> It is equivalent to Extract Surface
	ExtractSurface.SetInputData(rgrid)		   
	ExtractSurface.Update() 

	return ExtractSurface


############################################
######### ADD FIELD DATA FUNCTION ##########
############################################

def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName):			 								# Defines a function that adds the field data to the existing points in the Paraview's output
	numberOfComponents = int(numTuples)									   									# We deal only with 1-component arrays
	dataArray = vtk.vtkDoubleArray()   
	dataArray.SetNumberOfComponents(numberOfComponents)
	dataArray.SetName(ArrayName)											   								# "ArrayName" is a string used to name the array in Paraview 
	data = DataArray																						# DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
	for value in data:
		dataArray.InsertNextValue(float(value))							  									# Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 
	
	dataSetAttributes.AddArray(dataArray)

############################################
############## GREEN FUNCTION ##############
############################################

def GREEN(i,y,z,C,A,B,N,Yw,Zw):   																			# Geen function: G=[(Y-A*Yc-K*Yw)^2 + (Z-B*Zc-K*Zw)^2]^N   with: N=[-1,1], A=[-1,1], B=[-1,1] and K=0 if A=1 or B=1 (else K=2)   Yc/Zc=YZ cell center coordinates  Yw/Zw=YZ wall coordinates
	if A==1:
		Ka=0
	if A==-1:
		Ka=2
	if B==1:
		Kb=0
	if B==-1:
		Kb=2
	Logarithm = ((y[i]-A*C[:,1]-Ka*Yw)**2+(z[i]-B*C[:,2]-Kb*Zw)**2)**N
	return Logarithm
	

############################################
######## GREEN WIND TUNNEL FUNCTION ########
############################################

def GREEN_WTT(i,y,z,C,Ny,Nz,W,H,DeltaY,DeltaZ,E,Case):
	Ny = int(Ny)
	Nz = int(Nz)
	if Ny % 2 == 0: 																						# Check if Ny is an even number
		Ky = 1 
	if Ny % 2 != 0:  																						# It is an odd number
		Ky = -1
	if Ky == 1:
		Ky2 = 0
	if Ky == -1:
		Ky2 = 2	
	if Nz % 2 == 0: 																						# Check if Nz is an even number
		Kz = 1 
	if Nz % 2 != 0:  																						# It is an odd number
		Kz = -1
	if Kz == 1:
		Kz2 = 0
	if Kz == -1:
		Kz2 = 2		
	
	if Case == 4:  																							# Full model 
		Logarithm = ((y[i]-(Ky*C[:,1]+Ny*W+Ky2*DeltaY))**2+(z[i]-(Kz*C[:,2]+Nz*H+Kz2*DeltaZ))**2)**E
	if Case == 5:   																						# Half model Y+
		Ky2 = 0
		Logarithm = ((y[i]-(Ky*C[:,1]+2*int((Ny+1)/2)*W+Ky2*DeltaY))**2+(z[i]-(Kz*C[:,2]+Nz*H+Kz2*DeltaZ))**2)**E
	if Case == 6:   																						# Half model Y-
		Ky2 = 0
		Logarithm = ((y[i]-(Ky*C[:,1]+2*int(Ny/2)*W+Ky2*DeltaY))**2+(z[i]-(Kz*C[:,2]+Nz*H+Kz2*DeltaZ))**2)**E
	
	return Logarithm
	
	
############################################
########  GET DATA BOUNDS FUNCTION  ########
############################################

def GET_DATA_BOUNDS(InputData):												   								# Get the YZ bounds coordinates of a given data set
	import vtk
	from vtk.numpy_interface import dataset_adapter as dsa
	import numpy as np
	
	outline = vtk.vtkOutlineFilter()												 						# Gets the outline square that sourrounds the input data
	outline.SetInputConnection(InputData.GetOutputPort())
	outline.Update()

	data = dsa.WrapDataObject(outline.GetOutput())   														# Converts the VTK element into a Python object
	POINTS = data.Points   																					# Get points coordinates
	XminData = np.min(POINTS[:,0])													 						# Get minimum x-coordinate of the outline square
	XmaxData = np.max(POINTS[:,0])
	YminData = np.min(POINTS[:,1])
	YmaxData = np.max(POINTS[:,1])
	ZminData = np.min(POINTS[:,2])
	ZmaxData = np.max(POINTS[:,2])

	return XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData
    

###################################
#######  WRITE VTP FUNCTION  ######
###################################

def WRITE_VTP(outputObject,File_Path_Output):     					
	import vtk
	if File_Path_Output:
				
		OutputVTP = vtk.vtkXMLPolyDataWriter()                  
		OutputVTP.SetInputData(outputObject.VTKObject)    			    # Alternative command: SetInput           vtkObject.GetOutput()
		OutputVTP.SetFileName(File_Path_Output)                         # File path to the folder where to store the output data 
		OutputVTP.Write()    



############################################################################################################
######################################### INPUT DATA PRETREATMENT ##########################################
############################################################################################################

INPUT = inputs[0]																							# Load input data
TRANSFORMATION = INPUT.FieldData["TRANSFORMATION"]   														# Transformation matrix
T_t = np.transpose(TRANSFORMATION)
T_inverse = np.dot(T_t,np.dot(TRANSFORMATION,T_t))  														# It calculates the inverse of the TRANSFORMATION matrix


if Slice_Input == 1:

	# Define plane normal:
	if Select_Normal == 0:																					# Aerodynamic x-axis
		normal = np.array([1,0,0])																			# Define the normal of the survey plane on the aero reference frame
		Slice_Normal = np.dot(T_inverse,normal)   															# Express the normal in the mesh reference frame

	# Create slicing plane:
	plane = vtk.vtkPlane()																
	plane.SetOrigin(Slice_Center[0], Slice_Center[1], Slice_Center[2])
	plane.SetNormal(Slice_Normal[0], Slice_Normal[1], Slice_Normal[2])

	# Slice input data with plane:
	cutter = vtk.vtkCutter()													
	cutter.SetCutFunction(plane)
	cutter.SetInputData(INPUT.VTKObject)
	cutter.Update()

	INPUT = dsa.WrapDataObject(cutter.GetOutput())															# Updates the input data to process with this filter


if Clip_By_Scalar == 1:

	# Clip input data by scalar:
	#Parameter = INPUT.PointData[Parameter]   																# "Parameter" is the aerodynamic variable selected by the user in the GUI to perform the clip by scalar
	ClippedData = CLIP_BY_SCALAR(INPUT.VTKObject,Parameter,Clip_Value,Invert_Clip)

	INPUT = dsa.WrapDataObject(ClippedData.GetOutput())														# Updates the input data to process with this filter


if Select_Action == 0:																						# Display sliced plane
	self.GetOutput().ShallowCopy(cutter.GetOutput())

if Select_Action == 1:																						# Display clipped wake
	self.GetOutput().ShallowCopy(ClippedData.GetOutput())

if Select_Action >= 2:																						# Proceed to Poisson solver

	# Convert to vtkPolyData:
	if INPUT.IsA("vtkUnstructuredGrid"):
		input_PD = vtk.vtkGeometryFilter()					 
		input_PD.SetInputData(INPUT.VTKObject)	
		input_PD.Update()
		INPUT = dsa.WrapDataObject(input_PD.GetOutput())

	# Check if input is planar:
	Surface_Normals = vtk.vtkPolyDataNormals()
	Surface_Normals.SetInputData(INPUT.VTKObject)
	Surface_Normals.SetFeatureAngle(0)
	Surface_Normals.ComputeCellNormalsOn()
	Surface_Normals.ComputePointNormalsOn()
	Surface_Normals.AutoOrientNormalsOn()                                                                   # If not all the normals are pointing towards the same direction, it flips the normals of the reversed cells (so, all the normals are consistent on the entire surface)
	Surface_Normals.FlipNormalsOn()
	Surface_Normals.Update()
	INPUT = dsa.WrapDataObject(Surface_Normals.GetOutput())
		
	Surface_Normals = INPUT.CellData["Normals"]             #vertex_normal(INPUT)
	NBcells = INPUT.GetNumberOfCells()                        #INPUT.GetNumberOfPoints()
	N_a = round(Surface_Normals[0][0], 2)                                                                   # X-component of the point normal at ID=0. It is rounded to 2 decimals to ignore numerical errors       
	N_b = round(Surface_Normals[int(0.2*NBcells)][0], 2)                                                      # Take some control points at different ID's
	N_c = round(Surface_Normals[int(0.4*NBcells)][0], 2)
	N_d = round(Surface_Normals[int(0.6*NBcells)][0], 2)
	N_e = round(Surface_Normals[int(0.8*NBcells)][0], 2)
	N_f = round(Surface_Normals[NBcells-1][0], 2)
	if N_a!=N_b or N_a!=N_c or N_a!=N_d or N_a!=N_e or N_a!=N_f:                                            # Compare the X-component of the normal at the control points. If all are equal, then the input data is a plane
		DataFormatCheck = 0 
		print ('ERROR: Epsilon Poisson Solver can only process planar data')	
	else:
		DataFormatCheck = 1	
    
	# Get the input mesh: 
	WakeWithoutArrays=vtk.vtkPassArrays()																	# Delete all the arrays already contained in the input
	WakeWithoutArrays.SetInputDataObject(INPUT.VTKObject)
	WakeWithoutArrays.ClearArrays()
	WakeWithoutArrays.AddPointDataArray("")
	WakeWithoutArrays.AddCellDataArray("")
	WakeWithoutArrays.AddFieldDataArray("")
	WakeWithoutArrays.Update()



############################################################################################################
############################################# LOADING INPUT DATA ###########################################
############################################################################################################

if Select_Action >= 2 and DataFormatCheck == 1:																# Proceed to Poisson solver

	Sref = INPUT.FieldData["Sref"]
	TRANSFORMATION = INPUT.FieldData["TRANSFORMATION"]
	y = INPUT.Points[:,1]									   												# Wake slice Y-coordinates	   
	z = INPUT.Points[:,2]									   												# Wake slice Z-coordinates 
	Zita = INPUT.PointData["AERO_Vorticity"][:,0]		 													# x-vorticity (aero reference frame)
	Vector_Speed = INPUT.PointData["AERO_Velocity_Vector"]		 											# Velocity vector in the mesh reference frame
	u = INPUT.PointData["AERO_u"]							   												# x-velocity component
	V0 = INPUT.FieldData["V0"]																				# Upstream velocity 
	Ps = INPUT.PointData["AERO_Ps"]							 												# Static pressure
	Ps0 = INPUT.FieldData["Ps0"]						   													# Static pressure upstream
	Pt = INPUT.PointData["AERO_Pt"]							 												# Total pressure
	Pt0 = INPUT.FieldData["Pt0"]							   												# Total pressure upstream
	PtPt0 = INPUT.PointData["AERO_Pt/Pt0"]					  												# Total pressure ratio 
	Tt = INPUT.PointData["AERO_Tt"]							 												# Total temperature
	Tt0 = INPUT.FieldData["Tt0"]							   												# Total temperature upstream
	Rho = INPUT.PointData["AERO_Rho"]						   												# Density
	Rho0 = INPUT.FieldData["Rho0"]							  												# Upstream density
	Rho_star = INPUT.PointData["AGUIRRE_AERO_Rho_star"] 													# Isentropic density
	M0 = INPUT.FieldData["M0"]																				# Upstream Mach number
	AdimF = INPUT.FieldData["AdimF"]																		# Force adimensionalization factor 
	AdimP = INPUT.FieldData["AdimP"]																		# Power adimensionalization factor 

	if VelocityPotential > 0:
		if INPUT.GetPointData().HasArray("AERO_V_gradient")==0:					   							# This check if "AERO_V_gradient" array does not exists in the input				
			print ("ERROR: 'AERO_V_gradient' array is missing at the input data")							# It means that AERO_V_gradient array does not exists
			print ("Velocity gradients are required to calculate the velocity potential") 
			print ("SOLUTION: activate 'Near-field' method in Epsilon Aero Data")
		else:																								# If the array exists, then reads the array data directly from input
			du_dx = INPUT.PointData["AERO_V_gradient"][:,0,0]		  										# x-velocity gradient along x-axis (wind axes)
			dv_dy = INPUT.PointData["AERO_V_gradient"][:,1,1]		  										# y-velocity gradient along y-axis (wind axes)
			dw_dz = INPUT.PointData["AERO_V_gradient"][:,2,2]		  										# z-velocity gradient along z-axis (wind axes)



############################################################################################################
######################################### CELL-BASED CALCULATIONS ##########################################
############################################################################################################

if Select_Action >= 2 and DataFormatCheck == 1:																# Proceed to Poisson solver

	# CALCULATES CELL AREA:
	area = algs.area(INPUT)																					# Cell-based array

	# FIND CELL CENTERS:
	cen = vtk.vtkCellCenters()
	cen.SetInputDataObject(INPUT.VTKObject)		 # Alternative: self.GetInputDataObject(0, 0)  
	cen.Update()
	C = dsa.WrapDataObject(cen.GetOutput()).Points  														# C=cell centers (Cell-based array)

	# GET SOME ARRAYS FROM THE INPUT IN VTK FORMAT:															# These arrays will be converted later from point data to cell data
	myArrays=vtk.vtkPassArrays()								 											# use pass arrays to extract a copy of the input data (only some specific arrays needs to be converted)
	myArrays.SetInputDataObject(INPUT.VTKObject)		
	myArrays.ClearArrays()
	myArrays.AddPointDataArray('AERO_Vorticity')
	if VelocityPotential > 0:   																			# Only adds velocity gradient if this array exists at the input
		myArrays.AddPointDataArray('AERO_V_gradient')
	myArrays.AddCellDataArray('')
	myArrays.AddFieldDataArray('')
	myArrays.Update()

	# POINT DATA TO CELL DATA CONVERSION:
	p2c = vtk.vtkPointDataToCellData()			  				
	p2c.SetInputConnection(myArrays.GetOutputPort())
	p2c.Update()

	# CONVERT THE CELL-BASED VTK ARRAYS INTO NUMPY ARRAYS:
	Wrapped = dsa.WrapDataObject(p2c.GetOutput()) 															# Get the cell data
	Zita_cell = Wrapped.CellData["AERO_Vorticity"][:,0]

	if VelocityPotential > 0:   
		V_grad_cell = Wrapped.CellData["AERO_V_gradient"]

	# CALCULATING THE SOURCE TERM:
	if VelocityPotential==0:
		Sigma = 0*Zita																						# A zero array is created
		Sigma_cell = 0*Zita_cell
	if VelocityPotential==1:
		Sigma = -1*du_dx
		Sigma_cell = -1* V_grad_cell[:,0,0]
	if VelocityPotential==2:
		Sigma = dv_dy + dw_dz
		Sigma_cell =  V_grad_cell[:,1,1]+V_grad_cell[:,2,2]



############################################################################################################
######################################### SOLVING POISSON EQUATION #########################################
############################################################################################################

##############################################
########### GREEN FUNCTION SOLVER ############
##############################################

if Select_Action >= 2 and DataFormatCheck == 1:																# Proceed to Poisson solver
	
	if Method == 0:																							# 0=Green functions

		if Select_Action == 2:   																			# 2=Display walls
		
			if Case==0:																						# Free-air/Full wake
				self.GetOutput().ShallowCopy(INPUT.VTKObject)
				print ("No walls/symmetry plane to show")
				
			if Case==1:	 																					# Free-air/Wall or Symmetry at XZ plane --> the Green function takes into account the mirror domain
				# CREATE LINE AT THE XZ PLANE:
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data
				XZplane = Line_Function(XminData,Yw,ZminData-(ZmaxData-ZminData),XminData,Yw,ZmaxData+(ZmaxData-ZminData),10)  		# Draw a line representing the symmetry or wall plane
				
				# COMBINE THE XZplane AND THE WAKE INTO A SINGLE VARIABLE	
				XZwall = vtk.vtkAppendFilter()							  									# Output: vtkUnstructuredGrid
				XZwall.AddInputData(XZplane.GetOutput())
				XZwall.AddInputData(WakeWithoutArrays.GetOutput())
				XZwall.Update()		
				
				# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
				XZwall_VTP = vtk.vtkGeometryFilter()								  						# Converts vtkMultiblockDataSet to vtkPolyData
				XZwall_VTP.SetInputData(XZwall.GetOutput())							 
				XZwall_VTP.Update()
				
				self.GetOutput().ShallowCopy(XZwall_VTP.GetOutput())		
				print ("XZ wall/symmetry plane created successfully")
				
			if Case==2:	 																					# Free-air/Wall or Symmetry at XY plane
				# CREATE LINE AT THE XY PLANE:
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data
				XYplane = Line_Function(XminData,YminData-(YmaxData-YminData),Zw,XminData,YmaxData+(YmaxData-YminData),Zw,10)  		# Draw a line representing the symmetry or wall plane
				
				# COMBINE THE XZplane AND THE WAKE INTO A SINGLE VARIABLE	
				XYwall = vtk.vtkAppendFilter()							  									# Output: vtkUnstructuredGrid
				XYwall.AddInputData(XYplane.GetOutput())
				XYwall.AddInputData(WakeWithoutArrays.GetOutput())
				XYwall.Update()		
				
				# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
				XYwall_VTP = vtk.vtkGeometryFilter()								  						# Converts vtkMultiblockDataSet to vtkPolyData
				XYwall_VTP.SetInputData(XYwall.GetOutput())							 
				XYwall_VTP.Update()
				
				self.GetOutput().ShallowCopy(XYwall_VTP.GetOutput())		
				print ("XY wall/symmetry plane created successfully")

			if Case==3:	 																					# Free-air/Double wall or symmetry at XY and XZ planes
				# CREATE LINE AT THE XZ PLANE:
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data
				XZplane = Line_Function(XminData,Yw,Zw,XminData,Yw,ZmaxData+(ZmaxData-ZminData),10)  		# Draw a line representing the symmetry or wall plane
				
				# CREATE LINE AT THE XY PLANE:
				if 0.5*(YmaxData+YminData) > Yw:
					XYplane = Line_Function(XminData,Yw,Zw,XminData,YmaxData+(YmaxData-YminData),Zw,10)	 # Draw a line representing the symmetry or wall plane
				if 0.5*(YmaxData+YminData) < Yw:
					XYplane = Line_Function(XminData,YminData-(YmaxData-YminData),Zw,XminData,Yw,Zw,10)	 # Draw a line representing the symmetry or wall plane
							
				# COMBINE THE XZplane AND THE WAKE INTO A SINGLE VARIABLE	
				XYZwall = vtk.vtkAppendFilter()							  									# Output: vtkUnstructuredGrid
				XYZwall.AddInputData(XZplane.GetOutput())
				XYZwall.AddInputData(XYplane.GetOutput())
				XYZwall.AddInputData(WakeWithoutArrays.GetOutput())
				XYZwall.Update()		
				
				# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
				XYZwall_VTP = vtk.vtkGeometryFilter()								  						# Converts vtkMultiblockDataSet to vtkPolyData
				XYZwall_VTP.SetInputData(XYZwall.GetOutput())							 
				XYZwall_VTP.Update()
				
				self.GetOutput().ShallowCopy(XYZwall_VTP.GetOutput())		
				print ("XY/XZ walls/symmetry planes created successfully")
		
			if Case==4 or Case==5 or Case==6:   															# Wind tunnel
		
				# TUNNEL WALL EDGES (Defined in counter-clockwise direction)
				Tunnel_Floor = Line_Function(C[0,0],Tunnel_Y[0],Tunnel_Z[0],C[0,0],Tunnel_Y[1],Tunnel_Z[0],10)
				Tunnel_Right = Line_Function(C[0,0],Tunnel_Y[1],Tunnel_Z[0],C[0,0],Tunnel_Y[1],Tunnel_Z[1],10)
				Tunnel_Ceiling = Line_Function(C[0,0],Tunnel_Y[1],Tunnel_Z[1],C[0,0],Tunnel_Y[0],Tunnel_Z[1],10)
				Tunnel_Left = Line_Function(C[0,0],Tunnel_Y[0],Tunnel_Z[1],C[0,0],Tunnel_Y[0],Tunnel_Z[0],10)

				# COMBINE THE WAKE AND EDGES INTO A SINGLE VARIABLE
				WindTunnel = vtk.vtkAppendFilter()							  								# Output: vtkUnstructuredGrid
				WindTunnel.AddInputData(Tunnel_Floor.GetOutput())
				WindTunnel.AddInputData(Tunnel_Right.GetOutput())
				WindTunnel.AddInputData(Tunnel_Ceiling.GetOutput())
				WindTunnel.AddInputData(Tunnel_Left.GetOutput())
				WindTunnel.AddInputData(WakeWithoutArrays.GetOutput())
				WindTunnel.Update()
				
				# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
				WindTunnel_VTP = vtk.vtkGeometryFilter()									  				# Converts vtkMultiblockDataSet to vtkPolyData
				WindTunnel_VTP.SetInputData(WindTunnel.GetOutput())							 
				WindTunnel_VTP.Update()

				self.GetOutput().ShallowCopy(WindTunnel_VTP.GetOutput())		
				print ("Wind tunnel walls created successfully")	
		
		if Select_Action == 3:   																			#3=Full plane Phi/Psi solution

			"""	
			# IMPORTANT : CASES 0, 1, 2 AND 3 ARE KEPT HERE FOR DEBUGGIN PURPOSES BUT THEY ARE NOT EXPOSED IN THE OUTPUT NOR ACCESIBLE THROUGH THE GUI            
			if Case==0:																						# Free-air/Full wake	
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data
		
				# GENERATE GRID PLANE:
				XXX = [C[0,0]]
				YYY = np.linspace(YminData-(YmaxData-YminData), YmaxData+(YmaxData-YminData), Tunnel_Y[2])
				ZZZ = np.linspace(ZminData-(ZmaxData-ZminData), ZmaxData+(ZmaxData-ZminData), Tunnel_Z[2])
				PLANE = CREATE_GRID(XXX,YYY,ZZZ)
						
				# GET GRID POINT COORDINATES:
				GridPoints = dsa.WrapDataObject(PLANE.GetOutput())
				POINTSgrid = GridPoints.Points										   						# Get points coordinates
				Xg = POINTSgrid[:,0] 
				Yg = POINTSgrid[:,1] 
				Zg = POINTSgrid[:,2]
		
				# COMPUTE GREEN FUNCTION:
				Psi  = np.zeros(Xg.size)   																	# Stream function initialized as zero everywhere
				Phi  = np.zeros(Xg.size)   																	# Potential function initialized as zero everywhere	
				for i in range(np.size(Psi)):	 															# Iterates at each point of the mesh and calculates there the Green function. This function changes depending if the wake data correspond to a half model or full model
					Logarithm = GREEN(i,Yg,Zg,C,1,1,1,Yw,Zw)
					Psi[i] =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))
					if VelocityPotential>0:
						Phi[i] =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))		

				self.GetOutput().ShallowCopy(PLANE.GetOutput())
				addfielddata(1, output.GetPointData(),Psi,"Stream_Function")
				
			if Case==1:	 																					# Free-air/Wall or Symmetry at XZ plane --> the Green function takes into account the mirror domain
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data
				
				# GENERATE GRID PLANE:
				XXX = [C[0,0]]
				if 0.5*(YmaxData+YminData) > Yw:
					YYY = np.linspace(Yw, YmaxData+(YmaxData-YminData), Tunnel_Y[2])
				if 0.5*(YmaxData+YminData) < Yw:
					YYY = np.linspace(YminData-(YmaxData-YminData), Yw, Tunnel_Y[2])
				ZZZ = np.linspace(ZminData-(ZmaxData-ZminData), ZmaxData+(ZmaxData-ZminData), Tunnel_Z[2])
				PLANE = CREATE_GRID(XXX,YYY,ZZZ)
						
				# GET GRID POINT COORDINATES:
				GridPoints = dsa.WrapDataObject(PLANE.GetOutput())
				POINTSgrid = GridPoints.Points										   						# Get points coordinates
				Xg = POINTSgrid[:,0] 
				Yg = POINTSgrid[:,1] 
				Zg = POINTSgrid[:,2]			
				
				# COMPUTE GREEN FUNCTION:
				Psi  = np.zeros(Xg.size)   																	# Stream function initialized as zero everywhere
				Phi  = np.zeros(Xg.size)   																	# Potential function initialized as zero everywhere	
				for i in range(np.size(Psi)):	 															# Iterates at each point of the mesh and calculates there the Green function. This function changes depending if the wake data correspond to a half model or full model
					Logarithm = GREEN(i,Yg,Zg,C,1,1,1,Yw,Zw) * GREEN(i,Yg,Zg,C,-1,1,-1,Yw,Zw)
					Psi[i] =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))
					if VelocityPotential>0:
						Phi[i] =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))		

				self.GetOutput().ShallowCopy(PLANE.GetOutput())
				addfielddata(1, output.GetPointData(),Psi,"Stream_Function")

			if Case==2:	 																					# Free-air/Wall or Symmetry at XY plane
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data

				# GENERATE GRID PLANE:
				XXX = [C[0,0]]
				YYY = np.linspace(YminData-(YmaxData-YminData), YmaxData+(YmaxData-YminData), Tunnel_Y[2])
				ZZZ = np.linspace(Zw, ZmaxData+(ZmaxData-ZminData), Tunnel_Z[2])
				PLANE = CREATE_GRID(XXX,YYY,ZZZ)
						
				# GET GRID POINT COORDINATES:
				GridPoints = dsa.WrapDataObject(PLANE.GetOutput())
				POINTSgrid = GridPoints.Points										   						# Get points coordinates
				Xg = POINTSgrid[:,0] 
				Yg = POINTSgrid[:,1] 
				Zg = POINTSgrid[:,2]
		
				# COMPUTE GREEN FUNCTION:
				Psi  = np.zeros(Xg.size)   																	# Stream function initialized as zero everywhere
				Phi  = np.zeros(Xg.size)   																	# Potential function initialized as zero everywhere	
				for i in range(np.size(Psi)):	 															# Iterates at each point of the mesh and calculates there the Green function. This function changes depending if the wake data correspond to a half model or full model
					Logarithm = GREEN(i,Yg,Zg,C,1,1,1,Yw,Zw) * GREEN(i,Yg,Zg,C,1,-1,-1,Yw,Zw)
					Psi[i] =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))
					if VelocityPotential>0:
						Phi[i] =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))		

				self.GetOutput().ShallowCopy(PLANE.GetOutput())
				addfielddata(1, output.GetPointData(),Psi,"Stream_Function")

			if Case==3:	 																					# Free-air/Double wall or symmetry at XY and XZ planes
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data

				# GENERATE GRID PLANE:
				XXX = [C[0,0]]
				if 0.5*(YmaxData+YminData) > Yw:
					YYY = np.linspace(Yw, YmaxData+(YmaxData-YminData), Tunnel_Y[2])
				if 0.5*(YmaxData+YminData) < Yw:
					YYY = np.linspace(YminData-(YmaxData-YminData), Yw, Tunnel_Y[2])
				ZZZ = np.linspace(Zw, ZmaxData+(ZmaxData-ZminData), Tunnel_Z[2])
				PLANE = CREATE_GRID(XXX,YYY,ZZZ)
						
				# GET GRID POINT COORDINATES:
				GridPoints = dsa.WrapDataObject(PLANE.GetOutput())
				POINTSgrid = GridPoints.Points										   						# Get points coordinates
				Xg = POINTSgrid[:,0] 
				Yg = POINTSgrid[:,1] 
				Zg = POINTSgrid[:,2]
		
				# COMPUTE GREEN FUNCTION:
				Psi  = np.zeros(Xg.size)   																	# Stream function initialized as zero everywhere
				Phi  = np.zeros(Xg.size)   																	# Potential function initialized as zero everywhere	
				for i in range(np.size(Psi)):	 															# Iterates at each point of the mesh and calculates there the Green function. This function changes depending if the wake data correspond to a half model or full model
					Logarithm = GREEN(i,Yg,Zg,C,1,1,1,Yw,Zw) * GREEN(i,Yg,Zg,C,-1,-1,1,Yw,Zw) * GREEN(i,Yg,Zg,C,-1,1,-1,Yw,Zw) * GREEN(i,Yg,Zg,C,1,-1,-1,Yw,Zw)
					Psi[i] =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))
					if VelocityPotential>0:
						Phi[i] =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))		

				self.GetOutput().ShallowCopy(PLANE.GetOutput())
				addfielddata(1, output.GetPointData(),Psi,"Stream_Function")

			"""
			if Case==4 or Case==5 or Case==6:   															# Wind tunnel cases: 4=Full model   5=Half model Y+   6=Half model Y-
				XminData,XmaxData,YminData,YmaxData,ZminData,ZmaxData = GET_DATA_BOUNDS(WakeWithoutArrays)  # Get the bound limits coordinates of the box sourrounding the wake data
		
				# GENERATE GRID PLANE:
				XXX = [C[0,0]]
				YYY = np.linspace(Tunnel_Y[0], Tunnel_Y[1], Tunnel_Y[2])
				ZZZ = np.linspace(Tunnel_Z[0], Tunnel_Z[1], Tunnel_Z[2])
				PLANE = CREATE_GRID(XXX,YYY,ZZZ)
						
				# GET GRID POINT COORDINATES:
				GridPoints = dsa.WrapDataObject(PLANE.GetOutput())
				POINTSgrid = GridPoints.Points										   						# Get points coordinates
				Xg = POINTSgrid[:,0] 
				Yg = POINTSgrid[:,1] 
				Zg = POINTSgrid[:,2]
		
				# CREATE EXPONENTS MATRIX:  																# Exponents matrix (A matrix with alternated +1 and -1 --> defines the exponent of the logarithm of each image system)
				if Case==4:   																				# Case 4=Full model
					ny = 2*Layers+1                                                                         # Size of the matrices
					nz = 2*Layers+1			
					E = -1*np.ones((nz,ny), np.int32)                                                       # Exponent matrix
					E[::2,::2] = 1
					E[1::2,1::2] = 1
					
				if Case==5:   																				# Case 5=Half model Y+
					ny = 2*Layers+2
					nz = 2*Layers+1			
					E = 1*np.ones((nz,ny), np.int32)
					E[::2,::2] = -1
					E[1::2,1::2] = -1	
			
				if Case==6:   																				# Case 6=Half model Y-
					ny = 2*Layers+2
					nz = 2*Layers+1			
					E = -1*np.ones((nz,ny), np.int32)
					E[::2,::2] = 1
					E[1::2,1::2] = 1
				
				# CREATES IMAGE LAYER NUMBER MATRICES Ny/Nz:												# Matrix with the Y- and Z-layer number for each image system	
				Ny = []
				Nz = []
				
				if Case==4:   																				# Case 4=Full model
					for i in range(2*Layers+1):
						Ny.append(np.arange(-Layers,Layers+1))
					Ny = np.array(Ny)   																	# Matrix with the Y-layer number for each image system

					for i in np.arange(-Layers,Layers+1)[::-1]:
						Nz.append(i*np.ones(2*Layers+1))
					Nz = np.array(Nz) 
			
				if Case==5:   																				# Case 5=Half model Y+
					for i in range(2*Layers+1):
						Ny.append(np.arange(-Layers-1,Layers+1))
					Ny = np.array(Ny)   														
					
					for i in np.arange(-Layers,Layers+1)[::-1]:
						Nz.append(i*np.ones(2*Layers+2))
					Nz = np.array(Nz) 
					
				if Case==6:   																				# Case 6=Half model Y-
					for i in range(2*Layers+1):
						Ny.append(np.arange(-Layers,Layers+2))
					Ny = np.array(Ny)   														
					
					for i in np.arange(-Layers,Layers+1)[::-1]:
						Nz.append(i*np.ones(2*Layers+2))
					Nz = np.array(Nz) 				
				
				# COMPUTES TEST SECTION SIZE:
				W = Tunnel_Y[1]-Tunnel_Y[0]   																# Test section width
				H = Tunnel_Z[1]-Tunnel_Z[0]   																# Test section height	
				
				# COMPUTES TUNNEL REFERENCE SYSTEM OFFSET:
				DeltaZ = 0.5*(Tunnel_Z[1]+Tunnel_Z[0])
				if Case==4:
					DeltaY = 0.5*(Tunnel_Y[1]+Tunnel_Y[0])				
				if Case==5 or Case==6:
					DeltaY = 0
				
				# COMPUTE GREEN FUNCTION:			
				Psi  = np.zeros(Xg.size)   																	# Stream function initialized as zero everywhere
				Phi  = np.zeros(Xg.size)   																	# Potential function initialized as zero everywhere 			
				for i in range(np.size(Psi)):	 															# Iterates at each point "i" of the mesh and calculates there the Green function. This function changes depending if the wake data correspond to a half model or full model			
					Logarithm = []
					for j in range(E.flatten().size):   													# Iterates at each image system "j" 
						LOGA = GREEN_WTT(i,Yg,Zg,C,Ny.flatten()[j],Nz.flatten()[j],W,H,DeltaY,DeltaZ,E.flatten()[j],Case)  # Calculates the distance[(y-yc)^2+(z-zc)^2]^E from each wake cell of the image "j" to the point "i" --> output: an array with the same lenght than the amount of wake cells
						Logarithm.append(LOGA)
					Logarithm = np.prod(Logarithm,axis=0)
					Psi[i] =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))

					if VelocityPotential>0:
						Phi[i] =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:])) 

				# Control point:
				Logarithm_CTRL = []
				for j in range(E.flatten().size):   													# Iterates at each image system "j"                     
					LOGA_CTRL = GREEN_WTT(0,[Tunnel_Y[1]],[Tunnel_Z[1]],C,Ny.flatten()[j],Nz.flatten()[j],W,H,DeltaY,DeltaZ,E.flatten()[j],Case)  # Calculates the distance[(y-yc)^2+(z-zc)^2]^E from each wake cell to the control point lying at the WTT bounbdary
					Logarithm_CTRL.append(LOGA_CTRL)                       
				Logarithm_CTRL = np.prod(Logarithm_CTRL,axis=0) 
				
				Psi_CTRL =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm_CTRL))                    
				Psi = Psi-Psi_CTRL
				if VelocityPotential>0:
					Phi_CTRL =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm_CTRL))                    
					Phi = Phi-Phi_CTRL
                
				self.GetOutput().ShallowCopy(PLANE.GetOutput())
                
                
		if Select_Action == 4:   																					#4=Wake calculations	
		
			print ("Starting Poisson solver...")
			start = time.time()											 									# This is ised to measure the required time to solve the Poisson equation

			if Case==4 or Case==5 or Case==6:																# Wind tunnel cases: 4=Full model   5=Half model Y+   6=Half model Y-
				# CREATE EXPONENTS MATRIX:  																# Exponents matrix (A matrix with alternated +1 and -1 --> defines the exponent of the logarithm of each image system)
				if Case==4:   																				# Case 4=Full model
					ny = 2*Layers+1
					nz = 2*Layers+1			
					E = -1*np.ones((nz,ny), np.int32)
					E[::2,::2] = 1
					E[1::2,1::2] = 1
					
				if Case==5:   																				# Case 5=Half model Y+
					ny = 2*Layers+2
					nz = 2*Layers+1			
					E = 1*np.ones((nz,ny), np.int32)
					E[::2,::2] = -1
					E[1::2,1::2] = -1	
			
				if Case==6:   																				# Case 6=Half model Y-
					ny = 2*Layers+2
					nz = 2*Layers+1			
					E = -1*np.ones((nz,ny), np.int32)
					E[::2,::2] = 1
					E[1::2,1::2] = 1
				
				# CREATES IMAGE LAYER NUMBER MATRICES Ny/Nz:												# Matrix with the Y- and Z-layer number for each image system	
				Ny = []
				Nz = []
				
				if Case==4:   																				# Case 4=Full model
					for i in range(2*Layers+1):
						Ny.append(np.arange(-Layers,Layers+1))
					Ny = np.array(Ny)   																	# Matrix with the Y-layer number for each image system
					
					for i in np.arange(-Layers,Layers+1)[::-1]:
						Nz.append(i*np.ones(2*Layers+1))
					Nz = np.array(Nz) 
				
				if Case==5:   																				# Case 5=Half model Y+
					for i in range(2*Layers+1):
						Ny.append(np.arange(-Layers-1,Layers+1))
					Ny = np.array(Ny)   														
					
					for i in np.arange(-Layers,Layers+1)[::-1]:
						Nz.append(i*np.ones(2*Layers+2))
					Nz = np.array(Nz) 
					
				if Case==6:   																				# Case 6=Half model Y-
					for i in range(2*Layers+1):
						Ny.append(np.arange(-Layers,Layers+2))
					Ny = np.array(Ny)   														
					
					for i in np.arange(-Layers,Layers+1)[::-1]:
						Nz.append(i*np.ones(2*Layers+2))
					Nz = np.array(Nz) 		
				
				# COMPUTES TEST SECTION SIZE:
				W = Tunnel_Y[1]-Tunnel_Y[0]   																# Test section width
				H = Tunnel_Z[1]-Tunnel_Z[0]   																# Test section height	
				
				# COMPUTES TUNNEL REFERENCE SYSTEM OFFSET:
				DeltaZ = 0.5*(Tunnel_Z[1]+Tunnel_Z[0])
				if Case==4:
					DeltaY = 0.5*(Tunnel_Y[1]+Tunnel_Y[0])				
				if Case==5 or Case==6:
					DeltaY = 0

			Psi  = np.zeros(Zita.size)   																	# Stream function initialized as zero everywhere
			Phi  = np.zeros(Zita.size)   																	# Potential function initialized as zero everywhere	
			for i in range(np.size(Psi)):	 																# Iterates at each point of the mesh and calculates there the Green function. This function changes depending if the wake data correspond to a half model or full model
				if Case==0:																					# Free-air/Full wake
					Logarithm = GREEN(i,y,z,C,1,1,1,Yw,Zw)
				if Case==1:	 																				# Free-air/Wall or Symmetry at XZ plane --> the Green function takes into account the mirror domain
					Logarithm = GREEN(i,y,z,C,1,1,1,Yw,Zw) * GREEN(i,y,z,C,-1,1,-1,Yw,Zw)
				if Case==2:	 																				# Free-air/Wall or Symmetry at XY plane
					Logarithm = GREEN(i,y,z,C,1,1,1,Yw,Zw) * GREEN(i,y,z,C,1,-1,-1,Yw,Zw)
				if Case==3:	 																				# Free-air/Double wall or symmetry at XY and XZ planes
					Logarithm = GREEN(i,y,z,C,1,1,1,Yw,Zw) * GREEN(i,y,z,C,-1,-1,1,Yw,Zw) * GREEN(i,y,z,C,-1,1,-1,Yw,Zw) * GREEN(i,y,z,C,1,-1,-1,Yw,Zw)
				if Case==4 or Case==5 or Case==6:															# Wind tunnel cases: 4=Full model   5=Half model Y+   6=Half model Y-
					Logarithm = []
					for j in range(E.flatten().size):                                                       # Iterates foe each mirror singularity			
						LOGA = GREEN_WTT(i,y,z,C,Ny.flatten()[j],Nz.flatten()[j],W,H,DeltaY,DeltaZ,E.flatten()[j],Case)
						Logarithm.append(LOGA)
					Logarithm = np.prod(Logarithm,axis=0)

				Psi[i] =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))
				if VelocityPotential>0:
					Phi[i] =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm[:]))		

			if Case==4 or Case==5 or Case==6:
				# Control point:
				Logarithm_CTRL = []
				for j in range(E.flatten().size):   													# Iterates at each image system "j"                     
					LOGA_CTRL = GREEN_WTT(0,[Tunnel_Y[0]],[Tunnel_Z[0]],C,Ny.flatten()[j],Nz.flatten()[j],W,H,DeltaY,DeltaZ,E.flatten()[j],Case)  # Calculates the distance[(y-yc)^2+(z-zc)^2]^E from each wake cell to the control point lying at the WTT bounbdary
					Logarithm_CTRL.append(LOGA_CTRL)                       
				Logarithm_CTRL = np.prod(Logarithm_CTRL,axis=0) 
				
				Psi_CTRL =  np.sum( -Zita_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm_CTRL))                    
				Psi = Psi-Psi_CTRL
				if VelocityPotential>0:
					Phi_CTRL =  np.sum( Sigma_cell[:]*area[:]*(0.25/np.pi)*np.log(Logarithm_CTRL))                    
					Phi = Phi-Phi_CTRL


			# Printing the time required to solve the Poisson equation:
			end = time.time()
			SolveTime = end - start
			print ("Calculation time", SolveTime, " Seconds (", SolveTime/60.0, " minutes)" ) 
			print ("")
			
			self.GetOutput().ShallowCopy(WakeWithoutArrays.GetOutput())
			#addfielddata(1, output.GetPointData(),Psi,"Stream_Function")
			


##############################################
########### SINGULARITIES METHOD  ############
##############################################

if Select_Action >= 2 and DataFormatCheck == 1:																# Proceed to Poisson solver

	if Method == 1:																							# 1=Singularities
		
		# TUNNEL WALL EDGES (Defined in counter-clockwise direction)
		Tunnel_Floor = Line_Function(C[0,0],Tunnel_Y[0],Tunnel_Z[0],C[0,0],Tunnel_Y[1],Tunnel_Z[0],Tunnel_Y[2])
		Tunnel_Right = Line_Function(C[0,0],Tunnel_Y[1],Tunnel_Z[0],C[0,0],Tunnel_Y[1],Tunnel_Z[1],Tunnel_Z[2])
		Tunnel_Ceiling = Line_Function(C[0,0],Tunnel_Y[1],Tunnel_Z[1],C[0,0],Tunnel_Y[0],Tunnel_Z[1],Tunnel_Y[2])
		Tunnel_Left = Line_Function(C[0,0],Tunnel_Y[0],Tunnel_Z[1],C[0,0],Tunnel_Y[0],Tunnel_Z[0],Tunnel_Z[2])

		# COMBINE THE EDGES INTO A SINGLE VARIABLE
		WindTunnel = vtk.vtkAppendFilter()							  										# Output: vtkUnstructuredGrid
		WindTunnel.AddInputData(Tunnel_Floor.GetOutput())
		WindTunnel.AddInputData(Tunnel_Right.GetOutput())
		WindTunnel.AddInputData(Tunnel_Ceiling.GetOutput())
		WindTunnel.AddInputData(Tunnel_Left.GetOutput())
		WindTunnel.Update()
		
		# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
		WindTunnel_VTP = vtk.vtkGeometryFilter()									  						# Converts vtkMultiblockDataSet to vtkPolyData
		WindTunnel_VTP.SetInputData(WindTunnel.GetOutput())							 
		WindTunnel_VTP.Update()


		if Select_Action == 2: 																						# 2=Display tunnel wall panels
		
			# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
			PDpoints = vtk.vtkPolyData()
			PDpoints.SetPoints(WindTunnel_VTP.GetOutput().GetPoints())

			Survey_Vertices = vtk.vtkVertexGlyphFilter()
			Survey_Vertices.AddInputData(PDpoints)
			Survey_Vertices.Update()
			
			# COMBINE THE EDGES, THE VERTICES AND THE WAKE INTO A SINGLE VARIABLE	
			WTpanels = vtk.vtkAppendFilter()							  									# Output: vtkUnstructuredGrid
			WTpanels.AddInputData(WindTunnel.GetOutput())
			WTpanels.AddInputData(Survey_Vertices.GetOutput())
			WTpanels.AddInputData(WakeWithoutArrays.GetOutput())
			WTpanels.Update()		
			
			# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA:
			WTpanels_VTP = vtk.vtkGeometryFilter()								  							# Converts vtkMultiblockDataSet to vtkPolyData
			WTpanels_VTP.SetInputData(WTpanels.GetOutput())							 
			WTpanels_VTP.Update()
			
			self.GetOutput().ShallowCopy(WTpanels_VTP.GetOutput())		
			print ("Wall panels created successfully")

		if Select_Action >= 3: 																			# Step=3: Full plane calculations   Step=4: Wake calculations

			#  REMOVE DUPLICATED WALL PANEL POINTS
			WindTunnelClean = vtk.vtkCleanPolyData()
			WindTunnelClean.SetInputConnection(WindTunnel_VTP.GetOutputPort())
			WindTunnelClean.Update()

			# GET PANEL POINT COORDINATES:
			PanelPoints = dsa.WrapDataObject(WindTunnelClean.GetOutput())
			POINTS = PanelPoints.Points											 							# Get points coordinates
			Xp = POINTS[:,0] 
			Yp = POINTS[:,1] 
			Zp = POINTS[:,2]
			
			# GET PANEL CENTER COORDINATES:
			Xc = 0.5*(Xp[:-1]+Xp[1:])
			Yc = 0.5*(Yp[:-1]+Yp[1:])
			Zc = 0.5*(Zp[:-1]+Zp[1:])
			
			Xc = np.append(Xc,0.5*(Xp[0]+Xp[-1]))
			Yc = np.append(Yc,0.5*(Yp[0]+Yp[-1]))
			Zc = np.append(Zc,0.5*(Zp[0]+Zp[-1]))
			
			# CALCULATES WALL MATRIX:
			Matrix=[] 
			for i in range(len(Yp)): 
				Array = [] 
				for j in range(len(Yc)): 
					Logarithm = np.log((Yp[i]-Yc[j])**2+(Zp[i]-Zc[j])**2)
					Array.append(Logarithm)
				Matrix.append(Array) 
			m = np.matrix(Matrix)  																			# Converts the 2D numpy array into a numpy matrix format
			
			# COMPUTES THE SECOND MEMBER OF THE WALL STREAM FUNCTION:
			S = np.zeros(len(Yp))                                                                           # Second member for stream function
			Sphi = np.zeros(len(Yp))                                                                        # Second member for velocity potential
			for i in range(len(Yp)):
				S[i] = -1*np.sum(np.log((Yp[i]-C[:,1])**2+(Zp[i]-C[:,2])**2)*Zita_cell[:]*area[:])   
				Sphi[i] = -1*np.sum(np.log((Yp[i]-C[:,1])**2+(Zp[i]-C[:,2])**2)*Sigma_cell[:]*area[:])		

			# COMPUTES THE VORTICITY AT EACH PANEL:
			# sigma = numpy.linalg.solve(A, b)                                                          # Alternative solver
            #VORTEX = np.linalg.lstsq(m,S)[0]															# Backup method for solving linear matrix equations --> but it is not robust!   
			Vortex = np.linalg.pinv(m).dot(S)																# Note: the pseudo-inverse is the most suited numpy tool to solve this kind of matrix lineal problem (all the other methods failed)
			SourceSink = np.linalg.pinv(m).dot(Sphi)
			VORTEX = []																						# Converts "Vortex" array into a list ("VORTEX")
			SOURCE_SINK = []	
			for i in range(len(Yp)):
				VORTEX = np.append(VORTEX,Vortex[0,i])	
				SOURCE_SINK = np.append(SOURCE_SINK,SourceSink[0,i])	
			
			if Select_Action == 3:																					# Compute stream function solution for the full plane
				
				# GENERATE GRID PLANE:
				XXX = [C[0,0]]
				YYY = np.linspace(Tunnel_Y[0], Tunnel_Y[1], Tunnel_Y[2])
				ZZZ = np.linspace(Tunnel_Z[0], Tunnel_Z[1], Tunnel_Z[2])
				PLANE = CREATE_GRID(XXX,YYY,ZZZ)
						
				# GET GRID POINT COORDINATES:
				GridPoints = dsa.WrapDataObject(PLANE.GetOutput())
				POINTSgrid = GridPoints.Points										   						# Get points coordinates
				Xg = POINTSgrid[:,0] 
				Yg = POINTSgrid[:,1] 
				Zg = POINTSgrid[:,2]
				
				print ("Starting Poisson solver...")
				start = time.time()											 								# This is used to measure the required time to solve the Poisson equation
				
				# COMPUTE THE STREAM FUNCTION AND POTENTIAL FIELD FUNCTIONS AT EACH GRID POINT:
				Psi  = np.zeros(Xg.size)   																	# Stream function initialized as zero everywhere at each wake point
				Phi  = np.zeros(Xg.size)   																	# Potential function initialized as zero everywhere  at each wake point
				
				for i in range(np.size(Psi)): 
					VorticityContribution = np.sum( np.log((Yg[i]-C[:,1])**2+(Zg[i]-C[:,2])**2)*Zita_cell[:]*area[:])   #"Yg" and "Zg" are the point coordinates on the entire tunnel grid
					SingularityContribution = np.sum(VORTEX[:]*np.log((Yg[i]-Yc[:])**2+(Zg[i]-Zc[:])**2)	)
					Psi[i] =  -(0.25/np.pi)*(VorticityContribution+SingularityContribution)
					
					if VelocityPotential != 0:
						PhiContribution = np.sum( np.log((Yg[i]-C[:,1])**2+(Zg[i]-C[:,2])**2)*Sigma_cell[:]*area[:])   #"Yg" and "Zg" are the point coordinates on the entire tunnel grid
						SingularityContribution_Phi = np.sum(SOURCE_SINK[:]*np.log((Yg[i]-Yc[:])**2+(Zg[i]-Zc[:])**2)	)
						Phi[i] =  (0.25/np.pi)*(PhiContribution+SingularityContribution_Phi)
				
				# Printing the time required to solve the Poisson equation:
				end = time.time()
				SolveTime = end - start
				print ("Calculation time", SolveTime, " Seconds (", SolveTime/60.0, " minutes)" ) 
				print ("")

				self.GetOutput().ShallowCopy(PLANE.GetOutput())

                
			if Select_Action == 4:																					# Compute stream function solution for the wake only	

				print ("Starting Poisson solver...")
				start = time.time()											 								# This is used to measure the required time to solve the Poisson equation
				
				# COMPUTES THE STREAM FUNCTION AND POTENTIAL FIELD FUNCTIONS:
				Psi  = np.zeros(Zita.size)   																# Stream function initialized as zero everywhere at each wake point
				Phi  = np.zeros(Zita.size)   																# Potential function initialized as zero everywhere  at each wake point
				
				for i in range(np.size(Psi)): 
					VorticityContribution = np.sum( np.log((y[i]-C[:,1])**2+(z[i]-C[:,2])**2)*Zita_cell[:]*area[:])   #"y" and "z" are the wake point coordinates
					SingularityContribution = np.sum(VORTEX[:]*np.log((y[i]-Yc[:])**2+(z[i]-Zc[:])**2)	)
					Psi[i] =  -(0.25/np.pi)*(VorticityContribution+SingularityContribution)
					if VelocityPotential != 0:
						PhiContribution = np.sum( np.log((y[i]-C[:,1])**2+(z[i]-C[:,2])**2)*Sigma_cell[:]*area[:])   #"Yg" and "Zg" are the point coordinates on the entire tunnel grid
						SingularityContribution_Phi = np.sum(SOURCE_SINK[:]*np.log((y[i]-Yc[:])**2+(z[i]-Zc[:])**2)	)
						Phi[i] =  (0.25/np.pi)*(PhiContribution+SingularityContribution_Phi)

				# Printing the time required to solve the Poisson equation:
				end = time.time()
				SolveTime = end - start
				print ("Calculation time", SolveTime, " Seconds (", SolveTime/60.0, " minutes)" ) 
				print ("")
				
				self.GetOutput().ShallowCopy(WakeWithoutArrays.GetOutput())



############################################################################################################
########################################## VORTEX DRAG CALCULATION #########################################
############################################################################################################

if Select_Action >= 2 and DataFormatCheck == 1:																# Proceed to Poisson solver

	if Select_Action == 4:

		# Maskell method:
		MASKELL_vortex_drag = 0.5*Rho0*(Psi*Zita-Phi*Sigma)/AdimF
		
		# Aguirre method:
		Gamma=1.4
		ETA = 1+0.5*(Gamma-1)*M0**2*(1-((u**2+Psi*Zita-Phi*Sigma)*Tt0/(V0**2*Tt)))
		Raiz = 1-((ETA*(Pt/Pt0)**((Gamma-1)/Gamma)-1)/(0.5*(Gamma-1)*M0**2)) - ((Psi*Zita-Phi*Sigma)/V0**2) # Term inside the root of DeltaUstar
		Ustar = V0*algs.sqrt(Raiz)																			# Deficit of velocity related to the potential effect
		AGUIRRE_Ev_wake = 0.5*Rho*(Psi*Zita-Phi*Sigma)*Ustar/AdimP
		AGUIRRE_vortex_drag_wake = 0.5*Rho0*Psi*Zita*V0/AdimP
		
		where_are_NaNs = isnan(AGUIRRE_Ev_wake)																# Sometimes the ResampleWithDataSet fails to set data in some points, leading to NaN value in Ev_Maskell, avoiding its integration on the survey plane. This eliminates this problem.
		AGUIRRE_Ev_wake[where_are_NaNs] = 0

		where_are_NaNs = isnan(AGUIRRE_vortex_drag_wake)													# Sometimes the ResampleWithDataSet fails to set data in some points, leading to NaN value in Ev_Maskell, avoiding its integration on the survey plane. This eliminates this problem.
		AGUIRRE_vortex_drag_wake[where_are_NaNs] = 0

		# Select the arrays to integrate:
		addfielddata(1, WakeWithoutArrays.GetOutput().GetPointData(),AGUIRRE_Ev_wake,"AGUIRRE_Ev_wake")
		addfielddata(1, WakeWithoutArrays.GetOutput().GetPointData(),AGUIRRE_vortex_drag_wake,"AGUIRRE_vortex_drag_wake") 
		addfielddata(1, WakeWithoutArrays.GetOutput().GetPointData(),MASKELL_vortex_drag,"MASKELL_vortex_drag")
 
		# Integrate variables:
		integrate = vtk.vtkIntegrateAttributes()
		integrate.SetInputData(WakeWithoutArrays.GetOutput())
		integrate.Update()

		# Add name and values to the output variables:
		integrated = integrate.GetOutput()
		pd = integrated.GetPointData()

		print("# INTEGRATED VALUES:")
		for j in range(pd.GetNumberOfArrays()):  
			arrayName = pd.GetArray(j).GetName()
			IntegralValue = np.array(pd.GetArray(j).GetTuple(0))
			output.FieldData.append(IntegralValue, arrayName)
			print(str(arrayName),"=",IntegralValue[0])

	if Select_Action == 3 or Select_Action == 4:
		# Convert Psi and Phi into Numpy arrays:
		MASK1 = Psi!=inf
		MASK2 = Psi!=-inf
		Psi = 1*MASK1*MASK2*Psi

		where_are_NaNs = isnan(Psi)																			# Sometimes the ResampleWithDataSet fails to set data in some points, leading to NaN value in Ev_Maskell, avoiding its integration on the survey plane. This eliminates this problem.
		Psi[where_are_NaNs] = 0
		Psi = np.array(Psi)                                                                                 # Converts Psi into a numpy array
		if Select_Action == 3:
			addfielddata(1, PLANE.GetOutput().GetPointData(),Psi,"Psi")    #self
		if Select_Action == 4:
			addfielddata(1, WakeWithoutArrays.GetOutput().GetPointData(),Psi,"Psi")    #self

		if VelocityPotential != 0:
			where_are_NaNs = isnan(Phi)																			# Sometimes the ResampleWithDataSet fails to set data in some points, leading to NaN value in Ev_Maskell, avoiding its integration on the survey plane. This eliminates this problem.
			Phi[where_are_NaNs] = 0
			Phi = np.array(Phi)
		if Select_Action == 3:
			addfielddata(1, PLANE.GetOutput().GetPointData(),Phi,"Phi")    #self
		if Select_Action == 4:
			addfielddata(1, WakeWithoutArrays.GetOutput().GetPointData(),Phi,"Phi")    #self

		# Compute Psi and Phi gradients:
		gradientFilter = vtk.vtkGradientFilter()
		if Select_Action == 3:
			gradientFilter.SetInputData(PLANE.GetOutput())
		if Select_Action == 4:
			gradientFilter.SetInputData(WakeWithoutArrays.GetOutput())      
		gradientFilter.SetInputArrayToProcess(0,0,0,0,'Psi')
		gradientFilter.SetResultArrayName('Psi_gradient')
		gradientFilter.Update()
		WrapData = dsa.WrapDataObject(gradientFilter.GetOutput())
  
		if VelocityPotential != 0:
			gradientFilterPhi = vtk.vtkGradientFilter()
			gradientFilterPhi.SetInputData(gradientFilter.GetOutput())
			gradientFilterPhi.SetInputArrayToProcess(0,0,0,0,'Phi')
			gradientFilterPhi.SetResultArrayName('Phi_gradient')
			gradientFilterPhi.Update()
			WrapData = dsa.WrapDataObject(gradientFilterPhi.GetOutput())

		# Compute induced velocities:
		Psi_gradient = WrapData.PointData["Psi_gradient"]												    # Gradient of Psi in the mesh reference axes				

		Psi_gradient = np.einsum('ij,kj->ik',Psi_gradient,TRANSFORMATION)	 								# Psi gradient in aero reference fram (vector transformation: M.Psi_gradient)
		Psi_v = Psi_gradient[:,2]   																		# induced y-velocity component 
		Psi_w = -Psi_gradient[:,1]
		Psi_Vabs = np.sqrt(Psi_v**2+Psi_w**2)
		Velocity_Vector_Psi = algs.make_vector(0*Psi_v,Psi_v,Psi_w)                                         # In-plane rotational velocity vector
		induced_v = Psi_v 
		induced_w = Psi_w

		if VelocityPotential != 0:
			Phi_gradient = WrapData.PointData["Phi_gradient"]												    # Gradient of Psi in the mesh reference axes				
			Phi_gradient = np.einsum('ij,kj->ik',Phi_gradient,TRANSFORMATION)	 								# Psi gradient in aero reference fram (vector transformation: M.Psi_gradient)
			Phi_v = Phi_gradient[:,1]   																		# induced y-velocity component 
			Phi_w = Phi_gradient[:,2]
			Phi_Vabs = np.sqrt(Phi_v**2+Phi_w**2)
			Velocity_Vector_Phi = algs.make_vector(0*Phi_v,Phi_v,Phi_w)                                         # In-plane source/sink velocity vector
			induced_v = induced_v + Phi_v
			induced_w = induced_w + Phi_w
			Velocity_Vector_Psi_Phi = Velocity_Vector_Psi + Velocity_Vector_Phi

		induced_abs = np.sqrt(induced_v**2+induced_w**2)



############################################################################################################
###############################################  DATA OUTPUT ###############################################
############################################################################################################



if Select_Action > 2 and DataFormatCheck == 1:																# Proceed to Poisson solver

	output.PointData.append(Psi , 'POISSON_Psi')
	output.PointData.append(Velocity_Vector_Psi, "Velocity_Vector_Psi") 
    
	if VelocityPotential != 0:
		output.PointData.append(Phi , 'POISSON_Phi')
		output.PointData.append(Velocity_Vector_Phi, "Velocity_Vector_Phi")        
		output.PointData.append(Velocity_Vector_Psi_Phi, "Velocity_Vector_Psi_Phi")           
        
	if Select_Action == 4:        
		output.PointData.append(PtPt0, "AERO_Pt/Pt0")
		output.PointData.append(Ustar, "AERO_Ustar")
		output.PointData.append(AGUIRRE_Ev_wake, "AGUIRRE_Ev_wake")
		output.PointData.append(AGUIRRE_vortex_drag_wake, "AGUIRRE_vortex_drag_wake")	
		output.PointData.append(MASKELL_vortex_drag , 'MASKELL_vortex_drag')
		output.PointData.append(Vector_Speed, "Velocity_Vector")
		output.PointData.append(Zita, "POISSON_Vorticity_x")
		output.FieldData.append(Sref, "Sref")
		output.FieldData.append(TRANSFORMATION, "TRANSFORMATION")
        
		if VelocityPotential != 0:
			output.PointData.append(Sigma, "POISSON_Sigma")
			output.PointData.append(Phi*Sigma, "POISSON_Phi*Sigma")




EpsilonIntegrationTool = []
if Select_Action > 2 and DataFormatCheck == 1:																# Proceed to Poisson solver
	if Select_Action == 4:  
		EpsilonIntegrationTool.append(["AERO_Pt/Pt0","Survey_Plane"]) 
		EpsilonIntegrationTool.append(["AGUIRRE_Ev_wake","Survey_Plane"]) 
		EpsilonIntegrationTool.append(["AGUIRRE_vortex_drag_wake","Survey_Plane"]) 
		EpsilonIntegrationTool.append(["MASKELL_vortex_drag","Survey_Plane"]) 
		EpsilonIntegrationTool.append(["POISSON_Vorticity_x","Survey_Plane"]) 

# Send "EpsilonIntegrationTool" array to the output:
NB_of_Arrays = np.shape(EpsilonIntegrationTool)[0]

array = vtk.vtkStringArray()
array.SetName("ARRAY")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][0])
output.GetFieldData().AddArray(array)


array = vtk.vtkStringArray()
array.SetName("ARRAY_Integration")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][1])
output.GetFieldData().AddArray(array)


############################################################################################################
############################################  SAVE DATA AS VTP #############################################
############################################################################################################

if Save_Output_Data==1:
	WRITE_VTP(output,File_Path_Output)
	print ("# Output file written successfully")



############################################################################################################
###############################################  END OF CODE ###############################################
############################################################################################################