# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################                Wake survey Plugin                #############################
############################################################################################################

# Abstract:
This Paraview's programmable source provides several tools for wind tunnel wake survey systems:
a) Real-time data visulization during wind tunnel run
b) Third-party probe data visualization

# IMPORTANT: accepted file format for the multi-hole probe data is ".CSV" (other formats are accepted with data 
parsing).Required format for the mesh input data are Tecplot ".dat" (for raw PIV files) and VTP/VTU/VTS (for 
PIV data from Epsilon_PIV as well as CFD data slices previously created in Paraview). Note that VTP/VTU/VTS files 
from Paraview are only accepted and NOT vtm (vtk multiblock file). Hence, prior to saving a CFD slice data into 
vtp/vtu/vts format, a merge block filter must be applied (converts multiblock to single block). Moreover, if any 
other single block data format is being used in Paraview, the simplest way to convert this format into vtp is by 
using the filter "Extract Surface" (it converts any file format into a polygonal mesh --> vtp file).

# IMPORTANT: The accepted standard header names for the 5HP variables are: X,Y,Z,R,Vx,Vy,Vz,Vrad,Vtan,Ps,Pt, Ts and Tt. 
Headers in lowercase (i.e., x,y,z) or including units are NOT supported. However, any 5HP data format and header
name can be accepted by using a txt file parser.
 
# IMPORTANT: data must be in SI units: [m], [m/s] and [Pa], otherwise, the calculated vorticity and gradients 
will be wrong. Moreover, Epsilon_Aero_Data also requires input data in SI units. Therefore, if input data is 
in another unit, a unit conversion is required (use the unit convertion options at the bottom of the properties panel).

# IMPORTANT: the user must provide data already expressed in the Aerodynamic Reference System. Local reference 
system like the in-plane XY PIV reference system is not allowed or may lead to errors in the calculations. In order
to change the reference system of PIV data, use Epsilon_PIV. To change the reference system of CFD data, use Epsilon_Aero_Data

# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""


print ("")
print ("")
print ('#########################################################################################################################################')
print ('##################################                                                                     ##################################')
print ('##################################                              EPSILON 3.0                            ##################################')
print ('##################################                        Aerodynamic analysis tool                    ##################################')
print ('##################################                                                                     ##################################')
print ('#########################################################################################################################################')
print ("")
print ('Developed at ISAE-SUPAERO, France, 2022')
print ("")
print ('#DISCLAIMER:')
print ('This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee of any kind, either express')
print ('or implied. Its use for teaching, research and industrial applications is allowed under the strict responsibility of the user. The authors')
print ('of this plugin are not responsible for any damages or errors related to the use of the plugin.')

print ("")
print ("")
print ("")
print ("####################################################")
print ("###########      EPSILON WAKE SURVEY     ###########")
print ("####################################################")
print ("")

############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

import vtk
from vtk.numpy_interface import algorithms as algs
from vtk.numpy_interface import dataset_adapter as dsa
import numpy as np
import vtk.vtkFiltersParallel
import paraview.vtk.util.numpy_support as vnp
import glob
import os
import os.path
from vtk.util import numpy_support



############################################################################################################
############################################# STRING TO FLOAT ##############################################
############################################################################################################

# Input parameters are defined as string in Paraview GUI, thus a string to float convertion is required:
Mode = float(Mode)                                              # 0=None  1=Survey plane visualization   2=Survey line visualization    3=Third-party probe data visualization
ActionGUI = float(Action)   									# 0=None  1=Visualize input mesh    2=Visualize points    3=Visualize delaunay   4=Interpolate data
Visualization_ModeGUI = float(Visualization_Mode)               # 0=None  1=Line   2=Extruded line   3=Axisymmetric
MeshTypeGUI = float(Mesh_Type)                                  # 0=None  1=Mapped grid   2=Delaunay


if Mode == 1:
	Action = ActionGUI
	Visualization_Mode = 0
	MeshType  = 0
if Mode == 2:
	Action = 0
	Visualization_Mode = Visualization_ModeGUI
	MeshType  = 0
if Mode == 3:
	Action = 0
	Visualization_Mode = 0
	MeshType  = MeshTypeGUI



File_Path_Mesh = str(File_Path_Mesh)                        # File path to the input data (wake mesh, PIV or CFD slice)  
File_Path_CSV = str(File_Path_CSV)                          # File path to the CSV file containing the multi-hole data (other formats accepted but requires file parser)
File_Parser = float(File_Parser)                            # 0=None  1=Labview 5HP ISAE   2=User defined parser
Ignore_Tt = float(Ignore_Tt)    							# 0=assembly Tt into mesh    1=Do not assembly Tt
Ignore_First_Line = float(Ignore_First_Line)    			# 0=do not ignore first data line    1=ignore
Ignore_Last_Line = float(Ignore_Last_Line)    				# 0=do not ignore last data line     1=ignore
Shifted_Point = float(Shifted_Point)    					# 0=Desabled   1=enable shifted-point wake survey technique
File_Path_CSV_1 = str(File_Path_CSV_1)    				    # File path to the probe data of the complementary wake survey 1
File_Path_CSV_2 = str(File_Path_CSV_2)    				    # File path to the probe data of the complementary wake survey 2
File_Path_CSV_3 = str(File_Path_CSV_3)    				    # File path to the probe data of the complementary wake survey 3
File_Path_CSV_4 = str(File_Path_CSV_4)    				    # File path to the probe data of the complementary wake survey 4
ProbeOrientation = ProbeOrientation                       	# 5HP reference system orientation: rotations about x-, y- and z-axis (WTT axes)
Extrusion_Direction = Extrusion_Direction                   # Vector with the extrusion direction   
Extrusion_Lenght = float(Extrusion_Lenght)                  # Extrusion lenght in [m]
Plane_orientation_3rdParty = float(Plane_orientation_3rdParty) # 0=YZ plane	1=XZ plane	2=XY plane
InputToTransform = float(InputToTransform)   				# 0=None   1=Mesh and probe   2=Mesh only   3=Probe only
Translation = Translation 									# Translation vector along X-, Y- and Z-axes
Rotate = Rotate                                             # Rotation vector about X-, Y- and Z-axes
Scale_Factors = Scale_Factors                               # Vector with scale factors for coordinates, velocities and pressure




############################################################################################################
#######################################    DEFINITION OF FUNCTIONS    ######################################
############################################################################################################

###################################
#######  READ CSV FUNCTION  #######
###################################

def READ_CSV(File_Path_CSV,Visualization_Mode,Scale_Factors,PLOT_VTK_LINES_VERTEX,addfielddata,File_Parser,Ignore_Tt,Ignore_First_Line,Ignore_Last_Line,ProbeOrientation,ROTATE_vector,Shifted_Point):
	import numpy as np
	import vtk
	from vtk.numpy_interface import algorithms as algs
	import sys
	sys.path.insert(0, r'C:\EPSILON\Readers')
	from Reader_Labview_5HP_ISAE import Reader_Labview_5HP_ISAE
	
	if File_Parser == 0:	
		ProbeData = np.genfromtxt(File_Path_CSV, dtype=None, names=True, delimiter=",", autostrip=True)     # Reads input data 
	
	if File_Parser == 1:
		DATA, Headers = Reader_Labview_5HP_ISAE(File_Path_CSV)
		DATAtoSTRING = [",".join([str(x) for x in m]) for m in DATA]
		ProbeData = np.genfromtxt(DATAtoSTRING, dtype=None, names=Headers, delimiter=",", autostrip=True)
	
	if Ignore_First_Line == 1:
		ProbeData = ProbeData[1:]

	if Ignore_Last_Line == 1:
		ProbeData = ProbeData[:-1]						
			
	if Visualization_Mode != 3:                                               	# This reads data for non-axisymmetric cases
		
		if Shifted_Point==0:
			# INITIALIZE ARRAYS:
			X = 0*ProbeData[ProbeData.dtype.names[0]]                               # Initialize the array with zeros. It takes the first array of data file as reference size
			Y = 0*ProbeData[ProbeData.dtype.names[0]]
			Z = 0*ProbeData[ProbeData.dtype.names[0]]
			Vx = 0*ProbeData[ProbeData.dtype.names[0]]
			Vy = 0*ProbeData[ProbeData.dtype.names[0]]
			Vz = 0*ProbeData[ProbeData.dtype.names[0]]
			
			# READ ARRAY DATA:
			for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
				if array=="X":                                                      # If the X array actually exists in the input data, it updates its values, otherwise a zero vector is retained
					X = ProbeData["X"]
				if array=="Y":
					Y = ProbeData["Y"]
				if array=="Z":
					Z = ProbeData["Z"]
				if array=="Vx":
					Vx = ProbeData["Vx"]
				if array=="Vy":
					Vy = ProbeData["Vy"]
				if array=="Vz":
					Vz = ProbeData["Vz"]
					
		if Shifted_Point==1:   														# This technique requires transfering the 5HP data at a point N+1 to the point N due to the measurement shift
			# INITIALIZE ARRAYS:
			X = 0*ProbeData[ProbeData.dtype.names[0]][:-1]                               # Initialize the array with zeros. It takes the first array of data file as reference size
			Y = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
			Z = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
			Vx = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
			Vy = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
			Vz = 0*ProbeData[ProbeData.dtype.names[0]][:-1]

			# READ ARRAY DATA:
			for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
				if array=="X":                                                      # If the X array actually exists in the input data, it updates its values, otherwise a zero vector is retained
					X = ProbeData["X"][:-1]
				if array=="Y":
					Y = ProbeData["Y"][:-1]
				if array=="Z":
					Z = ProbeData["Z"][:-1]
				if array=="Vx":
					Vx = ProbeData["Vx"][1:]
				if array=="Vy":
					Vy = ProbeData["Vy"][1:]
				if array=="Vz":
					Vz = ProbeData["Vz"][1:]		
		
		
		# CHANGE VELOCITY SCALE:
		Vx = Vx*Scale_Factors[1]
		Vy = Vy*Scale_Factors[1]	
		Vz = Vz*Scale_Factors[1]
		
		# CREATE VECTORS:
		PointsVector = algs.make_vector(X,Y,Z)
		Velocity = algs.make_vector(Vx,Vy,Vz)
		Velocity = ROTATE_vector(Velocity,ProbeOrientation)	
		
		CheckInputXYZ = 1
		CheckInputVxVyVz = 1
		# OUTPUT MESSAGES:
		if np.mean(X)==0 and np.mean(Y)==0 and np.mean(Z)==0:
			print ("Input probe data: coordinate arrays are unexistant or the headers are wrong (use 'X, Y and Z')")
			CheckInputXYZ = 0
		if np.mean(Vx)==0 and np.mean(Vy)==0 and np.mean(Vz)==0:
			print ("Input probe data: velocity arrays are unexistant or the headers are wrong (use 'Vx, Vy and Vz')")
			CheckInputVxVyVz = 0
		

	if Visualization_Mode == 3:                                             	# This applies only for annular measurement region reconstructed from a survey line data
		
		"""
		El reader de data axisimetrica falta mejorarlo !!! Hacerlo identico al lector de data no axisimetrica
		"""
		R = ProbeData["R"]                                                      # Radius in [m]
		R = R + InternalRadius                                                  # Grid translation
		X = 0*R + X0                                                            # By default, X-coordinate is set to zero
		Vx = ProbeData["Vx"]
		Vtan = ProbeData["Vtan"]                                                # Note: older Paraview versions do not support Polar Reference System. Hence a convertion to Cartesian grid wiil be made later in this code (See "CREATING ANNULUS" section)
		Vrad = ProbeData["Vrad"]

		# VELOCITY DIMENSIONALISATION:
		Vx = Vx*Scale_Factors[1]
		Vtan = Vtan*Scale_Factors[1]	
		Vrad = Vrad*Scale_Factors[1]
	
	"""
	Agregar opcion en GUI para lectura de datos 5HP sin termocupla
	"""
	if Shifted_Point==0: 
		# INITIALIZE PRESSURE AND TEMPERATURE ARRAYS:
		Ps = 0*ProbeData[ProbeData.dtype.names[0]]
		Pt = 0*ProbeData[ProbeData.dtype.names[0]]
		Tt = 0*ProbeData[ProbeData.dtype.names[0]]
		P1 = 0*ProbeData[ProbeData.dtype.names[0]]
		P2 = 0*ProbeData[ProbeData.dtype.names[0]]
		P3 = 0*ProbeData[ProbeData.dtype.names[0]]
		P4 = 0*ProbeData[ProbeData.dtype.names[0]]
		P5 = 0*ProbeData[ProbeData.dtype.names[0]]
		
		# READ ARRAY DATA:
		for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
			if array=="Ps": 
				Ps = ProbeData["Ps"]
			if array=="Pt":
				Pt = ProbeData["Pt"]
			if array=="Tt":
				Tt = ProbeData["Tt"]
			if array=="P1": 
				P1 = ProbeData["P1"]
			if array=="P2": 
				P2 = ProbeData["P2"]
			if array=="P3": 
				P3 = ProbeData["P3"]
			if array=="P4": 
				P4 = ProbeData["P4"]
			if array=="P5": 
				P5 = ProbeData["P5"]
				
	if Shifted_Point==1: 		
		# INITIALIZE PRESSURE AND TEMPERATURE ARRAYS:
		Ps = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		Pt = 0*ProbeData[ProbeData.dtype.names[0]][1:]
		Tt = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
		P1 = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
		P2 = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
		P3 = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
		P4 = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
		P5 = 0*ProbeData[ProbeData.dtype.names[0]][:-1]
		
		# READ ARRAY DATA:
		for array in ProbeData.dtype.names:                                     # "ProbeData.dtype.names" provides the names of the variables stored in the CSV data file
			if array=="Ps": 
				Ps = ProbeData["Ps"][1:]
			if array=="Pt":
				Pt = ProbeData["Pt"][1:]
			if array=="Tt":
				Tt = ProbeData["Tt"][:-1]
			if array=="P1": 
				P1 = ProbeData["P1"][:-1]
			if array=="P2": 
				P2 = ProbeData["P2"][:-1]
			if array=="P3": 
				P3 = ProbeData["P3"][:-1]
			if array=="P4": 
				P4 = ProbeData["P4"][:-1]
			if array=="P5": 
				P5 = ProbeData["P5"][:-1]
	
	# PRESSURE DIMENSIONALISATION:
	Ps = Ps*Scale_Factors[2]
	Pt = Pt*Scale_Factors[2]
	P1 = P1*Scale_Factors[2]
	P2 = P2*Scale_Factors[2]
	P3 = P3*Scale_Factors[2]
	P4 = P4*Scale_Factors[2]
	P5 = P5*Scale_Factors[2]
	
	# OUTPUT MESSAGES:
	if np.mean(Ps)==0 and np.mean(Pt)==0:
		print ("Input probe data: total and static pressure arrays are unexistant or the headers are wrong (use 'Pt and Ps')")
	if np.mean(Tt)==0:
		print ("Input probe data: total temperature array missing or the header is wrong (use 'Tt')")

	# CREATE THE VTK ELEMENT WITH THE PROBE POINTS
	Survey_Line,Survey_Vertices = PLOT_VTK_LINES_VERTEX(PointsVector)   		
	
	# ADD MEASURE 5HP DATA TO THE LINES:
	addfielddata(3, Survey_Line.GetOutput().GetPointData(),Velocity.ravel(),"Velocity")
	addfielddata(1, Survey_Line.GetOutput().GetPointData(),Ps,"Ps")
	addfielddata(1, Survey_Line.GetOutput().GetPointData(),Pt,"Pt")
	addfielddata(1, Survey_Line.GetOutput().GetPointData(),P1,"P1")
	addfielddata(1, Survey_Line.GetOutput().GetPointData(),P2,"P2")
	addfielddata(1, Survey_Line.GetOutput().GetPointData(),P3,"P3")
	addfielddata(1, Survey_Line.GetOutput().GetPointData(),P4,"P4")
	addfielddata(1, Survey_Line.GetOutput().GetPointData(),P5,"P5")
	
	if Ignore_Tt==0:
		addfielddata(1, Survey_Line.GetOutput().GetPointData(),Tt,"Tt")
	
	# ADD MEASURE 5HP DATA TO THE VERTICES:
	addfielddata(3, Survey_Vertices.GetOutput().GetPointData(),Velocity.ravel(),"Velocity")
	addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),Ps,"Ps")
	addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),Pt,"Pt")
	addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),P1,"P1")
	addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),P2,"P2")
	addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),P3,"P3")
	addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),P4,"P4")
	addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),P5,"P5")
	
	if Ignore_Tt==0:
		addfielddata(1, Survey_Vertices.GetOutput().GetPointData(),Tt,"Tt")
	
	# PRINT OUTPUT MESSAGES:
	if CheckInputXYZ == 0 and CheckInputVxVyVz == 0:
		print ("Error while reading input 5HP data")

	if CheckInputXYZ == 1 and CheckInputVxVyVz == 1:
		print ("Input 5HP data loaded successfully")
		print ("Activate 'Axes Grid' or 'Data Axes Grid' in the properties panel to verify if the input 5HP data dimensions are in meters")
		print ("IMPORTANT: If input 5HP CSV file is in [mm] it must be transformed to [m] by following these steps")
		print ("A) Select the 'Input to transform' in the section 'Options'")    
		print ("B) Set 'Scale XYZ' from 1 to 0.001")
	
	return Survey_Line,Survey_Vertices


###################################
#####  READ TECPLOT FUNCTION  #####
###################################

def READ_TECPLOT(File_Path_Mesh):
	import vtk
	MBTecplotReader = vtk.vtkTecplotReader()                                    # "vtkTecplotReader" gives a vtkMultiblockDataSet output that contains several blocks of vtkStructuredGrid type
	MBTecplotReader.SetFileName(File_Path_Mesh)
	MBTecplotReader.Update()
	
	InputMesh = vtk.vtkGeometryFilter()                                         # Converts vtkMultiblockDataSet to vtkPolyData
	InputMesh.SetInputData(MBTecplotReader.GetOutput().GetBlock(0))             # "MBTecplotReader" is vtkMultiblockDataSet type. With "GetBlock(0)" it gets the only PIV block of the data (it is vtkStructuredGrid type). Then "vtkGeometryFilter" converts it to vtkPolyData
	InputMesh.Update()
	#self.GetOutput().ShallowCopy(InputMesh.GetOutput())                        # It displays the mesh. Kept here for debugging purposes
	return InputMesh

###################################
#######  READ VTP FUNCTION  #######
###################################

def READ_VTP(File_Path_Mesh):
	import vtk
	InputMesh = vtk.vtkXMLPolyDataReader()                                      # Alternative reader: vtkPolyDataReader
	InputMesh.SetFileName(File_Path_Mesh)
	InputMesh.Update()                                                          # "InputMesh" is a vtkPolyData variable
	#output.ShallowCopy(InputMesh.GetOutput())
	return InputMesh
	

###################################
#######  READ VTU FUNCTION  #######
###################################

def READ_VTU(File_Path_Mesh):
	import vtk
	VTUreader = vtk.vtkXMLUnstructuredGridReader()                              # "vtkXMLUnstructuredGridReader" gives a vtkUnstructuredGrid output type
	VTUreader.SetFileName(File_Path_Mesh)
	VTUreader.Update()

	InputMesh = vtk.vtkGeometryFilter()                                         # Converts the vtkUnstructuredGrid to vtkPolyData
	InputMesh.SetInputData(VTUreader.GetOutput())
	InputMesh.Update()
	#self.GetOutput().ShallowCopy(InputMesh.GetOutput())                        # It displays the mesh. Kept here for debugging purposes
	return InputMesh

	
###################################
#######  READ VTS FUNCTION  #######
###################################

def READ_VTS(File_Path_Mesh):
	import vtk
	VTSreader = vtk.vtkXMLStructuredGridReader()                                # "vtkXMLStructuredGridReader" gives a vtkStructuredGrid output type
	VTSreader.SetFileName(File_Path_Mesh)
	VTSreader.Update()
	
	InputMesh = vtk.vtkGeometryFilter()                                         # Converts the vtkStructuredGrid to vtkPolyData
	InputMesh.SetInputData(VTSreader.GetOutput())
	InputMesh.Update()
	#self.GetOutput().ShallowCopy(InputMesh.GetOutput())                        # It displays the mesh. Kept here for debugging purposes
	return InputMesh

	
############################################
############  ROTATE FUNCTION  #############
############################################

def ROTATE(InputVTK, Rotation):
	import vtk

	RotX = vtk.vtkTransform()                                               	# Generates the VTK transformation function to rotate the PIV data about the x-axis
	RotX.RotateX(Rotation[0])
	RotX.Update()

	OutputVTK_Rx = vtk.vtkTransformFilter()                                    	# Applies the rotation to the PIV data and creates a new transformed output
	OutputVTK_Rx.SetTransform(RotX)
	OutputVTK_Rx.SetInputData(InputVTK.GetOutput())
	OutputVTK_Rx.Update()

	RotY = vtk.vtkTransform()                                                 	# Generates the VTK transformation function to rotate the PIV data about the y-axis
	RotY.RotateY(Rotation[1])
	RotY.Update()

	OutputVTK_Ry = vtk.vtkTransformFilter()                                   	# Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
	OutputVTK_Ry.SetTransform(RotY)
	OutputVTK_Ry.SetInputData(OutputVTK_Rx.GetOutput())
	OutputVTK_Ry.Update()

	RotZ = vtk.vtkTransform()                                               	# Generates the VTK transformation function to rotate the PIV data about the z-axis
	RotZ.RotateZ(Rotation[2])
	RotZ.Update()

	OutputVTK_Rz = vtk.vtkTransformFilter()                                   	# Applies the rotation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
	OutputVTK_Rz.SetTransform(RotZ)
	OutputVTK_Rz.SetInputData(OutputVTK_Ry.GetOutput())
	OutputVTK_Rz.Update()

	return OutputVTK_Rz


############################################
############ TRANSLATE FUNCTION ############
############################################

def TRANSLATE(InputVTK, Translation):
	import vtk
	
	Transform = vtk.vtkTransform()                                           	# Generates the VTK transformation function to translate the PIV data along the x-, y- and z-axes
	Transform.Translate(Translation[0], Translation[1], Translation[2])
	Transform.Update()

	OutputVTK_T = vtk.vtkTransformFilter()                                    	# Applies the translation to the previously transformed PIV data and creates a new transformed output (i.e., transformations are accumulated progressively)
	OutputVTK_T.SetTransform(Transform)
	OutputVTK_T.SetInputData(InputVTK.GetOutput())
	OutputVTK_T.Update()

	return OutputVTK_T

	
############################################
########## SCALE GEOMETRY FUNCTION #########
############################################

def SCALE_GEOMETRY(InputVTK, LenghtScaleFactor):
	import vtk

	transform1 = vtk.vtkTransform()                                           	# Generates the VTK transformation function to scale up/down the PIV data
	transform1.Scale(LenghtScaleFactor, LenghtScaleFactor, LenghtScaleFactor)
	transform1.Update()

	OutputVTK_T1 = vtk.vtkTransformFilter()                                  	# Applies the scale up/down to the input PIV data and creates a transformed output
	OutputVTK_T1.SetTransform(transform1)
	OutputVTK_T1.SetInputData(InputVTK.GetOutput())
	OutputVTK_T1.Update()
	
	return OutputVTK_T1


############################################
######### ADD FIELD DATA FUNCTION ##########
############################################

def addfielddata(numTuples, dataSetAttributes,DataArray,ArrayName):             # Defines a function that adds the field data to the existing points in the Paraview's output
	numberOfComponents = int(numTuples)                                       	# We deal only with 1-component arrays
	dataArray = vtk.vtkDoubleArray()   
	dataSetAttributes.AddArray(dataArray)
	dataArray.SetNumberOfComponents(numberOfComponents)
	dataArray.SetName(ArrayName)                                               	# "ArrayName" is a string used to name the array in Paraview 
	data = DataArray                                                        	# DataArray is the numpy array with the aerodynamic parameters that we want to associate to the points
	for value in data:
		dataArray.InsertNextValue(float(value))                              	# Iterates along every value of the array and add its value to the output array designed with the "ArrayName" name 


############################################
#########  ROTATE POINTS FUNCTION  #########
############################################

def ROTATE_vector(vector,Rotation):
	import numpy as np
	from vtk.numpy_interface import algorithms as algs
	
	AngleX = Rotation[0]
	AngleY = Rotation[1]
	AngleZ = Rotation[2]

	Rx = np.array([(1,0,0),(0,algs.cos(AngleX*np.pi/180),-algs.sin(AngleX*np.pi/180)),(0,algs.sin(AngleX*np.pi/180),algs.cos(AngleX*np.pi/180))])   #Rotation matrix about X-axis
	Ry = np.array([(algs.cos(AngleY*np.pi/180),0,algs.sin(AngleY*np.pi/180)),(0,1,0),(-algs.sin(AngleY*np.pi/180),0,algs.cos(AngleY*np.pi/180))])  	#Rotation matrix about Y-axis
	Rz = np.array([(algs.cos(AngleZ*np.pi/180),-algs.sin(AngleZ*np.pi/180),0),(algs.sin(AngleZ*np.pi/180),algs.cos(AngleZ*np.pi/180),0),(0,0,1)])  	#Rotation matrix about Z-axis

	Rx_T = np.transpose(Rx)
	Ry_T = np.transpose(Ry)
	Rz_T = np.transpose(Rz)

	MX = np.dot(Rx_T,np.dot(Rx,Rx_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh X-axis
	MY = np.dot(Ry_T,np.dot(Ry,Ry_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Y-axis
	MZ = np.dot(Rz_T,np.dot(Rz,Rz_T))  #Transformation matrix from the original reference system (mesh-XYZ) to the new reference system (aerodynamic-xyz) --> rotation about mesh Z-axis

	TRANSFORMATION = np.dot(MX,np.dot(MY,MZ))  #Total (combined) transformation matrix

	T_t = np.transpose(TRANSFORMATION)
	T_inverse = np.dot(T_t,np.dot(TRANSFORMATION,T_t))  	# It calculates the inverse of the TRANSFORMATION matrix
	TransformedVector = np.einsum('ij,kj->ik',vector,T_inverse) 
	return TransformedVector



############################################
#########  PLOT VTK LINES FUNCTION  ########
############################################

def PLOT_VTK_LINES_VERTEX(PointsVector):
	import numpy as np
	import vtk

	points = vtk.vtkPoints()                                                	# create a vtkPoints container to store all the point coordinates.
	points.SetData(dsa.numpyTovtkDataArray(PointsVector , "Points"))   					# numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()

	NBpts = np.shape(PointsVector)[0]                                             		# Number of mesh points
	
	polyLine = vtk.vtkPolyLine()                                            	# Creates a polyline and set the ID for each point
	polyLine.GetPointIds().SetNumberOfIds(int(NBpts)) 
	for i in range(0, int(NBpts)):
		polyLine.GetPointIds().SetId(i, i)

	cells = vtk.vtkCellArray()                                              	# Create a cell array to store the lines in and add the lines to it
	cells.InsertNextCell(polyLine)

	polyData = vtk.vtkPolyData()                                            	# Create a polydata to store everything in
	polyData.SetPoints(points)                                              	# Add the points to the dataset
	polyData.SetLines(cells)                                                	# Add the lines to the dataset

	Survey_Line = vtk.vtkGeometryFilter()                                   	# Converts the polydata to vtkPolyData  
	Survey_Line.SetInputData(polyData)             
	Survey_Line.Update()
	
	# CREATES VERTICES AT EACH POINT TO IMPROVE THE VISUALIZATION
	polydata2 = vtk.vtkPolyData()
	polydata2.SetPoints(points)

	Survey_Vertices = vtk.vtkVertexGlyphFilter()
	Survey_Vertices.AddInputData(polydata2)
	Survey_Vertices.Update()
	
	return Survey_Line,Survey_Vertices


	
###################################
#######  DELAUNAY FUNCTION  #######
###################################

def DELAUNAY(InputPoints): 
	import vtk
	
	pts = InputPoints.GetOutput().GetPoints()

	aPolyData = vtk.vtkPolyData()
	aPolyData.SetPoints(pts)
		 
	aCellArray = vtk.vtkCellArray()
		 
	boundary = vtk.vtkPolyData()
	boundary.SetPoints(aPolyData.GetPoints())
	boundary.SetPolys(aCellArray)

	delaunay = vtk.vtkDelaunay2D()
	delaunay.SetProjectionPlaneMode(vtk.VTK_BEST_FITTING_PLANE)

	delaunay.SetInputData(aPolyData)
	delaunay.SetSourceData(boundary)

	delaunay.Update()
	return delaunay
	


############################################################################################################
########################################## ERROR MESSAGE FOR USER ##########################################
############################################################################################################

if Visualization_Mode==0 and MeshType==0:
		
	if Mode == 1:
		if not (File_Path_Mesh and File_Path_CSV):
				print ("ERROR: GUI is not yet properly defined by the user: NO OUTPUT EXPECTED!")
				print ("The input mesh and probe data must be selected beforehand (and a file parser if required)")			
   
	
############################################################################################################
#################################### READ AND TRANSFORM INPUT MESH DATA ####################################
############################################################################################################

# This section reads any input mesh file format and converts it into a vtkPolyData format. Then it transforms the mesh geometry

# READ INPUT MESH:
if File_Path_Mesh.endswith('.dat'):
	InputMesh = READ_TECPLOT(File_Path_Mesh)

if File_Path_Mesh.endswith('.vtp'):
	InputMesh = READ_VTP(File_Path_Mesh)
	
if File_Path_Mesh.endswith('.vtu'):
	InputMesh = READ_VTU(File_Path_Mesh)
	
if File_Path_Mesh.endswith('.vts'):
	InputMesh = READ_VTS(File_Path_Mesh)	

# TRANSFORM THE INPUT MESH:
if InputToTransform==1 or InputToTransform==2:    								# The mesh geometry must be transformed

	if File_Path_Mesh.endswith('.dat') or File_Path_Mesh.endswith('.vtp') or File_Path_Mesh.endswith('.vtu') or File_Path_Mesh.endswith('.vts'):   # Only transforms a known file extension
		
		InputMesh = SCALE_GEOMETRY(InputMesh, Scale_Factors[0])
		InputMesh = TRANSLATE(InputMesh, Translation)
		InputMesh = ROTATE(InputMesh, Rotate)

		self.GetOutput().ShallowCopy(InputMesh.GetOutput())
		



############################################################################################################
#################################    DATA ASSEMBLY/DELAUNAY VISUALIZATION   ################################
############################################################################################################

if Mode == 1 or Mode == 3:   																# Mode=1(Data assembly)   Mode=3(Delaunay visualization)

	if Action==0 and MeshType==0:
		print ("NO OUTPUT EXPECTED!. Select an action in the data assembly drop down list")
		
	if Action==1:
		# OUTPUT MESSAGE:
		if not InputMesh:
			print ("Error while loading the input mesh")
		else:
			print ("Input mesh loaded successfully")
		
		# VISUALIZATION:
		self.GetOutput().ShallowCopy(InputMesh.GetOutput())
	
	if Action>1 or MeshType==2:
		# READ THE INPUT PROBE DATA:
		if File_Path_CSV.endswith('.csv') or File_Path_CSV.endswith('.moy'):
			Survey_Line,Data_5HP = READ_CSV(File_Path_CSV,Visualization_Mode,Scale_Factors,PLOT_VTK_LINES_VERTEX,addfielddata,File_Parser,Ignore_Tt,Ignore_First_Line,Ignore_Last_Line,ProbeOrientation,ROTATE_vector,Shifted_Point)   # Survey_Line not used in this application
			

			if File_Path_CSV_1.endswith('.csv') or File_Path_CSV_1.endswith('.moy'):
				Survey_Line_1,Data_5HP_1 = READ_CSV(File_Path_CSV_1,Visualization_Mode,Scale_Factors,PLOT_VTK_LINES_VERTEX,addfielddata,File_Parser,Ignore_Tt,Ignore_First_Line,Ignore_Last_Line,ProbeOrientation,ROTATE_vector,Shifted_Point)   # Survey_Line not used in this application
				
				appendFilter = vtk.vtkAppendFilter()   					# Add the complementary probe data to the primary wake survey
				appendFilter.AddInputData(Data_5HP.GetOutput())
				appendFilter.AddInputData(Data_5HP_1.GetOutput())
				appendFilter.Update()
				
				Data_5HP = appendFilter    								# Update the "Data_5HP" variable with the appended data included
				Data_5HP.Update()
				
			if File_Path_CSV_2.endswith('.csv') or File_Path_CSV_2.endswith('.moy'):
				Survey_Line_2,Data_5HP_2 = READ_CSV(File_Path_CSV_2,Visualization_Mode,Scale_Factors,PLOT_VTK_LINES_VERTEX,addfielddata,File_Parser,Ignore_Tt,Ignore_First_Line,Ignore_Last_Line,ProbeOrientation,ROTATE_vector,Shifted_Point)   # Survey_Line not used in this application

				appendFilter = vtk.vtkAppendFilter()   					# Add the complementary probe data to the primary wake survey
				appendFilter.AddInputData(Data_5HP.GetOutput())
				appendFilter.AddInputData(Data_5HP_2.GetOutput())
				appendFilter.Update()
				
				Data_5HP = appendFilter    								# Update the "Data_5HP" variable with the appended data included
				Data_5HP.Update()

			if File_Path_CSV_3.endswith('.csv') or File_Path_CSV_3.endswith('.moy'):
				Survey_Line_3,Data_5HP_3 = READ_CSV(File_Path_CSV_3,Visualization_Mode,Scale_Factors,PLOT_VTK_LINES_VERTEX,addfielddata,File_Parser,Ignore_Tt,Ignore_First_Line,Ignore_Last_Line,ProbeOrientation,ROTATE_vector,Shifted_Point)   # Survey_Line not used in this application

				appendFilter = vtk.vtkAppendFilter()   					# Add the complementary probe data to the primary wake survey
				appendFilter.AddInputData(Data_5HP.GetOutput())
				appendFilter.AddInputData(Data_5HP_3.GetOutput())
				appendFilter.Update()
				
				Data_5HP = appendFilter    								# Update the "Data_5HP" variable with the appended data included
				Data_5HP.Update()


			if File_Path_CSV_4.endswith('.csv') or File_Path_CSV_4.endswith('.moy'):
				Survey_Line_4,Data_5HP_4 = READ_CSV(File_Path_CSV_4,Visualization_Mode,Scale_Factors,PLOT_VTK_LINES_VERTEX,addfielddata,File_Parser,Ignore_Tt,Ignore_First_Line,Ignore_Last_Line,ProbeOrientation,ROTATE_vector,Shifted_Point)   # Survey_Line not used in this application

				appendFilter = vtk.vtkAppendFilter()   					# Add the complementary probe data to the primary wake survey
				appendFilter.AddInputData(Data_5HP.GetOutput())
				appendFilter.AddInputData(Data_5HP_4.GetOutput())
				appendFilter.Update()
				
				Data_5HP = appendFilter    								# Update the "Data_5HP" variable with the appended data included
				Data_5HP.Update()	


			# TRANSFORM THE INPUT POINTS:
			if InputToTransform==1 or InputToTransform==3:    								# The input probe data points must be transformed
				Data_5HP = SCALE_GEOMETRY(Data_5HP, Scale_Factors[0])
				Data_5HP = TRANSLATE(Data_5HP, Translation)
				Data_5HP = ROTATE(Data_5HP, Rotate)
		
		# VISUALIZATION:
		self.GetOutput().ShallowCopy(Data_5HP.GetOutput())
	
	if Action>2 or MeshType==2:
		# GENERATE A DELAUNAY MESH FROM THE INPUT 5HP POINTS:   (The 5HP point data cannot be interpolated directly into the input mesh. Instead, a delaunay mesh plane has to be created first from the 5HP point data. Then, this mesh plane is used to interpolate data into the input mesh)
		delaunay = DELAUNAY(Data_5HP)
		
		# GET ARRAY DATA FROM 5HP:
		Data_5HP_Python = dsa.WrapDataObject(Data_5HP.GetOutput())             				# Converts VTK element into a Python object
		Velocity = Data_5HP_Python.PointData["Velocity"]
		Ps = Data_5HP_Python.PointData["Ps"]
		Pt = Data_5HP_Python.PointData["Pt"]
		Tt = Data_5HP_Python.PointData["Tt"]
		P1 = Data_5HP_Python.PointData["P1"]
		P2 = Data_5HP_Python.PointData["P2"]
		P3 = Data_5HP_Python.PointData["P3"]
		P4 = Data_5HP_Python.PointData["P4"]
		P5 = Data_5HP_Python.PointData["P5"]

		# ADD MEASURED 5HP DATA INTO THE DELAUNAY MESH:     
		addfielddata(3, delaunay.GetOutput().GetPointData(),Velocity.ravel(),"Velocity")
		addfielddata(1, delaunay.GetOutput().GetPointData(),Ps,"Ps")
		addfielddata(1, delaunay.GetOutput().GetPointData(),Pt,"Pt")
		addfielddata(1, delaunay.GetOutput().GetPointData(),P1,"P1")		
		addfielddata(1, delaunay.GetOutput().GetPointData(),P2,"P2")
		addfielddata(1, delaunay.GetOutput().GetPointData(),P3,"P3")
		addfielddata(1, delaunay.GetOutput().GetPointData(),P4,"P4")
		addfielddata(1, delaunay.GetOutput().GetPointData(),P5,"P5")
		
		if Ignore_Tt == 0:
			addfielddata(1, delaunay.GetOutput().GetPointData(),Tt,"Tt")
		
		# VISUALIZATION:
		self.GetOutput().ShallowCopy(delaunay.GetOutput())
		
		# OUTPUT MESSAGE:
		print ("Change the visualization mode from 'Surface' to 'Surface with edges' to see the cells:")

	if Action>3:
		# ASSIGN VARIABLE VALUES FROM THE DELAUNAY MESH TO THE DESTINATION MESH     
		probe_vertices = vtk.vtkProbeFilter()
		probe_vertices.SetInputConnection(InputMesh.GetOutputPort())   
		probe_vertices.SetSourceData(delaunay.GetOutput())     		
		probe_vertices.Update()   
		
		# VISUALIZATION:
		self.GetOutput().ShallowCopy(probe_vertices.GetOutput())


############################################################################################################
########################################### VISUALIZE LINE DATA ############################################
############################################################################################################

if Mode == 2:   																# Plot line only
	# READ THE INPUT PROBE DATA:
	if File_Path_CSV.endswith('.csv') or File_Path_CSV.endswith('.moy'):
		Survey_Line,Data_5HP = READ_CSV(File_Path_CSV,Visualization_Mode,Scale_Factors,PLOT_VTK_LINES_VERTEX,addfielddata,File_Parser,Ignore_Tt,Ignore_First_Line,Ignore_Last_Line,ProbeOrientation,ROTATE_vector,Shifted_Point)		
		# TRANSFORM THE INPUT POINTS:
		if InputToTransform==1 or InputToTransform==3:    								# The input probe data points must be transformed
			Data_5HP = SCALE_GEOMETRY(Data_5HP, Scale_Factors[0])
			Data_5HP = TRANSLATE(Data_5HP, Translation)
			Data_5HP = ROTATE(Data_5HP, Rotate)
			Survey_Line = SCALE_GEOMETRY(Survey_Line, Scale_Factors[0])
			Survey_Line = TRANSLATE(Survey_Line, Translation)
			Survey_Line = ROTATE(Survey_Line, Rotate)
	
	if Visualization_Mode == 1:
		# COMBINE LINE AND VERTICES INTO A SINGLE VARIABLE:
		appendFilter = vtk.vtkAppendFilter()                                    #Output: vtkUnstructuredGrid
		appendFilter.AddInputData(Survey_Line.GetOutput())        
		appendFilter.AddInputData(Data_5HP.GetOutput())
		appendFilter.Update()

		# CONVERT THE COMBINED OUTPUT FORMAT TO VTKPOLYDATA  
		combined = vtk.vtkGeometryFilter()                                      # Converts vtkMultiblockDataSet to vtkPolyData
		combined.SetInputData(appendFilter.GetOutput())                             
		combined.Update()
		
		# VISUALIZATION:
		self.GetOutput().ShallowCopy(combined.GetOutput())

	if Visualization_Mode == 2:
	
		# EXTRUDE SURVEY LINE ALONG GIVEN DIRECTION:
		extrude = vtk.vtkLinearExtrusionFilter()
		extrude.CappingOn()
		extrude.SetExtrusionTypeToVectorExtrusion()
		extrude.SetVector(Extrusion_Direction)
		extrude.SetInputData(Survey_Line.GetOutput())
		extrude.SetScaleFactor(Extrusion_Lenght)
		extrude.Update()
		
		# VISUALIZATION:
		self.GetOutput().ShallowCopy(extrude.GetOutput())




############################################################################################################
############################################## CREATING PLANE ##############################################
############################################################################################################

###################################
#########  DELAUNAY MESH  #########
###################################

# Already included in the previous part


"""


############################################################################################################
############################################# CREATING ANNULUS #############################################
############################################################################################################

if Survey_Type==1 and Axisymmetric ==1:

	# Creation of the visualisation grid points in polar coordinates: 
	AngularGrid = np.linspace(0,2*np.pi,Resolution)    				# Defines a uniform distribution of meshpoints in the angular direction

	Xmesh = []
	Ymesh = []
	Zmesh = []
	Vx_mesh = []
	Vy_mesh = []	
	Vz_mesh = []
	Ps_mesh = []
	Pt_mesh = []	
	
	for i in AngularGrid:
		Ycoordinates = R*np.cos(i)+Y0                    			# Proyecting meshpoints onto XY coordinate system
		Zcoordinates = R*np.sin(i)+Z0
		Vy = Vrad*np.cos(i)-Vtan*np.sin(i)
		Vz = Vrad*np.sin(i)+Vtan*np.cos(i)
		Xmesh.append(X)
		Ymesh.append(Ycoordinates)
		Zmesh.append(Zcoordinates)
		Vx_mesh.append(Vx)
		Vy_mesh.append(Vy)
		Vz_mesh.append(Vz)
		Ps_mesh.append(Ps)
		Pt_mesh.append(Pt)
		
	Xmesh = np.array(Xmesh)
	Ymesh = np.array(Ymesh)
	Zmesh = np.array(Zmesh)
	Vx_mesh = np.array(Vx_mesh)
	Vy_mesh = np.array(Vy_mesh)
	Vz_mesh = np.array(Vz_mesh)
	Ps_mesh = np.array(Ps_mesh)
	Pt_mesh = np.array(Pt_mesh)
	
	Xmesh = Xmesh.ravel()
	Ymesh = Ymesh.ravel()											# Ravel command converts 2D array into 1D array (in order to have a single column for each variable in the Paraview output array)
	Zmesh = Zmesh.ravel()
	Vx = Vx_mesh.ravel()
	Vy = Vy_mesh.ravel()	
	Vz = Vz_mesh.ravel()	
	Ps = Ps_mesh.ravel()	
	Pt = Pt_mesh.ravel()
	
	numCells=(R.size-1)*(Resolution-1)                  			# Number of cells	
	
	numCells = int(numCells)
	
	POINTS = algs.make_vector(Xmesh, Ymesh, Zmesh)             		# convert the 3 arrays into a single 3 component array for use as the coordinates for the points.
	pts = vtk.vtkPoints()                                           # create a vtkPoints container to store all the point coordinates.
	pts.SetData(dsa.numpyTovtkDataArray(POINTS , "POINTS"))    		# numpyTovtkDataArray converts the NumPy array to a vtkDataArray, which is the format needed by SetData()
	output = self.GetOutput()
	output.SetPoints(pts)
	

	IDlist = []
	for j in range(int(Resolution)-1):                           	# We iterate over each cell in order to define the counter-clockwise point ID that defines the cell connectivity
		for i in range(int(R.size)-1):
			IDlist.append([i+j*R.size,i+1+j*R.size,i+1+(j+1)*R.size,i+(j+1)*R.size])
	IDlist = np.array(IDlist)
	
	
	output.Allocate(numCells, 2*numCells)                       	# Allocate the cells in the memory
	
	for cell in range(numCells):
		numCellPoints = int(4)                                    	# There are 4 points for each cell (vtk_quad cell type)
		pointIds = vtk.vtkIdList()                              	# Creates an empty array to store the cells ID's
		for pointId in range(numCellPoints):
			pointIds.InsertId(pointId, int(IDlist[cell][pointId]))
		output.InsertNextCell(9, pointIds)# 9 for VTK_quad
"""


############################################################################################################
#############################################    END OF CODE   #############################################
############################################################################################################