# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################                 Wave Drag Plugin                 #############################
############################################################################################################

# Abstract:
This Paraview's programmable filter performs the breakdown of the profile drag into its viscous, wave and jet
components. It takes the desired variable from the Epsilon_Aero_Data output and it  looks for negative drag 
points which are tagged as "jet" regions. Then the wave drag regions are detected by knowing that vorticity 
and drag are weak there. Finally the viscous drag is extracted. 

# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""

print ("")
print ("")
print ("")
print ("####################################################")
print ("############    EPSILON WAVE DRAG      #############")
print ("####################################################")
print ("")



############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

# We load the Python and Paraview standard modules:
import numpy as np
import vtk
from vtk.numpy_interface import algorithms as algs



############################################################################################################
############################################# INPUT PARAMETERS #############################################
############################################################################################################

# Input parameters are defined as string in Paraview's GUI, thus a string to float convertion is required:
Breakdown = float(Breakdown)                                            # 0=2D   1=3D
Slice_Plane = float(Slice_Plane)                                        # 0=Normal to upstream direction    1=Normal direction defined by user
Slice_Normal = Slice_Normal
Slice_Point = Slice_Point
Parameter = Parameter
Breakdown_Method = float(Breakdown_Method)						        # 0=Aguirre method   1=Kusunose method
Aphi_Threshold = float(Aphi_Threshold) 
Anergy_Threshold = float(Anergy_Threshold)
DragThreshold = float(Drag_Threshold) 
VorticityThreshold = float(Vorticity_Threshold)



############################################################################################################
########################################### SLICING THE 3D DOMAIN ##########################################
############################################################################################################

INPUT = inputs[0]

if Breakdown == 1:                                                      # Breakdown a slice of the domain
     
    if Slice_Plane == 0:                                                # 0=Normal to upstream direction
        Slice_Normal = INPUT.FieldData["Plane_NORMAL"]

    plane = vtk.vtkPlane()											    # Creates slice plane
    plane.SetOrigin(Slice_Point[0], Slice_Point[1], Slice_Point[2])
    plane.SetNormal(Slice_Normal[0], Slice_Normal[1], Slice_Normal[2])

    cutter = vtk.vtkCutter()										    # Slices data with slice plane
    cutter.SetCutFunction(plane)
    cutter.SetInputData(INPUT.VTKObject)
    cutter.Update()

    
    PassArrays=vtk.vtkPassArrays()                                      # Let pass only the arrays required for the analysis
    PassArrays.SetInputDataObject(cutter.GetOutput())
    PassArrays.ClearArrays()    
    PassArrays.AddCellDataArray("")    
    PassArrays.AddFieldDataArray("Sref")    
    PassArrays.AddFieldDataArray("TRANSFORMATION") 
    PassArrays.AddPointDataArray("ARNTZ_A")      
    PassArrays.AddPointDataArray("ARNTZ_Aphi")   
    PassArrays.AddPointDataArray("KUSUNOSE_profile_drag")   
    PassArrays.AddPointDataArray("AERO_Vorticity_Mag")   
    PassArrays.AddPointDataArray(Parameter)   
    PassArrays.Update()  
    
    # Convert vtkPolyData input to vtkStructured grid: 
    clean = vtk.vtkmCleanGrid()                         
    clean.SetInputData(PassArrays.GetOutput()) 
    clean.CompactPointsOn()                                             
    clean.Update()
  
    INPUT = dsa.WrapDataObject(clean.GetOutput())
    
# Send mesh to the output (without arrays):                             # This mesh is used as a canvas where all the new arrays (calculated below) will be added
mesh=vtk.vtkPassArrays()   
mesh.SetInputDataObject(INPUT.VTKObject)
mesh.AddPointDataArray("")                               	        
mesh.AddCellDataArray("")
mesh.AddFieldDataArray("")
mesh.Update()
self.GetOutput().ShallowCopy(mesh.GetOutput())                          # This send the mesh to the output of ParaView



############################################################################################################
############################################ READING INPUT DATA ############################################
############################################################################################################

# Verifying that the required input data is available:
if Breakdown_Method == 0:
    if INPUT.GetPointData().HasArray("ARNTZ_A")==0: 
        print("# ERROR: Activate 'Arntz' method in Epsilon_Aero_Data (or 'Exergy')")
if Breakdown_Method == 1:
    if INPUT.GetPointData().HasArray("KUSUNOSE_profile_drag")==0: 
        print("# ERROR: Activate 'Kusunose' method in Epsilon_Aero_Data (or 'Far-field method')")

# Reading data from the input:
Parameter = INPUT.PointData[Parameter]   					            # "Parameter" is the aerodynamic variable selected by the user in the GUI (which is intended to be split into two parts: one lying on the viscous region and other lying on the wave region)
if Breakdown_Method == 0:
    ARNTZ_A = INPUT.PointData["ARNTZ_A"]
    ARNTZ_Aphi = INPUT.PointData["ARNTZ_Aphi"]
if Breakdown_Method == 1:
	KUSUNOSE_profile_drag = INPUT.PointData["KUSUNOSE_profile_drag"]
	AERO_Vorticity_Mag = INPUT.PointData["AERO_Vorticity_Mag"]

Sref = INPUT.FieldData["Sref"]
TRANSFORMATION = INPUT.FieldData["TRANSFORMATION"]


############################################################################################################
######################################### BREAKDOWN OF PROFILE DRAG ########################################
############################################################################################################

Aphi_Threshold = float(Aphi_Threshold) 
Anergy_Threshold = float(Anergy_Threshold)


if Breakdown_Method == 0:                                               # Aguirre method:
	
	# DETECTING SHOCKWAVE WAKE REGIONS:                                 # It looks for regions with weak viscous anergy and weak total anergy. The points not satisfying this condition are masked (settled to zero)
	MASK1 = ARNTZ_Aphi < Aphi_Threshold
	MASK2 = ARNTZ_A < Anergy_Threshold 
	wave = 1*MASK1*MASK2*Parameter 
	
	# DETECTING VISCOUS WAKE REGIONS:	                                # It looks for regions with high viscous anergy and high total anergy. The points not satisfying this condition are masked (settled to zero)
	MASK3 = 1-(1*MASK1*MASK2)                                           # This allows tagging with 1 the points that where previously tagged with 0
	viscous = 1*MASK3*Parameter


if Breakdown_Method == 1:                                               # Kusunose method:

	# DETECTING SHOCKWAVE WAKE REGIONS:                                 # It looks for regions with weak vorticity and weak profile drag. The points not satisfying this condition are masked (settled to zero)
	MASK1 = AERO_Vorticity_Mag < Vorticity_Threshold
	MASK2 = KUSUNOSE_profile_drag < Drag_Threshold
	wave = 1*MASK1*MASK2*Parameter

	# DETECTING VISCOUS WAKE REGIONS:                                   # It looks for regions with high vorticity and high profile drag. The points not satisfying this condition are masked (settled to zero)
	MASK3 = 1-(1*MASK1*MASK2)                                           # This allows tagging with 1 the points that where previously tagged with 0
	viscous = 1*MASK3*Parameter



############################################################################################################
###############################################  DATA OUTPUT ###############################################
############################################################################################################

# Basic aero data:
output.FieldData.append(Sref, "Sref")
output.FieldData.append(TRANSFORMATION, "TRANSFORMATION")
output.PointData.append(INPUT.PointData["Points"], "Points")            # Required for Epsilon_Integration_Tool
output.PointData.append(Parameter, "Parameter")
output.PointData.append(viscous, "Parameter_viscous")
output.PointData.append(wave, "Parameter_wave")


if Breakdown_Method == 0:
	output.PointData.append(ARNTZ_Aphi, "ARNTZ_Aphi")
	output.PointData.append(ARNTZ_A, "ARNTZ_A")
if Breakdown_Method == 1:
	output.PointData.append(AERO_Vorticity_Mag, "AERO_Vorticity_Mag")
	output.PointData.append(KUSUNOSE_profile_drag, "KUSUNOSE_profile_drag")


EpsilonIntegrationTool = []
EpsilonIntegrationTool.append(["Parameter","Survey_Plane"]) 
EpsilonIntegrationTool.append(["Parameter_viscous","Survey_Plane"]) 
EpsilonIntegrationTool.append(["Parameter_wave","Survey_Plane"]) 

# Send "EpsilonIntegrationTool" array to the output:
NB_of_Arrays = np.shape(EpsilonIntegrationTool)[0]

array = vtk.vtkStringArray()
array.SetName("ARRAY")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][0])
output.GetFieldData().AddArray(array)


array = vtk.vtkStringArray()
array.SetName("ARRAY_Integration")
array.SetNumberOfTuples(NB_of_Arrays)
for i in range(NB_of_Arrays):
	array.SetValue(i,EpsilonIntegrationTool[i][1])
output.GetFieldData().AddArray(array)

############################################################################################################
######################################### INTEGRATE DRAG COMPONENTS ########################################
############################################################################################################

if Breakdown == 1:                                                      # Breakdown a slice of the domain

	# Select the arrays to integrate:
	WakeSlice=vtk.vtkPassArrays()																	# Delete all the arrays already contained in the input
	WakeSlice.SetInputDataObject(self.GetOutput())
	WakeSlice.ClearArrays()
	WakeSlice.AddPointDataArray("Parameter")
	WakeSlice.AddPointDataArray("Parameter_viscous")
	WakeSlice.AddPointDataArray("Parameter_wave")
	WakeSlice.AddCellDataArray("")
	WakeSlice.AddFieldDataArray("")
	WakeSlice.Update()

	# Integrate variables:
	integrate = vtk.vtkIntegrateAttributes()
	integrate.SetInputData(WakeSlice.GetOutput())
	integrate.Update()

	print("# INTEGRATED VALUES ON SLICE:")
	# Add name and values to the output variables:
	pd = integrate.GetOutput().GetPointData()

	for j in range(pd.GetNumberOfArrays()):  
		arrayName = pd.GetArray(j).GetName()
		IntegralValue = np.array(pd.GetArray(j).GetTuple(0))
		output.FieldData.append(IntegralValue, arrayName)
		print(str(arrayName),"=",IntegralValue[0])



############################################################################################################
###############################################  END OF CODE ###############################################
############################################################################################################

