# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################               Fluent CGNS reader                 #############################
############################################################################################################

# Abstract:
This python code is a module of the "Epsilon Aero Data" plugin that reads some variables from a FLUENT file
exported into a CGNS file format.

# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""


############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

# We load the Python and Paraview standard modules:
import numpy as np
import vtk
from vtk.numpy_interface import algorithms as algs


############################################################################################################
######################################## FLUENT CGNS READER FUNCTION #######################################
############################################################################################################

def Reader_Fluent_CGNS(INPUT):                                                              # It creates a function that will be called by the Epsion_Aero_Data.py file
	
	# Reading aero data from Fluent CGNS file:
	POINTS = INPUT.Points   										                            # It takes the points coordinates
	Vx = INPUT.PointData["Velocity"][:,0]                                                       # Speed component along x-axis [m/s]
	
	
	# Read Y_VELOCITY data:    
	try:
		Vy = INPUT.PointData["Velocity"][:,1]
	except:
		print ()
		print ('# You are loading a 2D data (No Y_VELOCITY data available)')                    # It means that Y_VELOCITY array does not exists 
		Vy = 0*Vx
	
	
	# Read Z_VELOCITY data:
	try:
		Vz = INPUT.PointData["Velocity"][:,2]
	except:
		print ()
		print ('# You are loading a 2D data (No Z_VELOCITY data available)')                    # It means that Z_VELOCITY array does not exists 
		Vz = 0*Vx
	
	print ()
	print ('# ATTENTION: verify if "Upwards Direction" has the right axis selected (Y-axis or Z-axis)')    
	
	
	# Read static pressure data:
	Ps = INPUT.PointData["Pressure"]                                                            # Static presure [Pa]
	
	
	# Read density data:
	Rho = INPUT.PointData["Density"]                                                            # Density [Kg/m^3]    
	
	
	# Read static temperature data:
	
	if INPUT.GetPointData().HasArray("Temperature")==0:     								    # This check if "TEMPERATURE" array does not exists in the input 
		Ts = 'None'                                                                                 
	else:
		Ts = INPUT.PointData["Temperature"]                                                     # If the array exists, then reads the array data directly from input     
	
	
	MU_LAM = INPUT.PointData["ViscosityMolecular"]                                              # Molecular viscosity [Kg/m*s]
	MU_TURB = INPUT.PointData["ViscosityEddy"]                                                  # Turbulent viscosity [Kg/m*s]

	
	
	
	Velocity_Vector = algs.make_vector(Vx,Vy,Vz)    			        	                    # Velocity vector in the mesh reference frame [m/s]
	
	# Read vorticity data:
	if INPUT.GetPointData().HasArray("VORTICITY")==0:                                       	# This check if "VORTICITY" array does not exists in the input 
		VORTICITY = 'None'  
	else:
		VORTICITY = INPUT.PointData["VORTICITY"]                                                # Vorticity vector in the mesh reference frame [1/s]
	
	
	return POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY
