# -*- coding: utf-8 -*-
"""
############################################################################################################
#############################                     EPSILON                      #############################
#############################             STAR CCM Ensight reader              #############################
############################################################################################################

# Abstract:
This python code is a module of the "Epsilon Aero Data" plugin that reads some variables from a STAR CCM file
exported into the Ensight Gold Case format.


# Disclaimer:
This plugin was developed for research purposes and released under GNU GPL V3 license without any guarantee
of any kind, either express or implied. Its use for teaching, research and industrial applications is allowed
under the strict responsibility of the user. The authors of this plugin are not responsible for any damages
or errors related to the use of the plugin.
"""


############################################################################################################
############################################### MODULES LOAD ###############################################
############################################################################################################

# We load the Python and Paraview standard modules:
import numpy as np
import vtk
from vtk.numpy_interface import algorithms as algs


############################################################################################################
##################################### STAR CCM ENSIGHT READER FUNCTION #####################################
############################################################################################################

def Reader_STAR_CCM_Ensight(INPUT):
	
	print ()
	
	# Reading aero data from Fluent file:
	POINTS = INPUT.Points   															# It takes the points coordinates
	
	if INPUT.GetPointData().HasArray("Velocityi")==1:     
		Vx = INPUT.PointData["Velocityi"] 							                    # Velocity component along x-axis [m/s]
	else:
		Vx = INPUT.PointData["Velocity[i] (m/s)"]
	
	# Read Velocityj data:
	if INPUT.GetPointData().HasArray("Velocityj")==1:                           		# This check if "Velocityj" array exists at the input  
		Vy = INPUT.PointData["Velocityj"]                                       		# If the array exists, then reads the array data directly from input
	elif INPUT.GetPointData().HasArray("Velocity[j] (m/s)")==1:                           		
		Vy = INPUT.PointData["Velocity[j] (m/s)"]                                       	        
	else:
		print ('# STAR CCM: You are loading a 2D data (No Velocityj data available)')   # It means that Velocityj array does not exists 
		Vy = 0*Vx                                                              			# If the array does not exists,it creates a null vector
	
	# Read Velocityk data:
	if INPUT.GetPointData().HasArray("Velocityk")==1:                           		# This check if "Velocityk" array exists at the input    
		Vz = INPUT.PointData["Velocityk"]                                       		# If the array exists, then reads the array data directly from input    
	elif INPUT.GetPointData().HasArray("Velocity[k] (m/s)")==1:                           		
		Vz = INPUT.PointData["Velocity[k] (m/s)"]
	else:
		print ('# STAR CCM: You are loading a 2D data (No Velocityk data available)')  	# It means that Velocityk array does not exists 
		Vz = 0*Vx
		
	# Read static pressure data:
	if INPUT.GetPointData().HasArray("AbsolutePressure")==1:                           	# This check if "AbsolutePressure" array exists at the input 
		Ps = INPUT.PointData["AbsolutePressure"]                                        # Static pressure [Pa]
	elif INPUT.GetPointData().HasArray("Absolute Pressure (Pa)")==1:                           	
		Ps = INPUT.PointData["Absolute Pressure (Pa)"]
	elif INPUT.GetPointData().HasArray("StaticPressure")==1:                           	
		Ps = INPUT.PointData["StaticPressure"]
	elif INPUT.GetPointData().HasArray("Static Pressure (Pa)")==1:
		Ps = INPUT.PointData["Static Pressure (Pa)"]     
	else:
		Ps = 'None'        
	
	# Read density data:
	if INPUT.GetPointData().HasArray("Density")==1: 
		Rho = INPUT.PointData["Density"]                                				# Density [Kg/m^3]
	elif INPUT.GetPointData().HasArray("Density (kg/m^3)")==1:                           	
		Rho = INPUT.PointData["Density (kg/m^3)"]
	else:
		Rho = 'None'
	
	# Read static temperature data:
	if INPUT.GetPointData().HasArray("Temperature")==1:                           		# This check if "Temperature" array exists at the input 
		Ts = INPUT.PointData["Temperature"]     
	elif INPUT.GetPointData().HasArray("StaticTemperature")==1:                     	# This check if "StaticTemperature" array exists at the input 
		Ts = INPUT.PointData["StaticTemperature"]                                 	    # If the array exists, then reads the array data directly from input 
	elif INPUT.GetPointData().HasArray("Temperature (K)")==1:                     	
		Ts = INPUT.PointData["Temperature (K)"]
	elif INPUT.GetPointData().HasArray("Static Temperature (K)")==1:                     	
		Ts = INPUT.PointData["Static Temperature (K)"]
	else:
		Ts = 'None'    
	
	# Read viscosity data:
	if INPUT.GetPointData().HasArray("TurbulentViscosity")==1:
		MU_TURB = INPUT.PointData["TurbulentViscosity"]                 				    # Turbulent viscosity [Kg/m*s]	
	else:
		MU_TURB = INPUT.PointData["Turbulent Viscosity (Pa-s)"]
	
	
	if INPUT.GetPointData().HasArray("Turbulent Viscosity Ratio")==0:                  	# This check if "TurbulentViscosityRatio" array does not  exists in the input 	
		MU_EFF = INPUT.PointData["EffectiveViscosity"] 									# Effective viscosity [Kg/m*s] 
		MU_LAM = MU_EFF - MU_TURB														# Molecular viscosity [Kg/m*s]
	else:
		MU_ratio = INPUT.PointData["Turbulent Viscosity Ratio"]
		MU_LAM = MU_TURB/MU_ratio
		MU_EFF = MU_LAM + MU_TURB
	
	
	Velocity_Vector = algs.make_vector(Vx,Vy,Vz)    			        	            # Velocity vector in the mesh reference frame [m/s]
	
	# Read vorticity data:
	if INPUT.GetPointData().HasArray("VORTICITY")==0:                                   # This check if "VORTICITY" array does not exists in the input 
		VORTICITY = 'None'
	else:
		VORTICITY = INPUT.PointData["VORTICITY"]                                        # Vorticity vector in the mesh reference frame [1/s]
	
	return POINTS,Ps,Ts,Rho,MU_LAM,MU_TURB,Velocity_Vector,VORTICITY